@@ -21,18 +21,24 @@
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA.
 
-   $Date: 2002/02/22 04:47:36 $ $Revision: 1.7 $
+   $Date: 2004/01/28 07:33:20 $ $Revision: 1.25 $
    */
 
-#include <config.h>
-#include <base/preface.h>
-#include <Foundation/NSArray.h>
-#include <base/behavior.h>
-#include <Foundation/NSException.h>
-#include <Foundation/NSPortCoder.h>
-#include <Foundation/NSDebug.h>
+#include "config.h"
+#include "GNUstepBase/preface.h"
+#include "Foundation/NSArray.h"
+#include "GNUstepBase/GSObjCRuntime.h"
+#include "Foundation/NSException.h"
+#include "Foundation/NSPortCoder.h"
+#include "Foundation/NSDebug.h"
+#include "Foundation/NSValue.h"
+// For private method _decodeArrayOfObjectsForKey:
+#include "Foundation/NSKeyedArchiver.h"
 
 static SEL	eqSel;
+static SEL	oaiSel;
+
+static Class	GSInlineArrayClass;
 
 @class	GSArrayEnumerator;
 @class	GSArrayEnumeratorReverse;
@@ -60,6 +66,10 @@ static SEL	eqSel;
 }
 @end
 
+@interface GSMutableArray (GSArrayBehavior)
+- (void) _raiseRangeExceptionWithIndex: (unsigned)index from: (SEL)sel;
+@end
+
 @interface GSPlaceholderArray : NSArray
 {
 }
@@ -67,12 +77,34 @@ static SEL	eqSel;
 
 @implementation GSArray
 
+- (void) _raiseRangeExceptionWithIndex: (unsigned)index from: (SEL)sel
+{
+  NSDictionary *info;
+  NSException  *exception;
+  NSString     *reason;
+
+  info = [NSDictionary dictionaryWithObjectsAndKeys:
+    [NSNumber numberWithUnsignedInt: index], @"Index",
+    [NSNumber numberWithUnsignedInt: _count], @"Count",
+    self, @"Array", nil, nil];
+                              
+  reason = [NSString stringWithFormat: @"Index %d is out of range %d (in '%@')",
+    index, _count, NSStringFromSelector(sel)];
+
+  exception = [NSException exceptionWithName: NSRangeException
+		                      reason: reason
+                                    userInfo: info];
+  [exception raise];
+}
+
 + (void) initialize
 {
   if (self == [GSArray class])
     {
       [self setVersion: 1];
       eqSel = @selector(isEqual:);
+      oaiSel = @selector(objectAtIndex:);
+      GSInlineArrayClass = [GSInlineArray class];
     }
 }
 
@@ -83,6 +115,11 @@ static SEL	eqSel;
   return array;
 }
 
+- (id) copyWithZone: (NSZone*)zone
+{
+  return RETAIN(self);	// Optimised version
+}
+
 - (void) dealloc
 {
   if (_contents_array)
@@ -121,7 +158,7 @@ static SEL	eqSel;
 	      _count = i;
 	      RELEASE(self);
 	      [NSException raise: NSInvalidArgumentException
-			  format: @"Tried to add nil"];
+			  format: @"Tried to init array with nil to object"];
 	    }
 	}
       _count = count;
@@ -131,31 +168,49 @@ static SEL	eqSel;
 
 - (void) encodeWithCoder: (NSCoder*)aCoder
 {
-  [aCoder encodeValueOfObjCType: @encode(unsigned)
-			     at: &_count];
-  if (_count > 0)
+  if ([aCoder allowsKeyedCoding])
     {
-      [aCoder encodeArrayOfObjCType: @encode(id)
-			      count: _count
-				 at: _contents_array];
+      [super encodeWithCoder: aCoder];
+    }
+  else
+    {
+      /* For performace we encode directly ... must exactly match the
+       * superclass implemenation. */
+      [aCoder encodeValueOfObjCType: @encode(unsigned)
+				 at: &_count];
+      if (_count > 0)
+	{
+	  [aCoder encodeArrayOfObjCType: @encode(id)
+				  count: _count
+				     at: _contents_array];
+	}
     }
 }
 
 - (id) initWithCoder: (NSCoder*)aCoder
 {
-  [aCoder decodeValueOfObjCType: @encode(unsigned)
-			     at: &_count];
-  if (_count > 0)
+  if ([aCoder allowsKeyedCoding])
     {
-      _contents_array = NSZoneCalloc([self zone], _count, sizeof(id));
-      if (_contents_array == 0)
+      self = [super initWithCoder: aCoder];
+    }
+  else
+    {
+      /* for performance, we decode directly into memory rather than
+       * using the superclass method. Must exactly match superclass. */
+      [aCoder decodeValueOfObjCType: @encode(unsigned)
+				 at: &_count];
+      if (_count > 0)
 	{
-	  [NSException raise: NSMallocException
-		      format: @"Unable to make array"];
+	  _contents_array = NSZoneCalloc([self zone], _count, sizeof(id));
+	  if (_contents_array == 0)
+	    {
+	      [NSException raise: NSMallocException
+			  format: @"Unable to make array"];
+	    }
+	  [aCoder decodeArrayOfObjCType: @encode(id)
+				  count: _count
+				     at: _contents_array];
 	}
-      [aCoder decodeArrayOfObjCType: @encode(id)
-			      count: _count
-				 at: _contents_array];
     }
   return self;
 }
@@ -218,6 +273,33 @@ static SEL	eqSel;
   return NSNotFound;
 }
 
+- (BOOL) isEqualToArray: (NSArray*)otherArray
+{
+  unsigned i;
+ 
+  if (self == (id)otherArray)
+    {
+      return YES;
+    }
+  if (_count != [otherArray count])
+    {
+      return NO;
+    }
+  if (_count > 0)
+    {
+      IMP	get1 = [otherArray methodForSelector: oaiSel];
+
+      for (i = 0; i < _count; i++)
+	{
+	  if (![_contents_array[i] isEqual: (*get1)(otherArray, oaiSel, i)])
+	    {
+	      return NO;
+	    }
+	}
+    }
+  return YES;
+}
+
 - (id) lastObject
 {
   if (_count)
@@ -231,8 +313,7 @@ static SEL	eqSel;
 {
   if (index >= _count)
     {
-      [NSException raise: NSRangeException
-		  format: @"Index out of bounds"];
+        [self _raiseRangeExceptionWithIndex: index from: _cmd];
     }
   return _contents_array[index];
 }
@@ -247,7 +328,7 @@ static SEL	eqSel;
     }
 }
 
-- (void) makeObjectsPerformSelector: (SEL)aSelector withObject:argument
+- (void) makeObjectsPerformSelector: (SEL)aSelector withObject: argument
 {
   unsigned i = _count;
 
@@ -278,7 +359,6 @@ static SEL	eqSel;
       aBuffer[j++] = _contents_array[i];
     }
 }
-
 @end
 
 @implementation	GSInlineArray
@@ -311,7 +391,7 @@ static SEL	eqSel;
 	      _count = i;
 	      RELEASE(self);
 	      [NSException raise: NSInvalidArgumentException
-			  format: @"Tried to add nil"];
+			  format: @"Tried to init array with nil object"];
 	    }
 	}
       _count = count;
@@ -327,7 +407,64 @@ static SEL	eqSel;
   if (self == [GSMutableArray class])
     {
       [self setVersion: 1];
-      behavior_class_add_class(self, [GSArray class]);
+      GSObjCAddClassBehavior(self, [GSArray class]);
+    }
+}
+
+- (void) addObject: (id)anObject
+{
+  if (anObject == nil)
+    {
+      [NSException raise: NSInvalidArgumentException
+		  format: @"Tried to add nil to array"];
+    }
+  if (_count >= _capacity)
+    {
+      id	*ptr;
+      size_t	size = (_capacity + _grow_factor)*sizeof(id);
+
+      ptr = NSZoneRealloc([self zone], _contents_array, size);
+      if (ptr == 0)
+	{
+	  [NSException raise: NSMallocException
+		      format: @"Unable to grow array"];
+	}
+      _contents_array = ptr;
+      _capacity += _grow_factor;
+      _grow_factor = _capacity/2;
+    }
+  _contents_array[_count] = RETAIN(anObject);
+  _count++;	/* Do this AFTER we have retained the object.	*/
+}
+
+/**
+ * Optimised code for copying
+ */
+- (id) copyWithZone: (NSZone*)zone
+{
+  NSArray       *copy;
+
+  copy = (id)NSAllocateObject(GSInlineArrayClass, sizeof(id)*_count, zone);
+  return [copy initWithObjects: _contents_array count: _count];
+}
+
+- (void) exchangeObjectAtIndex: (unsigned int)i1 
+             withObjectAtIndex: (unsigned int)i2
+{
+  if (i1 >= _count)
+    {
+      [self _raiseRangeExceptionWithIndex: i1 from: _cmd];
+    }
+  if (i2 >= _count)
+    {
+      [self _raiseRangeExceptionWithIndex: i2 from: _cmd];
+    }
+  if (i1 != i2)
+    {
+      id	tmp = _contents_array[i1];
+
+      _contents_array[i1] = _contents_array[i2];
+      _contents_array[i2] = tmp;
     }
 }
 
@@ -345,21 +482,31 @@ static SEL	eqSel;
 
 - (id) initWithCoder: (NSCoder*)aCoder
 {
-  unsigned    count;
+  if ([aCoder allowsKeyedCoding])
+    {
+      NSArray *array = [(NSKeyedUnarchiver*)aCoder _decodeArrayOfObjectsForKey: 
+						@"NS.objects"];
 
-  [aCoder decodeValueOfObjCType: @encode(unsigned)
-			     at: &count];
-  if ((self = [self initWithCapacity: count]) == nil)
-    {
-      [NSException raise: NSMallocException
-		  format: @"Unable to make array"];
+      [self initWithArray: array];
     }
-  if (count > 0)
+  else
     {
-      [aCoder decodeArrayOfObjCType: @encode(id)
-			      count: count
-				 at: _contents_array];
-      _count = count;
+	unsigned    count;
+
+	[aCoder decodeValueOfObjCType: @encode(unsigned)
+			           at: &count];
+	if ((self = [self initWithCapacity: count]) == nil)
+	  {
+	    [NSException raise: NSMallocException
+			format: @"Unable to make array while initializing from coder"];
+	  }
+	if (count > 0)
+	  {
+	    [aCoder decodeArrayOfObjCType: @encode(id)
+		                    count: count
+				       at: _contents_array];
+	    _count = count;
+	  }
     }
   return self;
 }
@@ -378,7 +525,7 @@ static SEL	eqSel;
 	      _count = i;
 	      RELEASE(self);
 	      [NSException raise: NSInvalidArgumentException
-			  format: @"Tried to add nil"];
+			  format: @"Tried to init array with nil object"];
 	    }
 	}
       _count = count;
@@ -392,13 +539,21 @@ static SEL	eqSel;
 
   if (!anObject)
     {
-      [NSException raise: NSInvalidArgumentException
-		  format: @"Tried to insert nil"];
+      NSException  *exception;
+      NSDictionary *info;
+
+      info = [NSDictionary dictionaryWithObjectsAndKeys:
+	[NSNumber numberWithUnsignedInt: index], @"Index",
+	self, @"Array", nil, nil];
+
+      exception = [NSException exceptionWithName: NSInvalidArgumentException
+	reason: @"Tried to insert nil to array"
+	userInfo: info];
+      [exception raise];
     }
   if (index > _count)
     {
-      [NSException raise: NSRangeException format:
-	      @"in insertObject:atIndex:, index %d is out of range", index];
+      [self _raiseRangeExceptionWithIndex: index from: _cmd];
     }
   if (_count == _capacity)
     {
@@ -429,30 +584,16 @@ static SEL	eqSel;
   _contents_array[index] = RETAIN(anObject);
 }
 
-- (void) addObject: (id)anObject
+- (id) makeImmutableCopyOnFail: (BOOL)force
 {
-  if (anObject == nil)
-    {
-      [NSException raise: NSInvalidArgumentException
-		  format: @"Tried to add nil"];
-    }
-  if (_count >= _capacity)
-    {
-      id	*ptr;
-      size_t	size = (_capacity + _grow_factor)*sizeof(id);
-
-      ptr = NSZoneRealloc([self zone], _contents_array, size);
-      if (ptr == 0)
-	{
-	  [NSException raise: NSMallocException
-		      format: @"Unable to grow"];
-	}
-      _contents_array = ptr;
-      _capacity += _grow_factor;
-      _grow_factor = _capacity/2;
-    }
-  _contents_array[_count] = RETAIN(anObject);
-  _count++;	/* Do this AFTER we have retained the object.	*/
+#ifndef NDEBUG
+  GSDebugAllocationRemove(isa, self);
+#endif
+  isa = [GSArray class];
+#ifndef NDEBUG
+  GSDebugAllocationAdd(isa, self);
+#endif
+  return self;
 }
 
 - (void) removeLastObject
@@ -464,6 +605,7 @@ static SEL	eqSel;
     }
   _count--;
   RELEASE(_contents_array[_count]);
+  _contents_array[_count] = 0;
 }
 
 - (void) removeObject: (id)anObject
@@ -504,6 +646,7 @@ static SEL	eqSel;
 		  _contents_array[pos-1] = _contents_array[pos];
 		}
 	      _count--;
+	      _contents_array[_count] = 0;
 	      RELEASE(obj);
 	    }
 	}
@@ -522,9 +665,7 @@ static SEL	eqSel;
 
   if (index >= _count)
     {
-      [NSException raise: NSRangeException
-		  format: @"in removeObjectAtIndex:, index %d is out of range",
-			index];
+      [self _raiseRangeExceptionWithIndex: index from: _cmd];
     }
   obj = _contents_array[index];
   _count--;
@@ -533,6 +674,7 @@ static SEL	eqSel;
       _contents_array[index] = _contents_array[index+1];
       index++;
     }
+  _contents_array[_count] = 0;
   RELEASE(obj);	/* Adjust array BEFORE releasing object.	*/
 }
 
@@ -560,6 +702,7 @@ static SEL	eqSel;
 	      _contents_array[pos-1] = _contents_array[pos];
 	    }
 	  _count--;
+	  _contents_array[_count] = 0;
 	  RELEASE(obj);
 	}
     }
@@ -571,9 +714,7 @@ static SEL	eqSel;
 
   if (index >= _count)
     {
-      [NSException raise: NSRangeException format:
-	    @"in replaceObjectAtIndex:withObject:, index %d is out of range",
-	    index];
+      [self _raiseRangeExceptionWithIndex: index from: _cmd];
     }
   /*
    *	Swap objects in order so that there is always a valid object in the
@@ -585,20 +726,21 @@ static SEL	eqSel;
   RELEASE(obj);
 }
 
-- (void) sortUsingFunction: (int(*)(id,id,void*))compare 
+- (void) sortUsingFunction: (NSComparisonResult(*)(id,id,void*))compare 
 		   context: (void*)context
 {
   /* Shell sort algorithm taken from SortingInAction - a NeXT example */
 #define STRIDE_FACTOR 3	// good value for stride factor is not well-understood
                         // 3 is a fairly good choice (Sedgewick)
-  unsigned	c,d, stride;
+  unsigned int	c;
+  unsigned int	d;
+  unsigned int	stride = 1;
   BOOL		found;
-  int		count = _count;
+  unsigned int	count = _count;
 #ifdef	GSWARN
   BOOL		badComparison = NO;
 #endif
 
-  stride = 1;
   while (stride <= count)
     {
       stride = stride * STRIDE_FACTOR + 1;
@@ -851,8 +993,6 @@ static SEL	eqSel;
 
 @implementation	GSPlaceholderArray
 
-static Class	GSInlineArrayClass;
-
 + (void) initialize
 {
   GSInlineArrayClass = [GSInlineArray class];
@@ -867,7 +1007,7 @@ static Class	GSInlineArrayClass;
 - (id) objectAtIndex: (unsigned)index
 {
   [NSException raise: NSInternalInconsistencyException
-	      format: @"attempt to use uninitialised array"];
+	      format: @"Attempt to use uninitialised array"];
   return 0;
 }
 
@@ -878,32 +1018,30 @@ static Class	GSInlineArrayClass;
 
 - (id) initWithCoder: (NSCoder*)aCoder
 {
-  GSInlineArray	*a;
-  unsigned	c;
-
-  [aCoder decodeValueOfObjCType: @encode(unsigned) at: &c];
-  a = (id)NSAllocateObject(GSInlineArrayClass, sizeof(id)*c,
-    GSObjCZone(self));
-  a->_contents_array = (id*)&a[1];
-  if (c > 0)
+  if ([aCoder allowsKeyedCoding])
     {
-#if	!GS_WITH_GC
-      unsigned	i;
-#endif
+      NSArray *array = [(NSKeyedUnarchiver*)aCoder _decodeArrayOfObjectsForKey: 
+						@"NS.objects"];
 
-      [aCoder decodeArrayOfObjCType: @encode(id)
-			      count: c
-				 at: a->_contents_array];
-#if	!GS_WITH_GC
-      for (i = 0; i < c; i++)
-	{
-	  a->_contents_array[i] = RETAIN(a->_contents_array[i]);
-	  a->_count = i;
+      return array;
+    }
+  else
+    {
+      GSInlineArray	*a;
+      unsigned	c;
+
+      [aCoder decodeValueOfObjCType: @encode(unsigned) at: &c];
+      a = (id)NSAllocateObject(GSInlineArrayClass, sizeof(id)*c, GSObjCZone(self));
+      a->_contents_array = (id*)&a[1];
+      if (c > 0)
+        {
+	  [aCoder decodeArrayOfObjCType: @encode(id)
+		                  count: c
+				  at: a->_contents_array];
 	}
-#endif
+      a->_count = c;
+      return a;
     }
-  a->_count = c;
-  return a;
 }
 
 - (id) initWithObjects: (id*)objects count: (unsigned)count
@@ -916,7 +1054,7 @@ static Class	GSInlineArrayClass;
 - (unsigned) count
 {
   [NSException raise: NSInternalInconsistencyException
-	      format: @"attempt to use uninitialised array"];
+	      format: @"Attempt to use uninitialised array"];
   return 0;
 }
 
