@@ -44,16 +44,17 @@
  */
 
 #include "config.h"
-#include <base/preface.h>
-#include <Foundation/NSAttributedString.h>
-#include <Foundation/NSException.h>
-#include <Foundation/NSRange.h>
-#include <Foundation/NSDebug.h>
-#include <Foundation/NSArray.h>
-#include <Foundation/NSLock.h>
-#include <Foundation/NSThread.h>
-#include <Foundation/NSNotification.h>
-#include <Foundation/NSZone.h>
+#include "GNUstepBase/preface.h"
+#include "GNUstepBase/GSLock.h"
+#include "Foundation/NSAttributedString.h"
+#include "Foundation/NSException.h"
+#include "Foundation/NSRange.h"
+#include "Foundation/NSDebug.h"
+#include "Foundation/NSArray.h"
+#include "Foundation/NSLock.h"
+#include "Foundation/NSThread.h"
+#include "Foundation/NSNotification.h"
+#include "Foundation/NSZone.h"
 
 #define		SANITY_CHECKS	0
 
@@ -103,7 +104,7 @@ static	NSDictionary	*blank;
 #define GSI_MAP_VTYPES	GSUNION_INT
 #define	GSI_MAP_NOCLEAN	1
 
-#include <base/GSIMap.h>
+#include "GNUstepBase/GSIMap.h"
 
 static NSLock		*attrLock = nil;
 static GSIMapTable_t	attrMap;
@@ -253,11 +254,11 @@ static SEL	oatSel;
 static SEL	remSel;
 
 static IMP	infImp;
-static void	(*addImp)();
-static unsigned (*cntImp)();
-static void	(*insImp)();
+static void	(*addImp)(NSMutableArray*,SEL,id);
+static unsigned (*cntImp)(NSArray*,SEL);
+static void	(*insImp)(NSMutableArray*,SEL,id,unsigned);
 static IMP	oatImp;
-static void	(*remImp)();
+static void	(*remImp)(NSMutableArray*,SEL,unsigned);
 
 #define	NEWINFO(Z,O,L)	((*infImp)(infCls, infSel, (Z), (O), (L)))
 #define	ADDOBJECT(O)	((*addImp)(_infoArray, addSel, (O)))
@@ -265,7 +266,7 @@ static void	(*remImp)();
 #define	OBJECTAT(I)	((*oatImp)(_infoArray, oatSel, (I)))
 #define	REMOVEAT(I)	((*remImp)(_infoArray, remSel, (I)))
 
-static void _setup()
+static void _setup(void)
 {
   if (infCls == 0)
     {
@@ -286,11 +287,13 @@ static void _setup()
 
       a = [NSMutableArray allocWithZone: NSDefaultMallocZone()];
       a = [a initWithCapacity: 1];
-      addImp = (void (*)())[a methodForSelector: addSel];
-      cntImp = (unsigned (*)())[a methodForSelector: cntSel];
-      insImp = (void (*)())[a methodForSelector: insSel];
+      addImp = (void (*)(NSMutableArray*,SEL,id))[a methodForSelector: addSel];
+      cntImp = (unsigned (*)(NSArray*,SEL))[a methodForSelector: cntSel];
+      insImp = (void (*)(NSMutableArray*,SEL,id,unsigned))
+	[a methodForSelector: insSel];
       oatImp = [a methodForSelector: oatSel];
-      remImp = (void (*)())[a methodForSelector: remSel];
+      remImp = (void (*)(NSMutableArray*,SEL,unsigned))
+	[a methodForSelector: remSel];
       RELEASE(a);
       d = [NSDictionary new];
       blank = cacheAttributes(d);
@@ -318,6 +321,7 @@ _setAttributesFrom(
   if (aRange.length == 0)
     {
       attr = blank;
+      range = aRange; /* Set to satisfy the loop condition below. */
     }
   else
     {
@@ -424,34 +428,15 @@ _attributesAtIndexEffectiveRange(
   return nil;
 }
 
-/*
- * If we are multi-threaded, we must guard access to the uniquing set.
- */
-+ (void) _becomeThreaded: (id)notification
-{
-  attrLock = [NSLock new];
-  lockSel = @selector(lock);
-  unlockSel = @selector(unlock);
-  lockImp = [attrLock methodForSelector: lockSel];
-  unlockImp = [attrLock methodForSelector: unlockSel];
-}
-
 + (void) initialize
 {
   _setup();
 
-  if ([NSThread isMultiThreaded])
-    {
-      [self _becomeThreaded: nil];
-    }
-  else
-    {
-      [[NSNotificationCenter defaultCenter]
-	addObserver: self
-	   selector: @selector(_becomeThreaded:)
-	       name: NSWillBecomeMultiThreadedNotification
-	     object: nil];
-    }
+  attrLock = [GSLazyLock new];
+  lockSel = @selector(lock);
+  unlockSel = @selector(unlock);
+  lockImp = [attrLock methodForSelector: lockSel];
+  unlockImp = [attrLock methodForSelector: unlockSel];
 }
 
 - (id) initWithString: (NSString*)aString
@@ -522,9 +507,15 @@ _attributesAtIndexEffectiveRange(
 
 #if	SANITY_CHECKS
 
-#define	SANITY()	[self sanity]
-	
-- (void) sanity
+#define	SANITY()	[self _sanity]
+
+#else
+#define	SANITY()	
+#endif
+
+/* We always compile in this method so that it is available from
+ * regression test cases.  */
+- (void) _sanity
 {
   GSAttrInfo	*info;
   unsigned	i;
@@ -543,9 +534,6 @@ _attributesAtIndexEffectiveRange(
       l = info->loc;
     }
 }
-#else
-#define	SANITY()	
-#endif
 
 + (void) initialize
 {
@@ -715,6 +703,7 @@ SANITY();
       if (info->attrs == attributes)
 	{
 	  unCacheAttributes(attributes);
+	  RELEASE(attributes);
 	}
       else
 	{
@@ -726,6 +715,7 @@ SANITY();
   else if (info->attrs == attributes)
     {
       unCacheAttributes(attributes);
+      RELEASE(attributes);
     }
   else
     {
@@ -802,7 +792,7 @@ SANITY();
       info = OBJECTAT(arrayIndex);
       if (info->loc < NSMaxRange(range))
 	{
-	  int	next = arrayIndex + 1;
+	  unsigned int	next = arrayIndex + 1;
 
 	  while (next < arraySize)
 	    {
@@ -825,23 +815,8 @@ SANITY();
 	}
       else
 	{
-	  if (arrayIndex != 0)
-	    {
-	      REMOVEAT(arrayIndex);
-	      arraySize--;
-	    }
-	  else
-	    {
-	      NSDictionary	*d = blank;
-
-	      unCacheAttributes(info->attrs);
-	      DESTROY(info->attrs);
-	      d = cacheAttributes(d);
-	      info->attrs = d;
-	      /* set location so it will be correct after being modified
-	      below */
-	      info->loc = NSMaxRange(range);
-	    }
+	  REMOVEAT(arrayIndex);
+	  arraySize--;
 	}
     }
 
@@ -860,7 +835,7 @@ SANITY();
 	&& effectiveRange.length == range.length)
 	{
 	  arrayIndex--;
-	  if (arrayIndex!=0)
+	  if (arrayIndex!=0 || arraySize > 1)
 	    {
 	      REMOVEAT(arrayIndex);
 	      arraySize--;
