@@ -22,16 +22,18 @@
    */
 
 
-#include <config.h>
-#include <Foundation/NSDictionary.h>
-#include <Foundation/NSAutoreleasePool.h>
-#include <Foundation/NSUtilities.h>
-#include <Foundation/NSString.h>
-#include <Foundation/NSException.h>
-#include <Foundation/NSPortCoder.h>
-#include <Foundation/NSDebug.h>
+#include "config.h"
+#include "Foundation/NSDictionary.h"
+#include "Foundation/NSAutoreleasePool.h"
+#include "Foundation/NSUtilities.h"
+#include "Foundation/NSString.h"
+#include "Foundation/NSException.h"
+#include "Foundation/NSPortCoder.h"
+#include "Foundation/NSDebug.h"
+// For private method _decodeArrayOfObjectsForKey:
+#include "Foundation/NSKeyedArchiver.h"
 
-#include <base/behavior.h>
+#include "GNUstepBase/GSObjCRuntime.h"
 
 /*
  *	The 'Fastmap' stuff provides an inline implementation of a mapping
@@ -41,10 +43,10 @@
 #define	GSI_MAP_VTYPES		GSUNION_OBJ
 #define	GSI_MAP_HASH(M, X)		[X.obj hash]
 #define	GSI_MAP_EQUAL(M, X,Y)		[X.obj isEqual: Y.obj]
-#define	GSI_MAP_RETAIN_KEY(M, X)	((id)(X).obj) = \
+#define	GSI_MAP_RETAIN_KEY(M, X)	((X).obj) = \
 				[((id)(X).obj) copyWithZone: map->zone]
 
-#include	<base/GSIMap.h>
+#include	"GNUstepBase/GSIMap.h"
 
 @interface GSDictionary : NSDictionary
 {
@@ -84,6 +86,11 @@ static SEL	objSel;
     }
 }
 
+- (id) copyWithZone: (NSZone*)zone
+{
+  return RETAIN(self);
+}
+
 - (unsigned) count
 {
   return map.nodeCount;
@@ -97,18 +104,26 @@ static SEL	objSel;
 
 - (void) encodeWithCoder: (NSCoder*)aCoder
 {
-  unsigned	count = map.nodeCount;
-  SEL		sel = @selector(encodeObject:);
-  IMP		imp = [aCoder methodForSelector: sel];
-  GSIMapEnumerator_t	enumerator = GSIMapEnumeratorForMap(&map);
-  GSIMapNode	node = GSIMapEnumeratorNextNode(&enumerator);
-
-  [aCoder encodeValueOfObjCType: @encode(unsigned) at: &count];
-  while (node != 0)
+  if ([aCoder allowsKeyedCoding])
     {
-      (*imp)(aCoder, sel, node->key.obj);
-      (*imp)(aCoder, sel, node->value.obj);
-      node = GSIMapEnumeratorNextNode(&enumerator);
+      self = [super initWithCoder: aCoder];
+    }
+  else
+    {
+      unsigned		count = map.nodeCount;
+      SEL		sel = @selector(encodeObject:);
+      IMP		imp = [aCoder methodForSelector: sel];
+      GSIMapEnumerator_t	enumerator = GSIMapEnumeratorForMap(&map);
+      GSIMapNode	node = GSIMapEnumeratorNextNode(&enumerator);
+
+      [aCoder encodeValueOfObjCType: @encode(unsigned) at: &count];
+      while (node != 0)
+	{
+	  (*imp)(aCoder, sel, node->key.obj);
+	  (*imp)(aCoder, sel, node->value.obj);
+	  node = GSIMapEnumeratorNextNode(&enumerator);
+	}
+      GSIMapEndEnumerator(&enumerator);
     }
 }
 
@@ -119,22 +134,29 @@ static SEL	objSel;
 
 - (id) initWithCoder: (NSCoder*)aCoder
 {
-  unsigned	count;
-  id		key;
-  id		value;
-  SEL		sel = @selector(decodeValueOfObjCType:at:);
-  IMP		imp = [aCoder methodForSelector: sel];
-  const char	*type = @encode(id);
-
-  [aCoder decodeValueOfObjCType: @encode(unsigned)
-			     at: &count];
-
-  GSIMapInitWithZoneAndCapacity(&map, GSObjCZone(self), count);
-  while (count-- > 0)
+  if ([aCoder allowsKeyedCoding])
     {
-      (*imp)(aCoder, sel, type, &key);
-      (*imp)(aCoder, sel, type, &value);
-      GSIMapAddPairNoRetain(&map, (GSIMapKey)key, (GSIMapVal)value);
+      self = [super initWithCoder: aCoder];
+    }
+  else
+    {
+      unsigned	count;
+      id		key;
+      id		value;
+      SEL		sel = @selector(decodeValueOfObjCType:at:);
+      IMP		imp = [aCoder methodForSelector: sel];
+      const char	*type = @encode(id);
+
+      [aCoder decodeValueOfObjCType: @encode(unsigned)
+	                         at: &count];
+
+      GSIMapInitWithZoneAndCapacity(&map, GSObjCZone(self), count);
+      while (count-- > 0)
+        {
+	  (*imp)(aCoder, sel, type, &key);
+	  (*imp)(aCoder, sel, type, &value);
+	  GSIMapAddPairNoRetain(&map, (GSIMapKey)key, (GSIMapVal)value);
+	}
     }
   return self;
 }
@@ -142,7 +164,7 @@ static SEL	objSel;
 /* Designated initialiser */
 - (id) initWithObjects: (id*)objs forKeys: (id*)keys count: (unsigned)c
 {
-  int	i;
+  unsigned int	i;
 
   GSIMapInitWithZoneAndCapacity(&map, GSObjCZone(self), c);
   for (i = 0; i < c; i++)
@@ -238,6 +260,42 @@ static SEL	objSel;
   return self;
 }
 
+- (BOOL) isEqualToDictionary: (NSDictionary*)other
+{
+  unsigned	count;
+
+  if (other == self)
+    {
+      return YES;
+    }
+  count = map.nodeCount;
+  if (count == [other count])
+    {
+      if (count > 0)
+	{
+	  GSIMapEnumerator_t	enumerator;
+	  GSIMapNode		node;
+	  IMP			otherObj = [other methodForSelector: objSel];
+
+	  enumerator = GSIMapEnumeratorForMap(&map);
+	  while ((node = GSIMapEnumeratorNextNode(&enumerator)) != 0)
+	    {
+	      id o1 = node->value.obj;
+	      id o2 = (*otherObj)(other, objSel, node->key.obj);
+
+	      if (o1 != o2 && [o1 isEqual: o2] == NO)
+		{
+		  GSIMapEndEnumerator(&enumerator);
+		  return NO;
+		}
+	    }
+	  GSIMapEndEnumerator(&enumerator);
+	}
+      return YES;
+    }
+  return NO;
+}
+
 - (NSEnumerator*) keyEnumerator
 {
   return AUTORELEASE([[GSDictionaryKeyEnumerator allocWithZone:
@@ -272,10 +330,17 @@ static SEL	objSel;
 {
   if (self == [GSMutableDictionary class])
     {
-      behavior_class_add_class(self, [GSDictionary class]);
+      GSObjCAddClassBehavior(self, [GSDictionary class]);
     }
 }
 
+- (id) copyWithZone: (NSZone*)zone
+{
+  NSDictionary	*copy = [GSDictionary allocWithZone: zone];
+
+  return [copy initWithDictionary: self copyItems: NO];
+}
+
 /* Designated initialiser */
 - (id) initWithCapacity: (unsigned)cap
 {
@@ -283,6 +348,18 @@ static SEL	objSel;
   return self;
 }
 
+- (id) makeImmutableCopyOnFail: (BOOL)force
+{
+#ifndef NDEBUG
+  GSDebugAllocationRemove(isa, self);
+#endif
+  isa = [GSDictionary class];
+#ifndef NDEBUG
+  GSDebugAllocationAdd(isa, self);
+#endif
+  return self;
+}
+
 - (void) setObject: (id)anObject forKey: (id)aKey
 {
   GSIMapNode	node;
@@ -350,6 +427,7 @@ static SEL	objSel;
 
 - (void) dealloc
 {
+  GSIMapEndEnumerator(&enumerator);
   RELEASE(dictionary);
   [super dealloc];
 }
