@@ -20,10 +20,11 @@
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA.
    */ 
-#include <Foundation/NSException.h>
-#include <Foundation/NSCoder.h>
-#include <Foundation/NSDistantObject.h>
-#include <base/GSInvocation.h>
+#include "Foundation/NSException.h"
+#include "Foundation/NSCoder.h"
+#include "Foundation/NSDistantObject.h"
+#include "Foundation/NSDebug.h"
+#include "GSInvocation.h"
 #include <config.h>
 #include <objc/objc-api.h>
 #include <avcall.h>
@@ -114,7 +115,7 @@ ReturnTypeEqualsReturnType (vacallReturnTypeInfo *a, vacallReturnTypeInfo *b)
 #define GSI_MAP_RELEASE_VAL(M, X)
 #define	GSI_MAP_NOCLEAN	1
 
-#include <base/GSIMap.h>
+#include "GNUstepBase/GSIMap.h"
 
 /* This determines the number of precomputed
    callback data entries.  The list is indexed
@@ -206,20 +207,18 @@ gs_splittable (const char *type)
  * we work around it.
  */
 
-static INLINE Method_t
+static INLINE GSMethod
 gs_method_for_receiver_and_selector (id receiver, SEL sel)
 {
   if (receiver)
     {
-      if (object_is_instance (receiver))
+      if (GSObjCIsInstance(receiver))
         {
-          return class_get_instance_method (object_get_class
-                                              (receiver), sel);
+          return GSGetInstanceMethod(GSObjCClass(receiver), sel);
         }
-      else if (object_is_class (receiver))
+      else if (GSObjCIsClass(receiver))
         {
-          return class_get_class_method (object_get_meta_class
-                                           (receiver), sel);
+          return GSGetClassMethod(receiver, sel);
         }
     }
 
@@ -248,7 +247,7 @@ gs_find_best_typed_sel (SEL sel)
 {
   if (!sel_get_type (sel))
     {
-      const char *name = sel_get_name (sel);
+      const char *name = GSNameFromSelector(sel);
       
       if (name)
 	{
@@ -278,7 +277,7 @@ gs_find_by_receiver_best_typed_sel (id receiver, SEL sel)
 
   if (receiver)
     {
-      Method_t method;
+      GSMethod method;
 
       method = gs_method_for_receiver_and_selector (receiver, sel);
       /* CHECKME:  Can we assume that:
@@ -377,14 +376,13 @@ static IMP gs_objc_msg_forward (SEL sel)
   void			*forwarding_callback;
 
   /*
-   * 1. determine return type.
+   * 1. determine return type.  The compiler should have provided us with
+   * a typed selector if possible, if not we have to assume an id return.
    */
-  sel = gs_find_best_typed_sel (sel);
   sel_type = sel_get_type (sel);
 
   if (!sel_type)
     {
-      NSLog(@"Invalid selector %s (no type information)", sel_get_name (sel));
       sel_type = "@";	// Default to id return type
     }
 
@@ -426,7 +424,7 @@ static IMP gs_objc_msg_forward (SEL sel)
 	  
 	  forwarding_callback
 	    = alloc_callback (&GSInvocationCallback, ret_info);
-	
+
 	  GSIMapAddPairNoRetain (&ff_callback_map, 
 	    (GSIMapKey) (void *) ret_info,
 	    (GSIMapVal) forwarding_callback);
@@ -450,7 +448,7 @@ static IMP gs_objc_msg_forward (SEL sel)
 	&returnTypeInfo [index]);
     }
   
-  GSIMapInitWithZoneAndCapacity (&ff_callback_map, 0, 9);
+  GSIMapInitWithZoneAndCapacity (&ff_callback_map, NSDefaultMallocZone(), 9);
 
   __objc_msg_forward = gs_objc_msg_forward;
 }
@@ -469,6 +467,11 @@ static IMP gs_objc_msg_forward (SEL sel)
  */
 - (id) initWithMethodSignature: (NSMethodSignature*)aSignature
 {
+  if (aSignature == nil)
+    {
+      RELEASE(self);
+      return nil;
+    }
   _sig = RETAIN(aSignature);
   _numArgs = [aSignature numberOfArguments];
   _info = [aSignature methodInfo];
@@ -483,7 +486,7 @@ static IMP gs_objc_msg_forward (SEL sel)
 void
 GSFFCallInvokeWithTargetAndImp(NSInvocation *_inv, id anObject, IMP imp)
 {
-  int			i;
+  unsigned int		i;
   av_alist		alist;
   NSInvocation_t	*inv = (NSInvocation_t*)_inv;
   void			*retval = inv->_retval;
@@ -640,15 +643,23 @@ GSFFCallInvokeWithTargetAndImp(NSInvocation *_inv, id anObject, IMP imp)
   id		old_target;
   IMP		imp;
 
+  CLEAR_RETURN_VALUE_IF_OBJECT;
+  _validReturn = NO;
+  
   /*
    *	A message to a nil object returns nil.
    */
   if (anObject == nil)
     {
       memset(_retval, '\0', _info[0].size);	/* Clear return value */
+      if (*_info[0].type != _C_VOID)
+        {
+          _validReturn = YES;
+        }
       return;
     }
 
+    
   NSAssert(_selector != 0, @"you must set the selector before invoking");
 
   /*
@@ -667,18 +678,18 @@ GSFFCallInvokeWithTargetAndImp(NSInvocation *_inv, id anObject, IMP imp)
 
       s.self = _target;
       if (GSObjCIsInstance(_target))
-	s.class = class_get_super_class(GSObjCClass(_target));
+	s.class = GSObjCSuper(GSObjCClass(_target));
       else
-	s.class = class_get_super_class((Class)_target);
+	s.class = GSObjCSuper((Class)_target);
       imp = objc_msg_lookup_super(&s, _selector);
     }
   else
     {
-      imp = method_get_imp(object_is_instance(_target) ?
-	class_get_instance_method(
+      imp = method_get_imp(GSObjCIsInstance(_target) ?
+	GSGetInstanceMethod(
 	  ((struct objc_class*)_target)->class_pointer, _selector)
-	: class_get_class_method(
-	  ((struct objc_class*)_target)->class_pointer, _selector));
+	: GSGetClassMethod(
+	  ((struct objc_class*)_target), _selector));
       /*
        * If fast lookup failed, we may be forwarding or something ...
        */
@@ -692,6 +703,8 @@ GSFFCallInvokeWithTargetAndImp(NSInvocation *_inv, id anObject, IMP imp)
   RELEASE(old_target);
   
   GSFFCallInvokeWithTargetAndImp(self, anObject, imp);
+
+  RETAIN_RETURN_VALUE;
   _validReturn = YES;
 }
 
@@ -701,6 +714,44 @@ GSFFCallInvokeWithTargetAndImp(NSInvocation *_inv, id anObject, IMP imp)
 }
 @end
 
+/*
+ * Return YES if the selector contains protocol qualifiers.
+ */
+static BOOL
+gs_protocol_selector(const char *types)
+{
+  if (types == 0)
+    {
+      return NO;
+    }
+  while (*types != '\0')
+    {
+      if (*types == '+' || isdigit(*types))
+	{
+	  types = objc_skip_offset(types);
+	}
+      while (*types == _C_CONST || *types == _C_GCINVISIBLE)
+	{
+	  types++;
+	}
+      if (*types == _C_IN
+	|| *types == _C_INOUT
+	|| *types == _C_OUT
+	|| *types == _C_BYCOPY
+	|| *types == _C_BYREF
+	|| *types == _C_ONEWAY)
+	{
+	  return YES;
+	}
+      if (*types == '\0')
+	{
+	  return NO;
+	}
+      types = objc_skip_typespec(types);
+    }
+  return NO;
+}
+
 /*
  * Wim Oudshoorn (6 aug 2001)
  *
@@ -727,9 +778,8 @@ GSInvocationCallback (void *callback_data, va_alist args)
   NSArgumentInfo	*info;
   GSFFCallInvocation	*invocation;
   NSMethodSignature	*sig;
-  Method_t               fwdInvMethod;
+  GSMethod               fwdInvMethod;
   
-    
   typeinfo = (vacallReturnTypeInfo *) callback_data;
     
   if (typeinfo->type != __VAstruct)
@@ -750,26 +800,77 @@ GSInvocationCallback (void *callback_data, va_alist args)
   
   if (!fwdInvMethod)
     {
-      NSCAssert2 (0, @"GSFFCallInvocation: Class '%s' does not respond"
-                  @" to forwardInvocation: for '%s'",
-                  object_get_class_name (obj), sel_get_name(selector));
+      [NSException raise: NSInvalidArgumentException
+		   format: @"GSFFCallInvocation: Class '%s(%s)'"
+		           @" does not respond"
+		           @" to forwardInvocation: for '%s'",
+		   GSClassNameFromObject(obj),
+		   GSObjCIsInstance(obj) ? "instance" : "class",
+		   selector ? GSNameFromSelector(selector) : "(null)"];
     }
-       
-  selector = gs_find_by_receiver_best_typed_sel (obj, selector);
-  sig = nil;
   
-  if (sel_get_type (selector))
+  sig = nil;
+  if (gs_protocol_selector(sel_get_type(selector)) == YES)
+    {
+      /*
+       * We already have protocol information locally, so we don't need
+       * to get it from the remote system.
+       */
+      sig = [NSMethodSignature signatureWithObjCTypes: sel_get_type(selector)];
+    }
+  if (sig == nil)
+    {
+      sig = [obj methodSignatureForSelector: selector];
+    }
+
+  /*
+   * If we got a method signature from the receiving object,
+   * ensure that the selector we are using matches the types.
+   */
+  if (sig != nil)
+    {
+      const char	*receiverTypes = [sig methodType];
+      const char	*runtimeTypes = sel_get_type (selector);
+
+      if (runtimeTypes == 0 || strcmp(receiverTypes, runtimeTypes) != 0)
+	{
+	  const char	*runtimeName = GSNameFromSelector(selector);
+
+	  selector = sel_get_typed_uid (runtimeName, receiverTypes);
+	  if (selector == 0)
+	    {
+	      selector = sel_register_typed_name (runtimeName, receiverTypes);
+	    }
+	  if (runtimeTypes != 0)
+	    {
+	      /*
+	       * FIXME ... if we have a typed selector, it probably came
+	       * from the compiler, and the types of the proxied method
+	       * MUST match those that the compiler supplied on the stack
+	       * and the type it expects to retrieve from the stack.
+	       * We should therefore discriminate between signatures where
+	       * type qalifiers and sizes differ, and those where the
+	       * actual types differ.
+	       */
+	      NSWarnFLog(@"Changed type signature '%s' to '%s' for '%s'",
+		runtimeTypes, receiverTypes, runtimeName);
+	    }
+	}
+    }
+  
+  if (sig == nil && sel_get_type (selector) != 0)
     {
       sig = [NSMethodSignature signatureWithObjCTypes: sel_get_type(selector)];
     }
 
-  if (!sig)
+  if (sig == nil)
     {
-      sig = [obj methodSignatureForSelector: selector];
+      [NSException raise: NSInvalidArgumentException
+		  format: @"%s(%s) does not recognize %s",
+	 GSClassNameFromObject(obj),
+	 GSObjCIsInstance(obj) ? "instance" : "class",
+	 selector ? GSNameFromSelector(selector) : "(null)"];
     }
-  
-  NSCAssert1(sig, @"No signature for selector %@", 
-    NSStringFromSelector(selector));
     
   invocation = [[GSFFCallInvocation alloc] initWithMethodSignature: sig];
   AUTORELEASE(invocation);
@@ -859,7 +960,7 @@ GSInvocationCallback (void *callback_data, va_alist args)
    *
    *   [obj forwardInvocation: invocation];
    *
-   * but we have already the Method_t for forwardInvocation
+   * but we have already the GSMethod for forwardInvocation
    * so the line below is somewhat faster. */
   fwdInvMethod->method_imp (obj, fwdInvMethod->method_name, invocation);
 
@@ -899,7 +1000,21 @@ GSInvocationCallback (void *callback_data, va_alist args)
 	_va_return_struct(args, info[0].size, info[0].align, retval);
 	break;
       case _C_VOID:
-	va_return_void(args);
+	/* FIXME ... evil hack ... where the compiler did not know
+	 * selector types, if may have had to assume a method returning
+	 * an id, but the actual method may have returned void ...
+	 * we check for that case here, and use the fact that in the case
+	 * of a void return value, passing retval back as a voipd will
+	 * look like the method actually returned nil.
+	 */
+	if (typeinfo->type == __VAvoidp)
+	  {
+	    va_return_ptr(args, void *, *(void **)retval);
+	  }
+	else
+	  {
+	    va_return_void(args);
+	  }
 	break;
       default:
 	NSCAssert1(0, @"GSFFCallInvocation: Return Type '%s' not implemented",
@@ -914,7 +1029,7 @@ GSInvocationCallback (void *callback_data, va_alist args)
    to send over the wire */
 - (BOOL) encodeWithDistantCoder: (NSCoder*)coder passPointers: (BOOL)passp
 {
-  int		i;
+  unsigned int	i;
   BOOL		out_parameters = NO;
   const char	*type = [_sig methodType];
 
@@ -1057,3 +1172,4 @@ GSInvocationCallback (void *callback_data, va_alist args)
   return out_parameters;
 }
 
+@end
