@@ -37,35 +37,41 @@
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA.
 */
 
-#include <config.h>
+#include "config.h"
 #include <stdio.h>
 #include <string.h>
-#include <base/preface.h>
-#include <Foundation/NSString.h>
-#include <Foundation/NSArray.h>
-#include <Foundation/NSCharacterSet.h>
-#include <Foundation/NSException.h>
-#include <Foundation/NSValue.h>
-#include <Foundation/NSDictionary.h>
-#include <Foundation/NSUserDefaults.h>
-#include <Foundation/NSFileManager.h>
-#include <Foundation/NSPortCoder.h>
-#include <Foundation/NSPathUtilities.h>
-#include <Foundation/NSRange.h>
-#include <Foundation/NSException.h>
-#include <Foundation/NSData.h>
-#include <Foundation/NSBundle.h>
-#include <Foundation/NSURL.h>
-#include <Foundation/NSMapTable.h>
-#include <Foundation/NSLock.h>
-#include <Foundation/NSZone.h>
-#include <Foundation/NSDebug.h>
-#include <base/GSFormat.h>
-#include <base/GSLocale.h>
+#include "GNUstepBase/preface.h"
+#include "Foundation/NSString.h"
+#include "Foundation/NSArray.h"
+#include "Foundation/NSCharacterSet.h"
+#include "Foundation/NSException.h"
+#include "Foundation/NSValue.h"
+#include "Foundation/NSDictionary.h"
+#include "Foundation/NSUserDefaults.h"
+#include "Foundation/NSFileManager.h"
+#include "Foundation/NSPortCoder.h"
+#include "Foundation/NSPathUtilities.h"
+#include "Foundation/NSRange.h"
+#include "Foundation/NSException.h"
+#include "Foundation/NSData.h"
+#include "Foundation/NSBundle.h"
+#include "Foundation/NSURL.h"
+#include "Foundation/NSMapTable.h"
+#include "Foundation/NSLock.h"
+#include "Foundation/NSZone.h"
+#include "Foundation/NSDebug.h"
+#include "GNUstepBase/GSLocale.h"
+#include "GSFormat.h"
+
+/* We need to define _GNU_SOURCE on systems (SuSE) to get LONG_LONG_MAX.  */
+#ifndef _GNU_SOURCE
+#define _GNU_SOURCE
+#endif
+
 #include <limits.h>
 #include <string.h>		// for strstr()
 #include <sys/stat.h>
-#if HAVE_UNISTD_H
+#ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif
 #include <sys/types.h>
@@ -98,9 +104,7 @@ typedef unsigned long long uintmax_t;
 #endif
 #endif
 
-#include <base/behavior.h>
-
-#include <base/Unicode.h>
+#include "GNUstepBase/Unicode.h"
 
 struct printf_info
 {
@@ -747,12 +751,12 @@ parse_one_spec (const unichar *format, size_t posn, struct printf_spec *spec,
     {									      \
       unsigned i;							      \
       									      \
-      if (s->len+(Len) >= s->size) {					      \
-      	s->size += s->size/2 > (Len)? s->size/2: (Len);			      \
+      if (s->len+((unsigned)(Len)) >= s->size) {			      \
+      	s->size += s->size/2 > ((unsigned)(Len))? s->size/2: (unsigned)(Len); \
 	s->buf = NSZoneRealloc(s->z, s->buf, s->size*sizeof(s->buf[0]));      \
       }									      \
-      for (i=0; i < (Len); i++) s->buf[s->len++] = (String)[i];		      \
-      done += (Len);							      \
+      for (i=0; i < ((unsigned)(Len)); i++) s->buf[s->len++] = (String)[i];   \
+      done += (unsigned)(Len);						      \
     }									      \
   while (0)
 
@@ -924,7 +928,7 @@ NSDictionary *locale)
     if (grouping == (const char *) -1)
       {
 	thousands_sep = [locale objectForKey: NSThousandsSeparator];
-	if (!thousands_sep) thousands_sep = @",";
+	if (thousands_sep == nil) thousands_sep = @",";
 
 	grouping = ""; // FIXME: grouping info missing in locale?
 	if (*grouping == '\0' || *grouping == CHAR_MAX)
@@ -1085,7 +1089,9 @@ NSDictionary *locale)
 	int left = specs[nspecs_done].info.left;
 	int showsign = specs[nspecs_done].info.showsign;
 	int group = specs[nspecs_done].info.group;
+#if defined(LONG_LONG_MAX) && (LONG_MAX != LONG_LONG_MAX)
 	int is_long_double = specs[nspecs_done].info.is_long_double;
+#endif
 	int is_short = specs[nspecs_done].info.is_short;
 	int is_char = specs[nspecs_done].info.is_char;
 	int is_long = specs[nspecs_done].info.is_long;
@@ -1127,10 +1133,12 @@ NSDictionary *locale)
 	  }
 
 	/* Maybe the buffer is too small.  */
-	if (MAX (prec, width) + 32 > sizeof (work_buffer) / sizeof (unichar))
-	  workend = ((unichar *) alloca ((MAX (prec, width) + 32)
-					* sizeof (unichar))
-		     + (MAX (prec, width) + 32));
+	if ((unsigned)(MAX (prec, width) + 32)
+	  > sizeof (work_buffer) / sizeof (unichar))
+	  {
+	    workend = ((unichar *) alloca ((MAX (prec, width) + 32)
+	      * sizeof (unichar)) + (MAX (prec, width) + 32));
+	  }
 
 	/* Process format specifiers.  */
 	while (1)
@@ -1403,6 +1411,7 @@ NSDictionary *locale)
 	NSString	*decimal_sep;
 
 	decimal_sep = [locale objectForKey: NSDecimalSeparator];
+	if (decimal_sep == nil) decimal_sep = @".";
 
 	bp = buf1;
 
@@ -1523,6 +1532,7 @@ NSDictionary *locale)
 	NSString	*decimal_sep;
 
 	decimal_sep = [locale objectForKey: NSDecimalSeparator];
+	if (decimal_sep == nil) decimal_sep = @".";
 
 	bp = buf1;
 
@@ -1647,14 +1657,8 @@ NSDictionary *locale)
 	  }
 	else
 	  {
-	    unichar nil_name[] = {'(','n','i','l',')','\0'};
-	    /* Write "(nil)" for a nil pointer.  */
-	    string = nil_name;
-	    /* Make sure the full string "(nil)" is printed.  */
-	    if (prec < 5)
-	      prec = 5;
-	    string_malloced = 0;
-	    goto LABEL (print_uni_string);
+	    string = NULL;
+	    goto LABEL (print_string);
 	  }
       }
       /* NOTREACHED */
@@ -1708,7 +1712,7 @@ NSDictionary *locale)
 
     LABEL (form_string):
       {
-	size_t len;
+	size_t len = 0;
 
 	/* The string argument could in fact be `char *' or `wchar_t *'.
 	   But this should not make a difference here.  */
@@ -1738,12 +1742,28 @@ NSDictionary *locale)
 	    /* This is complicated.  We have to transform the multibyte
 	       string into a unicode string.  */
 	    const char		*str = (const char*)string;
-	    unsigned		slen = strlen(str);
+	    unsigned		slen;
 	    NSStringEncoding	enc = GetDefEncoding();
 
-	    len = prec != -1 ? prec : slen;
-	    if (len > slen)
-	      len = slen;
+	    if (prec != -1)
+	      {
+		len = (unsigned)prec;
+		/*
+		 * If the actual length is less than the precision,
+		 * we use the actual length.
+		 */
+		for (slen = 0; slen < len; slen++)
+		  {
+		    if (str[slen] == 0)
+		      {
+			len = slen;
+		      }
+		  }
+	      }
+	    else
+	      {
+		len = strlen(str);
+	      }
 
 	    /* Allocate dynamically an array which definitely is long
 	       enough for the wide character version.  */
@@ -1786,18 +1806,18 @@ NSDictionary *locale)
 	    while (prc--) *sp = *wsp;
 	  }
 
-	LABEL (print_uni_string):
 	if ((width -= len) <= 0)
 	  {
 	    outstring (string, len);
-	    break;
 	  }
-
-	if (!left)
-	  PAD (' ');
-	outstring (string, len);
-	if (left)
-	  PAD (' ');
+	else
+	  {
+	    if (!left)
+	      PAD (' ');
+	    outstring (string, len);
+	    if (left)
+	      PAD (' ');
+	  }
 	if (string_malloced)
 	  NSZoneFree(s->z, string);
       }
@@ -1826,7 +1846,7 @@ NSDictionary *locale)
 	    NSRange r;
 
 	    len = [dsc length];
-	    if (prec >= 0 && prec < len) len = prec;
+	    if (prec >= 0 && prec < (int)len) len = prec;
 
 	    /* Allocate dynamically an array which definitely is long
 	       enough for the wide character version.  */
@@ -1845,14 +1865,15 @@ NSDictionary *locale)
 	if ((width -= len) < 0)
 	  {
 	    outstring (string, len);
-	    break;
 	  }
-
-	if (!left)
-	  PAD (' ');
-	outstring (string, len);
-	if (left)
-	  PAD (' ');
+	else
+	  {
+	    if (!left)
+	      PAD (' ');
+	    outstring (string, len);
+	    if (left)
+	      PAD (' ');
+	  }
 	if (string_malloced)
 	  NSZoneFree(s->z, string);
       }
@@ -2003,3 +2024,4 @@ group_number (unichar *w, unichar *rear_ptr, const char *grouping,
     }
   return w;
 }
+
