@@ -1,9 +1,9 @@
 /** GSHTTPURLHandle.m - Class GSHTTPURLHandle
    Copyright (C) 2000 Free Software Foundation, Inc.
    
-   Written by: 	Mark Allison <mark@brainstorm.co.uk>
-   Integrated:	Richard Frith-Macdonald <rfm@gnu.org>
-   Date:	November 2000 		
+   Written by: 		Mark Allison <mark@brainstorm.co.uk>
+   Integrated by:	Richard Frith-Macdonald <rfm@gnu.org>
+   Date:		November 2000 		
    
    This file is part of the GNUstep Library.
    
@@ -23,22 +23,23 @@
 */
 
 #include "config.h"
-#include <Foundation/NSArray.h>
-#include <Foundation/NSString.h>
-#include <Foundation/NSException.h>
-#include <Foundation/NSValue.h>
-#include <Foundation/NSData.h>
-#include <Foundation/NSURL.h>
-#include <Foundation/NSURLHandle.h>
-#include <Foundation/NSNotification.h>
-#include <Foundation/NSRunLoop.h>
-#include <Foundation/NSByteOrder.h>
-#include <Foundation/NSLock.h>
-#include <Foundation/NSFileHandle.h>
-#include <Foundation/NSDebug.h>
-#include <Foundation/GSMime.h>
+#include "Foundation/NSArray.h"
+#include "Foundation/NSString.h"
+#include "Foundation/NSException.h"
+#include "Foundation/NSValue.h"
+#include "Foundation/NSData.h"
+#include "Foundation/NSURL.h"
+#include "Foundation/NSURLHandle.h"
+#include "Foundation/NSNotification.h"
+#include "Foundation/NSRunLoop.h"
+#include "Foundation/NSByteOrder.h"
+#include "Foundation/NSLock.h"
+#include "Foundation/NSFileHandle.h"
+#include "Foundation/NSDebug.h"
+#include "Foundation/NSProcessInfo.h"
+#include "GNUstepBase/GSMime.h"
 #include <string.h>
-#if HAVE_UNISTD_H
+#ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif
 #include <sys/file.h>
@@ -70,7 +71,7 @@ char emp[64] = {
   NSMutableDictionary   *wProperties;
   NSData		*wData;
   NSMutableDictionary   *request;
-  unsigned int          contentLength;
+  unsigned int          bodyPos;
   enum {
     idle,
     connecting,
@@ -78,11 +79,89 @@ char emp[64] = {
     reading,
   } connectionState;
 }
-- (NSString*) encodebase64: (NSString*) input;
 - (void) setDebug: (BOOL)flag;
 @end
 
-
+/**
+ * <p>
+ *   This is a <em>PRIVATE</em> subclass of NSURLHandle.
+ *   It is documented here in order to give you information about the
+ *   default behavior of an NSURLHandle created to deal with a URL
+ *   that has either the <code>http</code> or <code>https</code> scheme.
+ *   The name and/or other implementation details of this class
+ *   may be changed at any time.
+ * </p>
+ * <p>
+ *   A GSHTTPURLHandle instance is used to manage connections to
+ *   <code>http</code> and <code>https</code> URLs.
+ *    Secure connections are handled automatically
+ *   (using openSSL) for URLs with the scheme <code>https</code>.
+ *   Connection via proxy server is supported, as is proxy tunneling
+ *   for secure connections.  Basic parsing of <code>http</code>
+ *   headers is performed to extract <code>http</code> status
+ *   information, cookies etc.  Cookies are
+ *   retained and automatically sent during subsequent requests where
+ *   the cookie is valid.
+ * </p>
+ * <p>
+ *   Header information from the current page may be obtained using
+ *   -propertyForKey and -propertyForKeyIfAvailable.  <code>HTTP</code>
+ *   status information can be retrieved as by calling either of these 
+ *   methods specifying one of the following keys:
+ * </p>
+ * <list>
+ *   <item>
+ *     NSHTTPPropertyStatusCodeKey - numeric status code
+ *   </item>
+ *   <item>
+ *     NSHTTPPropertyStatusReasonKey - text describing status
+ *   </item>
+ *   <item>
+ *     NSHTTPPropertyServerHTTPVersionKey - <code>http</code>
+ *     version supported by remote server
+ *   </item>
+ * </list>
+ * <p>
+ *   According to MacOS-X headers, the following should also
+ *   be supported, but currently are not:
+ * </p>
+ * <list>
+ *   <item>NSHTTPPropertyRedirectionHeadersKey</item>
+ *   <item>NSHTTPPropertyErrorPageDataKey</item>
+ * </list>
+ * <p>
+ *   The omission of these headers is not viewed as important at
+ *   present, since the MacOS-X public beta implementation doesn't
+ *   work either.
+ * </p>
+ * <p>
+ *   Other calls to -propertyForKey and -propertyForKeyIfAvailable may
+ *   be made specifying a <code>http</code> header field name.
+ *   For example specifying a key name of &quot;Content-Length&quot;
+ *   would return the value of the &quot;Content-Length&quot; header
+ *   field.
+ * </p>
+ * <p>
+ *   [GSHTTPURLHandle-writeProperty:forKey:]
+ *   can be used to specify the parameters
+ *   for the <code>http</code> request.  The default request uses the
+ *   &quot;GET&quot; method when fetching a page, and the
+ *   &quot;POST&quot; method when using -writeData:.
+ *   This can be over-ridden by calling -writeProperty:forKey: with
+ *   the key name &quot;GSHTTPPropertyMethodKey&quot; and specifying an 
+ *   alternative method (i.e &quot;PUT&quot;).
+ * </p>
+ * <p>
+ *   A Proxy may be specified by calling -writeProperty:forKey:
+ *   with the keys &quot;GSHTTPPropertyProxyHostKey&quot; and
+ *   &quot;GSHTTPPropertyProxyPortKey&quot; to set the host and port
+ *   of the proxy server respectively.  The GSHTTPPropertyProxyHostKey
+ *   property can be set to either the IP address or the hostname of
+ *   the proxy server.  If an attempt is made to load a page via a
+ *   secure connection when a proxy is specified, GSHTTPURLHandle will
+ *   attempt to open an SSL Tunnel through the proxy.
+ * </p>
+ */
 @implementation GSHTTPURLHandle
 
 static NSMutableDictionary	*urlCache = nil;
@@ -155,7 +234,8 @@ static void debugWrite(NSData *data)
       urlCache = [NSMutableDictionary new];
       urlLock = [NSLock new];
       debugLock = [NSLock new];
-      sprintf(debugFile, "/tmp/GSHTTP.%d", getpid());
+      sprintf(debugFile, "/tmp/GSHTTP.%d",
+	[[NSProcessInfo processInfo] processIdentifier]);
 #ifndef __MINGW__
       sslClass = [NSFileHandle sslClass];
 #endif
@@ -216,43 +296,108 @@ static void debugWrite(NSData *data)
   NSNotificationCenter	*nc = [NSNotificationCenter defaultCenter];
   NSDictionary		*dict = [not userInfo];
   NSData		*d;
+  NSRange		r;
 
   d = [dict objectForKey: NSFileHandleNotificationDataItem];
   if (debug == YES) debugRead(d);
 
-  [parser parse: d];
-  if ([parser isComplete] == YES)
+  if ([parser parse: d] == NO)
     {
-      NSDictionary	*info;
-      NSString		*val;
-
-      connectionState = idle;
-      [nc removeObserver: self
-	            name: NSFileHandleReadCompletionNotification
-                  object: sock];
-      [sock closeFile];
-      DESTROY(sock);
-      /*
-       * Retrieve essential keys from document
-       */
-      info = [document headerNamed: @"http"];
-      val = [info objectForKey: NSHTTPPropertyServerHTTPVersionKey];
-      if (val != nil)
-	[pageInfo setObject: val forKey: NSHTTPPropertyServerHTTPVersionKey];
-      val = [info objectForKey: NSHTTPPropertyStatusCodeKey];
-      if (val != nil)
-	[pageInfo setObject: val forKey: NSHTTPPropertyStatusCodeKey];
-      val = [info objectForKey: NSHTTPPropertyStatusReasonKey];
-      if (val != nil)
-	[pageInfo setObject: val forKey: NSHTTPPropertyStatusReasonKey];
-      /*
-       * Tell superclass that we have successfully loaded the data.
-       */
-      [self didLoadBytes: [parser data] loadComplete: YES];
+      if (debug == YES)
+	{
+	  NSLog(@"HTTP parse failure - %@", parser);
+	}
+      [self endLoadInBackground];
+      [self backgroundLoadDidFailWithReason: @"Response parse failed"];
     }
   else
     {
-      [sock readInBackgroundAndNotify];
+      BOOL	complete = [parser isComplete];
+
+      if (complete == NO && [parser isInHeaders] == NO)
+	{
+	  NSString	*enc;
+	  NSString	*len;
+	  int		ver;
+
+	  ver = [[[document headerNamed: @"http"]
+	    objectForKey: NSHTTPPropertyServerHTTPVersionKey] intValue];
+	  len = [[document headerNamed: @"content-length"] value];
+	  enc = [[document headerNamed: @"content-transfer-encoding"] value];
+	  if (enc == nil)
+	    {
+	      enc = [[document headerNamed: @"transfer-encoding"] value];
+	    }
+
+	  if ([enc isEqualToString: @"chunked"] == YES)	
+	    {
+	      complete = NO;	// Read chunked body data
+	    }
+	  else if (ver >= 1 && [len intValue] == 0)
+	    {
+	      complete = YES;	// No content
+	    }
+	  else
+	    {
+	      complete = NO;	// No
+	    }
+	}
+      if (complete == YES)
+	{
+	  GSMimeHeader	*info;
+	  NSString		*val;
+
+	  connectionState = idle;
+	  [nc removeObserver: self
+			name: NSFileHandleReadCompletionNotification
+		      object: sock];
+	  [sock closeFile];
+	  DESTROY(sock);
+
+	  /*
+	   * Retrieve essential keys from document
+	   */
+	  info = [document headerNamed: @"http"];
+	  val = [info objectForKey: NSHTTPPropertyServerHTTPVersionKey];
+	  if (val != nil)
+	    {
+	      [pageInfo setObject: val
+			   forKey: NSHTTPPropertyServerHTTPVersionKey];
+	    }
+	  val = [info objectForKey: NSHTTPPropertyStatusCodeKey];
+	  if (val != nil)
+	    {
+	      [pageInfo setObject: val forKey: NSHTTPPropertyStatusCodeKey];
+	    }
+	  val = [info objectForKey: NSHTTPPropertyStatusReasonKey];
+	  if (val != nil)
+	    {
+	      [pageInfo setObject: val forKey: NSHTTPPropertyStatusReasonKey];
+	    }
+	  /*
+	   * Tell superclass that we have successfully loaded the data.
+	   */
+	  d = [parser data];
+	  r = NSMakeRange(bodyPos, [d length] - bodyPos);
+	  bodyPos = 0;
+	  [self didLoadBytes: [d subdataWithRange: r]
+		loadComplete: YES];
+	}
+      else
+	{
+	  /*
+	   * Report partial data if possible.
+	   */
+	  if ([parser isInBody])
+	    {
+	      d = [parser data];
+	      r = NSMakeRange(bodyPos, [d length] - bodyPos);
+	      bodyPos = [d length];
+	      [self didLoadBytes: [d subdataWithRange: r]
+		    loadComplete: NO];
+	    }
+	  [sock readInBackgroundAndNotify];
+	}
     }
 }
 
@@ -273,11 +418,11 @@ static void debugWrite(NSData *data)
   [p parse: dat];
   if ([p isInBody] == YES || [d length] == 0)
     {
-      NSDictionary	*info;
+      GSMimeHeader	*info;
       NSString		*val;
 
       [p parse: nil];
-      info = [[p document] headerNamed: @"http"];
+      info = [[p mimeDocument] headerNamed: @"http"];
       val = [info objectForKey: NSHTTPPropertyServerHTTPVersionKey];
       if (val != nil)
 	[pageInfo setObject: val forKey: NSHTTPPropertyServerHTTPVersionKey];
@@ -319,18 +464,27 @@ static void debugWrite(NSData *data)
   RELEASE(document);
   RELEASE(parser);
   parser = [GSMimeParser new];
-  document = RETAIN([parser document]);
+  document = RETAIN([parser mimeDocument]);
   [self beginLoadInBackground];
   if (sock != nil)
     {
       [sock closeFile];
       DESTROY(sock);
     }
-  contentLength = 0;
   if ([[request objectForKey: GSHTTPPropertyProxyHostKey] length] == 0)
     {
+      NSNumber	*p;
+
       host = [url host];
-      port = [url scheme];
+      p = [url port];
+      if (p != nil)
+	{
+	  port = [NSString stringWithFormat: @"%u", [p unsignedIntValue]];
+	}
+      else
+	{
+	  port = [url scheme];
+	}
       if ([[url scheme] isEqualToString: @"https"])
 	{
 	  if (sslClass == 0)
@@ -434,6 +588,13 @@ static void debugWrite(NSData *data)
   NSString              *key;
   NSMutableData		*buf;
   NSString		*method;
+  NSString		*path;
+
+  path = [[url path] stringByTrimmingSpaces];
+  if ([path length] == 0)
+    {
+      path = @"/";
+    } 
 
   /*
    * See if the connection attempt caused an error.
@@ -556,23 +717,31 @@ static void debugWrite(NSData *data)
   if ([[request objectForKey: GSHTTPPropertyProxyHostKey] length] > 0
     && [[url scheme] isEqualToString: @"https"] == NO)
     {
-      s = [[NSMutableString alloc] initWithFormat: @"%@ http://%@%@", 
-	method, [url host], [url path]];
-      if ([[url query] length] > 0)
+      if ([url port] == nil)
 	{
-	  [s appendFormat: @"?%@", [url query]];
+	  s = [[NSMutableString alloc] initWithFormat: @"%@ http://%@%@", 
+	    method, [url host], path];
+	}
+      else
+	{
+	  s = [[NSMutableString alloc] initWithFormat: @"%@ http://%@:%@%@", 
+	    method, [url host], [url port], path];
 	}
-      [s appendFormat: @" HTTP/%@\r\n", httpVersion];
     }
   else    // no proxy
     {
       s = [[NSMutableString alloc] initWithFormat: @"%@ %@", 
-	method, [url path]];
-      if ([[url query] length] > 0)
-	{
-	  [s appendFormat: @"?%@", [url query]];
-	}
-      [s appendFormat: @" HTTP/%@\nHost: %@\r\n", httpVersion, [url host]];
+	method, path];
+    }
+  if ([[url query] length] > 0)
+    {
+      [s appendFormat: @"?%@", [url query]];
+    }
+  [s appendFormat: @" HTTP/%@\r\n", httpVersion];
+
+  if ([wProperties objectForKey: @"host"] == nil)
+    {
+      [wProperties setObject: [url host] forKey: @"host"];
     }
 
   if ([wData length] > 0)
@@ -588,31 +757,35 @@ static void debugWrite(NSData *data)
 			  forKey: @"content-type"];
 	}
     }
-  if ([url user] != nil)
+  if ([wProperties objectForKey: @"authorisation"] == nil)
     {
-      NSString	*auth;
-
-      if ([[url password] length] > 0)
-	{ 
-	  auth = [NSString stringWithFormat: @"%@:%@", 
-	    [url user], [url password]];
-	}
-      else
+      if ([url user] != nil)
 	{
-	  auth = [NSString stringWithFormat: @"%@", [url user]];
+	  NSString	*auth;
+
+	  if ([[url password] length] > 0)
+	    { 
+	      auth = [NSString stringWithFormat: @"%@:%@", 
+		[url user], [url password]];
+	    }
+	  else
+	    {
+	      auth = [NSString stringWithFormat: @"%@", [url user]];
+	    }
+	  auth = [NSString stringWithFormat: @"Basic %@",
+	    [GSMimeDocument encodeBase64String: auth]];
+	  [wProperties setObject: auth
+			  forKey: @"authorization"];
 	}
-      auth = [NSString stringWithFormat: @"Basic %@",
-	[self encodebase64: auth]];
-      [wProperties setObject: auth
-		      forKey: @"Authorization"];
     }
+
   wpEnumerator = [wProperties keyEnumerator];
   while ((key = [wpEnumerator nextObject]))
     {
       [s appendFormat: @"%@: %@\r\n", key, [wProperties objectForKey: key]];
     }
   [wProperties removeAllObjects];
-  [s appendString: @"\n"];
+  [s appendString: @"\r\n"];
   buf = [[s dataUsingEncoding: NSASCIIStringEncoding] mutableCopy];
 
   /*
@@ -683,6 +856,7 @@ static void debugWrite(NSData *data)
 	}
       else
 	{
+	  bodyPos = 0;
 	  [nc addObserver: self
 	         selector: @selector(bgdRead:)
 		     name: NSFileHandleReadCompletionNotification
@@ -693,6 +867,30 @@ static void debugWrite(NSData *data)
     }
 }
 
+/**
+ *  If necessary, this method calls -loadInForeground to send a
+ *  request to the webserver, and get a page back.  It then returns
+ *  the property for the specified key -
+ * <list>
+ *   <item>
+ *     NSHTTPPropertyStatusCodeKey - numeric status code returned
+ *     by the last request.
+ *   </item>
+ *   <item>
+ *     NSHTTPPropertyStatusReasonKey - text describing status of
+ *     the last request
+ *   </item>
+ *   <item>
+ *     NSHTTPPropertyServerHTTPVersionKey - <code>http</code>
+ *     version supported by remote server
+ *   </item>
+ *   <item>
+ *     Other keys are taken to be the names of <code>http</code>
+ *     headers and the corresponding header value (or nil if there
+ *     is none) is returned.
+ *   </item>
+ * </list>
+ */
 - (id) propertyForKey: (NSString*) propertyKey
 {
   if (document == nil)
@@ -715,17 +913,19 @@ static void debugWrite(NSData *data)
 	}
       else if ([array count] == 1)
 	{
-	  result = [[array objectAtIndex: 0] objectForKey: @"BaseValue"];
+	  GSMimeHeader	*hdr = [array objectAtIndex: 0];
+
+	  result = [hdr value];
 	}
       else
 	{
 	  NSEnumerator	*enumerator = [array objectEnumerator];
-	  NSDictionary	*val;
+	  GSMimeHeader	*val;
 
 	  result = [NSMutableArray arrayWithCapacity: [array count]];
 	  while ((val = [enumerator nextObject]) != nil)
 	    {
-	      [result addObject: [val objectForKey: @"BaseValue"]];
+	      [result addObject: [val value]];
 	    }
 	}
     }
@@ -737,12 +937,44 @@ static void debugWrite(NSData *data)
   debug = flag;
 }
 
+/**
+ * Writes the specified data as the body of an <code>http</code>
+ * or <code>https</code> request to the web server.
+ * Returns YES on success,
+ * NO on failure.  By default, this method performs a POST operation.
+ * On completion, the resource data for this handle is set to the
+ * page returned by the request.
+ */
 - (BOOL) writeData: (NSData*)d
 {
   ASSIGN(wData, d);
   return YES;
 }
 
+/**
+ * Sets a property to be used in the next request made by this handle.
+ * The property is set as a header in the next request, unless it is
+ * one of the following -
+ * <list>
+ *   <item>
+ *     GSHTTPPropertyBodyKey - set an NSData item to be sent to
+ *     the server as the body of the request.
+ *   </item>
+ *   <item>
+ *     GSHTTPPropertyMethodKey - override the default method of
+ *     the request (eg. &quot;PUT&quot;).
+ *   </item>
+ *   <item>
+ *     GSHTTPPropertyProxyHostKey - specify the name or IP address
+ *     of a host to proxy through.
+ *   </item>
+ *   <item>
+ *     GSHTTPPropertyProxyPortKey - specify the port number to 
+ *     connect to on the proxy host.  If not give, this defaults
+ *     to 8080 for <code>http</code> and 4430 for <code>https</code>.
+ *   </item>
+ * </list>
+ */
 - (BOOL) writeProperty: (id) property forKey: (NSString*) propertyKey
 {
   if (propertyKey == nil || [propertyKey isKindOfClass: [NSString class]] == NO)
@@ -776,41 +1008,5 @@ static void debugWrite(NSData *data)
   return YES;
 }
 
-- (NSString*) encodebase64: (NSString*) input
-{
-   char			*str = calloc([input length], sizeof(char));
-   char			*sptr = str;
-   NSMutableString	*nstr = [NSMutableString string];
-   int i;
-
-   strcpy(str, [input cString]);
- 
-   for (i=0; i < [input length]; i += 3) 
-     {
-       [nstr appendFormat: @"%c", emp[*sptr >> 2]];
-       [nstr appendFormat: @"%c", 
-	 emp[((*sptr << 4) & 060) | ((sptr[1] >> 4) & 017)]];
-       [nstr appendFormat: @"%c", 
-	 emp[((sptr[1] << 2) & 074) | ((sptr[2] >> 6) & 03)]];
-       [nstr appendFormat: @"%c", emp[sptr[2] & 077]];
-       sptr += 3;
-     }
- 
-   /* If len was not a multiple of 3, then we have encoded too
-    * many characters.  Adjust appropriately.
-    */
-   if (i == [input length] + 1) 
-     {
-       /* There were only 2 bytes in that last group */
-       [nstr deleteCharactersInRange: NSMakeRange([nstr length] - 1, 1)];
-     } 
-   else if (i == [input length] + 2) 
-     {
-       /* There was only 1 byte in that last group */
-       [nstr deleteCharactersInRange: NSMakeRange([nstr length] - 2, 2)];
-     }
-   free (str);
-   return (nstr);
-}
 @end
 
