@@ -21,11 +21,11 @@
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA.
 */
-#include <config.h>
-#include <base/GSLocale.h>
-#include <Foundation/NSDictionary.h>
-#include <Foundation/NSArray.h>
-#include <Foundation/NSLock.h>
+#include "config.h"
+#include "GNUstepBase/GSLocale.h"
+#include "Foundation/NSDictionary.h"
+#include "Foundation/NSArray.h"
+#include "Foundation/NSLock.h"
 
 #ifdef HAVE_LOCALE_H
 
@@ -33,8 +33,8 @@
 #ifdef HAVE_LANGINFO_H
 #include <langinfo.h>
 #endif
-#include <Foundation/NSUserDefaults.h>
-#include <Foundation/NSBundle.h>
+#include "Foundation/NSUserDefaults.h"
+#include "Foundation/NSBundle.h"
 
 /*
  * Function called by [NSObject +initialize] to setup locale information
@@ -43,19 +43,19 @@
  * make use of locale information.
  */
 const char*
-GSSetLocaleC(const char *loc)
+GSSetLocaleC(int category, const char *loc)
 {
-  return setlocale(LC_ALL, loc);
+  return setlocale(category, loc);
 }
 
 /* Set the locale for libc functions from the supplied string or from
    the environment if not specified. This function should be called
    as soon as possible after the start of the program. Passing
-   @"" will set the locale from the environment variables LC_ALL or LANG (or
+   @"" will set the locale from the environment variables LC_CTYPE or LANG (or
    whatever is specified by setlocale) Passing nil will just return the
    current locale. */
 NSString *
-GSSetLocale(NSString *locale)
+GSSetLocale(int category, NSString *locale)
 {
   const char *clocale;
 
@@ -64,7 +64,7 @@ GSSetLocale(NSString *locale)
     {
       clocale = [locale cString];
     }
-  clocale = GSSetLocaleC(clocale);
+  clocale = GSSetLocaleC(category, clocale);
 
   if (clocale == NULL || strcmp(clocale, "C") == 0 
     || strcmp(clocale, "POSIX") == 0) 
@@ -184,7 +184,11 @@ GSDomainFromDefaultLocale(void)
   
   /* FIXME: Get currency format from localeconv */
 
-  str1 = GSSetLocale(nil);
+#ifdef	LC_MESSAGES
+  str1 = GSSetLocale(LC_MESSAGES, nil);
+#else
+  str1 = nil;
+#endif
   if (str1 != nil)
     {
       [dict setObject: str1 forKey: NSLocale];
@@ -196,7 +200,14 @@ GSDomainFromDefaultLocale(void)
     }
 
   [gnustep_global_lock lock];
-  saved = [dict mutableCopy];
+  /*
+   * Another thread might have been faster in setting the static variable.
+   * If so, we just drop our dict.
+   */
+  if (saved == nil)
+    {
+      saved = [dict mutableCopy];
+    }
   [gnustep_global_lock unlock];
   return saved;
 #else /* HAVE_LANGINFO_H */
@@ -204,18 +215,35 @@ GSDomainFromDefaultLocale(void)
 #endif
 }
 
+#else /* HAVE_LOCALE_H */
+NSString *
+GSSetLocale(int category, NSString *locale)
+{
+  return nil;
+}
+
+NSDictionary *
+GSDomainFromDefaultLocale(void)
+{
+  return nil;
+}
+
+#endif /* !HAVE_LOCALE_H */
+
 NSString *
 GSLanguageFromLocale(NSString *locale)
 {
   NSString	*language = nil;
   NSString	*aliases = nil;
+  NSBundle      *gbundle;
 
   if (locale == nil || [locale isEqual: @"C"] || [locale isEqual: @"POSIX"])
     return @"English";
 
-  aliases = [NSBundle pathForGNUstepResource: @"Locale"
-		                      ofType: @"aliases"
-		                 inDirectory: @"Resources/Languages"];  
+  gbundle = [NSBundle bundleForLibrary: @"gnustep-base"];
+  aliases = [gbundle pathForResource: @"Locale"
+		              ofType: @"aliases"
+		         inDirectory: @"Languages"];  
   if (aliases != nil)
     {
       NSDictionary	*dict;
@@ -236,26 +264,3 @@ GSLanguageFromLocale(NSString *locale)
       
   return language;
 }
-
-#else /* HAVE_LOCALE_H */
-NSString *
-GSSetLocale(NSString *locale)
-{
-  return nil;
-}
-
-NSDictionary *
-GSDomainFromDefaultLocale(void)
-{
-  return nil;
-}
-
-NSString *
-GSLanguageFromLocale(NSString *locale)
-{
-  return nil;
-}
-
-#endif /* !HAVE_LOCALE_H */
-
-
