@@ -21,31 +21,34 @@
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA.
    */ 
 
-#include <config.h>
-#include <base/preface.h>
-#include <Foundation/NSArray.h>
-#include <Foundation/NSNotification.h>
-#include <Foundation/NSException.h>
-#include <Foundation/NSRunLoop.h>
-#include <Foundation/NSByteOrder.h>
-#include <Foundation/NSData.h>
-#include <Foundation/NSDate.h>
-#include <Foundation/NSHost.h>
-#include <Foundation/NSMapTable.h>
-#include <Foundation/NSPortMessage.h>
-#include <Foundation/NSPortNameServer.h>
-#include <Foundation/NSLock.h>
-#include <Foundation/NSHost.h>
-#include <Foundation/NSThread.h>
-#include <Foundation/NSDebug.h>
+#include "config.h"
+#include "GNUstepBase/preface.h"
+#include "GNUstepBase/GSLock.h"
+#include "Foundation/NSArray.h"
+#include "Foundation/NSNotification.h"
+#include "Foundation/NSException.h"
+#include "Foundation/NSRunLoop.h"
+#include "Foundation/NSByteOrder.h"
+#include "Foundation/NSData.h"
+#include "Foundation/NSDate.h"
+#include "Foundation/NSHost.h"
+#include "Foundation/NSMapTable.h"
+#include "Foundation/NSPortMessage.h"
+#include "Foundation/NSPortNameServer.h"
+#include "Foundation/NSLock.h"
+#include "Foundation/NSHost.h"
+#include "Foundation/NSThread.h"
+#include "Foundation/NSConnection.h"
+#include "Foundation/NSDebug.h"
 #include <stdio.h>
 #include <stdlib.h>
-#include <signal.h>
-#if HAVE_UNISTD_H
+#ifdef HAVE_UNISTD_H
 #include <unistd.h>		/* for gethostname() */
 #endif
 
-#ifndef __MINGW__
+#ifdef __MINGW__
+#define close closesocket
+#else
 #include <sys/param.h>		/* for MAXHOSTNAMELEN */
 #include <sys/types.h>
 #include <netinet/in.h>
@@ -65,7 +68,6 @@
 #include <sys/time.h>
 #include <sys/resource.h>
 #include <netdb.h>
-#include <signal.h>
 #include <sys/socket.h>
 #include <sys/file.h>
 /*
@@ -89,25 +91,24 @@
 #if	defined(__svr4__)
 #include <sys/stropts.h>
 #endif
+
+#define	SOCKET	int
+#define	SOCKET_ERROR	-1
+#define	INVALID_SOCKET	-1
+
 #endif /* !__MINGW__ */
 
-#ifdef __MINGW__
-#define close closesocket
-#endif
-
-static	BOOL	multi_threaded = NO;
-
 /*
  * Largest chunk of data possible in DO
  */
 static gsu32	maxDataLength = 10 * 1024 * 1024;
 
 #if 0
-#define	DO_LOCK(X) {NSDebugMLLog(@"GSTcpHandle",@"lock %@",X); [X lock];}
-#define	DO_UNLOCK(X) {NSDebugMLLog(@"GSTcpHandle",@"unlock %@",X); [X unlock];}
+#define	M_LOCK(X) {NSDebugMLLog(@"GSTcpHandle",@"lock %@",X); [X lock];}
+#define	M_UNLOCK(X) {NSDebugMLLog(@"GSTcpHandle",@"unlock %@",X); [X unlock];}
 #else
-#define	DO_LOCK(X) {[X lock];}
-#define	DO_UNLOCK(X) {[X unlock];}
+#define	M_LOCK(X) {[X lock];}
+#define	M_UNLOCK(X) {[X unlock];}
 #endif
 
 #define	GS_CONNECTION_MSG	0
@@ -184,7 +185,7 @@ typedef enum {
 
 @interface GSTcpHandle : NSObject <GCFinalization, RunLoopEvents>
 {
-  int			desc;		/* File descriptor for I/O.	*/
+  SOCKET		desc;		/* File descriptor for I/O.	*/
   unsigned		wItem;		/* Index of item being written.	*/
   NSMutableData		*wData;		/* Data object being written.	*/
   unsigned		wLength;	/* Ammount written so far.	*/
@@ -197,7 +198,7 @@ typedef enum {
   gsu32			rId;		/* Id of incoming message.	*/
   unsigned		nItems;		/* Number of items to be read.	*/
   GSHandleState		state;		/* State of the handle.		*/
-  int			addrNum;	/* Address number within host.	*/
+  unsigned int		addrNum;	/* Address number within host.	*/
 @public
   NSRecursiveLock	*myLock;	/* Lock for this handle.	*/
   BOOL			caller;		/* Did we connect to other end?	*/
@@ -208,10 +209,9 @@ typedef enum {
   NSString		*defaultAddress;
 }
 
-+ (GSTcpHandle*) handleWithDescriptor: (int)d;
++ (GSTcpHandle*) handleWithDescriptor: (SOCKET)d;
 - (BOOL) connectToPort: (GSTcpPort*)aPort beforeDate: (NSDate*)when;
 - (int) descriptor;
-- (void) dispatch;
 - (void) invalidate;
 - (BOOL) isValid;
 - (void) receivedEvent: (void*)data
@@ -234,7 +234,7 @@ typedef enum {
   NSHost		*host;		/* OpenStep host for this port.	*/
   NSString		*address;	/* Forced internet address.	*/
   gsu16			portNum;	/* TCP port in host byte order.	*/
-  int			listener;	/* Descriptor to listen on.	*/
+  SOCKET		listener;
   NSMapTable		*handles;	/* Handles indexed by socket.	*/
 }
 
@@ -247,7 +247,7 @@ typedef enum {
 
 - (void) addHandle: (GSTcpHandle*)handle forSend: (BOOL)send;
 - (NSString*) address;
-- (void) getFds: (int*)fds count: (int*)count;
+- (void) getFds: (SOCKET*)fds count: (int*)count;
 - (GSTcpHandle*) handleForPort: (GSTcpPort*)recvPort beforeDate: (NSDate*)when;
 - (void) handlePortMessage: (NSPortMessage*)m;
 - (NSHost*) host;
@@ -283,6 +283,12 @@ decodePort(NSData *data, NSString *defaultAddress)
   length = GSSwapBigI32ToHost(pih->length);
   pi = (GSPortInfo*)&pih[1];
   pnum = GSSwapBigI16ToHost(pi->num);
+  if (strncmp(pi->addr, "VER", 3) == 0)
+    {
+      NSLog(@"Remote version of GNUstep at %s:%d is more recent than this one",
+	pi->addr, pnum); 
+      return nil;
+    }
   addr = [NSString stringWithCString: pi->addr];
 
   NSDebugFLLog(@"NSPort", @"Decoded port as '%@:%d'", addr, pnum);
@@ -381,7 +387,7 @@ static Class	runLoopClass;
   return nil;
 }
 
-+ (GSTcpHandle*) handleWithDescriptor: (int)d
++ (GSTcpHandle*) handleWithDescriptor: (SOCKET)d
 {
   GSTcpHandle	*handle;
 #ifdef __MINGW__
@@ -390,14 +396,14 @@ static Class	runLoopClass;
   int		e;
 #endif /* __MINGW__ */
 
-  if (d < 0)
+  if (d == INVALID_SOCKET)
     {
       NSLog(@"illegal descriptor (%d) for Tcp Handle", d);
       return nil;
     }
 #ifdef __MINGW__
   dummy = 1;
-  if (ioctlsocket(d, FIONBIO, &dummy) < 0)
+  if (ioctlsocket(d, FIONBIO, &dummy) == SOCKET_ERROR)
     {
       NSLog(@"unable to set non-blocking mode on %d - %s",
 	d, GSLastErrorStr(errno));
@@ -421,13 +427,10 @@ static Class	runLoopClass;
       return nil;
     }
 #endif
-  handle = (GSTcpHandle*)NSAllocateObject(self,0,NSDefaultMallocZone());
+  handle = (GSTcpHandle*)NSAllocateObject(self, 0, NSDefaultMallocZone());
   handle->desc = d;
   handle->wMsgs = [NSMutableArray new];
-  if (multi_threaded == YES)
-    {
-      handle->myLock = [NSRecursiveLock new];
-    }
+  handle->myLock = [GSLazyRecursiveLock new];
   handle->valid = YES;
   return AUTORELEASE(handle);
 }
@@ -456,19 +459,36 @@ static Class	runLoopClass;
   BOOL			gotAddr = NO;
   NSRunLoop		*l;
 
-  DO_LOCK(myLock);
+  M_LOCK(myLock);
   NSDebugMLLog(@"GSTcpHandle", @"Connecting on 0x%x in thread 0x%x before %@",
     self, GSCurrentThread(), when);
   if (state != GS_H_UNCON)
     {
-      NSLog(@"attempting connect on connected handle");
-      DO_UNLOCK(myLock);
-      return YES;	/* Already connected.	*/
+      BOOL	result;
+
+      if (state == GS_H_CONNECTED)	/* Already connected.	*/
+	{
+	  NSLog(@"attempting connect on connected handle");
+	  result = YES;
+	}
+      else if (state == GS_H_ACCEPT)	/* Impossible.	*/
+	{
+	  NSLog(@"attempting connect with accepting handle");
+	  result = NO;
+	}
+      else				/* Already connecting.	*/
+	{
+	  NSLog(@"attempting connect while connecting");
+	  result = NO;
+	}
+      M_UNLOCK(myLock);
+      return result;
     }
+
   if (recvPort == nil || aPort == nil)
     {
       NSLog(@"attempting connect with port(s) unset");
-      DO_UNLOCK(myLock);
+      M_UNLOCK(myLock);
       return NO;	/* impossible.		*/
     }
 
@@ -493,7 +513,7 @@ static Class	runLoopClass;
 	{
 	  NSLog(@"run out of addresses to try (tried %d) for port %@",
 	    addrNum, aPort);
-	  DO_UNLOCK(myLock);
+	  M_UNLOCK(myLock);
 	  return NO;
 	}
       addr = [[addrs objectAtIndex: addrNum++] cString];
@@ -512,11 +532,14 @@ static Class	runLoopClass;
       else
 	{
 	  gotAddr = YES;
+	  NSDebugMLLog(@"GSTcpHandle", @"Connecting to %s:%d using desc %d",
+	    addr, [aPort portNumber], desc);
 	}
     }
   sockAddr.sin_port = GSSwapHostI16ToBig([aPort portNumber]);
 
-  if (connect(desc, (struct sockaddr*)&sockAddr, sizeof(sockAddr)) < 0)
+  if (connect(desc, (struct sockaddr*)&sockAddr, sizeof(sockAddr))
+    == SOCKET_ERROR)
     {
 #ifdef __MINGW__
       if (WSAGetLastError() != WSAEWOULDBLOCK)
@@ -532,12 +555,12 @@ static Class	runLoopClass;
 	      BOOL	result;
 
 	      result = [self connectToPort: aPort beforeDate: when];
-	      DO_UNLOCK(myLock);
+	      M_UNLOCK(myLock);
 	      return result;
 	    }
 	  else
 	    {
-	      DO_UNLOCK(myLock);
+	      M_UNLOCK(myLock);
 	      return NO;	/* Tried all addresses	*/
 	    }
 	}
@@ -548,21 +571,30 @@ static Class	runLoopClass;
   [l addEvent: (void*)(gsaddr)desc
 	 type: ET_WDESC
       watcher: self
-      forMode: NSDefaultRunLoopMode];
-  while (state == GS_H_TRYCON && [when timeIntervalSinceNow] > 0)
+      forMode: NSConnectionReplyMode];
+  [l addEvent: (void*)(gsaddr)desc
+	 type: ET_EDESC
+      watcher: self
+      forMode: NSConnectionReplyMode];
+  while (valid == YES && state == GS_H_TRYCON
+    && [when timeIntervalSinceNow] > 0)
     {
-      [l runMode: NSDefaultRunLoopMode beforeDate: when];
+      [l runMode: NSConnectionReplyMode beforeDate: when];
     }
   [l removeEvent: (void*)(gsaddr)desc
 	    type: ET_WDESC
-	 forMode: NSDefaultRunLoopMode
+	 forMode: NSConnectionReplyMode
+	     all: NO];
+  [l removeEvent: (void*)(gsaddr)desc
+	    type: ET_EDESC
+	 forMode: NSConnectionReplyMode
 	     all: NO];
 
   if (state == GS_H_TRYCON)
     {
       state = GS_H_UNCON;
       addrNum = 0;
-      DO_UNLOCK(myLock);
+      M_UNLOCK(myLock);
       return NO;	/* Timed out 	*/
     }
   else if (state == GS_H_UNCON)
@@ -576,12 +608,12 @@ static Class	runLoopClass;
 	   * that we haven't tried.
 	   */
 	  result = [self connectToPort: aPort beforeDate: when];
-	  DO_UNLOCK(myLock);
+	  M_UNLOCK(myLock);
 	  return result;
 	}
       addrNum = 0;
       state = GS_H_UNCON;
-      DO_UNLOCK(myLock);
+      M_UNLOCK(myLock);
       return NO;	/* connection failed	*/
     }
   else
@@ -589,7 +621,7 @@ static Class	runLoopClass;
       addrNum = 0;
       caller = YES;
       [aPort addHandle: self forSend: YES];
-      DO_UNLOCK(myLock);
+      M_UNLOCK(myLock);
       return YES;
     }
 }
@@ -611,38 +643,11 @@ static Class	runLoopClass;
     desc, inet_ntoa(sockAddr.sin_addr), ntohs(sockAddr.sin_port)];
 }
 
-- (int) descriptor
+- (SOCKET) descriptor
 {
   return desc;
 }
 
-/*
- * Method to pass an incoming message to the recieving port.
- * Our lock must be locked on entry to this method, and we
- * unlock it temporarily while actually sending the message.
- */ 
-- (void) dispatch
-{
-  NSPortMessage	*pm;
-  GSTcpPort	*rp = [self recvPort];
-
-  pm = [portMessageClass allocWithZone: NSDefaultMallocZone()];
-  pm = [pm initWithSendPort: [self sendPort]
-		receivePort: rp
-		 components: rItems];
-  [pm setMsgid: rId];
-  rId = 0;
-  DESTROY(rItems);
-  NSDebugMLLog(@"GSTcpHandle", @"got message %@ on 0x%x in thread 0x%x",
-    pm, self, GSCurrentThread());
-  RETAIN(rp);
-  DO_UNLOCK(myLock);
-  [rp handlePortMessage: pm];
-  DO_LOCK(myLock);
-  RELEASE(pm);
-  RELEASE(rp);
-}
-
 - (void) gcFinalize
 {
   [self invalidate];
@@ -652,31 +657,34 @@ static Class	runLoopClass;
 
 - (void) invalidate
 {
-  DO_LOCK(myLock);
   if (valid == YES)
-    { 
-      NSRunLoop	*l;
+    {
+      M_LOCK(myLock);
+      if (valid == YES)
+	{ 
+	  NSRunLoop	*l;
 
-      valid = NO;
-      l = [runLoopClass currentRunLoop];
-      [l removeEvent: (void*)(gsaddr)desc
-		type: ET_RDESC
-	     forMode: nil
-		 all: YES];
-      [l removeEvent: (void*)(gsaddr)desc
-		type: ET_WDESC
-	     forMode: nil
-		 all: YES];
-      [l removeEvent: (void*)(gsaddr)desc
-		type: ET_EDESC
-	     forMode: nil
-		 all: YES];
-      NSDebugMLLog(@"GSTcpHandle", @"invalidated 0x%x in thread 0x%x",
-	self, GSCurrentThread());
-      [[self recvPort] removeHandle: self];
-      [[self sendPort] removeHandle: self];
+	  valid = NO;
+	  l = [runLoopClass currentRunLoop];
+	  [l removeEvent: (void*)(gsaddr)desc
+		    type: ET_RDESC
+		 forMode: nil
+		     all: YES];
+	  [l removeEvent: (void*)(gsaddr)desc
+		    type: ET_WDESC
+		 forMode: nil
+		     all: YES];
+	  [l removeEvent: (void*)(gsaddr)desc
+		    type: ET_EDESC
+		 forMode: nil
+		     all: YES];
+	  NSDebugMLLog(@"GSTcpHandle", @"invalidated 0x%x in thread 0x%x",
+	    self, GSCurrentThread());
+	  [[self recvPort] removeHandle: self];
+	  [[self sendPort] removeHandle: self];
+	}
+      M_UNLOCK(myLock);
     }
-  DO_UNLOCK(myLock);
 }
 
 - (BOOL) isValid
@@ -703,7 +711,7 @@ static Class	runLoopClass;
    * If we have been invalidated (desc < 0) then we should ignore this
    * event and remove ourself from the runloop.
    */
-  if (desc < 0)
+  if (desc == INVALID_SOCKET)
     {
       NSRunLoop	*l = [runLoopClass currentRunLoop];
 
@@ -711,10 +719,14 @@ static Class	runLoopClass;
 		type: ET_WDESC
 	     forMode: mode
 		 all: YES];
+      [l removeEvent: data
+		type: ET_EDESC
+	     forMode: mode
+		 all: YES];
       return;
     }
 
-  DO_LOCK(myLock);
+  M_LOCK(myLock);
 
   if (type == ET_RPORT)
     {
@@ -752,18 +764,14 @@ static Class	runLoopClass;
        * Now try to fill the buffer with data.
        */
       bytes = [rData mutableBytes];
-#ifdef __MINGW__
       res = recv(desc, bytes + rLength, want - rLength, 0);
-#else
-      res = read(desc, bytes + rLength, want - rLength);
-#endif
       if (res <= 0)
 	{
 	  if (res == 0)
 	    {
 	      NSDebugMLLog(@"GSTcpHandle", @"read eof on 0x%x in thread 0x%x",
 		self, GSCurrentThread());
-	      DO_UNLOCK(myLock);
+	      M_UNLOCK(myLock);
 	      [self invalidate];
 	      return;
 	    }
@@ -772,7 +780,7 @@ static Class	runLoopClass;
 	      NSDebugMLLog(@"GSTcpHandle",
 		@"read failed - %s on 0x%x in thread 0x%x",
 		GSLastErrorStr(errno), self, GSCurrentThread());
-	      DO_UNLOCK(myLock);
+	      M_UNLOCK(myLock);
 	      [self invalidate];
 	      return;
 	    }
@@ -784,6 +792,8 @@ static Class	runLoopClass;
 
       while (valid == YES && rLength >= rWant)
 	{
+	  BOOL	shouldDispatch = NO;
+
 	  switch (rType)
 	    {
 	      case GSP_NONE:
@@ -804,7 +814,7 @@ static Class	runLoopClass;
 			{
 			  NSLog(@"%@ - unreasonable length (%u) for port",
 			    self, l);
-			  DO_UNLOCK(myLock);
+			  M_UNLOCK(myLock);
 			  [self invalidate];
 			  return;
 			}
@@ -836,8 +846,7 @@ static Class	runLoopClass;
 			  RELEASE(d);
 			  if (nItems == [rItems count])
 			    {
-			      [self dispatch];
-			      bytes = [rData mutableBytes];
+			      shouldDispatch = YES;
 			    }
 			}
 		      else
@@ -846,7 +855,7 @@ static Class	runLoopClass;
 			    {
 			      NSLog(@"%@ - unreasonable length (%u) for data",
 				self, l);
-			      DO_UNLOCK(myLock);
+			      M_UNLOCK(myLock);
 			      [self invalidate];
 			      return;
 			    }
@@ -869,7 +878,7 @@ static Class	runLoopClass;
 			{
 			  NSLog(@"%@ - unreasonable length (%u) for data",
 			    self, l);
-			  DO_UNLOCK(myLock);
+			  M_UNLOCK(myLock);
 			  [self invalidate];
 			  return;
 			}
@@ -888,7 +897,7 @@ static Class	runLoopClass;
 		  else
 		    {
 		      NSLog(@"%@ - bad data received on port handle", self);
-		      DO_UNLOCK(myLock);
+		      M_UNLOCK(myLock);
 		      [self invalidate];
 		      return;
 		    }
@@ -934,8 +943,7 @@ static Class	runLoopClass;
 		      rWant = sizeof(GSPortItemHeader);
 		      if (nItems == 1)
 			{
-			  [self dispatch];
-			  bytes = [rData mutableBytes];
+			  shouldDispatch = YES;
 			}
 		    }
 		  else
@@ -970,8 +978,7 @@ static Class	runLoopClass;
 		  rWant = sizeof(GSPortItemHeader);
 		  if (nItems == [rItems count])
 		    {
-		      [self dispatch];
-		      bytes = [rData mutableBytes];
+		      shouldDispatch = YES;
 		    }
 		}
 		break;
@@ -982,6 +989,13 @@ static Class	runLoopClass;
 
 		  rType = GSP_NONE;	/* ready for a new item	*/
 		  p = decodePort(rData, defaultAddress);
+		  if (p == nil)
+		    {
+		      NSLog(@"%@ - unable to decode remote port", self);
+		      M_UNLOCK(myLock);
+		      [self invalidate];
+		      return;
+		    }
 		  /*
 		   * Set up to read another item header.
 		   */
@@ -1010,16 +1024,50 @@ static Class	runLoopClass;
 		      [rItems addObject: p];
 		      if (nItems == [rItems count])
 			{
-			  [self dispatch];
-			  bytes = [rData mutableBytes];
+			  shouldDispatch = YES;
 			}
 		    }
 		}
 		break;
 	    }
+
+	  if (shouldDispatch == YES)
+	    {
+	      NSPortMessage	*pm;
+	      GSTcpPort		*rp = [self recvPort];
+
+	      pm = [portMessageClass allocWithZone: NSDefaultMallocZone()];
+	      pm = [pm initWithSendPort: [self sendPort]
+			    receivePort: rp
+			     components: rItems];
+	      [pm setMsgid: rId];
+	      rId = 0;
+	      DESTROY(rItems);
+	      NSDebugMLLog(@"GSTcpHandle",
+		@"got message %@ on 0x%x in thread 0x%x",
+		pm, self, GSCurrentThread());
+	      RETAIN(rp);
+	      M_UNLOCK(myLock);
+	      NS_DURING
+		{
+		  [rp handlePortMessage: pm];
+		}
+	      NS_HANDLER
+		{
+		  M_LOCK(myLock);
+		  RELEASE(pm);
+		  RELEASE(rp);
+		  [localException raise];
+		}
+	      NS_ENDHANDLER
+	      M_LOCK(myLock);
+	      RELEASE(pm);
+	      RELEASE(rp);
+	      bytes = [rData mutableBytes];
+	    }
 	}
     }
-  else if (type == ET_WDESC)
+  else
     {
       if (state == GS_H_TRYCON)	/* Connection attempt.	*/
 	{
@@ -1036,12 +1084,8 @@ static Class	runLoopClass;
 	    {
 	      NSData	*d = newDataWithEncodedPort([self recvPort]);
 
-#ifdef __MINGW__
 	      len = send(desc, [d bytes], [d length], 0);
-#else
-	      len = write(desc, [d bytes], [d length]);
-#endif
-	      if (len == [d length])
+	      if (len == (int)[d length])
 		{
 		  RELEASE(defaultAddress);
 		  defaultAddress = RETAIN([NSString stringWithCString:
@@ -1083,17 +1127,13 @@ static Class	runLoopClass;
 	    }
 	  b = [wData bytes];
 	  l = [wData length];
-#ifdef __MINGW__
 	  res = send(desc, b + wLength,  l - wLength, 0);
-#else
-	  res = write(desc, b + wLength,  l - wLength);
-#endif
 	  if (res < 0)
 	    {
 	      if (errno != EINTR && errno != EAGAIN)
 		{
 		  NSLog(@"write attempt failed - %s", GSLastErrorStr(errno));
-		  DO_UNLOCK(myLock);
+		  M_UNLOCK(myLock);
 		  [self invalidate];
 		  return;
 		}
@@ -1139,13 +1179,17 @@ static Class	runLoopClass;
 				type: ET_WDESC
 			     forMode: mode
 				 all: NO];
+		      [l removeEvent: data
+				type: ET_EDESC
+			     forMode: mode
+				 all: NO];
 		    }
 		}
 	    }
 	}
     }
 
-  DO_UNLOCK(myLock);
+  M_UNLOCK(myLock);
 }
 
 - (BOOL) sendMessage: (NSArray*)components beforeDate: (NSDate*)when
@@ -1157,7 +1201,7 @@ static Class	runLoopClass;
   NSDebugMLLog(@"GSTcpHandle",
     @"Sending message 0x%x %@ on 0x%x(%d) in thread 0x%x before %@",
     components, components, self, desc, GSCurrentThread(), when);
-  DO_LOCK(myLock);
+  M_LOCK(myLock);
   [wMsgs addObject: components];
 
   l = [runLoopClass currentRunLoop];
@@ -1167,13 +1211,14 @@ static Class	runLoopClass;
   [l addEvent: (void*)(gsaddr)desc
 	 type: ET_WDESC
       watcher: self
-      forMode: NSDefaultRunLoopMode];
-  while ([wMsgs indexOfObjectIdenticalTo: components] != NSNotFound
+      forMode: NSConnectionReplyMode];
+  while (valid == YES
+    && [wMsgs indexOfObjectIdenticalTo: components] != NSNotFound
     && [when timeIntervalSinceNow] > 0)
     {
-      DO_UNLOCK(myLock);
-      [l runMode: NSDefaultRunLoopMode beforeDate: when];
-      DO_LOCK(myLock);
+      M_UNLOCK(myLock);
+      [l runMode: NSConnectionReplyMode beforeDate: when];
+      M_LOCK(myLock);
     }
   /*
    * NB. We will remove ourself from the run loop when the message send
@@ -1183,7 +1228,7 @@ static Class	runLoopClass;
     {
       sent = YES;
     }
-  DO_UNLOCK(myLock);
+  M_UNLOCK(myLock);
   RELEASE(self);
   NSDebugMLLog(@"GSTcpHandle",
     @"Message send 0x%x on 0x%x in thread 0x%x status %d",
@@ -1228,79 +1273,22 @@ static NSRecursiveLock	*tcpPortLock = nil;
 static NSMapTable	*tcpPortMap = 0;
 static Class		tcpPortClass;
 
-/*
- *	When the system becomes multithreaded, we set a flag to say so and
- *	make sure that port and handle locking is enabled.
- */
-+ (void) _becomeThreaded: (NSNotification*)notification
-{
-  if (multi_threaded == NO)
-    {
-      NSMapEnumerator	pEnum;
-      NSMapTable	*m;
-      void		*dummy;
-
-      multi_threaded = YES;
-      if (tcpPortLock == nil)
-	{
-	  tcpPortLock = [NSRecursiveLock new];
-	}
-      pEnum = NSEnumerateMapTable(tcpPortMap);
-      while (NSNextMapEnumeratorPair(&pEnum, &dummy, (void**)&m))
-	{
-	  NSMapEnumerator	mEnum;
-	  GSTcpPort		*p;
-
-	  mEnum = NSEnumerateMapTable(m);
-	  while (NSNextMapEnumeratorPair(&mEnum, &dummy, (void**)&p))
-	    {
-	      if ([p isValid] == YES)
-		{
-		  NSMapEnumerator	hEnum;
-		  GSTcpHandle		*h;
-
-		  if (p->myLock == nil)
-		    {
-		      p->myLock = [NSRecursiveLock new];
-		    }
-		  hEnum = NSEnumerateMapTable(p->handles);
-		  while (NSNextMapEnumeratorPair(&hEnum, &dummy, (void**)&h))
-		    {
-		      if ([h isValid] == YES && h->myLock == nil)
-			{
-			  h->myLock = [NSRecursiveLock new];
-			}
-		    }
-		}
-	    }
-	}
-    }
-  [[NSNotificationCenter defaultCenter]
-    removeObserver: self
-	      name: NSWillBecomeMultiThreadedNotification
-	    object: nil];
-}
+#if NEED_WORD_ALIGNMENT
+static unsigned	wordAlign;
+#endif
 
 + (void) initialize
 {
   if (self == [GSTcpPort class])
     {
+#if NEED_WORD_ALIGNMENT
+      wordAlign = objc_alignof_type(@encode(gsu32));
+#endif
       tcpPortClass = self;
       tcpPortMap = NSCreateMapTable(NSIntMapKeyCallBacks,
 	NSNonOwnedPointerMapValueCallBacks, 0);
 
-      if ([NSThread isMultiThreaded])
-	{
-	  [self _becomeThreaded: nil];
-	}
-      else
-	{
-	  [[NSNotificationCenter defaultCenter]
-	    addObserver: self
-	       selector: @selector(_becomeThreaded:)
-		   name: NSWillBecomeMultiThreadedNotification
-		 object: nil];
-	}
+      tcpPortLock = [GSLazyRecursiveLock new];
     }
 }
 
@@ -1321,7 +1309,7 @@ static Class		tcpPortClass;
   GSTcpPort	*port = nil;
   NSMapTable	*thePorts;
 
-  [tcpPortLock lock];
+  M_LOCK(tcpPortLock);
 
   /*
    *	Get the map table of ports with the specified number.
@@ -1332,7 +1320,7 @@ static Class		tcpPortClass;
       port = (GSTcpPort*)NSMapGet(thePorts, (void*)aHost);
       IF_NO_GC(AUTORELEASE(RETAIN(port)));
     }
-  [tcpPortLock unlock];
+  M_UNLOCK(tcpPortLock);
   return port;
 }
 
@@ -1377,7 +1365,7 @@ static Class		tcpPortClass;
       return nil;
     }
 
-  [tcpPortLock lock];
+  M_LOCK(tcpPortLock);
 
   /*
    * First try to find a pre-existing port.
@@ -1390,22 +1378,21 @@ static Class		tcpPortClass;
 
   if (port == nil)
     {
-      port = (GSTcpPort*)NSAllocateObject(self,0,NSDefaultMallocZone());
+      port = (GSTcpPort*)NSAllocateObject(self, 0, NSDefaultMallocZone());
       port->listener = -1;
       port->host = RETAIN(aHost);
       port->address = [addr copy];
       port->handles = NSCreateMapTable(NSIntMapKeyCallBacks,
 	NSObjectMapValueCallBacks, 0);
-      if (multi_threaded == YES)
-	{
-	  port->myLock = [NSRecursiveLock new];
-	}
+      port->myLock = [GSLazyRecursiveLock new];
       port->_is_valid = YES;
 
       if (shouldListen == YES && [thisHost isEqual: aHost])
 	{
+#ifndef	BROKEN_SO_REUSEADDR
 	  int	reuse = 1;	/* Should we re-use ports?	*/
-	  int	desc;
+#endif
+	  SOCKET desc;
 	  BOOL	addrOk = YES;
 	  struct sockaddr_in	sockaddr;
 
@@ -1442,34 +1429,44 @@ static Class		tcpPortClass;
 	      NSLog(@"Bad address (%@) specified for listening port", addr);
 	      DESTROY(port);
 	    }
-	  else if ((desc = socket(AF_INET, SOCK_STREAM, PF_UNSPEC)) < 0)
+	  else if ((desc = socket(AF_INET, SOCK_STREAM, PF_UNSPEC))
+	    == INVALID_SOCKET)
 	    {
 	      NSLog(@"unable to create socket - %s", GSLastErrorStr(errno));
 	      DESTROY(port);
 	    }
+#ifndef	BROKEN_SO_REUSEADDR
+	  /*
+	   * Under decent systems, SO_REUSEADDR means that the port can be
+	   * reused immediately that this porcess exits.  Under some it means
+	   * that multiple processes can serve the same port simultaneously.
+	   * We don't want that broken behavior!
+	   */
 	  else if (setsockopt(desc, SOL_SOCKET, SO_REUSEADDR, (char*)&reuse,
-		sizeof(reuse)) < 0)
+	    sizeof(reuse)) < 0)
 	    {
 	      (void) close(desc);
               NSLog(@"unable to set reuse on socket - %s",
 		GSLastErrorStr(errno));
               DESTROY(port);
 	    }
+#endif
 	  else if (bind(desc, (struct sockaddr *)&sockaddr,
-	    sizeof(sockaddr)) < 0)
+	    sizeof(sockaddr)) == SOCKET_ERROR)
 	    {
 	      NSLog(@"unable to bind to port %s:%d - %s",
 		inet_ntoa(sockaddr.sin_addr), number, GSLastErrorStr(errno));
 	      (void) close(desc);
               DESTROY(port);
 	    }
-	  else if (listen(desc, 5) < 0)
+	  else if (listen(desc, 5) == SOCKET_ERROR)
 	    {
 	      NSLog(@"unable to listen on port - %s", GSLastErrorStr(errno));
 	      (void) close(desc);
 	      DESTROY(port);
 	    }
-	  else if (getsockname(desc, (struct sockaddr*)&sockaddr, &i) < 0)
+	  else if (getsockname(desc, (struct sockaddr*)&sockaddr, &i)
+	    == SOCKET_ERROR)
 	    {
 	      NSLog(@"unable to get socket name - %s", GSLastErrorStr(errno));
 	      (void) close(desc);
@@ -1538,13 +1535,13 @@ static Class		tcpPortClass;
     }
   IF_NO_GC(AUTORELEASE(port));
 
-  [tcpPortLock unlock];
+  M_UNLOCK(tcpPortLock);
   return port;
 }
 
 - (void) addHandle: (GSTcpHandle*)handle forSend: (BOOL)send
 {
-  DO_LOCK(myLock);
+  M_LOCK(myLock);
   if (send == YES)
     {
       if (handle->caller == YES)
@@ -1557,7 +1554,7 @@ static Class		tcpPortClass;
       handle->recvPort = GS_GC_HIDE(self);
     }
   NSMapInsert(handles, (void*)(gsaddr)[handle descriptor], (void*)handle);
-  DO_UNLOCK(myLock);
+  M_UNLOCK(myLock);
 }
 
 - (NSString*) address
@@ -1582,9 +1579,17 @@ static Class		tcpPortClass;
 {
   NSMutableString	*desc;
 
-  desc = [NSMutableString stringWithFormat: @"Host - %@\n", host];
-  [desc appendFormat: @"Addr - %@\n", address];
-  [desc appendFormat: @"Port - %d\n", portNum];
+  desc = [NSMutableString stringWithFormat: @"NSPort on host with details -\n"
+    @"%@\n", host];
+  if (address == nil)
+    {
+      [desc appendFormat: @"  IP address - any\n"];
+    }
+  else
+    {
+      [desc appendFormat: @"  IP address - %@\n", address];
+    }
+  [desc appendFormat: @"  TCP port - %d\n", portNum];
   return desc;
 }
 
@@ -1598,19 +1603,20 @@ static Class		tcpPortClass;
  * This is a callback method used by the NSRunLoop class to determine which
  * descriptors to watch for the port.
  */
-- (void) getFds: (int*)fds count: (int*)count
+- (void) getFds: (SOCKET*)fds count: (int*)count
 {
   NSMapEnumerator	me;
-  int			sock;
+  SOCKET		sock;
   GSTcpHandle		*handle;
   id			recvSelf;
 
-  DO_LOCK(myLock);
+  M_LOCK(myLock);
 
   /*
    * Make sure there is enough room in the provided array.
    */
-  NSAssert(*count > NSCountMapTable(handles), NSInternalInconsistencyException);
+  NSAssert(*count > (int)NSCountMapTable(handles),
+    NSInternalInconsistencyException);
 
   /*
    * Put in our listening socket.
@@ -1634,16 +1640,20 @@ static Class		tcpPortClass;
 	  fds[(*count)++] = sock;
 	}
     }
-  DO_UNLOCK(myLock);
+  NSEndMapTableEnumeration(&me);
+  M_UNLOCK(myLock);
 }
 
 - (GSTcpHandle*) handleForPort: (GSTcpPort*)recvPort beforeDate: (NSDate*)when
 {
   NSMapEnumerator	me;
-  int			sock;
+  SOCKET		sock;
+#ifndef	BROKEN_SO_REUSEADDR
+  int			opt = 1;
+#endif
   GSTcpHandle		*handle = nil;
 
-  DO_LOCK(myLock);
+  M_LOCK(myLock);
   /*
    * Enumerate all our socket handles, and look for one with port.
    */
@@ -1652,35 +1662,45 @@ static Class		tcpPortClass;
     {
       if ([handle recvPort] == recvPort)
 	{
-	  DO_UNLOCK(myLock);
+	  M_UNLOCK(myLock);
+	  NSEndMapTableEnumeration(&me);
 	  return handle;
 	}
     }
-  if (handle == nil)
-    {
-      int	opt = 1;
+  NSEndMapTableEnumeration(&me);
 
-      if ((sock = socket(AF_INET, SOCK_STREAM, PF_UNSPEC)) < 0)
-	{
-	  NSLog(@"unable to create socket - %s", GSLastErrorStr(errno));
-	}
-      else if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (char*)&opt,
-	sizeof(opt)) < 0)
-	{
-	  (void)close(sock);
-	  NSLog(@"unable to set reuse on socket - %s", GSLastErrorStr(errno));
-	}
-      else if ((handle = [GSTcpHandle handleWithDescriptor: sock]) == nil)
-	{
-	  (void)close(sock);
-	  NSLog(@"unable to create GSTcpHandle - %s", GSLastErrorStr(errno));
-	}
-      else
-	{
-	  [recvPort addHandle: handle forSend: NO];
-	}
+  /*
+   * Not found ... create a new handle.
+   */
+  handle = nil;
+  if ((sock = socket(AF_INET, SOCK_STREAM, PF_UNSPEC)) == INVALID_SOCKET)
+    {
+      NSLog(@"unable to create socket - %s", GSLastErrorStr(errno));
     }
-  DO_UNLOCK(myLock);
+#ifndef	BROKEN_SO_REUSEADDR
+  /*
+   * Under decent systems, SO_REUSEADDR means that the port can be reused
+   * immediately that this process exits.  Under some it means
+   * that multiple processes can serve the same port simultaneously.
+   * We don't want that broken behavior!
+   */
+  else if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (char*)&opt,
+    sizeof(opt)) < 0)
+    {
+      (void)close(sock);
+      NSLog(@"unable to set reuse on socket - %s", GSLastErrorStr(errno));
+    }
+#endif
+  else if ((handle = [GSTcpHandle handleWithDescriptor: sock]) == nil)
+    {
+      (void)close(sock);
+      NSLog(@"unable to create GSTcpHandle - %s", GSLastErrorStr(errno));
+    }
+  else
+    {
+      [recvPort addHandle: handle forSend: NO];
+    }
+  M_UNLOCK(myLock);
   /*
    * If we succeeded in creating a new handle - connect to remote host.
    */
@@ -1731,40 +1751,53 @@ static Class		tcpPortClass;
 
 - (void) invalidate
 {
-  DO_LOCK(myLock);
-
-  if ([self isValid])
+  if ([self isValid] == YES)
     {
-      NSMapTable	*thePorts;
-      NSArray		*handleArray;
-      unsigned		i;
+      M_LOCK(myLock);
 
-      [tcpPortLock lock];
-      thePorts = NSMapGet(tcpPortMap, (void*)(gsaddr)portNum);
-      if (thePorts != 0)
+      if ([self isValid] == YES)
 	{
-	  if (listener >= 0)
+	  NSMapTable	*thePorts;
+	  NSArray	*handleArray;
+	  unsigned	i;
+
+	  M_LOCK(tcpPortLock);
+	  thePorts = NSMapGet(tcpPortMap, (void*)(gsaddr)portNum);
+	  if (thePorts != 0)
 	    {
-	      (void) close(listener);
-	      listener = -1;
+	      if (listener >= 0)
+		{
+		  (void) close(listener);
+		  listener = -1;
+		}
+	      NSMapRemove(thePorts, (void*)host);
 	    }
-	  NSMapRemove(thePorts, (void*)host);
-	}
-      [tcpPortLock unlock];
+	  M_UNLOCK(tcpPortLock);
 
-      handleArray = NSAllMapTableValues(handles);
-      i = [handleArray count];
-      while (i-- > 0)
-	{
-	  GSTcpHandle	*handle = [handleArray objectAtIndex: i];
+	  if (handles != 0)
+	    {
+	      handleArray = NSAllMapTableValues(handles);
+	      i = [handleArray count];
+	      while (i-- > 0)
+		{
+		  GSTcpHandle	*handle = [handleArray objectAtIndex: i];
 
-	  [handle invalidate];
+		  [handle invalidate];
+		}
+	      /*
+	       * We permit mutual recursive invalidation, so the handles map
+	       * may already have been destroyed.
+	       */
+	      if (handles != 0)
+		{
+		  NSFreeMapTable(handles);
+		  handles = 0;
+		}
+	    }
+	  [super invalidate];
 	}
-      NSFreeMapTable(handles);
-      handles = 0;
-      [super invalidate];
+      M_UNLOCK(myLock);
     }
-  DO_UNLOCK(myLock);
 }
 
 - (BOOL) isEqual: (id)anObject
@@ -1795,7 +1828,7 @@ static Class		tcpPortClass;
 		 extra: (void*)extra
 	       forMode: (NSString*)mode
 {
-  int		desc = (int)(gsaddr)extra;
+  SOCKET	desc = (SOCKET)(gsaddr)extra;
   GSTcpHandle	*handle;
 
   if (desc == listener)
@@ -1804,7 +1837,7 @@ static Class		tcpPortClass;
       int			size = sizeof(sockAddr);
 
       desc = accept(listener, (struct sockaddr*)&sockAddr, &size);
-      if (desc < 0)
+      if (desc == INVALID_SOCKET)
         {
 	  NSDebugMLLog(@"NSPort", @"accept failed - handled in other thread?");
         }
@@ -1826,10 +1859,10 @@ static Class		tcpPortClass;
     }
   else
     {
-      DO_LOCK(myLock);
+      M_LOCK(myLock);
       handle = (GSTcpHandle*)NSMapGet(handles, (void*)(gsaddr)desc);
       IF_NO_GC(AUTORELEASE(RETAIN(handle)));
-      DO_UNLOCK(myLock);
+      M_UNLOCK(myLock);
       if (handle == nil)
 	{
 	  const char	*t;
@@ -1859,7 +1892,7 @@ static Class		tcpPortClass;
  */
 - (void) removeHandle: (GSTcpHandle*)handle
 {
-  DO_LOCK(myLock);
+  M_LOCK(myLock);
   if ([handle sendPort] == self)
     {
       if (handle->caller != YES)
@@ -1884,7 +1917,7 @@ static Class		tcpPortClass;
     {
       [self invalidate];
     }
-  DO_UNLOCK(myLock);
+  M_UNLOCK(myLock);
 }
 
 /*
@@ -2008,7 +2041,7 @@ static Class		tcpPortClass;
 		   * word boundary, so we work with an aligned buffer
 		   * and use memcmpy()
 		   */
-		  if ((hLength % __alignof__(gsu32)) != 0)
+		  if ((hLength % wordAlign) != 0)
 		    {
 		      GSPortItemHeader	itemHeader;
 
