@@ -21,15 +21,15 @@
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA.
 */
 
-#include <config.h>
-#include <Foundation/NSValue.h>
-#include <Foundation/NSString.h>
-#include <Foundation/NSData.h>
-#include <Foundation/NSException.h>
-#include <Foundation/NSCoder.h>
-#include <Foundation/NSZone.h>
-#include <Foundation/NSObjCRuntime.h>
-#include <base/preface.h>
+#include "config.h"
+#include "Foundation/NSValue.h"
+#include "Foundation/NSString.h"
+#include "Foundation/NSData.h"
+#include "Foundation/NSException.h"
+#include "Foundation/NSCoder.h"
+#include "Foundation/NSZone.h"
+#include "Foundation/NSObjCRuntime.h"
+#include "GNUstepBase/preface.h"
 
 @interface GSValue : NSValue
 {
