@@ -21,10 +21,10 @@
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA.
 
    <title>NSArchiver class reference</title>
-   $Date: 2002/02/21 13:31:13 $ $Revision: 1.53 $
+   $Date: 2003/08/20 12:13:34 $ $Revision: 1.59 $
    */
 
-#include <config.h>
+#include "config.h"
 /*
  *	Setup for inline operation of pointer map tables.
  */
@@ -36,21 +36,23 @@
 #define	GSI_MAP_EQUAL(M, X,Y)	((X).uint == (Y).uint)
 #define	GSI_MAP_NOCLEAN	1
 
-#include <base/GSIMap.h>
+#include "GNUstepBase/GSIMap.h"
 
 #define	_IN_NSARCHIVER_M
-#include <Foundation/NSArchiver.h>
+#include "Foundation/NSArchiver.h"
 #undef	_IN_NSARCHIVER_M
 
-#include <Foundation/NSObjCRuntime.h>
-#include <Foundation/NSCoder.h>
-#include <Foundation/NSData.h>
-#include <Foundation/NSException.h>
-#include <Foundation/NSUtilities.h>
-#include <Foundation/NSString.h>
+#include "Foundation/NSObjCRuntime.h"
+#include "Foundation/NSCoder.h"
+#include "Foundation/NSData.h"
+#include "Foundation/NSException.h"
+#include "Foundation/NSUtilities.h"
+#include "Foundation/NSString.h"
 
 typedef	unsigned char	uchar;
 
+NSString * const NSInconsistentArchiveException =
+  @"NSInconsistentArchiveException";
 
 #define	PREFIX		"GNUstep archive"
 
@@ -159,7 +161,7 @@ static Class	NSMutableDataMallocClass;
 	}
       NSZoneFree(_clsMap->zone, (void*)_clsMap);
     }
-  return [super dealloc];
+  [super dealloc];
 }
 
 + (NSData*) archivedDataWithRootObject: (id)rootObject
@@ -290,34 +292,22 @@ static Class	NSMutableDataMallocClass;
 
       case _C_STRUCT_B:
 	{
-	  int	offset = 0;
+	  struct objc_struct_layout layout;
 
 	  if (_initialPass == NO)
 	    {
 	      (*_tagImp)(_dst, tagSel, _GSC_STRUCT_B);
 	    }
-
-	  while (*type != _C_STRUCT_E && *type++ != '='); /* skip "<name>=" */
-
-	  for (;;)
+	  objc_layout_structure (type, &layout);
+	  while (objc_layout_structure_next_member (&layout))
 	    {
-	      (*_eValImp)(self, eValSel, type, (char*)buf + offset);
-	      offset += objc_sizeof_type(type);
-	      type = objc_skip_typespec(type);
-	      if (*type == _C_STRUCT_E)
-		{
-		  break;
-		}
-	      else
-		{
-		  int	align = objc_alignof_type(type);
-		  int	rem = offset % align;
+	      int		offset;
+	      int		align;
+	      const char	*ftype;
 
-		  if (rem != 0)
-		    {
-		      offset += align - rem;
-		    }
-		}
+	      objc_layout_structure_get_info (&layout, &offset, &align, &ftype);
+
+	      (*_eValImp)(self, eValSel, ftype, (char*)buf + offset);
 	    }
 	}
 	return;
@@ -328,7 +318,7 @@ static Class	NSMutableDataMallocClass;
 	    if (_initialPass == NO)
 	      {
 		/*
-		 *	Special case - a nul pointer gets an xref of zero
+		 *	Special case - a null pointer gets an xref of zero
 		 */
 		(*_tagImp)(_dst, tagSel, _GSC_PTR | _GSC_XREF | _GSC_X_0);
 	      }
@@ -397,7 +387,7 @@ static Class	NSMutableDataMallocClass;
 	if (*(Class*)buf == 0)
 	  {
 	    /*
-	     *	Special case - a nul pointer gets an xref of zero
+	     *	Special case - a null pointer gets an xref of zero
 	     */
 	    (*_tagImp)(_dst, tagSel, _GSC_CLASS | _GSC_XREF | _GSC_X_0);
 	  }
@@ -465,7 +455,7 @@ static Class	NSMutableDataMallocClass;
 	if (*(SEL*)buf == 0)
 	  {
 	    /*
-	     *	Special case - a nul pointer gets an xref of zero
+	     *	Special case - a null pointer gets an xref of zero
 	     */
 	    (*_tagImp)(_dst, tagSel, _GSC_SEL | _GSC_XREF | _GSC_X_0);
 	  }
@@ -496,7 +486,7 @@ static Class	NSMutableDataMallocClass;
 	if (*(char**)buf == 0)
 	  {
 	    /*
-	     *	Special case - a nul pointer gets an xref of zero
+	     *	Special case - a null pointer gets an xref of zero
 	     */
 	    (*_tagImp)(_dst, tagSel, _GSC_CHARPTR | _GSC_XREF | _GSC_X_0);
 	  }
