@@ -30,7 +30,7 @@
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA.
 
    <title>NSAttributedString class reference</title>
-   $Date: 2001/12/18 16:54:14 $ $Revision: 1.37 $
+   $Date: 2003/08/20 12:13:34 $ $Revision: 1.40 $
 */
 
 /* Warning -	[-initWithString:attributes:] is the designated initialiser,
@@ -47,14 +47,14 @@
  */
 
 #include "config.h"
-#include <base/preface.h>
-#include <base/Unicode.h>
+#include "GNUstepBase/preface.h"
+#include "GNUstepBase/Unicode.h"
 
-#include <Foundation/NSAttributedString.h>
-#include <Foundation/NSException.h>
-#include <Foundation/NSAutoreleasePool.h>
-#include <Foundation/NSPortCoder.h>
-#include <Foundation/NSRange.h>
+#include "Foundation/NSAttributedString.h"
+#include "Foundation/NSException.h"
+#include "Foundation/NSAutoreleasePool.h"
+#include "Foundation/NSPortCoder.h"
+#include "Foundation/NSRange.h"
 
 @class	GSAttributedString;
 @class	GSMutableAttributedString;
@@ -887,12 +887,12 @@ static Class GSMutableAttributedStringClass;
 
 - (void)getCharacters: (unichar*)buffer
 {
-  return [[_owner string] getCharacters: buffer];
+  [[_owner string] getCharacters: buffer];
 }
 
 - (void)getCharacters: (unichar*)buffer range: (NSRange)aRange
 {
-  return [[_owner string] getCharacters: buffer range: aRange];
+  [[_owner string] getCharacters: buffer range: aRange];
 }
 
 - (const char*) cString
@@ -922,7 +922,7 @@ static Class GSMutableAttributedStringClass;
 
 - (void) encodeWithCoder: (NSCoder*)aCoder
 {
-  return [[_owner string] encodeWithCoder: aCoder];
+  [[_owner string] encodeWithCoder: aCoder];
 }
 
 - (Class) classForCoder
