@@ -2,6 +2,7 @@
    Copyright (C) 1995, 1996, 1997 Free Software Foundation, Inc.
    
    Written by:  Andrew Kachites McCallum <mccallum@gnu.ai.mit.edu>
+   Written by:  Richard Frith-Macdonald <rfm@gnu.org>
    Date: January 1995
    
    This file is part of the GNUstep Base Library.
@@ -21,15 +22,15 @@
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA.
 
    <title>NSAutoreleasePool class reference</title>
-   $Date: 2002/01/30 15:28:50 $ $Revision: 1.46 $
+   $Date: 2003/07/31 23:49:31 $ $Revision: 1.52 $
    */ 
 
-#include <config.h>
-#include <base/preface.h>
-#include <Foundation/NSAutoreleasePool.h>
-#include <Foundation/NSException.h>
-#include <Foundation/NSThread.h>
-#include <Foundation/NSZone.h>
+#include "config.h"
+#include "GNUstepBase/preface.h"
+#include "Foundation/NSAutoreleasePool.h"
+#include "Foundation/NSException.h"
+#include "Foundation/NSThread.h"
+#include "Foundation/NSZone.h"
 #include <limits.h>
 
 /*
@@ -119,6 +120,84 @@ pop_pool_from_cache (struct autorelease_thread_vars *tv)
 }
 
 
+/**
+ * <p>
+ *   The standard OpenStep system of memory management employs retain counts.
+ *   When an object is created, it has a retain count of 1.  When an object
+ *   is retained, the retain count is incremented.  When it is released the
+ *   retain count is decremented, and when the retain count goes to zero the
+ *   object gets deallocated.
+ * </p>
+ * <p>
+ *   A simple retain/release mechanism is not very interesting ...
+ *   so it's spiced up with autorelease pools.  You can use the
+ *   AUTORELEASE() macro to call the [NSObject-autorelease]
+ *   method, which adds an object to the current autorelease pool by
+ *   calling [NSAutoreleasePool+addObject:].<br />
+ *   An autorelease pool simply maintains a reference to each object
+ *   added to it, and for each addition, the autorelease pool will
+ *   call the [NSObject-release] method of the object when the pool
+ *   is released.  So doing an AUTORELEASE() is just the same as
+ *   doing a RELEASE(), but deferred until the current autorelease
+ *   pool is deallocated.
+ * </p>
+ * <p>
+ *   The NSAutoreleasePool class maintains a separate stack of
+ *   autorelease pools objects in each thread.
+ * </p>
+ * <p>
+ *   When an autorelease pool is created, it is automatically
+ *   added to the stack of pools in the thread.
+ * </p>
+ * <p>
+ *   When a pool is destroyed, it (and any pool later in
+ *   the stack) is removed from the stack.
+ * </p>
+ * <p>
+ *   This mechanism provides a simple but controllable and reasonably
+ *   efficient way of managing temporary objects.  An object can be
+ *   autoreleased and then passed around and used until the topmost 
+ *   pool in the stack is destroyed.
+ * </p>   
+ * <p>
+ *   Most methods return objects which are either owned by autorelease
+ *   pools or by the receiver of the method, so the lifetime of the
+ *   returned object can be assumed to be the shorter of the lifetime
+ *   of the current autorelease pool, or that of the receiver on which
+ *   the method was called.<br />
+ *   The exceptions to this are those object returned by -
+ * </p>
+ * <deflist>
+ *   <term>[NSObject+alloc], [NSObject+allocWithZone:]</term>
+ *   <desc>
+ *     Methods whose names begin with alloc return an uninitialised
+ *     object, owned by the caller.
+ *   </desc>
+ *   <term>[NSObject-init]</term>
+ *   <desc>
+ *     Methods whose names begin with init return an initialised
+ *     version of the receiving object, owned by the caller.<br />
+ *     NB. The returned object may not actualy be the same as the
+ *     receiver ... sometimes an init method releases the original
+ *     receiver and returns an alternative.
+ *   </desc>
+ *   <term>[NSObject+new]</term>
+ *   <desc>
+ *     Methods whose names begin with new combine the effects of
+ *     allocation and initialisation.
+ *   </desc>
+ *   <term>[NSObject-copy], [(NSCopying)-copyWithZone:]</term>
+ *   <desc>
+ *     Methods whose names begin with copy create a copy of the receiver
+ *     which is owned by the caller.
+ *   </desc>
+ *   <term>[NSObject-mutableCopy], [(NSMutableCopying)-mutableCopyWithZone:]</term>
+ *   <desc>
+ *     Methods whose names begin with mutableCopy create a copy of the receiver
+ *     which is owned by the caller.
+ *   </desc>
+ * </deflist>
+ */
 @implementation NSAutoreleasePool
 
 static IMP	allocImp;
@@ -133,10 +212,13 @@ static IMP	initImp;
     }
 }
 
+/**
+ * Allocate and return an autorelease pool instance.<br />
+ * If there is an already-allocated NSAutoreleasePool available,
+ * save time by just returning that, rather than allocating a new one.
+ */
 + (id) allocWithZone: (NSZone*)zone
 {
-  /* If there is an already-allocated NSAutoreleasePool available,
-     save time by just returning that, rather than allocating a new one. */
   struct autorelease_thread_vars *tv = ARP_THREAD_VARS;
   if (tv->pool_cache_count)
     return pop_pool_from_cache (tv);
@@ -202,7 +284,9 @@ static IMP	initImp;
   return _parent;
 }
 
-/* This method not in OpenStep */
+/*
+ * Return the number of objects in this pool.
+ */
 - (unsigned) autoreleaseCount
 {
   unsigned count = 0;
@@ -215,12 +299,21 @@ static IMP	initImp;
   return count;
 }
 
-/* This method not in OpenStep */
+/**
+ * <p>
+ *   Counts the number of times that the specified object occurs
+ *   in autorelease pools in the current thread.
+ * </p>
+ * <p>
+ *   This method is <em>slow</em> and should probably only be
+ *   used for debugging purposes.
+ * </p>
+ */
 - (unsigned) autoreleaseCountForObject: (id)anObject
 {
   unsigned count = 0;
   struct autorelease_array_list *released = _released_head;
-  int i;
+  unsigned int i;
 
   while (released && released->count)
     {
@@ -232,9 +325,16 @@ static IMP	initImp;
   return count;
 }
 
-/* This method not in OpenStep */
-/* xxx This count should be made for *all* threads, but currently is 
-   only madefor the current thread! */
+/**
+ * <p>
+ *   Counts the number of times that the specified object occurs
+ *   in autorelease pools in the current thread.
+ * </p>
+ * <p>
+ *   This method is <em>slow</em> and should probably only be
+ *   used for debugging purposes.
+ * </p>
+ */
 + (unsigned) autoreleaseCountForObject: (id)anObject
 {
   unsigned count = 0;
@@ -247,11 +347,19 @@ static IMP	initImp;
   return count;
 }
 
+/**
+ * Return the currently active autorelease pool.
+ */
 + (id) currentPool
 {
   return ARP_THREAD_VARS->current_pool;
 }
 
+/**
+ * Adds the specified object to the current autorelease pool.
+ * If there is no autorelease pool in the thread,
+ * a warning is logged and the object is leaked (ie it will not be released).
+ */
 + (void) addObject: (id)anObj
 {
   NSAutoreleasePool	*pool = ARP_THREAD_VARS->current_pool;
@@ -278,6 +386,9 @@ static IMP	initImp;
     }
 }
 
+/**
+ * Adds the specified object to this autorelease pool.
+ */
 - (void) addObject: (id)anObj
 {
   /* If the global, static variable AUTORELEASE_ENABLED is not set,
@@ -329,6 +440,9 @@ static IMP	initImp;
   _released_count++;
 }
 
+/**
+ * Raises an exception ... pools should not be retained.
+ */
 - (id) retain
 {
   [NSException raise: NSGenericException
@@ -336,6 +450,9 @@ static IMP	initImp;
   return self;
 }
 
+/**
+ * Destroys the receiver (calls -dealloc).
+ */
 - (oneway void) release
 {
   [self dealloc];
@@ -360,7 +477,7 @@ static IMP	initImp;
      releasing. */
   {
     struct autorelease_array_list *released = _released_head;
-    int i;
+    unsigned int i;
 
     while (released)
       {
@@ -416,6 +533,9 @@ static IMP	initImp;
   [super dealloc];
 }
 
+/**
+ * Raises an exception - pools should not be autoreleased.
+ */
 - (id) autorelease
 {
   [NSException raise: NSGenericException
@@ -423,6 +543,11 @@ static IMP	initImp;
   return self;
 }
 
+/**
+ * Destroys all the autorelease pools in the thread.
+ * You should not call this directly, it's called automatically
+ * when a thread exits.
+ */
 + (void) _endThread: (NSThread*)thread
 {
   struct autorelease_thread_vars *tv;
@@ -439,29 +564,73 @@ static IMP	initImp;
   free_pool_cache(tv);
 }
 
+/**
+ * Resets (to zero) the count of autoreleased objects in the current thread.
+ */
 + (void) resetTotalAutoreleasedObjects
 {
   ARP_THREAD_VARS->total_objects_count = 0;
 }
 
+/**
+ * Returns the number of objects which have been autoreleased in the
+ * current thread since the last call to +resetTotalAutoreleasedObjects.<br />
+ * NB. This is not normally supported ... enable it as a compile time option
+ * by editing NSAutoreleasePool.m when building the base library.
+ */
 + (unsigned) totalAutoreleasedObjects
 {
   return ARP_THREAD_VARS->total_objects_count;
 }
 
+/**
+ * <p>
+ *   Specifies whether objects contained in autorelease pools are to
+ *   be released when the pools are deallocated (by default YES).
+ * </p>
+ * <p>
+ *   You can set this to NO for debugging purposes.
+ * </p>
+ */
 + (void) enableRelease: (BOOL)enable
 {
   autorelease_enabled = enable;
 }
 
+/**
+ * <p>
+ *   When autorelease pools are deallocated, the memory they used
+ *   is retained in a cache for re-use so that new polls can be
+ *   created very quickly.
+ * </p>
+ * <p>
+ *   This method may be used to empty that cache, ensuring that
+ *   the minimum memory is used by the application.
+ * </p>
+ */
 + (void) freeCache
 {
   free_pool_cache(ARP_THREAD_VARS);
 }
 
+/**
+ * <p>
+ *   Specifies a limit to the number of objects that may be added to
+ *   an autorelease pool.  When this limit is reached an exception is
+ *   raised.
+ * </p>
+ * <p>
+ *   You can set this to a smallish value to catch problems with code
+ *   that autoreleases too many objects to operate efficiently.
+ * </p>
+ * <p>
+ *   Default value is maxint.
+ * </p>
+ */
 + (void) setPoolCountThreshhold: (unsigned)c
 {
   pool_count_warning_threshhold = c;
 }
 
 @end
+
