@@ -27,29 +27,29 @@
 
 
    <title>NSBundle class reference</title>
-   $Date: 2002/02/26 18:06:51 $ $Revision: 1.101 $
+   $Date: 2004/02/27 14:08:00 $ $Revision: 1.128 $
 */
 
-#include <config.h>
-#include <base/preface.h>
-#include <Foundation/objc-load.h>
-#include <Foundation/NSBundle.h>
-#include <Foundation/NSException.h>
-#include <Foundation/NSString.h>
-#include <Foundation/NSArray.h>
-#include <Foundation/NSDebug.h>
-#include <Foundation/NSDictionary.h>
-#include <Foundation/NSProcessInfo.h>
-#include <Foundation/NSObjCRuntime.h>
-#include <Foundation/NSUserDefaults.h>
-#include <Foundation/NSNotification.h>
-#include <Foundation/NSLock.h>
-#include <Foundation/NSMapTable.h>
-#include <Foundation/NSAutoreleasePool.h>
-#include <Foundation/NSFileManager.h>
-#include <Foundation/NSPathUtilities.h>
-#include <Foundation/NSValue.h>
-#if HAVE_UNISTD_H
+#include "config.h"
+#include "GNUstepBase/preface.h"
+#include "objc-load.h"
+#include "Foundation/NSBundle.h"
+#include "Foundation/NSException.h"
+#include "Foundation/NSString.h"
+#include "Foundation/NSArray.h"
+#include "Foundation/NSDebug.h"
+#include "Foundation/NSDictionary.h"
+#include "Foundation/NSProcessInfo.h"
+#include "Foundation/NSObjCRuntime.h"
+#include "Foundation/NSUserDefaults.h"
+#include "Foundation/NSNotification.h"
+#include "Foundation/NSLock.h"
+#include "Foundation/NSMapTable.h"
+#include "Foundation/NSAutoreleasePool.h"
+#include "Foundation/NSFileManager.h"
+#include "Foundation/NSPathUtilities.h"
+#include "Foundation/NSValue.h"
+#ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif
 #include <string.h>
@@ -87,6 +87,10 @@ static NSBundle		*_gnustep_bundle = nil;
 static NSRecursiveLock	*load_lock = nil;
 static BOOL		_strip_after_loading = NO;
 
+/* List of framework linked in the _loadingBundle */
+static NSMutableArray	*_loadingFrameworks = nil;
+static NSString         *_currentFrameworkName = nil;
+
 static NSString	*gnustep_target_dir = 
 #ifdef GNUSTEP_TARGET_DIR
   @GNUSTEP_TARGET_DIR;
@@ -117,7 +121,8 @@ static NSString	*library_combo =
 const char *
 objc_executable_location (void)
 {
-  return [[_executable_path stringByDeletingLastPathComponent] cString];
+  return [[_executable_path stringByDeletingLastPathComponent] 
+	     fileSystemRepresentation];
 }
 
 static BOOL
@@ -218,13 +223,13 @@ _bundle_name_first_match(NSString* directory, NSString* name)
 }
 
 @interface NSBundle (Private)
-+ (void) _addFrameworkFromClass:(Class)frameworkClass;
++ (void) _addFrameworkFromClass: (Class)frameworkClass;
 - (NSArray *) _bundleClasses;
 @end
 
 @implementation NSBundle (Private)
 
-/* Nicola:
+/* Nicola & Mirko:
 
    Frameworks can be used in an application in two different ways:
 
@@ -242,17 +247,23 @@ _bundle_name_first_match(NSString* directory, NSString* name)
    disk, and we have no way of knowing what the class list is, or the
    version.  So the trick we use in this case to work around those
    problems is that gnustep-make generates a 'NSFramework_xxx' class
-   and compiles it into each framework.  By asking the class, we can
-   get the version information, the list of classes which were
-   compiled into the framework, and the path were the framework bundle
-   was supposed to be installed (this is not completely reliable as
-   the programmer might change it later on by specifying a different
+   and compiles it into each framework.  By asking to the class, we
+   can get the version information and the list of classes which were
+   compiled into the framework.  To get the location of the framework
+   on disk, we try using advanced dynamic linker features to get the
+   shared object file on disk from which the NSFramework_xxx class was
+   loaded.  If that doesn't work, because the dynamic linker can't
+   provide this information on this platform (or maybe because the
+   framework was statically linked into the application), we have a
+   fallback trick :-) we can ask to the NSFramework_xxx class the path
+   to were the framework bundle was supposed to be installed (this is
+   recorded the first time when the NSFramework_xxx class is generated
+   at compile time; it's not completely reliable - for example the
+   programmer might change it later on by specifying a different
    GNUSTEP_INSTALLATION_DIR when installing ... (TODO - modify
    gnustep-make so that it issues a warning if you do that, suggesting
-   you to recompile before installing!)). (FIXME/TODO: If the linker
-   supports it, we should be asking the path on disk to the shared
-   object containing the NSFramework_xxx class ... and we can very
-   reliably guess the framework path from that path!!!!).
+   you to recompile before installing!)), and search for the framework
+   there.
 
    So at startup, we scan all classes which were compiled into the
    application.  For each NSFramework_ class, we call the following
@@ -260,79 +271,168 @@ _bundle_name_first_match(NSString* directory, NSString* name)
    the classes belonging to it, and tries to determine the path
    on disk to the framework bundle.
 
+   Bundles (and frameworks if dynamically loaded as bundles) could
+   depend on other frameworks (linked togheter on platform that
+   supports this behaviour) so whenever we dynamically load a bundle,
+   we need to spot out any additional NSFramework_* classes which are
+   loaded, and call this method (exactly as for frameworks linked into
+   the main application) to record them, and try finding the path on
+   disk to those framework bundles.
+
 */
-+ (void) _addFrameworkFromClass:(Class)frameworkClass
++ (void) _addFrameworkFromClass: (Class)frameworkClass
 {
-  NSBundle	 *bundle;
+  NSBundle	*bundle = nil;
   NSString	**fmClasses;
-  NSString	 *bundlePath = nil;
-  int		  len;
+  NSString	*bundlePath = nil;
+  unsigned int	len;
 
   if (frameworkClass == Nil)
     {
       return;
     }
 
-  /* FIXME NICOLA :-) Hmmm ... consider the case that the framework
-     was linked (using xxx_LIBRARIES_DEPEND_UPON) into a bundle.  */
-  if (_loadingBundle != nil)
-    {
-      return;
-    }
-  
   len = strlen (frameworkClass->name);
-  
+
   if (len > 12 * sizeof(char)
-      && !strncmp("NSFramework_", frameworkClass->name, 12))
+    && !strncmp("NSFramework_", frameworkClass->name, 12))
     {
-      NSString *varEnv, *path, *name;
-      
       /* The name of the framework.  */
-      name = [NSString stringWithCString: &frameworkClass->name[12]];
-      
-      /* Create the framework bundle if the bundle was linked into the 
-	 application.  */
+      NSString *name = [NSString stringWithCString: &frameworkClass->name[12]];
 
-      /* NICOLA: I think if we could get the path to the NSFramework_
-	 class from the linker, then we could build the framework path
-	 reliably.  I consider the following only a hack.  */
+      /* Try getting the path to the framework using the dynamic
+       * linker.  When it works it's really cool :-) This is the only
+       * really universal way of getting the framework path ... we can
+       * locate the framework no matter where it is on disk!
+       */
+      bundlePath = objc_get_symbol_path (frameworkClass, NULL);
 
-      /* varEnv is something like GNUSTEP_LOCAL_ROOT.  */
-      varEnv = [frameworkClass frameworkEnv];
-      if (varEnv != nil  &&  [varEnv length] > 0)
+      if ([bundlePath isEqualToString: _executable_path])
 	{
-	  /* FIXME - I don't think we should be reading it from
-	     the environment directly!  */
-	  bundlePath = [[[NSProcessInfo processInfo] environment]
-			 objectForKey: varEnv];
+	  /* Ops ... the NSFramework_xxx class is linked in the main
+	   * executable.  Maybe the framework was statically linked
+	   * into the application ... resort to searching the
+	   * framework bundle on the filesystem manually.
+	   */
+	  bundlePath = nil;
 	}
       
-      /* FIXME - path is something like ?.  */
-      path = [frameworkClass frameworkPath];
-      if (path && [path length])
+      if (bundlePath != nil)
 	{
-	  if (bundlePath)
+	  NSString *pathComponent;
+
+	  /* Dereference symlinks, and standardize path.  This will
+	   * only work properly if the original bundlePath is
+	   * absolute.  This should normally be the case if, as
+	   * recommended, you use only absolute paths in
+	   * LD_LIBRARY_PATH.
+	   */
+	  bundlePath = [bundlePath stringByStandardizingPath];
+
+	  /* We now have the location of the shared library object
+	   * file inside the framework directory.  We need to walk up
+	   * the directory tree up to the top of the framework.  To do
+	   * so, we need to chop off the extra subdirectories, the
+	   * library combo and the target cpu/os if they exist.  The
+	   * framework and this library should match so we can use the
+	   * compiled-in settings.
+	   */
+	  /* library name */
+	  bundlePath = [bundlePath stringByDeletingLastPathComponent];
+	  /* library combo */
+	  pathComponent = [bundlePath lastPathComponent];
+	  if ([pathComponent isEqual: library_combo])
 	    {
-	      bundlePath = [bundlePath stringByAppendingPathComponent: path];
+	      bundlePath = [bundlePath stringByDeletingLastPathComponent];
+	    }
+	  /* target os */
+	  pathComponent = [bundlePath lastPathComponent];
+	  if ([pathComponent isEqual: gnustep_target_os])
+	    {
+	      bundlePath = [bundlePath stringByDeletingLastPathComponent];
+	    }
+	  /* target cpu */
+	  pathComponent = [bundlePath lastPathComponent];
+	  if ([pathComponent isEqual: gnustep_target_cpu])
+	    {
+	      bundlePath = [bundlePath stringByDeletingLastPathComponent];
+	    }
+	  /* version name */
+	  bundlePath = [bundlePath stringByDeletingLastPathComponent];
+
+	  pathComponent = [bundlePath lastPathComponent];
+          if ([pathComponent isEqual: @"Versions"])
+	    {
+	      bundlePath = [bundlePath stringByDeletingLastPathComponent];
+	      pathComponent = [bundlePath lastPathComponent];
+	      
+	      if ([pathComponent isEqualToString: 
+				   [NSString stringWithFormat: @"%@%@",
+					     name, @".framework"]])
+		{
+		  /* Try creating the bundle.  */
+		  bundle = [[self alloc] initWithPath: bundlePath];
+		}
+	    }
+	  
+	  /* Failed - buu - try the fallback trick.  */
+	  if (bundle == nil)
+	    {
+	      bundlePath = nil;
+	    }
+	}
+
+      if (bundlePath == nil)
+	{
+	  /* NICOLA: In an ideal world, the following is just a hack
+	   * for when objc_get_symbol_path() fails!  But in real life
+	   * objc_get_symbol_path() is risky (some platforms don't
+	   * have it at all!), so this hack might be used a lot!  It
+	   * must be quite robust.  We try to look for the framework
+	   * in the standard GNUstep installation dirs.  This should
+	   * be reasonably safe if the user is not being too clever
+	   * ... :-)
+	  */
+	  NSString *varEnv, *path;
+	  
+	  /* varEnv is something like GNUSTEP_LOCAL_ROOT.  */
+	  varEnv = [frameworkClass frameworkEnv];
+	  if (varEnv != nil  &&  [varEnv length] > 0)
+	    {
+	      /* FIXME - I don't think we should be reading it from
+		 the environment directly!  */
+	      bundlePath = [[[NSProcessInfo processInfo] environment]
+			     objectForKey: varEnv];
+	    }
+	  
+	  /* FIXME - path is something like ?.  */
+	  path = [frameworkClass frameworkPath];
+	  if (path && [path length])
+	    {
+	      if (bundlePath)
+		{
+		  bundlePath = [bundlePath stringByAppendingPathComponent: 
+					     path];
+		}
+	      else
+		{
+		  bundlePath = path;
+		}
 	    }
 	  else
 	    {
-	      bundlePath = path;
+	      bundlePath = [bundlePath stringByAppendingPathComponent: 
+					 @"Library/Frameworks"];
 	    }
+	  
+	  bundlePath = [bundlePath stringByAppendingPathComponent:
+				     [NSString stringWithFormat: 
+						 @"%@.framework", 
+					       name]];
+
+	  /* Try creating the bundle.  */
+	  bundle = [[self alloc] initWithPath: bundlePath];
 	}
-      else
-	{
-	  bundlePath = [bundlePath stringByAppendingPathComponent: 
-				     @"Library/Frameworks"];
-	}
-      
-      bundlePath = [bundlePath stringByAppendingPathComponent:
-				 [NSString stringWithFormat: 
-					     @"%@.framework", 
-					   name]];
-      
-      /* Try creating the bundle.  */
-      bundle = [[self alloc] initWithPath: bundlePath];
       
       if (bundle == nil)
 	{
@@ -340,8 +440,7 @@ _bundle_name_first_match(NSString* directory, NSString* name)
 	     location.  NICOLA: We should be trying to locate it in
 	     some other obvious location, iterating over
 	     user/network/local/system dirs ...  TODO.  */
-	  NSLog (@"Problem locating framework locating for framework %@",
-		 name);
+	  NSLog (@"Problem locating framework %@", name);
 	  return;
 	}
       
@@ -354,18 +453,33 @@ _bundle_name_first_match(NSString* directory, NSString* name)
       /* A NULL terminated list of class names - the classes contained
 	 in the framework.  */
       fmClasses = [frameworkClass frameworkClasses];
-      
+
       while (*fmClasses != NULL)
 	{
 	  NSValue *value;
 	  Class    class = NSClassFromString(*fmClasses);
-	  
+
 	  value = [NSValue valueWithNonretainedObject: class];
 	  
 	  [(NSMutableArray *)[bundle _bundleClasses] addObject: value];
 	  
 	  fmClasses++;
 	}
+
+      /* If _loadingBundle is not nil, it means we reached this point
+       * while loading a bundle.  This can happen if the framework is
+       * linked into the bundle (then, the dynamic linker
+       * automatically drags in the framework when the bundle is
+       * loaded).  But then, the classes in the framework should be
+       * removed from the list of classes in the bundle. Check that
+       * _loadingBundle != bundle which happens on Windows machines when
+       * loading in Frameworks.
+       */
+      if (_loadingBundle != nil && _loadingBundle != bundle)
+	{
+	  [(NSMutableArray *)[_loadingBundle _bundleClasses]
+			     removeObjectsInArray: [bundle _bundleClasses]];
+	}
     }
 }
 
@@ -376,10 +490,20 @@ _bundle_name_first_match(NSString* directory, NSString* name)
 
 @end
 
+/*
+  Mirko:
+
+  The gnu-runtime calls the +load method of each class before the
+  _bundle_load_callback() is called and we can't provide the list of classes
+  ready for this method.
+
+ */
+
 void
 _bundle_load_callback(Class theClass, struct objc_category *theCategory)
 {
   NSCAssert(_loadingBundle, NSInternalInconsistencyException);
+  NSCAssert(_loadingFrameworks, NSInternalInconsistencyException);
 
   /* We never record categories - if this is a category, just do nothing.  */
   if (theCategory != 0)
@@ -388,10 +512,22 @@ _bundle_load_callback(Class theClass, struct objc_category *theCategory)
     }
 
   /* Don't store the internal NSFramework_xxx class into the list of
-     bundle classes.  */
+     bundle classes, but store the linked frameworks in _loadingFrameworks  */
   if (strlen (theClass->name) > 12   &&  !strncmp ("NSFramework_", 
 						   theClass->name, 12))
     {
+      if (_currentFrameworkName)
+	{
+	  const char *frameworkName;
+
+	  frameworkName = [_currentFrameworkName cString];
+
+	  if (!strcmp(theClass->name, frameworkName))
+	    return;
+	}
+
+      [_loadingFrameworks
+	addObject: [NSValue valueWithNonretainedObject: (id)theClass]];
       return; 
     }
 
@@ -451,11 +587,29 @@ _bundle_load_callback(Class theClass, struct objc_category *theCategory)
 	  _executable_path = [[NSFileManager defaultManager]
 	    pathContentOfSymbolicLinkAtPath:
               [NSString stringWithCString: PROCFS_EXE_LINK]];
+
+	  /*
+	  On some systems, the link is of the form "[device]:inode", which
+	  can be used to open the executable, but is useless if you want
+	  the path to it. Thus we check that the path is an actual absolute
+	  path. (Using '/' here is safe; it isn't the path separator
+	  everywhere, but it is on all systems that have PROCFS_EXE_LINK.)
+	  */
+	  if ([_executable_path length] > 0 &&
+	      [_executable_path characterAtIndex: 0] != '/')
+	    {
+	      _executable_path = nil;
+	    }
 #endif
 	  if (_executable_path == nil || [_executable_path length] == 0)
 	    {
+	      const char	*tmp;
+
 	      _executable_path =
 		[[[NSProcessInfo processInfo] arguments] objectAtIndex: 0];
+	      tmp = [_executable_path UTF8String];
+	      _executable_path = [[NSFileManager defaultManager]
+		stringWithFileSystemRepresentation: tmp length: strlen(tmp)];
 	      _executable_path = 
 		[self _absolutePathOfExecutable: _executable_path];
 	      NSAssert(_executable_path, NSInternalInconsistencyException);
@@ -485,10 +639,41 @@ _bundle_load_callback(Class theClass, struct objc_category *theCategory)
 	    free(classes);
 	  }
 #else
-	  while ((class = objc_next_class(&state)))
-	    {
-	      [self _addFrameworkFromClass: class];
+	  {
+	    int i, numBufClasses = 10, numClasses = 0;
+	    Class *classes;
+
+	    classes = malloc(sizeof(Class) * numBufClasses);
+
+	    while ((class = objc_next_class(&state)))
+	      {
+		unsigned int len = strlen (class->name);
+
+		if (len > 12 * sizeof(char)
+		  && !strncmp("NSFramework_", class->name, 12))
+		  {
+		    classes[numClasses++] = class;
+		  }
+		if (numClasses == numBufClasses)
+		  {
+		    Class *ptr;
+
+		    numClasses += 10;
+		    ptr = realloc(classes, sizeof(Class) * numClasses);
+
+		    if (!ptr)
+		      break;
+
+		    classes = ptr;
+		  }
 	    }
+
+	    for (i = 0; i < numClasses; i++)
+	      {
+		[self _addFrameworkFromClass: classes[i]];
+	      }
+	    free(classes);
+	  }
 #endif
 #if 0
 	  //		  _bundle_load_callback(class, NULL);
@@ -501,29 +686,46 @@ _bundle_load_callback(Class theClass, struct objc_category *theCategory)
     }
 }
 
+/**
+ * Returns an array of all the bundles which do not belong to frameworks.<br />
+ * This always contains the main bundle.
+ */
 + (NSArray *) allBundles
 {
-  NSMapEnumerator	enumerate;
   NSMutableArray	*array = [NSMutableArray arrayWithCapacity: 2];
-  void			*key;
-  NSBundle		*bundle;
 
   [load_lock lock];
-  enumerate = NSEnumerateMapTable(_bundles);
-  while (NSNextMapEnumeratorPair(&enumerate, &key, (void **)&bundle))
+  if (!_mainBundle) 
     {
-      if (bundle->_bundleType == NSBUNDLE_FRAMEWORK)
-	continue;
+      [self mainBundle];
+    }
+  if (_bundles != 0)
+    {
+      NSMapEnumerator	enumerate;
+      void		*key;
+      NSBundle		*bundle;
 
-      if ([array indexOfObjectIdenticalTo: bundle] == NSNotFound)
+      enumerate = NSEnumerateMapTable(_bundles);
+      while (NSNextMapEnumeratorPair(&enumerate, &key, (void **)&bundle))
 	{
-	  [array addObject: bundle];
+	  if (bundle->_bundleType == NSBUNDLE_FRAMEWORK)
+	    {
+	      continue;
+	    }
+	  if ([array indexOfObjectIdenticalTo: bundle] == NSNotFound)
+	    {
+	      [array addObject: bundle];
+	    }
 	}
+      NSEndMapTableEnumeration(&enumerate);
     }
   [load_lock unlock];
   return array;
 }
 
+/**
+ * Returns an array containing all the known bundles representing frameworks.
+ */
 + (NSArray *) allFrameworks
 {
   NSMapEnumerator  enumerate;
@@ -541,32 +743,34 @@ _bundle_load_callback(Class theClass, struct objc_category *theCategory)
 	  [array addObject: bundle];
 	}
     }
+  NSEndMapTableEnumeration(&enumerate);
   [load_lock unlock];
   return array;
 }
 
-/* For an application, returns the main bundle of the application.
-   For a tool, returns the main bundle associated with the tool (this
-   is experimental and not yet supported by gnustep-make, but will
-   soon be).
-   
-   For an application, the structure is as follows - 
-   
-   The executable is Gomoku.app/ix86/linux-gnu/gnu-gnu-gnu/Gomoku
-   and the main bundle directory is Gomoku.app/.
-   
-   For a tool, the structure is as follows - 
-   
-   The executable is xxx/Tools/ix86/linux-gnu/gnu-gnu-gnu/Control
-   and the main bundle directory is xxx/Tools/Resources/Control.
-   
-   (when the tool has not yet been installed, it's similar - 
-   xxx/shared_obj/ix86/linux-gnu/gnu-gnu-gnu/Control
-   and the main bundle directory is xxx/Resources/Control).
-   
-   (For a flattened structure, the structure is the same without the
-   ix86/linux-gnu/gnu-gnu-gnu directories).  */
-+ (NSBundle *)mainBundle
+/**
+ * For an application, returns the main bundle of the application.<br />
+ * For a tool, returns the main bundle associated with the tool.<br />
+ * <br />
+ * For an application, the structure is as follows - 
+ * <p> 
+ * The executable is Gomoku.app/ix86/linux-gnu/gnu-gnu-gnu/Gomoku
+ * and the main bundle directory is Gomoku.app/.
+ * </p> 
+ * For a tool, the structure is as follows - 
+ * <p>
+ * The executable is xxx/Tools/ix86/linux-gnu/gnu-gnu-gnu/Control
+ * and the main bundle directory is xxx/Tools/Resources/Control.
+ * </p> 
+ * <p>(when the tool has not yet been installed, it's similar - 
+ * xxx/shared_obj/ix86/linux-gnu/gnu-gnu-gnu/Control
+ * and the main bundle directory is xxx/Resources/Control).
+ * </p> 
+ * <p>(For a flattened structure, the structure is the same without the
+ * ix86/linux-gnu/gnu-gnu-gnu directories).
+ * </p>
+ */
++ (NSBundle *) mainBundle
 {
   [load_lock lock];
   if (!_mainBundle) 
@@ -583,7 +787,10 @@ _bundle_load_callback(Class theClass, struct objc_category *theCategory)
          the executable name here - just in case it turns out it's a
          tool.  */
       NSString *toolName = [_executable_path lastPathComponent];
-
+#if defined(__WIN32__)
+      toolName = [toolName stringByDeletingPathExtension];
+#endif
+      
       /* Strip off the name of the program */
       path = [_executable_path stringByDeletingLastPathComponent];
 
@@ -618,13 +825,17 @@ _bundle_load_callback(Class theClass, struct objc_category *theCategory)
              non-yet-installed tool.  */
 	  isApplication = NO;
 	}
-      
+
       if (isApplication == YES)
 	{
 	  s = [path lastPathComponent];
-	  if (([s hasSuffix: @".app"]  == NO)
-	      && ([s hasSuffix: @".debug"] == NO)
-	      && ([s hasSuffix: @".profile"] == NO))
+	      
+	  if ((([s hasSuffix: @".app"]  == NO)
+	    && ([s hasSuffix: @".debug"] == NO)
+	    && ([s hasSuffix: @".profile"] == NO))
+	    // GNUstep Web
+	    && (([s hasSuffix: @".gswa"] == NO)
+		&& ([s hasSuffix: @".woa"] == NO)))
 	    {
 	      isApplication = NO;
 	    }
@@ -649,7 +860,10 @@ _bundle_load_callback(Class theClass, struct objc_category *theCategory)
   return _mainBundle;
 }
 
-/* NB: We will not find a class if the bundle has not been loaded yet!  */
+/**
+ * Returns the bundle whose code contains the specified class.<br />
+ * NB: We will not find a class if the bundle has not been loaded yet!
+ */
 + (NSBundle *) bundleForClass: (Class)aClass
 {
   void*     key;
@@ -715,7 +929,7 @@ _bundle_load_callback(Class theClass, struct objc_category *theCategory)
 
   /* Check if we were already initialized for this directory */
   [load_lock lock];
-  if (_bundles) 
+  if (_bundles)
     {
       NSBundle* bundle = (NSBundle *)NSMapGet(_bundles, path);
       if (bundle)
@@ -866,46 +1080,63 @@ _bundle_load_callback(Class theClass, struct objc_category *theCategory)
   return _principalClass;
 }
 
+/**
+ * Returns YES if the receiver's code is loaded, otherwise, returns NO.
+ */
+- (BOOL) isLoaded
+{
+  return _codeLoaded;
+}
+
 - (BOOL) load
 {
-  if (self == _mainBundle || self == _gnustep_bundle) 
+  if (self == _mainBundle || self == _gnustep_bundle)
     {
       _codeLoaded = YES;
       return YES;
     }
-
   [load_lock lock];
+
   if (!_codeLoaded)
     {
-      NSString       *object, *path;
+      NSString       *object;
       NSEnumerator   *classEnumerator;
       NSMutableArray *classNames;
       NSValue        *class;
 
-      object = [[self infoDictionary] objectForKey: @"NSExecutable"];
+      object = [self executablePath];
       if (object == nil || [object length] == 0)
 	{
 	  [load_lock unlock];
 	  return NO;
 	}
-      if (_bundleType == NSBUNDLE_FRAMEWORK)
-	{
-	  path = [_path stringByAppendingPathComponent:@"Versions/Current"];
-	}
-      else
-	{
-	  path = _path;
-	}
-      object = bundle_object_name(path, object);
       _loadingBundle = self;
       _bundleClasses = RETAIN([NSMutableArray arrayWithCapacity: 2]);
-      if (objc_load_module([object cString], 
+      _loadingFrameworks = RETAIN([NSMutableArray arrayWithCapacity: 2]);
+
+      /* This code is executed twice if a class linked in the bundle call a
+	 NSBundle method inside +load (-principalClass). To avoid this we set
+	 _codeLoaded before loading the bundle. */
+      _codeLoaded = YES;
+
+      if (objc_load_module([object fileSystemRepresentation], 
 			   stderr, _bundle_load_callback, NULL, NULL))
 	{
+	  _codeLoaded = NO;
+	  DESTROY(_loadingFrameworks);
+	  DESTROY(_currentFrameworkName);
 	  [load_lock unlock];
 	  return NO;
 	}
-      _codeLoaded = YES;
+
+      /* We now construct the list of bundles from frameworks linked with
+	 this one */
+      classEnumerator = [_loadingFrameworks objectEnumerator];
+      while ((class = [classEnumerator nextObject]) != nil)
+	{
+	  [NSBundle _addFrameworkFromClass: [class nonretainedObjectValue]];
+	}
+
       /* After we load code from a bundle, we retain the bundle until
 	 we unload it (because we never unload bundles, that is
 	 forever).  The reason why we retain it is that we need it!
@@ -914,6 +1145,9 @@ _bundle_load_callback(Class theClass, struct objc_category *theCategory)
 	 +allBundles.  */
       RETAIN (self);
       _loadingBundle = nil;
+
+      DESTROY(_loadingFrameworks);
+      DESTROY(_currentFrameworkName);
       
       classNames = [NSMutableArray arrayWithCapacity: [_bundleClasses count]];
       classEnumerator = [_bundleClasses objectEnumerator];
@@ -941,10 +1175,10 @@ _bundle_load_callback(Class theClass, struct objc_category *theCategory)
    constructs an array of paths, where each path is a possible location
    for a resource in the bundle.  The current algorithm for searching goes:
 
-     <main bundle>/Resources/<bundlePath>
-     <main bundle>/Resources/<bundlePath>/<language.lproj>
-     <main bundle>/<bundlePath>
-     <main bundle>/<bundlePath>/<language.lproj>
+     <rootPath>/Resources/<bundlePath>
+     <rootPath>/Resources/<bundlePath>/<language.lproj>
+     <rootPath>/<bundlePath>
+     <rootPath>/<bundlePath>/<language.lproj>
 */
 + (NSArray *) _bundleResourcePathsWithRootPath: (NSString *)rootPath
 				       subPath: (NSString *)bundlePath
@@ -1000,8 +1234,8 @@ _bundle_load_callback(Class theClass, struct objc_category *theCategory)
       if (ext && [ext length] != 0)
 	{
 	  fullpath = [path stringByAppendingPathComponent:
-	    [NSString stringWithFormat: @"%@.%@", name, ext]];
-	  if ( bundle_file_readable(fullpath) )
+			     [NSString stringWithFormat: @"%@.%@", name, ext]];
+	  if (bundle_file_readable(fullpath))
 	    {
 	      if (gnustep_target_os)
 		{
@@ -1087,6 +1321,13 @@ _bundle_load_callback(Class theClass, struct objc_category *theCategory)
 		   withVersion: _version];
 }
 
++ (NSArray*) pathsForResourcesOfType: (NSString*)extension
+			 inDirectory: (NSString*)bundlePath
+{
+  [self notImplemented: _cmd];
+  return nil;
+}
+
 - (NSArray *) pathsForResourcesOfType: (NSString *)extension
 			  inDirectory: (NSString *)bundlePath
 {
@@ -1096,6 +1337,8 @@ _bundle_load_callback(Class theClass, struct objc_category *theCategory)
   NSEnumerator *pathlist;
   NSFileManager	*mgr = [NSFileManager defaultManager];
     
+  /* Not sure if this is correct since it will only search in 
+     lprojs that are user preferences. FIXME. */
   pathlist = [[NSBundle _bundleResourcePathsWithRootPath: [self bundlePath]
 			subPath: bundlePath] objectEnumerator];
   resources = [NSMutableArray arrayWithCapacity: 2];
@@ -1117,6 +1360,106 @@ _bundle_load_callback(Class theClass, struct objc_category *theCategory)
   return resources;
 }
 
+- (NSArray*) pathsForResourcesOfType: (NSString*)extension
+			 inDirectory: (NSString*)bundlePath
+		     forLocalization: (NSString*)localizationName
+{
+  [self notImplemented: _cmd];
+  return nil;
+}
+
+- (NSString*) pathForResource: (NSString*)name
+		       ofType: (NSString*)ext
+		  inDirectory: (NSString*)bundlePath
+	      forLocalization: (NSString*)localizationName
+{
+  [self notImplemented: _cmd];
+  return nil;
+}
+
++ (NSArray *) preferredLocalizationsFromArray: (NSArray *)localizationsArray
+{
+  return [self preferredLocalizationsFromArray: localizationsArray
+	         forPreferences: [NSUserDefaults userLanguages]];
+}
+
++ (NSArray *) preferredLocalizationsFromArray: (NSArray *)localizationsArray 
+			       forPreferences: (NSArray *)preferencesArray
+{
+  NSString *locale;
+  NSMutableArray* array;
+  NSEnumerator* enumerate;
+
+  array = [NSMutableArray arrayWithCapacity: 2];
+  enumerate = [preferencesArray objectEnumerator];
+  while ((locale = [enumerate nextObject]))
+    {
+      if ([localizationsArray indexOfObject: locale] != NSNotFound)
+	[array addObject: locale];
+    }
+  /* I guess this is arbitrary if we can't find a match? */
+  if ([array count] == 0 && [localizationsArray count] > 0)
+    [array addObject: [localizationsArray objectAtIndex: 0]];
+  return [array makeImmutableCopyOnFail: NO];
+}
+
+- (NSDictionary *)localizedInfoDictionary
+{
+  NSString  *path;
+  NSArray   *locales;
+  NSString  *locale = nil;
+  NSDictionary *dict = nil;
+
+  locales = [self preferredLocalizations];
+  if ([locales count] > 0)
+    locale = [locales objectAtIndex: 0];
+  path = [self pathForResource: @"Info-gnustep" 
+                        ofType: @"plist"
+                   inDirectory: nil 
+               forLocalization: locale];
+  if (path)
+    {
+      dict = [[NSDictionary alloc] initWithContentsOfFile: path];
+    }
+  else
+    {
+      path = [self pathForResource: @"Info" 
+                            ofType: @"plist"
+                       inDirectory: nil 
+                   forLocalization: locale];
+      if (path)
+	{
+	  dict = [[NSDictionary alloc] initWithContentsOfFile: path];
+	}
+    }
+  if (dict == nil)
+    dict = [self infoDictionary];
+  return dict;
+}
+
+- (NSArray *)localizations
+{
+  NSString *locale;
+  NSArray *localizations;
+  NSEnumerator* enumerate;
+  NSMutableArray *array = [NSMutableArray arrayWithCapacity: 2];
+
+  localizations = [self pathsForResourcesOfType: @"lproj"
+	                            inDirectory: nil];
+  enumerate = [localizations objectEnumerator];
+  while ((locale = [enumerate nextObject]))
+    {
+      locale = [[locale lastPathComponent] stringByDeletingPathExtension];
+      [array addObject: locale];
+    }
+  return [array makeImmutableCopyOnFail: NO];
+}
+
+- (NSArray *)preferredLocalizations
+{
+  return [NSBundle preferredLocalizationsFromArray: [self localizations]];
+}
+
 - (NSString *) localizedStringForKey: (NSString *)key	
 			       value: (NSString *)value
 			       table: (NSString *)tableName
@@ -1169,8 +1512,10 @@ _bundle_load_callback(Class theClass, struct objc_category *theCategory)
 	  NS_ENDHANDLER
 	}
       else
-	NSLog(@"Failed to locate strings file %@", tableName);
-	
+	{
+	  NSDebugMLLog(@"NSBundle", @"Failed to locate strings file %@",
+	    tableName);
+	}
       /*
        * If we couldn't found and parsed the strings table, we put it in
        * the cache of strings tables in this bundle, otherwise we will just
@@ -1208,6 +1553,31 @@ _bundle_load_callback(Class theClass, struct objc_category *theCategory)
   _strip_after_loading = flag;
 }
 
+- (NSString *)executablePath
+{
+  NSString *object, *path;
+  object = [[self infoDictionary] objectForKey: @"NSExecutable"];
+  if (object == nil || [object length] == 0)
+    {
+      return nil;
+    }
+  if (_bundleType == NSBUNDLE_FRAMEWORK)
+    {
+      path = [_path stringByAppendingPathComponent:@"Versions/Current"];
+      
+      _currentFrameworkName = RETAIN(([NSString stringWithFormat:
+						  @"NSFramework_%@",
+						object]));
+    }
+  else
+    {
+      path = _path;
+    }
+  
+  object = bundle_object_name(path, object);
+  return object;
+}
+
 - (NSString *) resourcePath
 {
   NSString *version = _frameworkVersion;
@@ -1236,26 +1606,34 @@ _bundle_load_callback(Class theClass, struct objc_category *theCategory)
 
   path = [self pathForResource: @"Info-gnustep" ofType: @"plist"];
   if (path)
-    _infoDict = [[NSDictionary alloc] initWithContentsOfFile: path];
+    {
+      _infoDict = [[NSDictionary alloc] initWithContentsOfFile: path];
+    }
   else
     {
       path = [self pathForResource: @"Info" ofType: @"plist"];
       if (path)
-	_infoDict = [[NSDictionary alloc] initWithContentsOfFile: path];
+	{
+	  _infoDict = [[NSDictionary alloc] initWithContentsOfFile: path];
+	}
       else
-	_infoDict = RETAIN([NSDictionary dictionary]);
+	{
+	  _infoDict = RETAIN([NSDictionary dictionary]);
+	}
     }
   return _infoDict;
 }
 
+- (NSString *)bundleIdentifier
+{
+    return [[self infoDictionary] objectForKey:@"CFBundleIdentifier"];
+}
+
 - (unsigned)bundleVersion
 {
   return _version;
 }
 
-/* Since I don't know how version numbers should behave - the version
-   number is not used. (FIXME)
-*/
 - (void)setBundleVersion:(unsigned)version
 {
   _version = version;
@@ -1372,7 +1750,7 @@ _bundle_load_callback(Class theClass, struct objc_category *theCategory)
   return _gnustep_bundle;
 }
 
-+ (NSString *) pathForGNUstepResource: (NSString *)name
++ (NSString *) pathForLibraryResource: (NSString *)name
 			       ofType: (NSString *)ext	
 			  inDirectory: (NSString *)bundlePath
 {
@@ -1383,7 +1761,7 @@ _bundle_load_callback(Class theClass, struct objc_category *theCategory)
   NSEnumerator	*enumerator;
 
   /* Gather up the paths */
-  paths = NSSearchPathForDirectoriesInDomains(GSLibrariesDirectory,
+  paths = NSSearchPathForDirectoriesInDomains(NSLibraryDirectory,
                                               NSAllDomainsMask, YES);
 
   enumerator = [paths objectEnumerator];
@@ -1398,6 +1776,16 @@ _bundle_load_callback(Class theClass, struct objc_category *theCategory)
   return path;
 }
 
++ (NSString *) pathForGNUstepResource: (NSString *)name
+			       ofType: (NSString *)ext	
+			  inDirectory: (NSString *)bundlePath
+{
+  NSLog(@"Warning: Deprecated method %@ called. Use +pathForLibraryResource:ofType:inDirectory: or +bundleForLibrary: instead",
+	NSStringFromSelector(_cmd));
+  return [self pathForLibraryResource: name ofType: ext 
+	                  inDirectory: bundlePath];
+}
+
 + (NSString*) _gnustep_target_cpu
 {
   return gnustep_target_cpu;
