@@ -21,13 +21,13 @@
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA.
 
    <title>NSClassDescription class reference</title>
-   $Date: 2002/02/12 12:33:57 $ $Revision: 1.6 $
+   $Date: 2003/06/07 01:24:40 $ $Revision: 1.7 $
 */
 
-#include <Foundation/NSClassDescription.h>
-#include <Foundation/NSLock.h>
-#include <Foundation/NSMapTable.h>
-#include <Foundation/NSNotification.h>
+#include "Foundation/NSClassDescription.h"
+#include "Foundation/NSLock.h"
+#include "Foundation/NSMapTable.h"
+#include "Foundation/NSNotification.h"
 
 
 @implementation NSClassDescription
