@@ -22,15 +22,14 @@
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA.
 
    <title>NSCoder class reference</title>
-   $Date: 2001/12/18 16:54:14 $ $Revision: 1.34 $
+   $Date: 2004/01/25 18:39:20 $ $Revision: 1.40 $
    */
 
-#include <config.h>
-#include <base/preface.h>
-#include <base/behavior.h>
-#include <Foundation/NSData.h>
-#include <Foundation/NSCoder.h>
-#include <Foundation/NSSerialization.h>
+#include "config.h"
+#include "GNUstepBase/preface.h"
+#include "Foundation/NSData.h"
+#include "Foundation/NSCoder.h"
+#include "Foundation/NSSerialization.h"
 
 @implementation NSCoder
 
@@ -274,6 +273,140 @@
   va_end(ap);
 }
 
+// Managing Zones
+
+- (NSZone*) objectZone
+{
+  return NSDefaultMallocZone();
+}
+
+- (void) setObjectZone: (NSZone*)zone
+{
+  ;
+}
+
+
+// Getting a Version
+
+- (unsigned int) systemVersion
+{
+  return (((GNUSTEP_BASE_MAJOR_VERSION * 100)
+    + GNUSTEP_BASE_MINOR_VERSION) * 100) + GNUSTEP_BASE_SUBMINOR_VERSION;
+}
+
+
+// Keyed archiving extensions
+
+- (BOOL) allowsKeyedCoding
+{
+  return NO;
+}
+
+- (BOOL) containsValueForKey: (NSString*)aKey
+{
+  [self subclassResponsibility: _cmd];
+  return NO;
+}
+
+- (BOOL) decodeBoolForKey: (NSString*)aKey
+{
+  [self subclassResponsibility: _cmd];
+  return NO;
+}
+
+- (const uint8_t*) decodeBytesForKey: (NSString*)aKey
+		      returnedLength: (unsigned*)alength
+{
+  [self subclassResponsibility: _cmd];
+  return 0;
+}
+
+- (double) decodeDoubleForKey: (NSString*)aKey
+{ 
+  [self subclassResponsibility: _cmd];
+  return 0.0;
+}
+
+- (float) decodeFloatForKey: (NSString*)aKey
+{ 
+  [self subclassResponsibility: _cmd];
+  return 0.0;
+}
+
+- (int) decodeIntForKey: (NSString*)aKey
+{ 
+  [self subclassResponsibility: _cmd];
+  return 0;
+}
+
+- (int32_t) decodeInt32ForKey: (NSString*)aKey
+{ 
+  [self subclassResponsibility: _cmd];
+  return 0;
+}
+
+- (int64_t) decodeInt64ForKey: (NSString*)aKey
+{ 
+  [self subclassResponsibility: _cmd];
+  return 0;
+}
+
+- (id) decodeObjectForKey: (NSString*)aKey
+{ 
+  [self subclassResponsibility: _cmd];
+  return nil;
+}
+
+- (void) encodeBool: (BOOL) aBool forKey: (NSString*)aKey
+{
+  [self subclassResponsibility: _cmd];
+}
+
+- (void) encodeBytes: (const uint8_t*)aPointer
+	      length: (unsigned)length
+	      forKey: (NSString*)aKey
+{
+  [self subclassResponsibility: _cmd];
+}
+
+- (void) encodeConditionalObject: (id)anObject forKey: (NSString*)aKey
+{
+  [self subclassResponsibility: _cmd];
+}
+
+- (void) encodeDouble: (double)aDouble forKey: (NSString*)aKey
+{
+  [self subclassResponsibility: _cmd];
+}
+
+- (void) encodeFloat: (float)aFloat forKey: (NSString*)aKey
+{
+  [self subclassResponsibility: _cmd];
+}
+
+- (void) encodeInt: (int)anInteger forKey: (NSString*)aKey
+{
+  [self subclassResponsibility: _cmd];
+}
+
+- (void) encodeInt32: (int32_t)anInteger forKey: (NSString*)aKey
+{
+  [self subclassResponsibility: _cmd];
+}
+
+- (void) encodeInt64: (int64_t)anInteger forKey: (NSString*)aKey
+{
+  [self subclassResponsibility: _cmd];
+}
+
+- (void) encodeObject: (id)anObject forKey: (NSString*)aKey
+{
+  [self subclassResponsibility: _cmd];
+}
+
+@end
+
+@implementation NSCoder (GNUstep)
 - (void) decodeArrayOfObjCType: (const char*)type
 		         count: (unsigned)count
 			    at: (void*)buf
@@ -356,25 +489,85 @@
   [self encodeValueOfObjCType: @encode(id) at: anObject withName: name];
 }
 
-// Managing Zones
-
-- (NSZone*) objectZone
-{
-  return NSDefaultMallocZone();
-}
-
-- (void) setObjectZone: (NSZone*)zone
-{
-  ;
-}
-
-
-// Getting a Version
-
-- (unsigned int) systemVersion
-{
-  return (((GNUSTEP_BASE_MAJOR_VERSION * 100)
-    + GNUSTEP_BASE_MINOR_VERSION) * 100) + GNUSTEP_BASE_SUBMINOR_VERSION;
-}
-
 @end
+
+
+
+#include	"GSPrivate.h"
+
+@implementation	_NSKeyedCoderOldStyleArray
+- (const void*) bytes
+{
+  return _a;
+}
+- (unsigned) count
+{
+  return _c;
+}
+- (void) dealloc
+{
+  DESTROY(_d);
+  [super dealloc];
+}
+- (id) initWithCoder: (NSCoder*)aCoder
+{
+  id		o;
+  void		*address;
+  unsigned	i;
+
+  _c = [aCoder decodeIntForKey: @"NS.count"];
+  _t[0] = (char)[aCoder decodeIntForKey: @"NS.type"];
+  _t[1] = '\0';
+
+  /*
+   * We decode the size from the remote end, but discard it as we
+   * are probably safer to use the local size of the datatype involved.
+   */
+  _s = [aCoder decodeIntForKey: @"NS.size"];
+  _s = objc_sizeof_type(_t);
+
+  _d = o = [[NSMutableData alloc] initWithLength: _c * _s];
+  _a = address = [o mutableBytes];
+  for (i = 0; i < _c; i++)
+    {
+      [aCoder decodeValueOfObjCType: _t at: address];
+      address += _s;
+    }
+  return self;
+}
+
+- (id) initWithObjCType: (const char*)t count: (int)c at: (const void*)a
+{
+  _t[0] = *t;
+  _t[1] = '\0';
+  _s = objc_sizeof_type(_t);
+  _c = c;
+  _a = a;
+  return self;
+}
+
+- (void) encodeWithCoder: (NSCoder*)aCoder
+{
+  int	i;
+
+  [aCoder encodeInt: _c forKey: @"NS.count"];
+  [aCoder encodeInt: *_t forKey: @"NS.type"];
+  [aCoder encodeInt: _s forKey: @"NS.size"];
+  for (i = 0; i < _c; i++)
+    {
+      [aCoder encodeValueOfObjCType: _t at: _a];
+      _a += _s;
+    }
+}
+
+- (unsigned) size
+{
+  return _s;
+}
+
+- (const char*) type
+{
+  return _t;
+}
+@end
+
