@@ -25,19 +25,19 @@
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA.
 */
 
-#include <config.h>
-#include <base/preface.h>
-#include <Foundation/NSObjCRuntime.h>
-#include <GSConfig.h>
-#include <Foundation/NSConcreteNumber.h>
-#include <Foundation/NSString.h>
-#include <Foundation/NSException.h>
-#include <Foundation/NSCoder.h>
-#include <Foundation/NSPortCoder.h>
-#include <Foundation/NSCoder.h>
+#include "config.h"
+#include "GNUstepBase/preface.h"
+#include "GSConfig.h"
+#include "Foundation/NSObjCRuntime.h"
+#include "Foundation/NSString.h"
+#include "Foundation/NSException.h"
+#include "Foundation/NSCoder.h"
+#include "Foundation/NSPortCoder.h"
+#include "Foundation/NSCoder.h"
+#include "NSConcreteNumber.h"
 
 /* This file should be run through a preprocessor with the macro TYPE_ORDER
-   defined to a number from 0 to 12 cooresponding to each number type */
+   defined to a number from 0 to 12 corresponding to each number type */
 #if TYPE_ORDER == 0
 #  define NumberTemplate	NSBoolNumber
 #  define TYPE_TYPE	BOOL
@@ -84,7 +84,7 @@
 - (id) initWithBytes: (const void*)value objCType: (const char*)type
 {
   typedef __typeof__(data) _dt;
-  data = *(_dt*)value;
+  data = *(const _dt*)value;
   return self;
 }
 
@@ -213,9 +213,25 @@
 	      BOOL	oData;
 
 	      (*(info->getValue))(other, @selector(getValue:), (void*)&oData);
-	      if (data == oData)
+#undef PT
+#if	(TYPE_ORDER & 1) == 1
+#if GS_SIZEOF_CHAR < GS_SIZEOF_SHORT && TYPE_ORDER < 3 
+#define	PT (short)
+#elif GS_SIZEOF_CHAR < GS_SIZEOF_INT && TYPE_ORDER < 5
+#define	PT (int)
+#elif GS_SIZEOF_CHAR < GS_SIZEOF_LONG && TYPE_ORDER < 7
+#define	PT (long)
+#elif GS_SIZEOF_CHAR < GS_SIZEOF_LONG_LONG && TYPE_ORDER < 9
+#define	PT (long long)
+#else
+#define	PT (double)
+#endif
+#else
+#define	PT
+#endif
+	      if (PT data == PT oData)
 		return NSOrderedSame;
-	      else if (data < oData)
+	      else if (PT data < PT oData)
 		return NSOrderedAscending;
 	      else
 		return NSOrderedDescending;
@@ -225,9 +241,25 @@
 	      signed char	oData;
 
 	      (*(info->getValue))(other, @selector(getValue:), (void*)&oData);
-	      if (data == oData)
+#undef PT
+#if	(TYPE_ORDER & 1) == 0
+#if GS_SIZEOF_CHAR < GS_SIZEOF_SHORT && TYPE_ORDER < 3
+#define	PT (short)
+#elif GS_SIZEOF_CHAR < GS_SIZEOF_INT && TYPE_ORDER < 5
+#define	PT (int)
+#elif GS_SIZEOF_CHAR < GS_SIZEOF_LONG && TYPE_ORDER < 7
+#define	PT (long)
+#elif GS_SIZEOF_CHAR < GS_SIZEOF_LONG_LONG && TYPE_ORDER < 9
+#define	PT (long long)
+#else
+#define	PT (double)
+#endif
+#else
+#define	PT
+#endif
+	      if (PT data == PT oData)
 		return NSOrderedSame;
-	      else if (data < oData)
+	      else if (PT data < PT oData)
 		return NSOrderedAscending;
 	      else
 		return NSOrderedDescending;
@@ -237,9 +269,25 @@
 	      unsigned char	oData;
 
 	      (*(info->getValue))(other, @selector(getValue:), (void*)&oData);
-	      if (data == oData)
+#undef PT
+#if	(TYPE_ORDER & 1) == 1
+#if GS_SIZEOF_CHAR < GS_SIZEOF_SHORT && TYPE_ORDER < 3
+#define	PT (short)
+#elif GS_SIZEOF_CHAR < GS_SIZEOF_INT && TYPE_ORDER < 5
+#define	PT (int)
+#elif GS_SIZEOF_CHAR < GS_SIZEOF_LONG && TYPE_ORDER < 7
+#define	PT (long)
+#elif GS_SIZEOF_CHAR < GS_SIZEOF_LONG_LONG && TYPE_ORDER < 9
+#define	PT (long long)
+#else
+#define	PT (double)
+#endif
+#else
+#define	PT
+#endif
+	      if (PT data == PT oData)
 		return NSOrderedSame;
-	      else if (data < oData)
+	      else if (PT data < PT oData)
 		return NSOrderedAscending;
 	      else
 		return NSOrderedDescending;
@@ -249,9 +297,23 @@
 	      signed short	oData;
 
 	      (*(info->getValue))(other, @selector(getValue:), (void*)&oData);
-	      if (data == oData)
+#undef PT
+#if	(TYPE_ORDER & 1) == 0
+#if GS_SIZEOF_SHORT < GS_SIZEOF_INT && TYPE_ORDER < 5
+#define	PT (int)
+#elif GS_SIZEOF_SHORT < GS_SIZEOF_LONG && TYPE_ORDER < 7
+#define	PT (long)
+#elif GS_SIZEOF_SHORT < GS_SIZEOF_LONG_LONG && TYPE_ORDER < 9
+#define	PT (long long)
+#else
+#define	PT (double)
+#endif
+#else
+#define	PT
+#endif
+	      if (PT data == PT oData)
 		return NSOrderedSame;
-	      else if (data < oData)
+	      else if (PT data < PT oData)
 		return NSOrderedAscending;
 	      else
 		return NSOrderedDescending;
@@ -261,9 +323,23 @@
 	      unsigned short	oData;
 
 	      (*(info->getValue))(other, @selector(getValue:), (void*)&oData);
-	      if (data == oData)
+#undef PT
+#if	(TYPE_ORDER & 1) == 1
+#if GS_SIZEOF_SHORT < GS_SIZEOF_INT && TYPE_ORDER < 5
+#define	PT (int)
+#elif GS_SIZEOF_SHORT < GS_SIZEOF_LONG && TYPE_ORDER < 7
+#define	PT (long)
+#elif GS_SIZEOF_SHORT < GS_SIZEOF_LONG_LONG && TYPE_ORDER < 9
+#define	PT (long long)
+#else
+#define	PT (double)
+#endif
+#else
+#define	PT
+#endif
+	      if (PT data == PT oData)
 		return NSOrderedSame;
-	      else if (data < oData)
+	      else if (PT data < PT oData)
 		return NSOrderedAscending;
 	      else
 		return NSOrderedDescending;
@@ -273,9 +349,21 @@
 	      signed int	oData;
 
 	      (*(info->getValue))(other, @selector(getValue:), (void*)&oData);
-	      if (data == oData)
+#undef PT
+#if	(TYPE_ORDER & 1) == 0
+#if GS_SIZEOF_INT < GS_SIZEOF_LONG && TYPE_ORDER < 7
+#define	PT (long)
+#elif GS_SIZEOF_INT < GS_SIZEOF_LONG_LONG && TYPE_ORDER < 9
+#define	PT (long long)
+#else
+#define	PT (double)
+#endif
+#else
+#define	PT
+#endif
+	      if (PT data == PT oData)
 		return NSOrderedSame;
-	      else if (data < oData)
+	      else if (PT data < PT oData)
 		return NSOrderedAscending;
 	      else
 		return NSOrderedDescending;
@@ -285,9 +373,21 @@
 	      unsigned int	oData;
 
 	      (*(info->getValue))(other, @selector(getValue:), (void*)&oData);
-	      if (data == oData)
+#undef PT
+#if	(TYPE_ORDER & 1) == 1
+#if GS_SIZEOF_INT < GS_SIZEOF_LONG && TYPE_ORDER < 7
+#define	PT (long)
+#elif GS_SIZEOF_INT < GS_SIZEOF_LONG_LONG && TYPE_ORDER < 9
+#define	PT (long long)
+#else
+#define	PT (double)
+#endif
+#else
+#define	PT
+#endif
+	      if (PT data == PT oData)
 		return NSOrderedSame;
-	      else if (data < oData)
+	      else if (PT data < PT oData)
 		return NSOrderedAscending;
 	      else
 		return NSOrderedDescending;
@@ -297,9 +397,19 @@
 	      signed long	oData;
 
 	      (*(info->getValue))(other, @selector(getValue:), (void*)&oData);
-	      if (data == oData)
+#undef PT
+#if	(TYPE_ORDER & 1) == 0
+#if GS_SIZEOF_LONG < GS_SIZEOF_LONG_LONG && TYPE_ORDER < 9
+#define	PT (long long)
+#else
+#define	PT (double)
+#endif
+#else
+#define	PT
+#endif
+	      if (PT data == PT oData)
 		return NSOrderedSame;
-	      else if (data < oData)
+	      else if (PT data < PT oData)
 		return NSOrderedAscending;
 	      else
 		return NSOrderedDescending;
@@ -309,9 +419,19 @@
 	      unsigned long	oData;
 
 	      (*(info->getValue))(other, @selector(getValue:), (void*)&oData);
-	      if (data == oData)
+#undef PT
+#if	(TYPE_ORDER & 1) == 1
+#if GS_SIZEOF_LONG < GS_SIZEOF_LONG_LONG && TYPE_ORDER < 9
+#define	PT (long long)
+#else
+#define	PT (double)
+#endif
+#else
+#define	PT
+#endif
+	      if (PT data == PT oData)
 		return NSOrderedSame;
-	      else if (data < oData)
+	      else if (PT data < PT oData)
 		return NSOrderedAscending;
 	      else
 		return NSOrderedDescending;
@@ -321,9 +441,15 @@
 	      signed long long	oData;
 
 	      (*(info->getValue))(other, @selector(getValue:), (void*)&oData);
-	      if (data == oData)
+#undef PT
+#if	(TYPE_ORDER & 1) == 0
+#define	PT (double)
+#else
+#define	PT
+#endif
+	      if (PT data == PT oData)
 		return NSOrderedSame;
-	      else if (data < oData)
+	      else if (PT data < PT oData)
 		return NSOrderedAscending;
 	      else
 		return NSOrderedDescending;
@@ -333,9 +459,15 @@
 	      unsigned long long	oData;
 
 	      (*(info->getValue))(other, @selector(getValue:), (void*)&oData);
-	      if (data == oData)
+#undef PT
+#if	(TYPE_ORDER & 1) == 1
+#define	PT (double)
+#else
+#define	PT
+#endif
+	      if (PT data == PT oData)
 		return NSOrderedSame;
-	      else if (data < oData)
+	      else if (PT data < PT oData)
 		return NSOrderedAscending;
 	      else
 		return NSOrderedDescending;
@@ -345,9 +477,15 @@
 	      float	oData;
 
 	      (*(info->getValue))(other, @selector(getValue:), (void*)&oData);
-	      if (data == oData)
+#undef PT
+#if TYPE_ORDER != 11
+#define	PT (double)
+#else
+#define	PT
+#endif
+	      if (PT data == PT oData)
 		return NSOrderedSame;
-	      else if (data < oData)
+	      else if (PT data < PT oData)
 		return NSOrderedAscending;
 	      else
 		return NSOrderedDescending;
@@ -357,9 +495,11 @@
 	      double	oData;
 
 	      (*(info->getValue))(other, @selector(getValue:), (void*)&oData);
-	      if (data == oData)
+#undef PT
+#define	PT (double)
+	      if (PT data == PT oData)
 		return NSOrderedSame;
-	      else if (data < oData)
+	      else if (PT data < PT oData)
 		return NSOrderedAscending;
 	      else
 		return NSOrderedDescending;
@@ -376,7 +516,7 @@
 - (NSString*) descriptionWithLocale: (NSDictionary*)locale
 {
 #if TYPE_ORDER == 0
-  return (data) ? @"true" : @"false";
+  return (data) ? @"YES" : @"NO";
 #else
   NSString	*result = [NSString alloc];
 
