@@ -21,12 +21,12 @@
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA.
 
    <title>NSCopyObject class reference</title>
-   $Date: 2001/12/18 16:54:14 $ $Revision: 1.12 $
+   $Date: 2003/07/31 23:49:31 $ $Revision: 1.14 $
    */ 
 
-#include <config.h>
-#include <base/preface.h>
-#include <Foundation/NSObject.h>
+#include "config.h"
+#include "GNUstepBase/preface.h"
+#include "Foundation/NSObject.h"
 
 NSObject *NSCopyObject(NSObject *anObject, unsigned extraBytes, NSZone *zone)
 {
