@@ -21,20 +21,20 @@
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA.
 
    <title>NSCountedSet class reference</title>
-   $Date: 2001/12/18 16:54:14 $ $Revision: 1.26 $
+   $Date: 2004/02/11 16:51:00 $ $Revision: 1.34 $
    */
 
-#include <config.h>
-#include <base/behavior.h>
-#include <Foundation/NSSet.h>
-#include <Foundation/NSCoder.h>
-#include <Foundation/NSArray.h>
-#include <Foundation/NSUtilities.h>
-#include <Foundation/NSString.h>
-#include <Foundation/NSLock.h>
-#include <Foundation/NSNotification.h>
-#include <Foundation/NSThread.h>
-#include <Foundation/NSObjCRuntime.h>
+#include "config.h"
+#include "GNUstepBase/GSLock.h"
+#include "Foundation/NSSet.h"
+#include "Foundation/NSCoder.h"
+#include "Foundation/NSArray.h"
+#include "Foundation/NSUtilities.h"
+#include "Foundation/NSString.h"
+#include "Foundation/NSLock.h"
+#include "Foundation/NSNotification.h"
+#include "Foundation/NSThread.h"
+#include "Foundation/NSObjCRuntime.h"
 
 @class	GSCountedSet;
 
@@ -48,10 +48,17 @@ static IMP		lockImp = 0;
 static IMP		unlockImp = 0;
 static BOOL		uniquing = NO;
 
-@interface	NSCountedSet (GSThreading)
-+ (void) _becomeThreaded: (id)notification;
-@end
-
+/**
+ * <p>
+ *   The NSCountedSet class is used to maintain a set of objects where
+ *   the number of times each object has been added (wiithout a
+ *   corresponding removal) is kept track of.
+ * </p>
+ * <p>
+ *   In GNUstep, extra methods are provided to make use of a counted
+ *   set for <em>uniquing</em> objects easier.
+ * </p>
+ */
 @implementation NSCountedSet 
 
 static Class NSCountedSet_abstract_class;
@@ -63,18 +70,9 @@ static Class NSCountedSet_concrete_class;
     {
       NSCountedSet_abstract_class = self;
       NSCountedSet_concrete_class = [GSCountedSet class];
-      if ([NSThread isMultiThreaded])
-	{
-	  [self _becomeThreaded: nil];
-	}
-      else
-	{
-	  [[NSNotificationCenter defaultCenter]
-	    addObserver: self
-	       selector: @selector(_becomeThreaded:)
-		   name: NSWillBecomeMultiThreadedNotification
-		 object: nil];
-	}
+      uniqueLock = [GSLazyLock new];
+      lockImp = [uniqueLock methodForSelector: @selector(lock)];
+      unlockImp = [uniqueLock methodForSelector: @selector(unlock)];
     }
 }
 
@@ -90,6 +88,11 @@ static Class NSCountedSet_concrete_class;
     }
 }
 
+/**
+ * Returns the number of times that an object that is equal to the
+ * specified object (as determined byt the [-isEqual:] method) has
+ * been added to the set and not removed from it.
+ */
 - (unsigned int) countForObject: (id)anObject
 {
   [self subclassResponsibility: _cmd];
@@ -138,6 +141,7 @@ static Class NSCountedSet_concrete_class;
 	  {
 	    (*addImp)(self, @selector(addObject:), objs[i]);
 	  }
+	RELEASE(objs[i]);
       }
   }
   return self;
@@ -201,6 +205,17 @@ static Class NSCountedSet_concrete_class;
   return self;
 }
 
+/**
+ * <p>
+ *   This method removes from the set all objects whose count is
+ *   less than or equal to the specified value.
+ * </p>
+ * <p>
+ *   This is useful where a counted set is used for uniquing objects.
+ *   The set can be periodically purged of objects that have only
+ *   been added once - and are therefore simply wasting space.
+ * </p>
+ */
 - (void) purge: (int)level
 {
   if (level > 0)
@@ -224,7 +239,7 @@ static Class NSCountedSet_concrete_class;
 	    {
 	      unsigned	c = (*cImp)(self, @selector(countForObject:), obj);
 
-	      if (c <= level)
+	      if (c <= (unsigned int)level)
 		{
 		  while (c-- > 0)
 		    {
@@ -236,6 +251,22 @@ static Class NSCountedSet_concrete_class;
     }
 }
 
+/**
+ * <p>
+ *   If the supplied object (or one equal to it as determined by
+ *   the [-isEqual:] method) is already present in the set, the
+ *   count for that object is incremented, the supplied object
+ *   is released, and the object in the set is retained and returned.
+ *   Otherwise, the supplied object is added to the set and returned.
+ * </p>
+ * <p> 
+ *   This method is useful for uniquing objects - the init method of
+ *   a class need simply end with -
+ *   <code>
+ *     return [myUniquingSet unique: self];
+ *   </code>
+ * </p>
+ */
 - (id) unique: (id)anObject
 {
   id	o = [self member: anObject];
@@ -256,35 +287,35 @@ static Class NSCountedSet_concrete_class;
 }
 @end
 
-@implementation	NSCountedSet (GSThreading)
-/*
- * If we are multi-threaded, we must guard access to the uniquing set.
+/**
+ * This function purges the global NSCountedSet object used for
+ * uniquing.  It handles locking as necessary.  It can be used to
+ * purge the set even when uniquing is turned off.
  */
-+ (void) _becomeThreaded: (id)notification
-{
-  uniqueLock = [NSLock new];
-  lockImp = [uniqueLock methodForSelector: @selector(lock)];
-  unlockImp = [uniqueLock methodForSelector: @selector(unlock)];
-}
-@end
-
-
 void
-GSUPurge(unsigned level)
+GSUPurge(unsigned count)
 {
   if (uniqueLock != nil)
     {
       (*lockImp)(uniqueLock, @selector(lock));
     }
-  [uniqueSet purge: level];
+  [uniqueSet purge: count];
   if (uniqueLock != nil)
     {
       (*unlockImp)(uniqueLock, @selector(unlock));
     }
 }
 
+/**
+ * This function sets the count for the specified object.  If the
+ * count for the object is set to zero then the object is removed
+ * from the global uniquing set.  The object is added to the set
+ * if necessary.  The object returned is the one stored in the set.
+ * The function handles locking as necessary.  It can be used to
+ * alter the set even when uniquing is turned off.
+ */
 id
-GSUSet(id obj, unsigned level)
+GSUSet(id anObject, unsigned count)
 {
   id		found;
   unsigned	i;
@@ -293,29 +324,29 @@ GSUSet(id obj, unsigned level)
     {
       (*lockImp)(uniqueLock, @selector(lock));
     }
-  found = [uniqueSet member: obj];
+  found = [uniqueSet member: anObject];
   if (found == nil)
     {
-      found = obj;
-      for (i = 0; i < level; i++)
+      found = anObject;
+      for (i = 0; i < count; i++)
 	{
-	  [uniqueSet addObject: obj];
+	  [uniqueSet addObject: anObject];
 	}
     }
   else
     {
       i = [uniqueSet countForObject: found];
-      if (i < level)
+      if (i < count)
 	{
-	  while (i < level)
+	  while (i < count)
 	    {
 	      [uniqueSet addObject: found];
 	      i++;
 	    }
 	}
-      else if (i > level)
+      else if (i > count)
 	{
-	  while (i > level)
+	  while (i > count)
 	    {
 	      [uniqueSet removeObject: found];
 	      i--;
@@ -329,8 +360,14 @@ GSUSet(id obj, unsigned level)
   return found;
 }
 
+/**
+ * This function <em>uniques</em> the supplied argument, returning
+ * the result.  It works by using the [-unique:] method of a global
+ * NSCountedSet object.  It handles locking as necessary.
+ * If uniquing is turned off, it simply returns its argument.
+ */
 id
-GSUnique(id obj)
+GSUnique(id anObject)
 {
   if (uniquing == YES)
     {
@@ -338,15 +375,21 @@ GSUnique(id obj)
 	{
 	  (*lockImp)(uniqueLock, @selector(lock));
 	}
-      obj = (*uniqueImp)(uniqueSet, @selector(unique:), obj);
+      anObject = (*uniqueImp)(uniqueSet, @selector(unique:), anObject);
       if (uniqueLock != nil)
 	{
 	  (*unlockImp)(uniqueLock, @selector(unlock));
 	}
     }
-  return obj;
+  return anObject;
 }
 
+/**
+ * This function sets the state of a flag that determines the
+ * behavior of the GSUnique() function.  If the flag is on,
+ * uniquing is performed, if it is off the function has no effect.
+ * The default is for uniquing to be turned off.
+ */
 void
 GSUniquing(BOOL flag)
 {
