@@ -23,26 +23,26 @@
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA.
 
    <title>NSDate class reference</title>
-   $Date: 2002/03/13 09:58:43 $ $Revision: 1.69 $
+   $Date: 2003/07/31 23:49:31 $ $Revision: 1.79 $
    */
 
-#include <config.h>
-#include <Foundation/NSArray.h>
-#include <Foundation/NSDictionary.h>
-#include <Foundation/NSDate.h>
-#include <Foundation/NSCalendarDate.h>
-#include <Foundation/NSTimeZone.h>
-#include <Foundation/NSString.h>
-#include <Foundation/NSCoder.h>
-#include <Foundation/NSException.h>
-#include <Foundation/NSCharacterSet.h>
-#include <Foundation/NSScanner.h>
-#include <Foundation/NSObjCRuntime.h>
-#include <Foundation/NSPortCoder.h>
-#include <Foundation/NSUserDefaults.h>
-#include <base/preface.h>
-#include <base/behavior.h>
-#if HAVE_SYS_TIME_H
+#include "config.h"
+#include "Foundation/NSArray.h"
+#include "Foundation/NSDictionary.h"
+#include "Foundation/NSDate.h"
+#include "Foundation/NSCalendarDate.h"
+#include "Foundation/NSTimeZone.h"
+#include "Foundation/NSString.h"
+#include "Foundation/NSCoder.h"
+#include "Foundation/NSException.h"
+#include "Foundation/NSCharacterSet.h"
+#include "Foundation/NSScanner.h"
+#include "Foundation/NSObjCRuntime.h"
+#include "Foundation/NSPortCoder.h"
+#include "Foundation/NSUserDefaults.h"
+#include "GNUstepBase/preface.h"
+#include "GNUstepBase/GSObjCRuntime.h"
+#ifdef HAVE_SYS_TIME_H
 #include <sys/time.h>
 #endif
 #include <time.h>
@@ -50,20 +50,13 @@
 #include <stdlib.h>
 #include "GSPrivate.h"
 
-/* The number of seconds between 1/1/2001 and 1/1/1970 = -978307200. */
-/* This number comes from:
--(((31 years * 365 days) + 8 days for leap years) =total number of days
-  * 24 hours
-  * 60 minutes
-  * 60 seconds)
-  This ignores leap-seconds. */
-#define UNIX_REFERENCE_INTERVAL -978307200.0
-
 /* I hope 100,000 years is distant enough. */
 #define DISTANT_YEARS 100000.0
 #define DISTANT_FUTURE	(DISTANT_YEARS * 365.0 * 24 * 60 * 60)
 #define DISTANT_PAST	(-DISTANT_FUTURE)
 
+const NSTimeInterval NSTimeIntervalSince1970 = 978307200.0;
+
 
 
 static BOOL	debug = NO;
@@ -71,6 +64,16 @@ static Class	abstractClass = nil;
 static Class	concreteClass = nil;
 static Class	calendarClass = nil;
 
+/**
+ * Our concrete base class - NSCalendar date must share the ivar layout.
+ */
+@interface NSGDate : NSDate
+{
+@public
+  NSTimeInterval _seconds_since_ref;
+}
+@end
+
 @interface	GSDateSingle : NSGDate
 @end
 
@@ -117,15 +120,18 @@ otherTime(NSDate* other)
     return [other timeIntervalSinceReferenceDate];
 }
 
+/**
+ * Returns the current time (seconds since reference date) as an NSTimeInterval.
+ */
 NSTimeInterval
-GSTimeNow()
+GSTimeNow(void)
 {
 #if !defined(__MINGW__)
   NSTimeInterval interval;
   struct timeval tp;
 
   gettimeofday (&tp, NULL);
-  interval = UNIX_REFERENCE_INTERVAL;
+  interval = -NSTimeIntervalSince1970;
   interval += tp.tv_sec;
   interval += (double)tp.tv_usec / 1000000.0;
   return interval;
@@ -161,8 +167,11 @@ GSTimeNow()
 #endif /* __MINGW__ */
 }
 
-/* The implementation of NSDate. */
-
+/**
+ * NSDate is the abstract base class of a cluster of date classes.<br />
+ * An NSDate object encapsulates a constant date/time to a high
+ * resolution represented by the NSTimeInterval typedef.
+ */
 @implementation NSDate
 
 + (void) initialize
@@ -192,25 +201,83 @@ GSTimeNow()
     return NSAllocateObject(self, 0, z);
 }
 
+/**
+ * Returns the time interval between the current date and the
+ * reference date (1 January 2001, GMT).
+ */
 + (NSTimeInterval) timeIntervalSinceReferenceDate
 {
   return GSTimeNow();
 }
 
-// Allocation and initializing
-
+/**
+ * Returns an autoreleased instance representing the current date/time.
+ */
 + (id) date
 {
   return AUTORELEASE([[self allocWithZone: NSDefaultMallocZone()]
 	initWithTimeIntervalSinceReferenceDate: GSTimeNow()]);
 }
 
+/**
+ * Returns an autoreleased instance representing the date and time given
+ * by string. The value of string may be a 'natural' specification as
+ * specified by the preferences in the user defaults database, allowing
+ * phrases like 'last tuesday'
+ */
 + (id) dateWithNaturalLanguageString: (NSString*)string
 {
   return [self dateWithNaturalLanguageString: string
 				      locale: nil];
 }
 
+/**
+ * <p>Returns an autoreleased instance representing the date and time given
+ * by string. The value of string may be a 'natural' specification as
+ * specified by the preferences in the user defaults database, allowing
+ * phrases like 'last tuesday'
+ * </p>
+ * The locale contains keys such as -
+ * <deflist>
+ *   <term>NSDateTimeOrdering</term>
+ *   <desc>Controls the use of ambiguous numbers. This is done as a
+ *   sequence of the letters D(ay), M(onth), Y(ear), and H(our).
+ *   YMDH means that the first number encountered is assumed to be a
+ *   year, the second a month, the third a day, and the last an hour.
+ *   </desc>
+ *   <term>NSEarlierTimeDesignations</term>
+ *   <desc>An array of strings for times in the past.<br />
+ *   Defaults are <em>ago</em>, <em>last</em>, <em>past</em>, <em>prior</em>
+ *   </desc>
+ *   <term>NSHourNameDesignations</term>
+ *   <desc>An array of arrays of strings identifying the time of day.
+ *   Each array has an hour as its first value, and one or more words
+ *   as subsequent values.<br />
+ *   Defaults are: (0, midnight), (10, mornint), (12, noon, lunch),
+ *   (14, afternoon), (19, dinner).
+ *   </desc>
+ *   <term>NSLaterTimeDesignations</term>
+ *   <desc>An array of strings for times in the future.<br />
+ *   Default is <em>next</em>
+ *   </desc>
+ *   <term>NSNextDayDesignations</term>
+ *   <desc>The day after today. Default is <em>tomorrow.</em>
+ *   </desc>
+ *   <term>NSNextNextDayDesignations</term>
+ *   <desc>The day after tomorrow. Default is <em>nextday.</em>
+ *   </desc>
+ *   <term>NSPriorDayDesignations</term>
+ *   <desc>The day before today. Default is <em>yesterday.</em>
+ *   </desc>
+ *   <term>NSThisDayDesignations</term>
+ *   <desc>Identifies the current day. Default is <em>today.</em>
+ *   </desc>
+ *   <term>NSYearMonthWeekDesignations</term>
+ *   <desc>An array giving the word for year, month, and week.<br />
+ *   Defaults are <em>year</em>, <em>month</em> and <em>week</em>.
+ *   </desc>
+ * </deflist>
+ */
 + (id) dateWithNaturalLanguageString: (NSString*)string
                               locale: (NSDictionary*)locale
 {
@@ -243,8 +310,9 @@ GSTimeNow()
   unsigned		dtoIndex;
 
   if (locale == nil)
-    locale = GSUserDefaultsDictionaryRepresentation();
-
+    {
+      locale = GSUserDefaultsDictionaryRepresentation();
+    }
   ws = [NSCharacterSet whitespaceAndNewlineCharacterSet];
   digits = [NSCharacterSet decimalDigitCharacterSet];
   scanner = [NSScanner scannerWithString: string];
@@ -508,14 +576,18 @@ GSTimeNow()
   if (dto == nil)
     {
       if (debug)
-	NSLog(@"no NSDateTimeOrdering - default to DMYH.\n");
+	{
+	  NSLog(@"no NSDateTimeOrdering - default to DMYH.");
+	}
       dto = @"DMYH";
     }
   length = [dto length];
   if (length > 4)
     {
       if (debug)
-	NSLog(@"too many characters in NSDateTimeOrdering - truncating.\n");
+	{
+	  NSLog(@"too many characters in NSDateTimeOrdering - truncating.");
+	}
       length = 4;
     }
 
@@ -544,9 +616,13 @@ GSTimeNow()
 		  if (num > 23)
 		    {
 		      if (debug)
-			NSLog(@"hour (%d) too large - ignored.\n", num);
+			{
+			  NSLog(@"hour (%d) too large - ignored.", num);
+			}
 		      else
-			return nil;
+			{
+			  return nil;
+			}
 		    }
 		  else
 		    {
@@ -561,9 +637,13 @@ GSTimeNow()
 		  if (num > 59)
 		    {
 		      if (debug)
-			NSLog(@"minute (%d) too large - ignored.\n", num);
+			{
+			  NSLog(@"minute (%d) too large - ignored.", num);
+			}
 		      else
-			return nil;
+			{
+			  return nil;
+			}
 		    }
 		  else
 		    {
@@ -577,9 +657,13 @@ GSTimeNow()
 		  if (num > 59)
 		    {
 		      if (debug)
-			NSLog(@"second (%d) too large - ignored.\n", num);
+			{
+			  NSLog(@"second (%d) too large - ignored.", num);
+			}
 		      else
-			return nil;
+			{
+			  return nil;
+			}
 		    }
 		  else
 		    {
@@ -590,7 +674,9 @@ GSTimeNow()
 	      else
 		{
 		  if (debug)
-		    NSLog(@"odd time spec - excess numbers ignored.\n");
+		    {
+		      NSLog(@"odd time spec - excess numbers ignored.");
+		    }
 		}
 
 	      done = YES;
@@ -648,8 +734,10 @@ GSTimeNow()
 
 		  default:
 		    if (debug)
-		      NSLog(@"odd char (unicode %d) in NSDateTimeOrdering.\n",
-			    [dto characterAtIndex: dtoIndex]);
+		      {
+			NSLog(@"odd char (unicode %d) in NSDateTimeOrdering.",
+			  [dto characterAtIndex: dtoIndex]);
+		      }
 		    dtoIndex++;
 		    break;
 		}
@@ -657,7 +745,9 @@ GSTimeNow()
 	  if (dtoIndex >= [dto length])
 	    {
 	      if (debug)
-		NSLog(@"odd date specification - excess numbers ignored.\n");
+		{
+		  NSLog(@"odd date specification - excess numbers ignored.");
+		}
 	      break;
 	    }
 	  switch ([dto characterAtIndex: dtoIndex])
@@ -666,16 +756,24 @@ GSTimeNow()
 		if (num < 1)
 		  {
 		    if (debug)
-		      NSLog(@"day (0) too small - ignored.\n");
+		      {
+			NSLog(@"day (0) too small - ignored.");
+		      }
 		    else
-		      return nil;
+		      {
+			return nil;
+		      }
 		  }
 		else if (num > 31)
 		  {
 		    if (debug)
-		      NSLog(@"day (%d) too large - ignored.\n", num);
+		      {
+			NSLog(@"day (%d) too large - ignored.", num);
+		      }
 		    else
-		      return nil;
+		      {
+			return nil;
+		      }
 		  }
 		else
 		  {
@@ -687,16 +785,24 @@ GSTimeNow()
 		if (num < 1)
 		  {
 		    if (debug)
-		      NSLog(@"month (0) too small - ignored.\n");
+		      {
+			NSLog(@"month (0) too small - ignored.");
+		      }
 		    else
-		      return nil;
+		      {
+			return nil;
+		      }
 		  }
 		else if (num > 12)
 		  {
 		    if (debug)
-		      NSLog(@"month (%d) too large - ignored.\n", num);
+		      {
+			NSLog(@"month (%d) too large - ignored.", num);
+		      }
 		    else
-		      return nil;
+		      {
+			return nil;
+		      }
 		  }
 		else
 		  {
@@ -716,7 +822,9 @@ GSTimeNow()
 			Y = num + 1900;
 		      }
 		    if (debug)
-		      NSLog(@"year (%d) adjusted to %d.\n", num, Y);
+		      {
+			NSLog(@"year (%d) adjusted to %d.", num, Y);
+		      }
 		  }
 		else
 		  {
@@ -744,10 +852,13 @@ GSTimeNow()
 			  if (num > 11)
 			    {
 			      if (debug)
-				NSLog(@"hour (%d) too large - ignored.\n",
-				      num);
+				{
+				  NSLog(@"hour (%d) too large - ignored.", num);
+				}
 			      else
-				return nil;
+				{
+				  return nil;
+				}
 			      shouldIgnore = YES;
 			    }
 			  else if (mod == [ampm objectAtIndex: 1])
@@ -761,9 +872,13 @@ GSTimeNow()
 		      if (num > 23)
 			{
 			  if (debug)
-			    NSLog(@"hour (%d) too large - ignored.\n", num);
+			    {
+			      NSLog(@"hour (%d) too large - ignored.", num);
+			    }
 			  else
-			    return nil;
+			    {
+			      return nil;
+			    }
 			}
 		      else
 			{
@@ -775,8 +890,10 @@ GSTimeNow()
 		}
 	      default:
 		if (debug)
-		  NSLog(@"unexpected char (unicode%d) in NSDateTimeOrdering.\n",
-		    [dto characterAtIndex: dtoIndex]);
+		  {
+		    NSLog(@"unexpected char (unicode%d) in NSDateTimeOrdering.",
+		      [dto characterAtIndex: dtoIndex]);
+		  }
 		break;
 	    }
 	}
@@ -830,7 +947,9 @@ GSTimeNow()
   if (hadWeekDay && [theDate dayOfWeek] != weekDay)
     {
       if (debug)
-	NSLog(@"Date resulted in wrong day of week.\n");
+	{
+	  NSLog(@"Date resulted in wrong day of week.");
+	}
       return nil;
     }
   if (theDate == nil)
@@ -844,27 +963,49 @@ GSTimeNow()
     }
 }
 
+/**
+ * Returns an autoreleased instance with the date and time value given
+ * by the string using the ISO standard format YYYY-MM-DD HH:MM:SS +/-HHHMM
+ * (all the fields of which must be present).
+ */
 + (id) dateWithString: (NSString*)description
 {
   return AUTORELEASE([[self alloc] initWithString: description]);
 }
 
+/**
+ * Returns an autoreleased instance with the offset from the current
+ * date/time given by seconds (which may be fractional).
+ */
 + (id) dateWithTimeIntervalSinceNow: (NSTimeInterval)seconds
 {
   return AUTORELEASE([[self alloc] initWithTimeIntervalSinceNow: seconds]);
 }
 
-+ (id)dateWithTimeIntervalSince1970: (NSTimeInterval)seconds
+/**
+ * Returns an autoreleased instance with the offset from the unix system
+ * reference date of 1 January 1970, GMT.
+ */
++ (id) dateWithTimeIntervalSince1970: (NSTimeInterval)seconds
 {
   return AUTORELEASE([[self alloc] initWithTimeIntervalSinceReferenceDate:
-		       UNIX_REFERENCE_INTERVAL + seconds]);
+		       -NSTimeIntervalSince1970 + seconds]);
 }
 
+/**
+ * Returns an autoreleased instance with the offset from the OpenStep
+ * reference date of 1 January 2001, GMT.
+ */
 + (id) dateWithTimeIntervalSinceReferenceDate: (NSTimeInterval)seconds
 {
-  return AUTORELEASE([[self alloc] initWithTimeIntervalSinceReferenceDate: seconds]);
+  return AUTORELEASE([[self alloc] initWithTimeIntervalSinceReferenceDate:
+    seconds]);
 }
 
+/**
+ * Returns an autoreleased instance with th date/time set in the far
+ * future.
+ */
 + (id) distantFuture
 {
   if (_distantFuture == nil)
@@ -872,6 +1013,10 @@ GSTimeNow()
   return _distantFuture;
 }
 
+/**
+ * Returns an autoreleased instance with th date/time set in the far
+ * past.
+ */
 + (id) distantPast
 {
   if (_distantPast == nil)
@@ -929,11 +1074,19 @@ GSTimeNow()
   return o;
 }
 
+/**
+ * Returns an instance initialised with the current date/time.
+ */
 - (id) init
 {
   return [self initWithTimeIntervalSinceReferenceDate: GSTimeNow()];
 }
 
+/**
+ * Returns an instance with the date and time value given
+ * by the string using the ISO standard format YYYY-MM-DD HH:MM:SS +/-HHHMM
+ * (all the fields of which must be present).
+ */
 - (id) initWithString: (NSString*)description
 {
   // Easiest to just have NSCalendarDate do the work for us
@@ -953,6 +1106,9 @@ GSTimeNow()
     }
 }
 
+/**
+ * Returns an instance with the given offset from anotherDate.
+ */
 - (id) initWithTimeInterval: (NSTimeInterval)secsToBeAdded
 		  sinceDate: (NSDate*)anotherDate
 {
@@ -967,6 +1123,9 @@ GSTimeNow()
     otherTime(anotherDate) + secsToBeAdded];
 }
 
+/**
+ * Returns an instance with the offset from the current date/time.
+ */
 - (id) initWithTimeIntervalSinceNow: (NSTimeInterval)secsToBeAdded
 {
   // Get the current time, add the secs and init thyself
@@ -974,20 +1133,31 @@ GSTimeNow()
     GSTimeNow() + secsToBeAdded];
 }
 
-- (id)initWithTimeIntervalSince1970: (NSTimeInterval)seconds
+/**
+ * Returns an instance with the offset from the unix system
+ * reference date of 1 January 1970, GMT.
+ */
+- (id) initWithTimeIntervalSince1970: (NSTimeInterval)seconds
 {
   return [self initWithTimeIntervalSinceReferenceDate:
-    UNIX_REFERENCE_INTERVAL + seconds];
+    -NSTimeIntervalSince1970 + seconds];
 }
 
+/** <init />
+ * Returns an instance with the given offset from the OpenStep
+ * reference date of 1 January 2001, GMT.
+ */
 - (id) initWithTimeIntervalSinceReferenceDate: (NSTimeInterval)secs
 {
   [self subclassResponsibility: _cmd];
   return self;
 }
 
-// Converting to NSCalendar
-
+/**
+ * Returns an autoreleased instance of the [NSCalendarDate] class whose
+ * date/time value is the same as that of the receiver, and which uses
+ * the formatString and timeZone specified.
+ */
 - (NSCalendarDate *) dateWithCalendarFormat: (NSString*)formatString
 				   timeZone: (NSTimeZone*)timeZone
 {
@@ -998,8 +1168,10 @@ GSTimeNow()
   return AUTORELEASE(d);
 }
 
-// Representing dates
-
+/**
+ * Returns a string representation of the receiver formatted according
+ * to the default format string, time zone, and locale.
+ */
 - (NSString*) description
 {
   // Easiest to just have NSCalendarDate do the work for us
@@ -1011,6 +1183,10 @@ GSTimeNow()
   return s;
 }
 
+/**
+ * Returns a string representation of the receiver formatted according
+ * to the specified format string, time zone, and locale.
+ */
 - (NSString*) descriptionWithCalendarFormat: (NSString*)format
 				   timeZone: (NSTimeZone*)aTimeZone
 				     locale: (NSDictionary*)l
@@ -1033,6 +1209,10 @@ GSTimeNow()
   return s;
 }
 
+/**
+ * Returns a string representation of the receiver formatted according
+ * to the default format string and time zone, but using the given locale.
+ */
 - (NSString *) descriptionWithLocale: (NSDictionary *)locale
 {
   // Easiest to just have NSCalendarDate do the work for us
@@ -1044,8 +1224,10 @@ GSTimeNow()
   return s;
 }
 
-// Adding and getting intervals
-
+/**
+ * Returns an autoreleased NSDate instance whose value if offset from
+ * that of the receiver by seconds.
+ */
 - (id) addTimeInterval: (NSTimeInterval)seconds
 {
   /* xxx We need to check for overflow? */
@@ -1053,11 +1235,20 @@ GSTimeNow()
     otherTime(self) + seconds];
 }
 
+/**
+ * Returns the time interval between the receivers value and the
+ * unix system reference date of 1 January 1970, GMT.
+ */
 - (NSTimeInterval) timeIntervalSince1970
 {
-  return otherTime(self) - UNIX_REFERENCE_INTERVAL;
+  return otherTime(self) + NSTimeIntervalSince1970;
 }
 
+/**
+ * Returns the time interval between the receivers value and that of the
+ * otherDate argument.  If otherDate is earlier than the receiver, the
+ * returned value will be positive, if it is later it will be negative.
+ */
 - (NSTimeInterval) timeIntervalSinceDate: (NSDate*)otherDate
 {
   if (otherDate == nil)
@@ -1068,19 +1259,32 @@ GSTimeNow()
   return otherTime(self) - otherTime(otherDate);
 }
 
+/**
+ * Returns the time interval between the receivers value and the
+ * current date/time.  If the receiver represents a date/time in
+ * the past this will be negative, if it is in the future the
+ * returned value will be positive.
+ */
 - (NSTimeInterval) timeIntervalSinceNow
 {
   return otherTime(self) - GSTimeNow();
 }
 
+/**
+ * Returns the time interval between the receivers value and the
+ * OpenStep reference date of 1 Jan 2001 GMT.
+ */
 - (NSTimeInterval) timeIntervalSinceReferenceDate
 {
   [self subclassResponsibility: _cmd];
   return 0;
 }
 
-// Comparing dates
-
+/**
+ * Returns NSOrderedDescending if the receiver is later than otherDate,
+ * Returns NSOrderedAscending if the receiver is earlier than otherDate,
+ * Otherwise, returns NSOrderedSame.
+ */
 - (NSComparisonResult) compare: (NSDate*)otherDate
 {
   if (otherDate == self)
@@ -1103,6 +1307,10 @@ GSTimeNow()
   return NSOrderedSame;
 }
 
+/**
+ * Returns the earlier of the receiver and otherDate.<br />
+ * If the two represent identical date/time values, returns the receiver.
+ */
 - (NSDate*) earlierDate: (NSDate*)otherDate
 {
   if (otherDate == nil)
@@ -1134,6 +1342,10 @@ GSTimeNow()
   return NO;
 }
 
+/**
+ * Returns the earlier of the receiver and otherDate.<br />
+ * If the two represent identical date/time values, returns the receiver.
+ */
 - (NSDate*) laterDate: (NSDate*)otherDate
 {
   if (otherDate == nil)
@@ -1182,7 +1394,7 @@ GSTimeNow()
 
 - (NSTimeInterval) timeIntervalSince1970
 {
-  return _seconds_since_ref - UNIX_REFERENCE_INTERVAL;
+  return _seconds_since_ref + NSTimeIntervalSince1970;
 }
 
 - (NSTimeInterval) timeIntervalSinceDate: (NSDate*)otherDate
@@ -1287,7 +1499,7 @@ GSTimeNow()
   if (self == [GSDateSingle class])
     {
       [self setVersion: 1];
-      behavior_class_add_class(self, [NSGDate class]);
+      GSObjCAddClassBehavior(self, [NSGDate class]);
     }
 }
 
