@@ -21,17 +21,17 @@
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA.
 
    <title>NSDateFormatter class reference</title>
-   $Date: 2001/12/18 16:54:14 $ $Revision: 1.8 $
+   $Date: 2003/06/07 01:24:40 $ $Revision: 1.9 $
    */
 
-#include <config.h>
-#include <Foundation/NSDate.h>
-#include <Foundation/NSCalendarDate.h>
-#include <Foundation/NSTimeZone.h>
-#include <Foundation/NSFormatter.h>
-#include <Foundation/NSDateFormatter.h>
-#include <Foundation/NSString.h>
-#include <Foundation/NSCoder.h>
+#include "config.h"
+#include "Foundation/NSDate.h"
+#include "Foundation/NSCalendarDate.h"
+#include "Foundation/NSTimeZone.h"
+#include "Foundation/NSFormatter.h"
+#include "Foundation/NSDateFormatter.h"
+#include "Foundation/NSString.h"
+#include "Foundation/NSCoder.h"
 
 @implementation NSDateFormatter
 
