@@ -22,34 +22,37 @@
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA.
 
-   <title>NSDebug class reference</title>
-   $Date: 2002/02/13 22:25:38 $ $Revision: 1.18 $
+   <title>NSDebug utilities reference</title>
+   $Date: 2004/02/08 09:42:38 $ $Revision: 1.31 $
    */
 
-#include <config.h>
+#include "config.h"
 #include <stdio.h>
-#include <Foundation/NSDebug.h>
-#include <Foundation/NSString.h>
-#include <Foundation/NSLock.h>
-#include <Foundation/NSNotificationQueue.h>
-#include <Foundation/NSThread.h>
+#include "GNUstepBase/GSLock.h"
+#include "Foundation/NSData.h"
+#include "Foundation/NSDebug.h"
+#include "Foundation/NSString.h"
+#include "Foundation/NSLock.h"
+#include "Foundation/NSNotificationQueue.h"
+#include "Foundation/NSThread.h"
 
 typedef struct {
   Class	class;
   /* The following are used for statistical info */
-  int	count;
-  int	lastc;
-  int	total;
-  int   peak;
+  unsigned int	count;
+  unsigned int	lastc;
+  unsigned int	total;
+  unsigned int   peak;
   /* The following are used to record actual objects */
   BOOL  is_recording;
   id    *recorded_objects;
-  int   num_recorded_objects;
-  int   stack_size;
+  id    *recorded_tags;
+  unsigned int   num_recorded_objects;
+  unsigned int   stack_size;
 } table_entry;
 
-static	int	num_classes = 0;
-static	int	table_size = 0;
+static	unsigned int	num_classes = 0;
+static	unsigned int	table_size = 0;
 
 static table_entry*	the_table = 0;
 
@@ -58,38 +61,68 @@ static BOOL	debug_allocation = NO;
 static NSLock	*uniqueLock;
 
 static const char*	_GSDebugAllocationList(BOOL difference);
-static const char*	_GSDebugAllocationListAll();
+static const char*	_GSDebugAllocationListAll(void);
+
+void _GSDebugAllocationAdd(Class c, id o);
+void _GSDebugAllocationRemove(Class c, id o);
+
+void (*_GSDebugAllocationAddFunc)(Class c, id o) = _GSDebugAllocationAdd;
+void (*_GSDebugAllocationRemoveFunc)(Class c, id o) = _GSDebugAllocationRemove;
 
 @interface GSDebugAlloc : NSObject
 + (void) initialize;
-+ (void) _becomeThreaded: (NSNotification*)notification;
 @end
 
 @implementation GSDebugAlloc
-
 + (void) initialize
 {
-  if ([NSThread isMultiThreaded])
+  uniqueLock = [GSLazyRecursiveLock new];
+}
+@end
+
+/**
+ * This functions allows to set own function backcalls for debugging allocation
+ * of objects. Useful if you intend to write your own objectalloc.
+ */
+void 
+GSSetDebugAllocationFunctions(void (*newAddObjectFunc)(Class c, id o),
+  void (*newRemoveObjectFunc)(Class c, id o))
+{
+  if (uniqueLock != nil)
     {
-      [self _becomeThreaded: nil];
+      [uniqueLock lock];
+    }
+
+  if (newAddObjectFunc && newRemoveObjectFunc)
+    {	   	
+      _GSDebugAllocationAddFunc = newAddObjectFunc;
+      _GSDebugAllocationRemoveFunc = newRemoveObjectFunc;
     }
   else
     {
-      [[NSNotificationCenter defaultCenter]
-	addObserver: self
-	selector: @selector(_becomeThreaded:)
-	name: NSWillBecomeMultiThreadedNotification
-	object: nil];
+      // Back to default
+      _GSDebugAllocationAddFunc = _GSDebugAllocationAdd;
+      _GSDebugAllocationRemoveFunc = _GSDebugAllocationRemove;
+    }
+  
+  if (uniqueLock != nil)
+    {
+      [uniqueLock unlock];
     }
 }
 
-+ (void) _becomeThreaded: (NSNotification*)notification
-{
-  uniqueLock = [NSRecursiveLock new];
-}
-
-@end
-
+/**
+ * This function activates or deactivates object allocation debugging.<br />
+ * Returns the previous state.<br />
+ * You should call this function to activate
+ * allocation debugging before using any of the other allocation
+ * debugging functions such as GSDebugAllocationList() or
+ * GSDebugAllocationTotal().<br />
+ * Object allocation debugging
+ * should not affect performance too much, and is very useful
+ * as it allows you to monitor how many objects of each class
+ * your application has allocated.
+ */
 BOOL
 GSDebugAllocationActive(BOOL active)
 {
@@ -100,10 +133,26 @@ GSDebugAllocationActive(BOOL active)
   return old;
 }
 
+/**
+ * This function activates tracking all allocated instances of
+ * the specified class c.<br />
+ * This tracking can slow your
+ * application down, so you should use it only when you are
+ * into serious debugging.  Usually, you will monitor your
+ * application by using the functions GSDebugAllocationList()
+ * and similar, which do not slow things down much and return
+ * the number of allocated instances; when
+ * (if) by studying the reports generated by these functions
+ * you have found a leak of objects of a certain class, and
+ * if you can't figure out how to fix it by looking at the
+ * code, you can use this function to start tracking
+ * allocated instances of that class, and the following one
+ * can sometime allow you to list the leaked objects directly.
+ */
 void
 GSDebugAllocationActiveRecordingObjects(Class c)
 {
-  unsigned i;
+  unsigned int i;
   
   GSDebugAllocationActive(YES);
 
@@ -149,6 +198,7 @@ GSDebugAllocationActiveRecordingObjects(Class c)
   the_table[num_classes].peak = 0;
   the_table[num_classes].is_recording = YES;
   the_table[num_classes].recorded_objects = NULL;
+  the_table[num_classes].recorded_tags = NULL;
   the_table[num_classes].num_recorded_objects = 0;
   the_table[num_classes].stack_size = 0;
   num_classes++;
@@ -158,10 +208,16 @@ GSDebugAllocationActiveRecordingObjects(Class c)
 
 void
 GSDebugAllocationAdd(Class c, id o)
+{
+	(*_GSDebugAllocationAddFunc)(c,o);
+}
+
+void
+_GSDebugAllocationAdd(Class c, id o)
 {
   if (debug_allocation)
     {
-      int	i;
+      unsigned int	i;
 
       for (i = 0; i < num_classes; i++)
 	{
@@ -177,21 +233,33 @@ GSDebugAllocationAdd(Class c, id o)
 		}
 	      if (the_table[i].is_recording == YES)
 		{
-		  if (the_table[i].num_recorded_objects >= 
-		      the_table[i].stack_size)
+		  if (the_table[i].num_recorded_objects
+		    >= the_table[i].stack_size)
 		    {
-		      int more = the_table[i].stack_size + 128;
+		      int	more = the_table[i].stack_size + 128;
 		      id	*tmp;
+		      id	*tmp1;
       
 		      tmp = NSZoneMalloc(NSDefaultMallocZone(), 
 					 more * sizeof(id));
-
 		      if (tmp == 0)
 			{
 			  if (uniqueLock != nil)
 			    [uniqueLock unlock];
 			  return;
 			}
+
+		      tmp1 = NSZoneMalloc(NSDefaultMallocZone(), 
+					 more * sizeof(id));
+		      if (tmp1 == 0)
+			{
+			  NSZoneFree(NSDefaultMallocZone(),  tmp);
+			  if (uniqueLock != nil)
+			    [uniqueLock unlock];
+			  return;
+			}
+
+
 		      if (the_table[i].recorded_objects != NULL)
 			{
 			  memcpy(tmp, the_table[i].recorded_objects, 
@@ -199,13 +267,21 @@ GSDebugAllocationAdd(Class c, id o)
 				 * sizeof(id));
 			  NSZoneFree(NSDefaultMallocZone(), 
 				     the_table[i].recorded_objects);
+			  memcpy(tmp1, the_table[i].recorded_tags, 
+				 the_table[i].num_recorded_objects 
+				 * sizeof(id));
+			  NSZoneFree(NSDefaultMallocZone(), 
+				     the_table[i].recorded_tags);
 			}
 		      the_table[i].recorded_objects = tmp;
+		      the_table[i].recorded_tags = tmp1;
 		      the_table[i].stack_size = more;
 		    }
 		  
 		  (the_table[i].recorded_objects)
 		    [the_table[i].num_recorded_objects] = o;
+		  (the_table[i].recorded_tags)
+		    [the_table[i].num_recorded_objects] = nil;
 		  the_table[i].num_recorded_objects++;
 		}
 	      if (uniqueLock != nil)
@@ -217,7 +293,7 @@ GSDebugAllocationAdd(Class c, id o)
 	[uniqueLock lock];
       if (num_classes >= table_size)
 	{
-	  int		more = table_size + 128;
+	  unsigned int	more = table_size + 128;
 	  table_entry	*tmp;
 	  
 	  tmp = NSZoneMalloc(NSDefaultMallocZone(), more * sizeof(table_entry));
@@ -243,6 +319,7 @@ GSDebugAllocationAdd(Class c, id o)
       the_table[num_classes].peak = 1;
       the_table[num_classes].is_recording = NO;
       the_table[num_classes].recorded_objects = NULL;
+      the_table[num_classes].recorded_tags = NULL;
       the_table[num_classes].num_recorded_objects = 0;
       the_table[num_classes].stack_size = 0;
       num_classes++;
@@ -251,10 +328,33 @@ GSDebugAllocationAdd(Class c, id o)
     }
 }
 
+/**
+ * <p>
+ *   Returns the number
+ *   of instances of the specified class which are currently
+ *   allocated.  This number is very important to detect memory
+ *   leaks.  If you notice that this number is constantly
+ *   increasing without apparent reason, it is very likely a
+ *   memory leak - you need to check that you are correctly
+ *   releasing objects of this class, otherwise when your
+ *   application runs for a long time, it will eventually
+ *   allocate so many objects as to eat up all your system's
+ *   memory ...
+ * </p>
+ * <p> 
+ *   This function, like the ones below, returns the number of
+ *   objects allocated/released from the time when
+ *   GSDebugAllocationActive() was first called.  A negative
+ *   number means that in total, there are less objects of this
+ *   class allocated now than there were when you called
+ *   GSDebugAllocationActive(); a positive one means there are
+ *   more.
+ * </p>
+ */
 int
 GSDebugAllocationCount(Class c)
 {
-  int	i;
+  unsigned int	i;
 
   for (i = 0; i < num_classes; i++)
     {
@@ -266,10 +366,27 @@ GSDebugAllocationCount(Class c)
   return 0;
 }
 
+/**
+ * Returns the total
+ * number of instances of the specified class c which have been
+ * allocated - basically the number of times you have
+ * allocated an object of this class.  If this number is very
+ * high, it means you are creating a lot of objects of this
+ * class; even if you are releasing them correctly, you must
+ * not forget that allocating and deallocating objects is
+ * usually one of the slowest things you can do, so you might
+ * want to consider whether you can reduce the number of
+ * allocations and deallocations that you are doing - for
+ * example, by recycling objects of this class, uniquing
+ * them, and/or using some sort of flyweight pattern.  It
+ * might also be possible that you are unnecessarily creating
+ * too many objects of this class.  Well - of course some times
+ * there is nothing you can do about it.
+ */
 int
 GSDebugAllocationTotal(Class c)
 {
-  int	i;
+  unsigned int	i;
 
   for (i = 0; i < num_classes; i++)
     {
@@ -281,10 +398,22 @@ GSDebugAllocationTotal(Class c)
   return 0;
 }
 
+/**
+ * Returns the peak
+ * number of instances of the specified class which have been
+ * concurrently allocated.  If this number is very high, it
+ * means at some point in time you had a situation with a
+ * huge number of objects of this class allocated - this is
+ * an indicator that probably at some point in time your
+ * application was using a lot of memory - so you might want
+ * to investigate whether you can prevent this problem by
+ * inserting autorelease pools in your application's
+ * processing loops.
+ */
 int
 GSDebugAllocationPeak(Class c)
 {
-  int	i;
+  unsigned int	i;
 
   for (i = 0; i < num_classes; i++)
     {
@@ -296,12 +425,21 @@ GSDebugAllocationPeak(Class c)
   return 0;
 }
 
+/**
+ * This function returns a NULL
+ * terminated array listing all the classes for which
+ * statistical information has been collected.  Usually, you
+ * call this function, and then loop on all the classes returned,
+ * and for each one you get current, peak and total count by
+ * using GSDebugAllocationCount(), GSDebugAllocationPeak() and
+ * GSDebugAllocationTotal().
+ */
 Class *
 GSDebugAllocationClassList()
 {
   Class *ans;
   size_t siz;
-  int	i;
+  unsigned int	i;
 
   if (uniqueLock != nil)
     [uniqueLock lock];
@@ -321,34 +459,45 @@ GSDebugAllocationClassList()
   return ans;
 }
 
-/*
- *	This function returns a string listing all those classes for which
- *	either objects are currently allocated (difference == 0), or there
- *	has been a change in the number of objects allocated since the last
- *	call (difference != 0).
+/**
+ * This function returns a newline
+ * separated list of the classes which have instances
+ * allocated, and the instance counts.  If the 'changeFlag'
+ * argument is YES then the list gives the number of
+ * instances allocated/deallocated since the function was
+ * last called.  This function only returns the current count
+ * of instances (not the peak or total count), but its output
+ * is ready to be displayed or logged.
  */
 const char*
 GSDebugAllocationList(BOOL changeFlag)
 {
-  const char *ans;
+  const char	*ans;
+  NSData	*d;
+
   if (debug_allocation == NO)
     {
       return "Debug allocation system is not active!\n";
     }
   if (uniqueLock != nil)
-    [uniqueLock lock];
+    {
+      [uniqueLock lock];
+    }
   ans = _GSDebugAllocationList(changeFlag);
+  d = [NSData dataWithBytes: ans length: strlen(ans) + 1];
   if (uniqueLock != nil)
-    [uniqueLock unlock];
-  return ans;
+    {
+      [uniqueLock unlock];
+    }
+  return (const char*)[d bytes];
 }
 
 static const char*
 _GSDebugAllocationList(BOOL difference)
 {
-  int		pos = 0;
-  int		i;
-  static int	siz = 0;
+  unsigned int	pos = 0;
+  unsigned int	i;
+  static unsigned int	siz = 0;
   static char	*buf = 0;
 
   for (i = 0; i < num_classes; i++)
@@ -415,28 +564,44 @@ _GSDebugAllocationList(BOOL difference)
   return buf;
 }
 
+/**
+ * This function returns a newline
+ * separated list of the classes which have had instances
+ * allocated at any point, and the total count of the number
+ * of instances allocated for each class.  The difference with
+ * GSDebugAllocationList() is that this function returns also
+ * classes which have no objects allocated at the moment, but
+ * which had in the past.
+ */
 const char*
 GSDebugAllocationListAll()
 {
-  const char *ans;
+  const char	*ans;
+  NSData	*d;
+
   if (debug_allocation == NO)
     {
       return "Debug allocation system is not active!\n";
     }
   if (uniqueLock != nil)
-    [uniqueLock lock];
+    {
+      [uniqueLock lock];
+    }
   ans = _GSDebugAllocationListAll();
+  d = [NSData dataWithBytes: ans length: strlen(ans)+1];
   if (uniqueLock != nil)
-    [uniqueLock unlock];
-  return ans;
+    {
+      [uniqueLock unlock];
+    }
+  return (const char*)[d bytes];
 }
 
 static const char*
-_GSDebugAllocationListAll()
+_GSDebugAllocationListAll(void)
 {
-  int		pos = 0;
-  int		i;
-  static int	siz = 0;
+  unsigned int	pos = 0;
+  unsigned int	i;
+  static unsigned int	siz = 0;
   static char	*buf = 0;
 
   for (i = 0; i < num_classes; i++)
@@ -487,15 +652,23 @@ _GSDebugAllocationListAll()
 
 void
 GSDebugAllocationRemove(Class c, id o)
+{
+	(*_GSDebugAllocationRemoveFunc)(c,o);
+}
+
+void
+_GSDebugAllocationRemove(Class c, id o)
 {
   if (debug_allocation)
     {
-      int	i;
+      unsigned int	i;
 
       for (i = 0; i < num_classes; i++)
 	{
 	  if (the_table[i].class == c)
 	    {
+	      id	tag = nil;
+
 	      if (uniqueLock != nil)
 		[uniqueLock lock];
 	      the_table[i].count--;
@@ -507,6 +680,7 @@ GSDebugAllocationRemove(Class c, id o)
 		    {
 		      if ((the_table[i].recorded_objects)[j] == o)
 			{
+			  tag = (the_table[i].recorded_tags)[j];
 			  break;
 			}
 		    }
@@ -518,6 +692,8 @@ GSDebugAllocationRemove(Class c, id o)
 			{
 			  (the_table[i].recorded_objects)[k] = 
 			    (the_table[i].recorded_objects)[k + 1];
+			  (the_table[i].recorded_tags)[k] = 
+			    (the_table[i].recorded_tags)[k + 1];
 			}
 		      the_table[i].num_recorded_objects--;
 		    }
@@ -531,17 +707,82 @@ GSDebugAllocationRemove(Class c, id o)
 		}
 	      if (uniqueLock != nil)
 		[uniqueLock unlock];
+	      RELEASE(tag);
 	      return;
 	    }
 	}
     }
 }
 
+/**
+ * This function associates the supplied tag with a recorded
+ * object and returns the tag which was previously associated
+ * with it (if any).<br />
+ * If the object was not recorded, the method returns nil<br />
+ * The tag is retained while it is associated with the object.
+ */
+id
+GSDebugAllocationTagRecordedObject(id object, id tag)
+{
+  Class c = [object class];
+  id	o = nil;
+  int	i;
+  int	j;
+
+  if (debug_allocation == NO)
+    {
+      return nil;
+    }
+  if (uniqueLock != nil)
+    [uniqueLock lock];
+  
+  for (i = 0; i < num_classes; i++)
+    {
+      if (the_table[i].class == c)
+      {
+	  break;
+	}
+    }
+  
+  if (i == num_classes
+    || the_table[i].is_recording == NO
+    || the_table[i].num_recorded_objects == 0)
+    {
+      if (uniqueLock != nil)
+	[uniqueLock unlock];
+      return nil;
+    }
+
+  for (j = 0; j < the_table[i].num_recorded_objects; j++)
+    {
+      if (the_table[i].recorded_objects[j] == object)
+	{
+	  o = the_table[i].recorded_tags[j];
+	  the_table[i].recorded_tags[j] = RETAIN(tag);
+	  break;
+	}
+    }
+
+  if (uniqueLock != nil)
+    [uniqueLock unlock];
+  return AUTORELEASE(o);
+}
+
+/**
+ * This function returns an array
+ * containing all the allocated objects of a certain class
+ * which have been recorded ... to start the recording, you need
+ * to invoke GSDebugAllocationActiveRecordingObjects().
+ * Presumably, you will immediately call [NSObject-description] on them
+ * to find out the objects you are leaking.  The objects are
+ * returned in an array, so until the array is autoreleased,
+ * the objects are not released.
+ */
 NSArray *
 GSDebugAllocationListRecordedObjects(Class c)
 {
   NSArray *answer;
-  unsigned i, k;
+  unsigned int i, k;
   id *tmp;
 
   if (debug_allocation == NO)
@@ -583,7 +824,6 @@ GSDebugAllocationListRecordedObjects(Class c)
 
   tmp = NSZoneMalloc(NSDefaultMallocZone(), 
 		     the_table[i].num_recorded_objects * sizeof(id));
-  
   if (tmp == 0)
     {
       if (uniqueLock != nil)
@@ -595,10 +835,6 @@ GSDebugAllocationListRecordedObjects(Class c)
   memcpy(tmp, the_table[i].recorded_objects, 
 	 the_table[i].num_recorded_objects * sizeof(id));
 
-  /* Then, we bravely unlock the lock */
-  if (uniqueLock != nil)
-    [uniqueLock unlock];
-  
   /* Retain all the objects - NB: if retaining one of the objects as a
      side effect releases another one of them , we are broken ... */
   for (k = 0; k < the_table[i].num_recorded_objects; k++)
@@ -606,6 +842,10 @@ GSDebugAllocationListRecordedObjects(Class c)
       RETAIN (tmp[k]);
     }
 
+  /* Then, we bravely unlock the lock */
+  if (uniqueLock != nil)
+    [uniqueLock unlock];
+  
   /* Only then we create an array with them - this is now safe as we
      have copied the objects out, unlocked, and retained them. */
   answer = [NSArray arrayWithObjects: tmp  
@@ -652,7 +892,7 @@ GSDebugMethodMsg(id obj, SEL sel, const char *file, int line, NSString *fmt)
   return message;
 }
 
-static void *_frameOffsets[100];
+// static void *_frameOffsets[100];
 
 #define _NS_FRAME_HACK(a) case a: return __builtin_frame_address(a + 1)
 #define _NS_RETURN_HACK(a) case a: return __builtin_return_address(a + 1)
