@@ -22,22 +22,25 @@
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA.
 
    <title>NSDictionary class reference</title>
-   $Date: 2002/03/13 09:58:43 $ $Revision: 1.86 $
+   $Date: 2004/02/24 19:43:34 $ $Revision: 1.104 $
    */
 
-#include <config.h>
-#include <base/behavior.h>
-#include <Foundation/NSDictionary.h>
-#include <Foundation/NSArray.h>
-#include <Foundation/NSUtilities.h>
-#include <Foundation/NSString.h>
-#include <Foundation/NSException.h>
-#include <Foundation/NSAutoreleasePool.h>
-#include <Foundation/NSFileManager.h>
-#include <Foundation/NSCoder.h>
-#include <Foundation/NSDebug.h>
-#include <Foundation/NSObjCRuntime.h>
-#include <Foundation/NSValue.h>
+#include "config.h"
+#include "Foundation/NSDictionary.h"
+#include "Foundation/NSArray.h"
+#include "Foundation/NSData.h"
+#include "Foundation/NSUtilities.h"
+#include "Foundation/NSString.h"
+#include "Foundation/NSException.h"
+#include "Foundation/NSAutoreleasePool.h"
+#include "Foundation/NSFileManager.h"
+#include "Foundation/NSCoder.h"
+#include "Foundation/NSDebug.h"
+#include "Foundation/NSObjCRuntime.h"
+#include "Foundation/NSValue.h"
+// For private method _decodeArrayOfObjectsForKey:
+#include "Foundation/NSKeyedArchiver.h"
+#include "GNUstepBase/GSCategories.h"
 #include "GSPrivate.h"
 
 @implementation NSDictionary 
@@ -45,6 +48,10 @@
 @class	GSDictionary;
 @class	GSMutableDictionary;
 
+extern BOOL	GSMacOSXCompatiblePropertyLists(void);
+extern void	GSPropertyListMake(id,NSDictionary*,BOOL,BOOL,unsigned,id*);
+
+
 static Class NSArray_class;
 static Class NSDictionaryClass;
 static Class NSMutableDictionaryClass;
@@ -89,7 +96,8 @@ static SEL	appSel;
     }
 }
 
-/* This is the designated initializer */
+/** <init />
+ */
 - (id) initWithObjects: (id*)objects
 	       forKeys: (id*)keys
 		 count: (unsigned)count
@@ -98,39 +106,69 @@ static SEL	appSel;
   return 0;
 }
 
+/**
+ * Returns an unsigned integer which is the number of elements
+ * stored in the dictionary.
+ */
 - (unsigned) count
 {
   [self subclassResponsibility: _cmd];
   return 0;
 }
 
+/**
+ * Returns the object in the dictionary corresponding to aKey, or nil if
+ * the key is not present.
+ */
 - (id) objectForKey: (id)aKey
 {
   [self subclassResponsibility: _cmd];
   return 0;
 }
 
+/**
+ * Return an enumerator object containing all the keys of the dictionary.
+ */
 - (NSEnumerator*) keyEnumerator
 {
   [self subclassResponsibility: _cmd];
   return nil;
 }
 
+/**
+ * Return an enumerator object containing all the objects of the dictionary.
+ */
 - (NSEnumerator*) objectEnumerator
 {
   [self subclassResponsibility: _cmd];
   return nil;
 }
 
+/**
+ * Returns a new copy of the receiver.<br />
+ * The default abstract implementation of a copy is to use the
+ * -initWithDictionary:copyItems: method with the flag set to YES.<br />
+ * Immutable subclasses generally simply retain and return the receiver.
+ */
 - (id) copyWithZone: (NSZone*)z
 {
-  return RETAIN(self);
+  NSDictionary	*copy = [NSDictionaryClass allocWithZone: z];
+
+  return [copy initWithDictionary: self copyItems: NO];
 }
 
+/**
+ * Returns a new instance containing the same objects as
+ * the receiver.<br />
+ * The default implementation does this by calling the
+ * -initWithDictionary:copyItems: method on a newly created object,
+ * and passing it NO to tell it just to retain the items.
+ */
 - (id) mutableCopyWithZone: (NSZone*)z
 {
-  return [[GSMutableDictionaryClass allocWithZone: z] 
-	  initWithDictionary: self];
+  NSMutableDictionary	*copy = [NSMutableDictionaryClass allocWithZone: z];
+
+  return [copy initWithDictionary: self copyItems: NO];
 }
 
 - (Class) classForCoder
@@ -142,52 +180,130 @@ static SEL	appSel;
 {
   unsigned	count = [self count];
 
-  [aCoder encodeValueOfObjCType: @encode(unsigned) at: &count];
-  if (count > 0)
+  if ([aCoder allowsKeyedCoding])
     {
-      NSEnumerator	*enumerator = [self keyEnumerator];
-      id		key;
-      IMP		enc;
-      IMP		nxt;
-      IMP		ofk;
-
-      nxt = [enumerator methodForSelector: @selector(nextObject)];
-      enc = [aCoder methodForSelector: @selector(encodeObject:)];
-      ofk = [self methodForSelector: @selector(objectForKey:)];
-
-      while ((key = (*nxt)(enumerator, @selector(nextObject))) != nil)
+      if (count > 0)
 	{
-	  id	val = (*ofk)(self, @selector(objectForKey:), key);
+	  NSEnumerator	*enumerator = [self keyEnumerator];
+	  id		key;
 
-	  (*enc)(aCoder, @selector(encodeObject:), key);
-	  (*enc)(aCoder, @selector(encodeObject:), val);
+	  if ([aCoder class] == [NSKeyedArchiver class])
+	    {
+	      NSArray	*keys = [self allKeys];
+	      id	objects = [NSMutableArray arrayWithCapacity: count];
+	      unsigned	i;
+
+	      for (i = 0; i < count; i++)
+		{
+		  key = [keys objectAtIndex: i];
+		  [objects addObject: [self objectForKey: key]];
+		}
+	      [(NSKeyedArchiver*)aCoder _encodeArrayOfObjects: keys
+						       forKey: @"NS.keys"];
+	      [(NSKeyedArchiver*)aCoder _encodeArrayOfObjects: objects
+						       forKey: @"NS.objects"];
+	    }
+	  else
+	    {
+	      unsigned	i = 0;
+
+	      while ((key = [enumerator nextObject]) != nil)
+		{
+		  NSString	*s;
+
+		  s = [NSString stringWithFormat: @"NS.key.%u", i];
+		  [aCoder encodeObject: key forKey: s];
+		  s = [NSString stringWithFormat: @"NS.object.%u", i];
+		  [aCoder encodeObject: [self objectForKey: key] forKey: s];
+		  i++;
+		}
+	    }
+	}
+    }
+  else
+    {
+      [aCoder encodeValueOfObjCType: @encode(unsigned) at: &count];
+      if (count > 0)
+	{
+	  NSEnumerator	*enumerator = [self keyEnumerator];
+	  id		key;
+	  IMP		enc;
+	  IMP		nxt;
+	  IMP		ofk;
+
+	  nxt = [enumerator methodForSelector: @selector(nextObject)];
+	  enc = [aCoder methodForSelector: @selector(encodeObject:)];
+	  ofk = [self methodForSelector: @selector(objectForKey:)];
+
+	  while ((key = (*nxt)(enumerator, @selector(nextObject))) != nil)
+	    {
+	      id	val = (*ofk)(self, @selector(objectForKey:), key);
+
+	      (*enc)(aCoder, @selector(encodeObject:), key);
+	      (*enc)(aCoder, @selector(encodeObject:), val);
+	    }
 	}
     }
 }
 
 - (id) initWithCoder: (NSCoder*)aCoder
 {
-  unsigned	count;
-
-  [aCoder decodeValueOfObjCType: @encode(unsigned) at: &count];
-  if (count > 0)
+  if ([aCoder allowsKeyedCoding])
     {
-      id	*keys = NSZoneMalloc(NSDefaultMallocZone(), sizeof(id)*count);
-      id	*vals = NSZoneMalloc(NSDefaultMallocZone(), sizeof(id)*count);
-      unsigned	i;
-      IMP	dec;
+      id	keys;
+      id	objects;
 
-      dec = [aCoder methodForSelector: @selector(decodeObject)];
-      for (i = 0; i < count; i++)
+      keys = [(NSKeyedUnarchiver*)aCoder _decodeArrayOfObjectsForKey: 
+					       @"NS.keys"];
+      objects = [(NSKeyedUnarchiver*)aCoder _decodeArrayOfObjectsForKey: 
+						  @"NS.objects"];
+      if (keys == nil)
 	{
-	  keys[i] = (*dec)(aCoder, @selector(decodeObject));
-	  vals[i] = (*dec)(aCoder, @selector(decodeObject));
-	}
-      self = [self initWithObjects: vals forKeys: keys count: count];
-      NSZoneFree(NSDefaultMallocZone(), keys);
-      NSZoneFree(NSDefaultMallocZone(), vals);
-    }
+	  unsigned	i = 0;
+	  NSString	*key;
+	  id		val;
 
+	  keys = [NSMutableArray arrayWithCapacity: 2];
+	  objects = [NSMutableArray arrayWithCapacity: 2];
+	  key = [NSString stringWithFormat: @"NS.object.%u", i];
+	  val = [(NSKeyedUnarchiver*)aCoder decodeObjectForKey: key];
+
+	  while (val != nil)
+	    {
+	      [objects addObject: val];
+	      key = [NSString stringWithFormat: @"NS.key.%u", i];
+	      val = [(NSKeyedUnarchiver*)aCoder decodeObjectForKey: key];
+	      [keys addObject: val];
+	      i++;
+	      key = [NSString stringWithFormat: @"NS.object.%u", i];
+	      val = [(NSKeyedUnarchiver*)aCoder decodeObjectForKey: key];
+	    }
+	}
+      self = [self initWithObjects: objects forKeys: keys];
+    }
+  else
+    {
+      unsigned	count;
+
+      [aCoder decodeValueOfObjCType: @encode(unsigned) at: &count];
+      if (count > 0)
+        {
+	  id	*keys = NSZoneMalloc(NSDefaultMallocZone(), sizeof(id)*count);
+	  id	*vals = NSZoneMalloc(NSDefaultMallocZone(), sizeof(id)*count);
+	  unsigned	i;
+	  IMP	dec;
+	  
+	  dec = [aCoder methodForSelector: @selector(decodeObject)];
+	  for (i = 0; i < count; i++)
+	    {
+	      keys[i] = (*dec)(aCoder, @selector(decodeObject));
+	      vals[i] = (*dec)(aCoder, @selector(decodeObject));
+	    }
+	  self = [self initWithObjects: vals forKeys: keys count: count];
+	  NSZoneFree(NSDefaultMallocZone(), keys);
+	  NSZoneFree(NSDefaultMallocZone(), vals);
+	}
+    }
   return self;
 }
 
@@ -196,12 +312,23 @@ static SEL	appSel;
   return AUTORELEASE([[self allocWithZone: NSDefaultMallocZone()] init]);
 }
 
+/**
+ * Returns a newly created dictionary with the keys and objects
+ * of otherDictionary.
+ * (The keys and objects are not copied.)
+ */
 + (id) dictionaryWithDictionary: (NSDictionary*)otherDictionary
 {
   return AUTORELEASE([[self allocWithZone: NSDefaultMallocZone()]
     initWithDictionary: otherDictionary]);
 }
 
+/**
+ * Returns a dictionary created using the given objects and keys.
+ * The two arrays must have the same size.
+ * The n th element of the objects array is associated with the n th
+ * element of the keys array.
+ */
 + (id) dictionaryWithObjects: (id*)objects 
 		     forKeys: (id*)keys
 		       count: (unsigned)count
@@ -215,6 +342,12 @@ static SEL	appSel;
   return [self count];
 }
 
+/**
+ * Initialises a dictionary created using the given objects and keys.
+ * The two arrays must have the same size.
+ * The n th element of the objects array is associated with the n th
+ * element of the keys array.
+ */
 - (id) initWithObjects: (NSArray*)objects forKeys: (NSArray*)keys
 {
   unsigned	objectCount = [objects count];
@@ -231,110 +364,30 @@ static SEL	appSel;
   return [self initWithObjects: os forKeys: ks count: objectCount];
 }
 
+/**
+ * Initialises a dictionary created using the list given as argument.
+ * The list is alernately composed of objects and keys.
+ * Thus, the list's length must be pair.
+ */
 - (id) initWithObjectsAndKeys: (id)firstObject, ...
 {
-  va_list ap;
-  int	capacity = 16;
-  int	num_pairs = 0;
-  id	*objects;
-  id	*keys;
-  id	arg;
-  int	argi = 1;
-
-  va_start (ap, firstObject);
-  if (firstObject == nil)
-    {
-      return [self init];
-    }
-  /* Gather all the arguments in a simple array, in preparation for
-     calling the designated initializer. */
-  objects = (id*)NSZoneMalloc(NSDefaultMallocZone(), sizeof(id) * capacity);
-  keys = (id*)NSZoneMalloc(NSDefaultMallocZone(), sizeof(id) * capacity);
-
-  objects[num_pairs] = firstObject;
-  /* Keep grabbing arguments until we get a nil... */
-  while ((arg = va_arg (ap, id)))
-    {
-      if (num_pairs >= capacity)
-	{
-	  /* Must increase capacity in order to fit additional ARG's. */
-	  capacity *= 2;
-	  objects = (id*)NSZoneRealloc(NSDefaultMallocZone(), objects,
-	    sizeof(id) * capacity);
-	  keys = (id*)NSZoneRealloc(NSDefaultMallocZone(), keys,
-	    sizeof(id) * capacity);
-	}
-      /* ...and alternately dump them into OBJECTS and KEYS */
-      if (argi++ % 2 == 0)
-	objects[num_pairs] = arg;
-      else
-	{
-	  keys[num_pairs] = arg;
-	  num_pairs++;
-	}
-    }
-  if (argi %2 != 0)
-    {
-      NSZoneFree(NSDefaultMallocZone(), objects);
-      NSZoneFree(NSDefaultMallocZone(), keys);
-      [NSException raise: NSInvalidArgumentException
-		  format: @"init dictionary with nil key"];
-    }
-  self = [self initWithObjects: objects forKeys: keys count: num_pairs];
-  NSZoneFree(NSDefaultMallocZone(), objects);
-  NSZoneFree(NSDefaultMallocZone(), keys);
+  GS_USEIDPAIRLIST(firstObject,
+    self = [self initWithObjects: __objects forKeys: __pairs count: __count/2]);
   return self;
 }
 
+/**
+ * Returns a dictionary created using the list given as argument.
+ * The list is alernately composed of objects and keys.
+ * Thus, the list's length must be pair.
+ */
 + (id) dictionaryWithObjectsAndKeys: (id)firstObject, ...
 {
-  va_list ap;
-  int	capacity = 16;
-  int	num_pairs = 0;
-  id	*objects;
-  id	*keys;
-  id	arg;
-  int	argi = 1;
+  id	o = [self allocWithZone: NSDefaultMallocZone()];
 
-  va_start (ap, firstObject);
-  /* Gather all the arguments in a simple array, in preparation for
-     calling the designated initializer. */
-  objects = (id*)NSZoneMalloc(NSDefaultMallocZone(), sizeof(id) * capacity);
-  keys = (id*)NSZoneMalloc(NSDefaultMallocZone(), sizeof(id) * capacity);
-  if (firstObject != nil)
-    {
-      NSDictionary *d;
-      objects[num_pairs] = firstObject;
-      /* Keep grabbing arguments until we get a nil... */
-      while ((arg = va_arg (ap, id)))
-	{
-	  if (num_pairs >= capacity)
-	    {
-	      /* Must increase capacity in order to fit additional ARG's. */
-	      capacity *= 2;
-	      objects = (id*)NSZoneRealloc(NSDefaultMallocZone(), objects,
-		sizeof(id) * capacity);
-	      keys = (id*)NSZoneRealloc(NSDefaultMallocZone(), keys,
-		sizeof(id) * capacity);
-	    }
-	  /* ...and alternately dump them into OBJECTS and KEYS */
-	  if (argi++ % 2 == 0)
-	    objects[num_pairs] = arg;
-	  else
-	    {
-	      keys[num_pairs] = arg;
-	      num_pairs++;
-	    }
-	}
-      NSAssert (argi % 2 == 0, NSInvalidArgumentException);
-      d = AUTORELEASE([[self allocWithZone: NSDefaultMallocZone()]
-	initWithObjects: objects forKeys: keys count: num_pairs]);
-      NSZoneFree(NSDefaultMallocZone(), objects);
-      NSZoneFree(NSDefaultMallocZone(), keys);
-      return d;
-    }
-  /* FIRSTOBJECT was nil; just return an empty NSDictionary object. */
-  return [self dictionary];
+  GS_USEIDPAIRLIST(firstObject,
+    o = [o initWithObjects: __objects forKeys: __pairs count: __count/2]);
+  return AUTORELEASE(o);
 }
 
 + (id) dictionaryWithObjects: (NSArray*)objects forKeys: (NSArray*)keys
@@ -343,6 +396,10 @@ static SEL	appSel;
     initWithObjects: objects forKeys: keys]);
 }
 
+/**
+ * Returns a dictionary containing only one object which is associated
+ * with a key.
+ */
 + (id) dictionaryWithObject: (id)object forKey: (id)key
 {
   return AUTORELEASE([[self allocWithZone: NSDefaultMallocZone()]
@@ -412,13 +469,32 @@ static SEL	appSel;
   return self;
 }
 
+/**
+ * <p>Initialises the dictionary with the contents of the specified file,
+ * which must contain a dictionary in property-list format.
+ * </p>
+ * <p>In GNUstep, the property-list format may be either the OpenStep
+ * format (ASCII data), or the MacOS-X format (URF8 XML data) ... this
+ * method will recognise which it is.
+ * </p>
+ * <p>If there is a failure to load the file for any reason, the receiver
+ * will be released and the method will return nil.
+ * </p>
+ * <p>Works by invoking [NSString-initWithContentsOfFile:] and
+ * [NSString-propertyList] then checking that the result is a dictionary.  
+ * </p>
+ */
 - (id) initWithContentsOfFile: (NSString*)path
 {
   NSString 	*myString;
 
   myString = [[NSString allocWithZone: NSDefaultMallocZone()]
     initWithContentsOfFile: path];
-  if (myString)
+  if (myString == nil)
+    {
+      DESTROY(self);
+    }
+  else
     {
       id result;
 
@@ -435,20 +511,90 @@ static SEL	appSel;
       if ([result isKindOfClass: NSDictionaryClass])
 	{
 	  self = [self initWithDictionary: result];
-	  return self;
+	}
+      else
+	{
+	  NSWarnMLog(@"Contents of file '%@' does not contain a dictionary",
+	    path);
+	  DESTROY(self);
 	}
     }
-  NSWarnMLog(@"Contents of file '%@' does not contain a dictionary", path);
-  RELEASE(self);
-  return nil;
+  return self;
 }
 
-+ (id) dictionaryWithContentsOfFile: (NSString *)path
+/**
+ * <p>Initialises the dictionary with the contents of the specified URL,
+ * which must contain a dictionary in property-list format.
+ * </p>
+ * <p>In GNUstep, the property-list format may be either the OpenStep
+ * format (ASCII data), or the MacOS-X format (URF8 XML data) ... this
+ * method will recognise which it is.
+ * </p>
+ * <p>If there is a failure to load the URL for any reason, the receiver
+ * will be released and the method will return nil.
+ * </p>
+ * <p>Works by invoking [NSString-initWithContentsOfURL:] and
+ * [NSString-propertyList] then checking that the result is a dictionary.  
+ * </p>
+ */
+- (id) initWithContentsOfURL: (NSURL*)aURL
+{
+  NSString 	*myString;
+
+  myString = [[NSString allocWithZone: NSDefaultMallocZone()]
+    initWithContentsOfURL: aURL];
+  if (myString == nil)
+    {
+      DESTROY(self);
+    }
+  else
+    {
+      id result;
+
+      NS_DURING
+	{
+	  result = [myString propertyList];
+	}
+      NS_HANDLER
+	{
+          result = nil;
+	}
+      NS_ENDHANDLER
+      RELEASE(myString);
+      if ([result isKindOfClass: NSDictionaryClass])
+	{
+	  self = [self initWithDictionary: result];
+	}
+      else
+	{
+	  NSWarnMLog(@"Contents of URL '%@' does not contain a dictionary",
+	    aURL);
+	  DESTROY(self);
+	}
+    }
+  return self;
+}
+
+/**
+ * Returns a dictionary using the file located at path.
+ * The file must be a property list containing a dictionary as its root object.
+ */
++ (id) dictionaryWithContentsOfFile: (NSString*)path
 {
   return AUTORELEASE([[self allocWithZone: NSDefaultMallocZone()]
     initWithContentsOfFile: path]);
 }
 
+/**
+ * Returns a dictionary using the contents of aURL.
+ * The URL must be a property list containing a dictionary as its root object.
+ */
++ (id) dictionaryWithContentsOfURL: (NSURL*)aURL
+{
+  return AUTORELEASE([[self allocWithZone: NSDefaultMallocZone()]
+    initWithContentsOfURL: aURL]);
+}
+
 - (BOOL) isEqual: other
 {
   if (other == self)
@@ -495,6 +641,9 @@ static SEL	appSel;
   return NO;
 }
 
+/**
+ * Returns an array containing all the dictionary's keys.
+ */
 - (NSArray*) allKeys
 {
   unsigned	c = [self count];
@@ -520,6 +669,9 @@ static SEL	appSel;
     }
 }
 
+/**
+ * Returns an array containing all the dictionary's objects.
+ */
 - (NSArray*) allValues
 {
   unsigned	c = [self count];
@@ -544,6 +696,10 @@ static SEL	appSel;
     }
 }
 
+/**
+ * Returns an array containing all the dictionary's keys that are
+ * associated with anObject.
+ */
 - (NSArray*) allKeysForObject: (id)anObject
 {
   unsigned	c;
@@ -630,313 +786,141 @@ compareIt(id o1, id o2, void* context)
     }
 }
 
+/**
+ * <p>Writes the contents of the dictionary to the file specified by path.
+ * The file contents will be in property-list format ... under GNUstep
+ * this is either OpenStep style (ASCII characters using \U hexadecimal
+ * escape sequences for unicode), or MacOS-X style (XML in the UTF8
+ * character set).
+ * </p>
+ * <p>If the useAuxiliaryFile flag is YES, the file write operation is
+ * atomic ... the data is written to a temporary file, which is then
+ * renamed to the actual file name.
+ * </p>
+ * <p>If the conversion of data into the correct property-list format fails
+ * or the write operation fails, the method returns NO, otherwise it
+ * returns YES.
+ * </p>
+ * <p>NB. The fact that the file is in property-list format does not
+ * necessarily mean that it can be used to reconstruct the dictionary using
+ * the -initWithContentsOfFile: method.  If the original dictionary contains
+ * non-property-list objects, the descriptions of those objects will
+ * have been written, and reading in the file as a property-list will
+ * result in a new dictionary containing the string descriptions.
+ * </p>
+ */
 - (BOOL) writeToFile: (NSString *)path atomically: (BOOL)useAuxiliaryFile
 {
-  extern BOOL	GSMacOSXCompatiblePropertyLists();
-  NSDictionary	*loc;
-  NSString	*desc;
-
-  loc = GSUserDefaultsDictionaryRepresentation();
+  NSDictionary	*loc = GSUserDefaultsDictionaryRepresentation();
+  NSString	*desc = nil;
+  NSData	*data;
 
   if (GSMacOSXCompatiblePropertyLists() == YES)
     {
-      extern NSString	*GSXMLPlMake(id obj, NSDictionary *loc);
-
-      desc = GSXMLPlMake(self, loc);
+      GSPropertyListMake(self, loc, YES, NO, 2, &desc);
+      data = [desc dataUsingEncoding: NSUTF8StringEncoding];
     }
   else
     {
-      NSMutableString	*result;
-
-      result = AUTORELEASE([[NSMutableString alloc] initWithCapacity:
-	20*[self count]]);
-      [self descriptionWithLocale: loc
-			   indent: 0
-			       to: (id<GNUDescriptionDestination>)result];
-      desc = result;
+      GSPropertyListMake(self, loc, NO, NO, 2, &desc);
+      data = [desc dataUsingEncoding: NSASCIIStringEncoding];
     }
-
-  return [desc writeToFile: path atomically: useAuxiliaryFile];
+  return [data writeToFile: path atomically: useAuxiliaryFile];
 }
 
+/**
+ * <p>Writes the contents of the dictionary to the specified url.
+ * This functions just like -writeToFile:atomically: except that the
+ * output may be written to any URL, not just a local file.
+ * </p>
+ */
 - (BOOL) writeToURL: (NSURL *)url atomically: (BOOL)useAuxiliaryFile
 {
-  extern BOOL	GSMacOSXCompatiblePropertyLists();
-  NSDictionary	*loc;
-  NSString	*desc;
-
-  loc = GSUserDefaultsDictionaryRepresentation();
+  NSDictionary	*loc = GSUserDefaultsDictionaryRepresentation();
+  NSString	*desc = nil;
+  NSData	*data;
 
   if (GSMacOSXCompatiblePropertyLists() == YES)
     {
-      extern NSString	*GSXMLPlMake(id obj, NSDictionary *loc);
-
-      desc = GSXMLPlMake(self, loc);
+      GSPropertyListMake(self, loc, YES, NO, 2, &desc);
+      data = [desc dataUsingEncoding: NSUTF8StringEncoding];
     }
   else
     {
-      NSMutableString	*result;
-
-      result = AUTORELEASE([[NSMutableString alloc] initWithCapacity:
-	20*[self count]]);
-      [self descriptionWithLocale: loc
-			   indent: 0
-			       to: (id<GNUDescriptionDestination>)result];
-      desc = result;
+      GSPropertyListMake(self, loc, NO, NO, 2, &desc);
+      data = [desc dataUsingEncoding: NSASCIIStringEncoding];
     }
 
-  return [desc writeToURL: url atomically: useAuxiliaryFile];
+  return [data writeToURL: url atomically: useAuxiliaryFile];
 }
 
+/**
+ * Returns the result of invoking -descriptionWithLocale:indent: with a nil
+ * locale and zero indent.
+ */
 - (NSString*) description
 {
   return [self descriptionWithLocale: nil indent: 0];
 }
 
+/**
+ * Returns the receiver as a text property list strings file format.<br />
+ * See [NSString-propertyListFromStringsFileFormat] for details.<br />
+ * The order of the items is undefined.
+ */
 - (NSString*) descriptionInStringsFileFormat
 {
-  NSMutableString	*result;
+  NSMutableString	*result = nil;
   NSEnumerator		*enumerator = [self keyEnumerator];
   IMP			nxtObj = [enumerator methodForSelector: nxtSel];
   IMP			myObj = [self methodForSelector: objSel];
-  IMP			appImp;
   id                    key;
 
-  result = AUTORELEASE([[NSMutableString alloc] initWithCapacity: 1024]);
-  appImp = [(NSObject*)result methodForSelector: appSel];
   while ((key = (*nxtObj)(enumerator, nxtSel)) != nil)
     {
       id val = (*myObj)(self, objSel, key);
 
-      [key descriptionWithLocale: nil
-			  indent: 0
-                              to: (id<GNUDescriptionDestination>)result];
+      GSPropertyListMake(key, nil, NO, YES, 0, &result);
       if (val != nil && [val isEqualToString: @""] == NO)
         {
-	  (*appImp)(result, appSel, @" = ");
-          [val descriptionWithLocale: nil
-			      indent: 0
-                                  to: (id<GNUDescriptionDestination>)result];
+	  [result appendString: @" = "];
+	  GSPropertyListMake(val, nil, NO, YES, 0, &result);
         }
-      (*appImp)(result, appSel, @";\n");
+      [result appendString: @";\n"];
     }
 
   return result;
 }
 
+/**
+ * Returns the result of invoking -descriptionWithLocale:indent: with
+ * a zero indent.
+ */
 - (NSString*) descriptionWithLocale: (NSDictionary*)locale
 {
   return [self descriptionWithLocale: locale indent: 0];
 }
 
+/**
+ * Returns the receiver as a text property list in the traditional format.<br />
+ * See [NSString-propertyList] for details.<br />
+ * If locale is nil, no formatting is done, otherwise entries are formatted
+ * according to the locale, and indented according to level.<br />
+ * Unless locale is nil, a level of zero indents items by four spaces,
+ * while a level of one indents them by a tab.<br />
+ * If the keys in the dictionary respond to -compare:, the items are
+ * listed by key in ascending order.  If not, the order in which the
+ * items are listed is undefined.
+ */
 - (NSString*) descriptionWithLocale: (NSDictionary*)locale
 			     indent: (unsigned int)level
 {
-  NSMutableString	*result;
+  NSMutableString	*result = nil;
 
-  result = AUTORELEASE([[NSMutableString alloc] initWithCapacity:
-    20*[self count]]);
-  [self descriptionWithLocale: locale
-		       indent: level
-			   to: (id<GNUDescriptionDestination>)result];
+  GSPropertyListMake(self, locale, NO, YES, level == 1 ? 3 : 2, &result);
   return result;
 }
 
-static NSString	*indentStrings[] = {
-  @"",
-  @"    ",
-  @"\t",
-  @"\t    ",
-  @"\t\t",
-  @"\t\t    ",
-  @"\t\t\t",
-  @"\t\t\t    ",
-  @"\t\t\t\t",
-  @"\t\t\t\t    ",
-  @"\t\t\t\t\t",
-  @"\t\t\t\t\t    ",
-  @"\t\t\t\t\t\t"
-};
-
-- (void) descriptionWithLocale: (NSDictionary*)locale
-			indent: (unsigned int)level
-			    to: (id<GNUDescriptionDestination>)result
-{
-  IMP			myObj = [self methodForSelector: objSel];
-  unsigned		i;
-  NSArray		*keyArray = [self allKeys];
-  unsigned		numKeys = [keyArray count];
-  NSString		*plists[numKeys];
-  NSString		*keys[numKeys];
-  IMP			appImp;
-
-  appImp = [(NSObject*)result methodForSelector: appSel];
-
-  [keyArray getObjects: keys];
-
-  if (locale == nil)
-    {
-      for (i = 0; i < numKeys; i++)
-	{
-	  plists[i] = (*myObj)(self, objSel, keys[i]);
-	}
-
-      (*appImp)(result, appSel, @"{");
-      for (i = 0; i < numKeys; i++)
-	{
-	  id	o = plists[i];
-
-	  [keys[i] descriptionWithLocale: nil indent: 0 to: result];
-	  (*appImp)(result, appSel, @" = ");
-	  [o descriptionWithLocale: nil indent: 0 to: result];
-	  (*appImp)(result, appSel, @"; ");
-	}
-      (*appImp)(result, appSel, @"}");
-    }
-  else
-    {
-      NSString	*iBaseString;
-      NSString	*iSizeString;
-      BOOL	canCompare = YES;
-      Class	lastClass = 0;
-
-      if (level < sizeof(indentStrings)/sizeof(id))
-	{
-	  iBaseString = indentStrings[level];
-	}
-      else
-	{
-	  iBaseString = indentStrings[sizeof(indentStrings)/sizeof(id)-1];
-	}
-      level++;
-      if (level < sizeof(indentStrings)/sizeof(id))
-	{
-	  iSizeString = indentStrings[level];
-	}
-      else
-	{
-	  iSizeString = indentStrings[sizeof(indentStrings)/sizeof(id)-1];
-	}
-
-      for (i = 0; i < numKeys; i++)
-	{
-	  if (GSObjCClass(keys[i]) == lastClass)
-	    continue;
-	  if ([keys[i] respondsToSelector: @selector(compare:)] == NO)
-	    {
-	      canCompare = NO;
-	      break;
-	    }
-	  lastClass = GSObjCClass(keys[i]);
-	}
-
-      if (canCompare == YES)
-	{
-	  /*
-	   * Shell sort algorithm taken from SortingInAction - a NeXT example
-	   * good value for stride factor is not well-understood
-	   * 3 is a fairly good choice (Sedgewick)
-	   */
-#define STRIDE_FACTOR 3
-	  unsigned	c,d, stride;
-	  BOOL		found;
-	  NSComparisonResult	(*comp)(id, SEL, id) = 0;
-	  int		count = numKeys;
-#ifdef	GSWARN
-	  BOOL		badComparison = NO;
-#endif
-
-	  stride = 1;
-	  while (stride <= count)
-	    {
-	      stride = stride * STRIDE_FACTOR + 1;
-	    }
-	  lastClass = 0;
-	  while (stride > (STRIDE_FACTOR - 1))
-	    {
-	      // loop to sort for each value of stride
-	      stride = stride / STRIDE_FACTOR;
-	      for (c = stride; c < count; c++)
-		{
-		  found = NO;
-		  if (stride > c)
-		    {
-		      break;
-		    }
-		  d = c - stride;
-		  while (!found)	// move to left until correct place
-		    {
-		      id			a = keys[d + stride];
-		      id			b = keys[d];
-		      Class			x;
-		      NSComparisonResult	r;
-
-		      x = GSObjCClass(a);
-		      if (x != lastClass)
-			{
-			  lastClass = x;
-			  comp = (NSComparisonResult (*)(id, SEL, id))
-			    [a methodForSelector: @selector(compare:)];
-			}
-		      r = (*comp)(a, @selector(compare:), b);
-		      if (r < 0)
-			{
-#ifdef	GSWARN
-			  if (r != NSOrderedAscending)
-			    {
-			      badComparison = YES;
-			    }
-#endif
-			  keys[d + stride] = b;
-			  keys[d] = a;
-			  if (stride > d)
-			    {
-			      break;
-			    }
-			  d -= stride;		// jump by stride factor
-			}
-		      else
-			{
-#ifdef	GSWARN
-			  if (r != NSOrderedDescending && r != NSOrderedSame)
-			    {
-			      badComparison = YES;
-			    }
-#endif
-			  found = YES;
-			}
-		    }
-		}
-	    }
-#ifdef	GSWARN
-	  if (badComparison == YES)
-	    {
-	      NSWarnMLog(@"Detected bad return value from comparison");
-	    }
-#endif
-	}
-
-      for (i = 0; i < numKeys; i++)
-	{
-	  plists[i] = (*myObj)(self, objSel, keys[i]);
-	}
-
-      (*appImp)(result, appSel, @"{\n");
-      for (i = 0; i < numKeys; i++)
-	{
-	  id	o = plists[i];
-
-	  (*appImp)(result, appSel, iSizeString);
-	  [keys[i] descriptionWithLocale: nil indent: 0 to: result];
-	  (*appImp)(result, appSel, @" = ");
-	  [o descriptionWithLocale: locale indent: level to: result];
-	  (*appImp)(result, appSel, @";\n");
-	}
-      (*appImp)(result, appSel, iBaseString);
-      (*appImp)(result, appSel, @"}");
-    }
-}
-
 /**
  * Default implementation for this class is to return the value stored in
  * the dictionary under the specified key, or nil if there is no value.
@@ -1159,4 +1143,21 @@ static NSString	*indentStrings[] = {
       [self setObject: value forKey: key];
     }
 }
+
+/**
+ * Default implementation for this class is equivalent to the
+ * -setObject:forKey: method unless value is nil, in which case
+ * it is equivalent to -removeObjectForKey:
+ */
+- (void) setValue: (id)value forKey: (NSString*)key
+{
+  if (value == nil)
+    {
+      [self removeObjectForKey: key];
+    }
+  else
+    {
+      [self setObject: value forKey: key];
+    }
+}
 @end
