@@ -22,27 +22,28 @@
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA.
 
    <title>NSDistantObject class reference</title>
-   $Date: 2002/02/03 17:21:20 $ $Revision: 1.41 $
+   $Date: 2004/01/18 12:19:43 $ $Revision: 1.58 $
    */
 
-#include <config.h>
-#include <base/preface.h>
-#include <base/DistributedObjects.h>
-#include <Foundation/NSLock.h>
-#include <Foundation/NSPort.h>
-#include <Foundation/NSMethodSignature.h>
-#include <Foundation/NSException.h>
-#include <Foundation/NSObjCRuntime.h>
-#include <Foundation/NSInvocation.h>
+#include "config.h"
+#include "GNUstepBase/preface.h"
+#include "GNUstepBase/DistributedObjects.h"
+#include "Foundation/NSLock.h"
+#include "Foundation/NSPort.h"
+#include "Foundation/NSMethodSignature.h"
+#include "Foundation/NSException.h"
+#include "Foundation/NSObjCRuntime.h"
+#include "Foundation/NSInvocation.h"
 
-#define DO_FORWARD_INVOCATION(_SELX, _ARG1)			\
+#define DO_FORWARD_INVOCATION(_SELX, _ARG1) ({			\
   sig = [self methodSignatureForSelector: @selector(_SELX)];	\
+  if (sig != nil) {						\
   inv = [NSInvocation invocationWithMethodSignature: sig];	\
   [inv setSelector: @selector(_SELX)];				\
   [inv setTarget: self];					\
   [inv setArgument: (void*)&_ARG1 atIndex: 2];			\
   [self forwardInvocation: inv];				\
-  [inv getReturnValue: &m]
+  [inv getReturnValue: &m]; }})
 
 
 static int	debug_proxy = 0;
@@ -54,10 +55,13 @@ typedef struct {
 } NSDO;
 
 @interface NSConnection (DistantObjectHacks)
-- (BOOL) includesProxyForTarget: (unsigned)target;
-- (void) addLocalObject: (id)obj;
-- (void) addProxy: (id)obj;
+- (void) aquireProxyForTarget: (unsigned)target;
+- (NSDistantObject*) retainOrAddLocal: (NSDistantObject*)aProxy
+			    forObject: (id)anObject;
+- (NSDistantObject*) retainOrAddProxy: (NSDistantObject*)aProxy
+			    forTarget: (unsigned)aTarget;
 - (void) removeProxy: (id)obj;
+- (void) vendLocal: (NSDistantObject*)aProxy;
 @end
 
 /* This is the proxy tag; it indicates where the local object is,
@@ -114,7 +118,7 @@ enum
 
 + (BOOL) respondsToSelector: (SEL)sel
 {
-  return (IMP)class_get_instance_method(self, sel) != (IMP)0;
+  return GSGetClassMethod(self, sel) != (GSMethod)0;
 }
 
 + (id) initWithCoder: (NSCoder*)aCoder
@@ -122,6 +126,7 @@ enum
   gsu8			proxy_tag;
   unsigned		target;
   id			decoder_connection;
+  NSDistantObject	*o;
 
 /*
   if ([aCoder isKindOfClass: [NSPortCoder class]] == NO)
@@ -155,7 +160,8 @@ enum
 	  NSLog(@"Receiving a proxy for local object 0x%x "
 		@"connection 0x%x\n", target, (gsaddr)decoder_connection);
 
-        if (![[decoder_connection class] includesLocalTarget: target])
+	o = [decoder_connection locateLocalTarget: target];
+        if (o == nil)
 	  {
 	    [NSException raise: @"ProxyDecodedBadTarget"
 			format: @"No local object with given target (0x%x)",
@@ -163,15 +169,12 @@ enum
 	  }
 	else
 	  {
-	    NSDistantObject	*o;
-
-	    o = [decoder_connection includesLocalTarget: target];
 	    if (debug_proxy)
 	      {
 		NSLog(@"Local object is 0x%x (0x%x)\n",
 		  (gsaddr)o, (gsaddr)o ? ((NSDO*)o)->_object : 0);
 	      }
-	    return o ? RETAIN(((NSDO*)o)->_object) : nil;
+	    return RETAIN(((NSDO*)o)->_object);
 	  }
 
       case PROXY_LOCAL_FOR_SENDER:
@@ -187,8 +190,9 @@ enum
 	if (debug_proxy)
 	  NSLog(@"Receiving a proxy, was local 0x%x connection 0x%x\n",
 		  target, (gsaddr)decoder_connection);
-	return [self initWithTarget: target
-			 connection: decoder_connection];
+	o = [self initWithTarget: target
+		      connection: decoder_connection];
+	return o;
 
       case PROXY_REMOTE_FOR_BOTH:
         /*
@@ -260,12 +264,11 @@ enum
 	    NSInternalInconsistencyException);
 
 	  /*
-	   *	If we don't already have a proxy for the object on the
-	   *	remote system, we must tell the other end to retain its
-	   *	local object for our use.
+	   * We may not already have a proxy for the object on the
+	   * remote system, we must tell the connection to make sure
+	   * the other end knows we are creating one.
 	   */
-	  if ([proxy_connection includesProxyForTarget: target] == NO)
-	    [proxy_connection retainTarget: target];
+	  [proxy_connection aquireProxyForTarget: target];
 
 	  /*
 	   *	Finally - we get a proxy via a direct connection to the
@@ -274,8 +277,9 @@ enum
 	   *	and will therefore go away when the current autorelease
 	   *	pool is destroyed.
 	   */
-	  return [self initWithTarget: target
-			   connection: proxy_connection];
+	  o = [self initWithTarget: target
+			connection: proxy_connection];
+	  return o;
         }
 
     default:
@@ -297,14 +301,14 @@ enum
    *	If there already is a local proxy for this target/connection
    *	combination, don't create a new one, just return the old one.
    */
-  if ((proxy = [aConnection localForObject: anObject]))
+  proxy = [aConnection retainOrAddLocal: nil forObject: anObject];
+  if (proxy == nil)
     {
-      return RETAIN(proxy);
-    }
-
-  proxy = (NSDistantObject*)NSAllocateObject(distantObjectClass,
+      proxy = (NSDistantObject*)NSAllocateObject(distantObjectClass,
 	0, NSDefaultMallocZone());
-  return [proxy initWithLocal: anObject connection: aConnection];
+      proxy = [proxy initWithLocal: anObject connection: aConnection];
+    }
+  return proxy;
 }
 
 + (id) initWithTarget: (unsigned)target connection: (NSConnection*)aConnection
@@ -314,17 +318,17 @@ enum
   NSAssert([aConnection isValid], NSInternalInconsistencyException);
 
   /*
-   *	If there already is a local proxy for this target/connection
-   *	combination, don't create a new one, just return the old one.
+   * If there already is a local proxy for this target/connection
+   * combination, don't create a new one, just return the old one.
    */
-  if ((proxy = [aConnection proxyForTarget: target]))
+  proxy = [aConnection retainOrAddProxy: nil forTarget: target];
+  if (proxy == nil)
     {
-      return RETAIN(proxy);
-    }
-
-  proxy = (NSDistantObject*)NSAllocateObject(distantObjectClass,
+      proxy = (NSDistantObject*)NSAllocateObject(distantObjectClass,
 	0, NSDefaultMallocZone());
-  return [proxy initWithTarget: target connection: aConnection];
+      proxy = [proxy initWithTarget: target connection: aConnection];
+    }
+  return proxy;
 }
 @end
 
@@ -339,6 +343,11 @@ enum
 }
 @end
 
+/**
+ * Instances of this class act as proxies to remote objects using
+ * the Distributed Objects system.  They also hold references to
+ * local objects which are vended to remote processes.
+ */
 @implementation NSDistantObject
 
 + (void) initialize
@@ -354,6 +363,10 @@ enum
   return placeHolder;
 }
 
+/**
+ * Creates and returns a proxy associated with aConnection
+ * which will hold a reference to the local object anObject.
+ */
 + (NSDistantObject*) proxyWithLocal: (id)anObject
 			 connection: (NSConnection*)aConnection
 {
@@ -361,6 +374,11 @@ enum
 				     connection: aConnection]);
 }
 
+/**
+ * Creates and returns a proxy associated with aConnection
+ * which will provide a link to a remote object whose reference
+ * locally is anObject.
+ */
 + (NSDistantObject*) proxyWithTarget: (unsigned)anObject
 			  connection: (NSConnection*)aConnection
 {
@@ -368,6 +386,10 @@ enum
 				      connection: aConnection]);
 }
 
+/**
+ * Returns the [NSConnection] instance with which the receiver is
+ * associated.
+ */
 - (NSConnection*) connectionForProxy
 {
   return _connection;
@@ -423,6 +445,10 @@ enum
 
 	  [aRmc encodeValueOfObjCType: @encode(typeof(proxy_target))
 				   at: &proxy_target];
+	  /*
+	   * Tell connection this object is being vended.
+	   */
+	  [_connection vendLocal: self];
 	}
       else
 	{
@@ -460,10 +486,10 @@ enum
       proxy_tag = PROXY_REMOTE_FOR_BOTH;
 
       /*
-       *	Get a proxy to refer to self - we send this to the other
-       *	side so we will be retained until the other side has
-       *	obtained a proxy to the original object via a connection
-       *	to the original vendor.
+       * Get a proxy to refer to self - we send this to the other
+       * side so we will be retained until the other side has
+       * obtained a proxy to the original object via a connection
+       * to the original vendor.
        */
       localProxy = [NSDistantObject proxyWithLocal: self
 					connection: encoder_connection];
@@ -475,8 +501,8 @@ enum
 		proxy_target, (gsaddr)_connection);
 
       /*
-       *	It's remote here, so we need to tell other side where to form
-       *	triangle connection to
+       * It's remote here, so we need to tell other side where to form
+       * triangle connection to
        */
       [aRmc encodeValueOfObjCType: @encode(typeof(proxy_tag))
 			       at: &proxy_tag];
@@ -488,6 +514,11 @@ enum
 			       at: &proxy_target];
 
       [aRmc encodeBycopyObject: proxy_connection_out_port];
+
+      /*
+       * Tell connection that localProxy is being vended.
+       */
+      [encoder_connection vendLocal: localProxy];
     }
 }
 
@@ -511,212 +542,24 @@ enum
 
 - (id) initWithCoder: (NSCoder*)aCoder
 {
-  gsu8		proxy_tag;
-  unsigned	target;
-  id		decoder_connection;
-
-/*
-  if ([aCoder isKindOfClass: [NSPortCoder class]] == NO)
-    {
-      RELEASE(self);
-      [NSException raise: NSGenericException
-		  format: @"NSDistantObject objects only decode with "
-			  @"NSPortCoder class"];
-    }
-*/
-
-  decoder_connection = [(NSPortCoder*)aCoder connection];
-  NSAssert(decoder_connection, NSInternalInconsistencyException);
-
-  /* First get the tag, so we know what values need to be decoded. */
-  [aCoder decodeValueOfObjCType: @encode(typeof(proxy_tag))
-			     at: &proxy_tag];
-
-  switch (proxy_tag)
-    {
-      case PROXY_LOCAL_FOR_RECEIVER:
-	/*
-	 *	This was a proxy on the other side of the connection, but
-	 *	here it's local.
-	 *	Lookup the target handle to ensure that it exists here.
-	 *	Return a retained copy of the local target object.
-	 */
-	[aCoder decodeValueOfObjCType: @encode(typeof(target))
-				   at: &target];
-
-        if (debug_proxy)
-	  NSLog(@"Receiving a proxy for local object 0x%x "
-		@"connection 0x%x\n", target, (gsaddr)decoder_connection);
-
-        if (![[decoder_connection class] includesLocalTarget: target])
-	  {
-	    RELEASE(self);
-	    [NSException raise: @"ProxyDecodedBadTarget"
-			format: @"No local object with given target (0x%x)",
-				target];
-	  }
-	else
-	  {
-	    NSDistantObject	*o;
-
-	    o = [decoder_connection includesLocalTarget: target];
-	    if (debug_proxy)
-	      {
-		NSLog(@"Local object is 0x%x (0x%x)\n",
-		  (gsaddr)o, (gsaddr)o ? o->_object : 0);
-	      }
-	    RELEASE(self);
-	    return o ? RETAIN(o->_object) : nil;
-	  }
-
-      case PROXY_LOCAL_FOR_SENDER:
-        /*
-	 *	This was a local object on the other side of the connection,
-	 *	but here it's a proxy object.  Get the target address, and
-	 *	send [NSDistantObject +proxyWithTarget:connection:]; this will
-	 *	return the proxy object we already created for this target, or
-	 *	create a new proxy object if necessary.
-	 */
-	[aCoder decodeValueOfObjCType: @encode(typeof(target))
-				   at: &target];
-	if (debug_proxy)
-	  NSLog(@"Receiving a proxy, was local 0x%x connection 0x%x\n",
-		  target, (gsaddr)decoder_connection);
-        RELEASE(self);
-	return RETAIN([NSDistantObject proxyWithTarget: target
-					    connection: decoder_connection]);
-
-      case PROXY_REMOTE_FOR_BOTH:
-        /*
-	 *	This was a proxy on the other side of the connection, and it
-	 *	will be a proxy on this side too; that is, the local version
-	 *	of this object is not on this host, not on the host the
-	 *	NSPortCoder is connected to, but on a *third* host.
-	 *	This is why I call this a "triangle connection".  In addition
-	 *	to decoding the target, we decode the OutPort object that we
-	 *	will use to talk directly to this third host.  We send
-	 *	[NSConnection +newForInPort:outPort:ancestorConnection:]; this
-	 *	will either return the connection already created for this
-	 *	inPort/outPort pair, or create a new connection if necessary.
-	 */
-	{
-	  NSDistantObject	*result;
-	  NSConnection		*proxy_connection;
-	  NSPort		*proxy_connection_out_port = nil;
-	  unsigned		intermediary;
-
-	  /*
-	   *	There is an object on the intermediary host that is keeping
-	   *	that hosts proxy for the original object retained, thus
-	   *	ensuring that the original is not released.  We create a
-	   *	proxy for that intermediate proxy.  When we release this
-	   *	proxy, the intermediary will be free to release it's proxy
-	   *	and the original can then be released.  Of course, by that
-	   *	time we will have obtained our own proxy for the original
-	   *	object ...
-	   */
-	  [aCoder decodeValueOfObjCType: @encode(typeof(intermediary))
-				     at: &intermediary];
-	  [NSDistantObject proxyWithTarget: intermediary
-				connection: decoder_connection];
-
-	  /*
-	   *	Now we get the target number and port for the orignal object
-	   *	and (if necessary) get the originating process to retain the
-	   *	object for us.
-	   */
-	  [aCoder decodeValueOfObjCType: @encode(typeof(target))
-				     at: &target];
-
-	  [aCoder decodeValueOfObjCType: @encode(id)
-				     at: &proxy_connection_out_port];
-
-	  NSAssert(proxy_connection_out_port, NSInternalInconsistencyException);
-	  /*
-	   #	If there already exists a connection for talking to the
-	   *	out port, we use that one rather than creating a new one from
-	   *	our listening port. 
-	   *
-	   *	First we try for a connection from our receive port,
-	   *	Then we try any connection to the send port
-	   *	Finally we resort to creating a new connection - we don't
-	   *	release the newly created connection - it will get released
-	   *	automatically when no proxies are left on it.
-	   */
-	  proxy_connection = [[decoder_connection class]
-	    connectionWithReceivePort: [decoder_connection receivePort]
-			     sendPort: proxy_connection_out_port];
-
-	  if (debug_proxy)
-	    NSLog(@"Receiving a triangle-connection proxy 0x%x "
-		  @"connection 0x%x\n", target, (gsaddr)proxy_connection);
-
-	  NSAssert(proxy_connection != decoder_connection,
-	    NSInternalInconsistencyException);
-	  NSAssert([proxy_connection isValid],
-	    NSInternalInconsistencyException);
-
-	  /*
-	   *	If we don't already have a proxy for the object on the
-	   *	remote system, we must tell the other end to retain its
-	   *	local object for our use.
-	   */
-	  if ([proxy_connection includesProxyForTarget: target] == NO)
-	    [proxy_connection retainTarget: target];
-
-	  result = RETAIN([NSDistantObject proxyWithTarget: target
-						connection: proxy_connection]);
-	  RELEASE(self);
-
-	  /*
-	   *	Finally - we have a proxy via a direct connection to the
-	   *	originating server.  We have also created a proxy to an
-	   *	intermediate object - but this proxy has not been retained
-	   *	and will therefore go away when the current autorelease
-	   *	pool is destroyed.
-	   */
-	  return result;
-        }
-
-    default:
-      /* xxx This should be something different than NSGenericException. */
-      RELEASE(self);
-      [NSException raise: NSGenericException
-		  format: @"Bad proxy tag"];
-    }
-  /* Not reached. */
+  DESTROY(self);
+  [NSException raise: NSGenericException
+	      format: @"NSDistantObject decodes from placeholder"];
   return nil;
 }
 
+/**
+ * Initialises and returns a proxy associated with aConnection
+ * which will hold a reference to the local object anObject.
+ */
 - (id) initWithLocal: (id)anObject connection: (NSConnection*)aConnection
 {
-  NSDistantObject	*new_proxy;
-
   NSAssert([aConnection isValid], NSInternalInconsistencyException);
 
-  /*
-   *	If there already is a local proxy for this target/connection
-   *	combination, don't create a new one, just return the old one.
-   */
-  if ((new_proxy = [aConnection localForObject: anObject]))
-    {
-      RETAIN(new_proxy);
-      RELEASE(self);
-      return new_proxy;
-    }
-
-  /*
-   *	We don't need to retain the oibject here - the connection
-   *	will retain the proxies local object if necessary (and release it
-   *	when all proxies referring to it have been released).
-   */
-  _object = anObject;
-
-  /*
-   *	We register this proxy with the connection using it.
-   */
+  _object = RETAIN(anObject);
+  _handle = 0;
   _connection = RETAIN(aConnection);
-  [_connection addLocalObject: self];
+  self = [_connection retainOrAddLocal: self forObject: anObject];
 
   if (debug_proxy)
     NSLog(@"Created new local=0x%x object 0x%x target 0x%x connection 0x%x\n",
@@ -725,36 +568,24 @@ enum
   return self;
 }
 
+/**
+ * Initialises and returns a proxy associated with aConnection
+ * which will provide a link to a remote object whose reference
+ * locally is target.
+ */
 - (id) initWithTarget: (unsigned)target connection: (NSConnection*)aConnection
 {
-  NSDistantObject	*new_proxy;
-
   NSAssert([aConnection isValid], NSInternalInconsistencyException);
 
-  /*
-   *	If there already is a proxy for this target/connection combination,
-   *	don't create a new one, just return the old one.
-   */
-  if ((new_proxy = [aConnection proxyForTarget: target]))
-    {
-      RETAIN(new_proxy);
-      RELEASE(self);
-      return new_proxy;
-    }
-
   _object = nil;
   _handle = target;
-
-  /*
-   *	We retain our connection so it can't disappear while the app
-   *	may want to use it.
-   */
   _connection = RETAIN(aConnection);
 
   /*
-   *	We register this object with the connection using it.
+   * We register this object with the connection using it.
+   * Conceivably this could result in self being changed.
    */
-  [_connection addProxy: self];
+  self = [_connection retainOrAddProxy: self forTarget: target];
 
   if (debug_proxy)
       NSLog(@"Created new proxy=0x%x target 0x%x connection 0x%x\n",
@@ -768,6 +599,17 @@ enum
   return get_imp(GSObjCClass((id)self), aSelector);
 }
 
+/**
+ * <p>Returns the method signature describing the arguments and return
+ * types of the method in the object referred to by the receiver
+ * which implements the aSelecto message.
+ * </p>
+ * <p>This method may need to refer to another process (causing relatively
+ * slow network communication) and apprximately double the time taken for
+ * sending a distributed objects message, so you are advised to use the
+ * -setProtocolForProxy: method to avoid this occurring.
+ * </p>
+ */
 - (NSMethodSignature*) methodSignatureForSelector: (SEL)aSelector
 {
   if (_object != nil)
@@ -793,6 +635,22 @@ enum
 	    }
 	  return sig;
 	}
+      /*
+       * Simlarly, when we fetch a method signature from the remote end,
+       * we get a proxy, and when we build a local signature we need to
+       * ask the proxy for its types ... and must avoid recursion again.
+       */
+      if (sel_eq(aSelector, @selector(methodType)))
+	{
+	  static	NSMethodSignature	*sig = nil;
+
+	  if (sig == nil)
+	    {
+	      sig = [NSMethodSignature signatureWithObjCTypes: "r*@:"];
+	      RETAIN(sig);
+	    }
+	  return sig;
+	}
 
       if (_protocol != nil)
 	{
@@ -800,10 +658,34 @@ enum
 
 	  struct objc_method_description* mth;
 
-	  mth = [_protocol descriptionForInstanceMethod: aSelector];
+	  /* Older gcc versions may not initialise Protocol objects properly
+	   * so we have an evil hack which checks for a known bad value of
+	   * the class pointer, and uses an internal function
+	   * (implemented in NSObject.m) to examine the protocol contents
+	   * without sending any ObjectiveC message to it.
+	   */
+	  if ((int)GSObjCClass(_protocol) == 0x2)
+	    {
+	      extern struct objc_method_description*
+		GSDescriptionForInstanceMethod();
+	      mth = GSDescriptionForInstanceMethod(_protocol, aSelector);
+	    }
+	  else
+	    {
+	      mth = [_protocol descriptionForInstanceMethod: aSelector];
+	    }
 	  if (mth == 0)
 	    {
-	      mth = [_protocol descriptionForClassMethod: aSelector];
+	      if ((int)GSObjCClass(_protocol) == 0x2)
+		{
+		  extern struct objc_method_description*
+		    GSDescriptionForClassMethod();
+		  mth = GSDescriptionForClassMethod(_protocol, aSelector);
+		}
+	      else
+		{
+		  mth = [_protocol descriptionForClassMethod: aSelector];
+		}
 	    }
 	  if (mth != 0)
 	    {
@@ -817,9 +699,8 @@ enum
 	}
       else
 	{
-	  id		m;
-	  const char	*types;
-#ifdef USE_FFCALL
+	  id		m = nil;
+#if	defined(USE_FFCALL) || defined(USE_LIBFFI)
 	  id		inv;
 	  id		sig;
 
@@ -840,12 +721,56 @@ enum
 	  retframe = [self forward: _cmd : args];
 	  m = retframe_id(retframe);
 #endif
-	  types = [m methodType];
-	  return [NSMethodSignature signatureWithObjCTypes: types];
+	  if ([m isProxy] == YES)
+	    {
+	      const char	*types = [m methodType];
+
+	      m = [NSMethodSignature signatureWithObjCTypes: types];
+	    }
+	  return m;
 	}
     }
 }
 
+/**
+ * <p>A key method for Distributed Objects performance.  This sets the
+ * a protocol that the distant object referred to by the proxy should
+ * conform to.  When messages in that protocol are sent to the proxy,
+ * the proxy knows that it does not need to ask the remote object for
+ * the method siagnature in order to send the message to it, but can
+ * send the message straight away based on the local method signature
+ * information obtained from the protocol.
+ * </p>
+ * <example>
+ *   if ([anObj isProxy] == YES)
+ *     {
+ *       [anObj setProtocolForProxy: @protocol(MyProtocol)];
+ *     }
+ * </example>
+ * <p>It is <em>highly recommended</em> that you make use of this facility,
+ * but you must beware that versions of the compiler prior to 3.3 suffer a
+ * serious bug with respect to the @protocol directive.  If the protocol
+ * referred to is not declared and implemented in the file where @protocol
+ * is used to refer to the protocol by name, a runtime error will occur
+ * when you try to use it.
+ * </p>
+ * <p>Beware, if you don't use this method to set the protocol, the system
+ * might well ask the remote process for method signature information, and
+ * the remote process might get it <em>wrong</em>.  This is because the
+ * class of the remote object needs to have been declared to conform to the
+ * protocol in order for it to know about any protocol qualifiers (the
+ * keywords <code>bycopy, byref, in, out, inout,</code> and
+ * <code>oneway</code>).  If the author of the server process forgot to do
+ * this, the type information returned from that process may not be what
+ * you are expecting.
+ * </p>
+ * The class of the server object should be declared like this ...
+ * <example>
+ * @interface MyServerClass : NSObject &lt;MyProtocol&gt;
+ * ...
+ * @end
+ * </example>
+ */
 - (void) setProtocolForProxy: (Protocol*)aProtocol
 {
   _protocol = aProtocol;
@@ -855,6 +780,9 @@ enum
 
 @implementation NSDistantObject(GNUstepExtensions)
 
+/**
+ * Used by the garbage collection system to tidy up when a proxy is destroyed.
+ */
 - (void) gcFinalize
 {
   if (_connection)
@@ -863,33 +791,30 @@ enum
 	NSLog(@"retain count for connection (0x%x) is now %u\n",
 		(gsaddr)_connection, [_connection retainCount]);
       /*
-       * A proxy for local object does not retain it's target - the
-       * NSConnection class does that for us - so we need not release it.
+       * A proxy for local object retains its target - so we release it.
        * For a local object the connection also retains this proxy, so we
        * can't be deallocated unless we are already removed from the
-       * connection.
+       * connection, and there is no need to remove self from connection.
        *
-       * A proxy retains it's connection so that the connection will
-       * continue to exist as long as there is a something to use it.
+       * A proxy has a nil local object, and retains it's connection so
+       * that the connection will continue to exist as long as there is
+       * a something to use it.
        * So we release our reference to the connection here just as soon
        * as we have removed ourself from the connection.
        */
       if (_object == nil)
 	[_connection removeProxy: self];
+      else
+	DESTROY(_object);
       RELEASE(_connection);
     }
 }
 
-- (id) awakeAfterUsingCoder: (NSCoder*)aDecoder
-{
-  return self;
-}
-
 static inline BOOL class_is_kind_of (Class self, Class aClassObject)
 {
   Class class;
 
-  for (class = self; class!=Nil; class = class_get_super_class(class))
+  for (class = self; class!=Nil; class = GSObjCSuper(class))
     if (class==aClassObject)
       return YES;
   return NO;
@@ -897,6 +822,11 @@ static inline BOOL class_is_kind_of (Class self, Class aClassObject)
 
 
 
+/**
+ * For backward compatibility ... do not use this method.<br />
+ * Returns the type information ... the modern way of doing this is
+ * with the -methodSignatureForSelector: method.
+ */
 - (const char *) selectorTypeForProxy: (SEL)selector
 {
 #if NeXT_RUNTIME
@@ -910,10 +840,14 @@ static inline BOOL class_is_kind_of (Class self, Class aClassObject)
 #endif
 }
 
+/**
+ * For backward compatibility ... do not use this method.<br />
+ * Handle old fashioned forwarding to the proxy.
+ */
 - (id) forward: (SEL)aSel :(arglist_t)frame
 {
   if (debug_proxy)
-    NSLog(@"NSDistantObject forwarding %s\n", sel_get_name(aSel));
+    NSLog(@"NSDistantObject forwarding %s\n", GSNameFromSelector(aSel));
 
   if (![_connection isValid])
     [NSException
@@ -932,6 +866,9 @@ static inline BOOL class_is_kind_of (Class self, Class aClassObject)
   return object_get_class (self);
 }
 
+/**
+ * Returns the class of the receiver.
+ */
 - (Class) classForPortCoder
 {
   return object_get_class (self);
@@ -939,8 +876,8 @@ static inline BOOL class_is_kind_of (Class self, Class aClassObject)
 
 - (BOOL) conformsToProtocol: (Protocol*)aProtocol
 {
-#ifdef USE_FFCALL
-  BOOL m;
+#if	defined(USE_FFCALL) || defined(USE_LIBFFI)
+  BOOL m = NO;
   id inv, sig;
   DO_FORWARD_INVOCATION(conformsToProtocol:, aProtocol);
   return m;
@@ -964,8 +901,8 @@ static inline BOOL class_is_kind_of (Class self, Class aClassObject)
 
 - (BOOL) respondsToSelector: (SEL)aSelector
 {
-#ifdef USE_FFCALL
-  BOOL m;
+#if	defined(USE_FFCALL) || defined(USE_LIBFFI)
+  BOOL m = NO;
   id inv, sig;
   DO_FORWARD_INVOCATION(respondsToSelector:, aSelector);
   return m;
@@ -1004,7 +941,7 @@ static inline BOOL class_is_kind_of (Class self, Class aClassObject)
 {
 #if NeXT_runtime
   {
-    Method m = class_get_instance_method(isa, selector);
+    Method m = GSGetInstanceMethod(isa, selector);
     if (m)
       return m->method_types;
     else
