@@ -21,13 +21,13 @@
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA.
 
    <title>NSEnumerator class reference</title>
-   $Date: 2001/12/18 16:54:14 $ $Revision: 1.15 $
+   $Date: 2003/07/31 23:49:31 $ $Revision: 1.17 $
 */
 
-#include <config.h>
-#include <base/preface.h>
-#include <Foundation/NSUtilities.h>
-#include <Foundation/NSArray.h>
+#include "config.h"
+#include "GNUstepBase/preface.h"
+#include "Foundation/NSUtilities.h"
+#include "Foundation/NSArray.h"
 
 @implementation NSEnumerator
 
