@@ -20,72 +20,94 @@
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA.
 
-   <title>NSException class reference</title>
-   $Date: 2002/01/03 10:01:25 $ $Revision: 1.34 $
+   $Date: 2003/07/31 23:49:31 $ $Revision: 1.46 $
 */
 
-#include <config.h>
-#include <base/preface.h>
-#include <Foundation/NSException.h>
-#include <Foundation/NSString.h>
-#include <Foundation/NSArray.h>
-#include <Foundation/NSCoder.h>
-#include <Foundation/NSThread.h>
-#include <Foundation/NSDictionary.h>
+#include "config.h"
+#include "GNUstepBase/preface.h"
+#include "Foundation/NSException.h"
+#include "Foundation/NSString.h"
+#include "Foundation/NSArray.h"
+#include "Foundation/NSCoder.h"
+#include "Foundation/NSThread.h"
+#include "Foundation/NSDictionary.h"
 #include <stdio.h>
-#include <stdlib.h>		// for getenv()
+
+/**
+ * A generic exception for general purpose usage.
+ */
+NSString* const NSGenericException
+  = @"NSGenericException";
+
+/**
+ * An exception for caes where unexpected state is detected within an object
+ */
+NSString* const NSInternalInconsistencyException
+  = @"NSInternalInconsistencyException";
+
+/**
+ * An exception used when an invalid argument is passed to a method
+ * or function.
+ */
+NSString* const NSInvalidArgumentException
+  = @"NSInvalidArgumentException";
+
+/**
+ * An exception used when the system faols to allocate required memory.
+ */
+NSString* const NSMallocException
+  = @"NSMallocException";
+
+/**
+ * An exception used when an illegal range is encountered ... usually this
+ * is used to provide more information than an invalid argument exception.
+ */
+NSString* const NSRangeException
+ = @"NSRangeException";
+
+/**
+ * An exception when character set conversion fails.
+ */
+NSString* const NSCharacterConversionException
+  = @"NSCharacterConversionException";
+
+/**
+ * An exception used when some form of parsing fails.
+ */
+NSString* const NSParseErrorException
+  = @"NSParseErrorException";
+
+#include "GSPrivate.h"
 
 static void
 _preventRecursion (NSException *exception)
 {
+  fprintf(stderr, "recursion encountered handling uncaught exception\n");
+  fflush(stderr);	/* NEEDED UNDER MINGW */
 }
 
 static void
 _NSFoundationUncaughtExceptionHandler (NSException *exception)
 {
-  const char	*c = getenv("CRASH_ON_ABORT");
-  BOOL		a;
+  BOOL			a;
+  extern const char*	GSArgZero(void);
 
   _NSUncaughtExceptionHandler = _preventRecursion;
-  fprintf(stderr, "Uncaught exception %s, reason: %s\n",
-    	[[exception name] lossyCString], [[exception reason] lossyCString]);
-/* FIXME: need to implement this:
-  NSLogError("Uncaught exception %@, reason: %@",
-    	[exception name], [exception reason]);
-*/
+#if 1
+  fprintf(stderr, "%s: Uncaught exception %s, reason: %s\n", GSArgZero(),
+    [[exception name] lossyCString], [[exception reason] lossyCString]);
+  fflush(stderr);	/* NEEDED UNDER MINGW */
+#else
+  NSLog("Uncaught exception %@, reason: %@",
+    [exception name], [exception reason]);
+#endif
 
 #ifdef	DEBUG
   a = YES;		// abort() by default.
 #else
   a = NO;		// exit() by default.
 #endif
-  if (c != 0)
-    {
-      /*
-       * Use the CRASH_ON_ABORT environment variable ... if it's defined
-       * then we use abort(), unless it's 'no', 'false', or '0, in which
-       * case we use exit()
-       */
-      if (c[0] == '0' && c[1] == 0)
-	{
-	  a = NO;
-	}
-      else if ((c[0] == 'n' || c[0] == 'N') && (c[1] == 'o' || c[1] == 'O')
-	&& c[2] == 0)
-	{
-	  a = NO;
-	}
-      else if ((c[0] == 'f' || c[0] == 'F') && (c[1] == 'a' || c[1] == 'A')
-	&& (c[2] == 'l' || c[2] == 'L') && (c[3] == 's' || c[3] == 'S')
-	&& (c[4] == 'e' || c[4] == 'E') && c[5] == 0)
-	{
-	  a = NO;
-	}
-      else
-	{
-	  a = YES;
-	}
-    }
+  a = GSEnvironmentFlag("CRASH_ON_ABORT", a);
   if (a == YES)
     {
       abort();
@@ -96,8 +118,43 @@ _NSFoundationUncaughtExceptionHandler (NSException *exception)
     }
 }
 
+/**
+   <p>
+   The NSException class helps manage errors in a program. It provides
+   a mechanism for lower-level methods to provide information about
+   problems to higher-level methods, which more often than not, have a
+   better ability to decide what to do about the problems.
+   </p>
+   <p>
+   Exceptions are typically handled by enclosing a sensitive section
+   of code inside the macros NS_DURING and NS_HANDLER, and then
+   handling any problems after this, up to the NS_ENDHANDLER macro:
+   </p>
+   <example>
+   NS_DURING
+    code that might cause an exception
+   NS_HANDLER
+    code that deals with the exception. If this code cannot deal with
+    it, you can re-raise the exception like this
+    [localException raise]
+    so the next higher level of code can handle it
+   NS_ENDHANDLER
+   </example>
+   <p>
+   The local variable localException is the name of the exception
+   object you can use in the NS_HANDLER section.
+   The easiest way to cause an exeption is using the +raise:format:,...
+   method.
+   </p>
+*/
 @implementation NSException
 
+/**
+   Create an an exception object with a name, reason and a dictionary
+   userInfo which can be used to provide additional information or
+   access to objects needed to handle the exception. After the
+   exception is created you must -raise it.
+*/
 + (NSException*) exceptionWithName: (NSString*)name
 			    reason: (NSString*)reason
 			  userInfo: (NSDictionary*)userInfo
@@ -106,6 +163,11 @@ _NSFoundationUncaughtExceptionHandler (NSException *exception)
 				   userInfo: userInfo]);
 }
 
+/**
+   Creates an exception with a name and a reason using the
+   format string and any additional arguments. The exception is then
+   raised.
+ */
 + (void) raise: (NSString*)name
 	format: (NSString*)format,...
 {
@@ -113,10 +175,15 @@ _NSFoundationUncaughtExceptionHandler (NSException *exception)
 
   va_start(args, format);
   [self raise: name format: format arguments: args];
-  // FIXME: This probably doesn't matter, but va_end won't get called
+  // This probably doesn't matter, but va_end won't get called
   va_end(args);
 }
 
+/**
+   Creates an exception with a name and a reason string using the
+   format string and additional arguments specified as a variable
+   argument list argList. The exception is then raised.
+ */
 + (void) raise: (NSString*)name
 	format: (NSString*)format
      arguments: (va_list)argList
@@ -129,6 +196,10 @@ _NSFoundationUncaughtExceptionHandler (NSException *exception)
   [except raise];
 }
 
+/**
+   <init/>Initializes a newly allocated NSException object with a
+   name, reason and a dictionary userInfo.
+*/
 - (id) initWithName: (NSString*)name
 	     reason: (NSString*)reason
 	   userInfo: (NSDictionary*)userInfo
@@ -147,6 +218,13 @@ _NSFoundationUncaughtExceptionHandler (NSException *exception)
   [super dealloc];
 }
 
+/**
+   Raises the exception. All code following the raise will not be
+   executed and program control will be transfered to the closest
+   calling method which encapsulates the exception code in an
+   NS_DURING macro, or to the uncaught exception handler if there is no
+   other handling code.
+*/
 - (void) raise
 {
   NSThread	*thread;
@@ -170,6 +248,7 @@ _NSFoundationUncaughtExceptionHandler (NSException *exception)
   longjmp(handler->jumpState, 1);
 }
 
+/** Returns the name of the exception */
 - (NSString*) name
 {
   if (_e_name != nil)
@@ -182,6 +261,7 @@ _NSFoundationUncaughtExceptionHandler (NSException *exception)
     }
 }
 
+/** Returns the exception reason */
 - (NSString*) reason
 {
   if (_e_reason != nil)
@@ -194,6 +274,7 @@ _NSFoundationUncaughtExceptionHandler (NSException *exception)
     }
 }
 
+/** Returns the exception userInfo dictionary */
 - (NSDictionary*) userInfo
 {
   return _e_info;
