@@ -24,18 +24,18 @@
  * Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA.
  *
  * <title>NSHashTable class reference</title>
- * $Date: 2002/03/08 07:11:10 $ $Revision: 1.26 $
+ * $Date: 2003/07/31 23:49:31 $ $Revision: 1.33 $
  */ 
 
 /**** Included Headers *******************************************************/
 
-#include <config.h>
-#include <Foundation/NSZone.h>
-#include <Foundation/NSString.h>
-#include <Foundation/NSArray.h>
-#include <Foundation/NSException.h>
-#include <Foundation/NSHashTable.h>
-#include <Foundation/NSDebug.h>
+#include "config.h"
+#include "Foundation/NSZone.h"
+#include "Foundation/NSString.h"
+#include "Foundation/NSArray.h"
+#include "Foundation/NSException.h"
+#include "Foundation/NSHashTable.h"
+#include "Foundation/NSDebug.h"
 #include "NSCallBacks.h"
 
 
@@ -56,7 +56,7 @@
  (M->extra.retain)((NSHashTable*)M, X.ptr)
 #define GSI_MAP_ENUMERATOR	NSHashEnumerator
 
-#include <base/GSIMap.h>
+#include "GNUstepBase/GSIMap.h"
 
 /**
  * Returns an array of all the objects in the table.
@@ -122,16 +122,17 @@ NSCompareHashTables(NSHashTable *table1, NSHashTable *table2)
     }
   else
     {
-      GSIMapNode        n = t1->firstNode;
-
-	while (n != 0)
-	  {
+      NSHashEnumerator enumerator = GSIMapEnumeratorForMap((GSIMapTable)t1);
+      GSIMapNode n;
+      while ((n = GSIMapEnumeratorNextNode(&enumerator)) != 0)
+        {
           if (GSIMapNodeForKey(t2, n->key) == 0)
             {
+	      GSIMapEndEnumerator((GSIMapEnumerator)&enumerator);
               return NO;
             }
-          n = n->nextInMap;
         }
+      GSIMapEndEnumerator((GSIMapEnumerator)&enumerator);
       return YES;
     }
 }
@@ -144,7 +145,8 @@ NSCopyHashTableWithZone(NSHashTable *table, NSZone *zone)
 {
   GSIMapTable   t;
   GSIMapNode    n;
-
+  NSHashEnumerator enumerator;
+  
   if (table == 0)
     {
       NSWarnFLog(@"Nul table argument supplied");
@@ -154,12 +156,12 @@ NSCopyHashTableWithZone(NSHashTable *table, NSZone *zone)
   t = (GSIMapTable)NSZoneMalloc(zone, sizeof(GSIMapTable_t));
   GSIMapInitWithZoneAndCapacity(t, zone, ((GSIMapTable)table)->nodeCount);
   t->extra = ((GSIMapTable)table)->extra;
-  n = ((GSIMapTable)table)->firstNode;
-  while (n != 0)
+  enumerator = GSIMapEnumeratorForMap((GSIMapTable)table);
+  while ((n = GSIMapEnumeratorNextNode(&enumerator)) != 0)
     {
       GSIMapAddKey(t, n->key);
-      n = n->nextInMap;
     }
+  GSIMapEndEnumerator((GSIMapEnumerator)&enumerator);
 
   return (NSHashTable*)t;
 }
@@ -180,7 +182,9 @@ NSCountHashTable(NSHashTable *table)
 
 /**
  * Create a new hash table by calling NSCreateHashTableWithZone() using
- * NSDefaultMallocZone().
+ * NSDefaultMallocZone().<br />
+ * If capacity is small or 0, then the returned
+ * table has a reasonable (but still small) capacity. 
  */
 NSHashTable *
 NSCreateHashTable(
@@ -225,7 +229,7 @@ NSCreateHashTableWithZone(
 
 /**
  * Function to be called when finished with the enumerator.
- * Not required in GNUstep ... just provided for MacOS-X compatibility.
+ * This permits memory used by the enumerator to be released.
  */
 void
 NSEndHashTableEnumeration(NSHashEnumerator *enumerator)
@@ -233,19 +237,21 @@ NSEndHashTableEnumeration(NSHashEnumerator *enumerator)
   if (enumerator == 0)
     {
       NSWarnFLog(@"Nul enumerator argument supplied");
+      return;
     }
+  GSIMapEndEnumerator((GSIMapEnumerator)enumerator);
 }
 
 /**
  * Return an enumerator for stepping through a map table using the
- * NSNextHashEnumeratorPair() function.
+ * NSNextHashEnumeratorItem() function.
  */
 NSHashEnumerator
 NSEnumerateHashTable(NSHashTable *table)
 {
   if (table == 0)
     {
-      NSHashEnumerator	v = { 0, 0 };
+      NSHashEnumerator	v = { 0, 0, 0 };
 
       NSWarnFLog(@"Nul table argument supplied");
       return v;
@@ -257,7 +263,9 @@ NSEnumerateHashTable(NSHashTable *table)
 }
 
 /**
- * Destroy the hash table and relase its contents.
+ * Releases all the keys and values of table (using the callbacks
+ * specified at the time of table's creation), and then proceeds
+ * to deallocate the space allocated for table itself.
  */
 void
 NSFreeHashTable(NSHashTable *table)
@@ -276,7 +284,7 @@ NSFreeHashTable(NSHashTable *table)
 }
 
 /**
- * Returns the value for the specified element, or a nul pointer if the
+ * Returns the value for the specified element, or a null pointer if the
  * element is not found in the table.
  */
 void *
@@ -303,7 +311,7 @@ NSHashGet(NSHashTable *table, const void *element)
 /**
  * Adds the element to table.<br />
  * If an equal element is already in table, replaces it with the new one.<br />
- * If element is nul raises an NSInvalidArgumentException.
+ * If element is null raises an NSInvalidArgumentException.
  */
 void
 NSHashInsert(NSHashTable *table, const void *element)
@@ -314,12 +322,12 @@ NSHashInsert(NSHashTable *table, const void *element)
   if (table == 0)
     {
       [NSException raise: NSInvalidArgumentException
-                  format: @"Attempt to place value in nul hash table"];
+                  format: @"Attempt to place value in null hash table"];
     }
   if (element == 0)
     {
       [NSException raise: NSInvalidArgumentException
-                  format: @"Attempt to place nul in hash table"];
+                  format: @"Attempt to place null in hash table"];
     }
   n = GSIMapNodeForKey(t, (GSIMapKey)element);
   if (n == 0)
@@ -351,12 +359,12 @@ NSHashInsertIfAbsent(NSHashTable *table, const void *element)
   if (table == 0)
     {
       [NSException raise: NSInvalidArgumentException
-                  format: @"Attempt to place value in nul hash table"];
+                  format: @"Attempt to place value in null hash table"];
     }
   if (element == 0)
     {
       [NSException raise: NSInvalidArgumentException
-                  format: @"Attempt to place nul in hash table"];
+                  format: @"Attempt to place null in hash table"];
     }
   n = GSIMapNodeForKey(t, (GSIMapKey)element);
   if (n == 0)
@@ -373,7 +381,7 @@ NSHashInsertIfAbsent(NSHashTable *table, const void *element)
 /**
  * Adds the element to table and returns nul.<br />
  * If an equal element is already present, raises NSInvalidArgumentException.
- * <br />If element is nul raises an NSInvalidArgumentException.
+ * <br />If element is null raises an NSInvalidArgumentException.
  */
 void
 NSHashInsertKnownAbsent(NSHashTable *table, const void *element)
@@ -384,12 +392,12 @@ NSHashInsertKnownAbsent(NSHashTable *table, const void *element)
   if (table == 0)
     {
       [NSException raise: NSInvalidArgumentException
-                  format: @"Attempt to place value in nul hash table"];
+                  format: @"Attempt to place value in null hash table"];
     }
   if (element == 0)
     {
       [NSException raise: NSInvalidArgumentException
-                  format: @"Attempt to place nul in hash table"];
+                  format: @"Attempt to place null in hash table"];
     }
   n = GSIMapNodeForKey(t, (GSIMapKey)element);
   if (n == 0)
@@ -445,7 +453,7 @@ NSNextHashEnumeratorItem(NSHashEnumerator *enumerator)
 }
 
 /**
- * Empty the hash table, but preserve its capacity.
+ * Empty the hash table (releasing all elements), but preserve its capacity.
  */
 void
 NSResetHashTable(NSHashTable *table)
@@ -495,3 +503,73 @@ NSStringFromHashTable(NSHashTable *table)
   return string;
 }
 
+
+
+/* These are to increase readabilty locally. */
+typedef unsigned int (*NSHT_hash_func_t)(NSHashTable *, const void *);
+typedef BOOL (*NSHT_isEqual_func_t)(NSHashTable *, const void *, const void *);
+typedef void (*NSHT_retain_func_t)(NSHashTable *, const void *);
+typedef void (*NSHT_release_func_t)(NSHashTable *, void *);
+typedef NSString *(*NSHT_describe_func_t)(NSHashTable *, const void *);
+
+/** For sets of pointer-sized or smaller quantities. */
+const NSHashTableCallBacks NSIntHashCallBacks =
+{
+  (NSHT_hash_func_t) _NS_int_hash,
+  (NSHT_isEqual_func_t) _NS_int_is_equal,
+  (NSHT_retain_func_t) _NS_int_retain,
+  (NSHT_release_func_t) _NS_int_release,
+  (NSHT_describe_func_t) _NS_int_describe
+};
+
+/** For sets of pointers hashed by address. */
+const NSHashTableCallBacks NSNonOwnedPointerHashCallBacks = 
+{
+  (NSHT_hash_func_t) _NS_non_owned_void_p_hash,
+  (NSHT_isEqual_func_t) _NS_non_owned_void_p_is_equal,
+  (NSHT_retain_func_t) _NS_non_owned_void_p_retain,
+  (NSHT_release_func_t) _NS_non_owned_void_p_release,
+  (NSHT_describe_func_t) _NS_non_owned_void_p_describe
+};
+
+/** For sets of objects without retaining and releasing. */
+const NSHashTableCallBacks NSNonRetainedObjectHashCallBacks = 
+{
+  (NSHT_hash_func_t) _NS_non_retained_id_hash,
+  (NSHT_isEqual_func_t) _NS_non_retained_id_is_equal,
+  (NSHT_retain_func_t) _NS_non_retained_id_retain,
+  (NSHT_release_func_t) _NS_non_retained_id_release,
+  (NSHT_describe_func_t) _NS_non_retained_id_describe
+};
+
+/** For sets of objects; similar to [NSSet]. */
+const NSHashTableCallBacks NSObjectHashCallBacks = 
+{
+  (NSHT_hash_func_t) _NS_id_hash,
+  (NSHT_isEqual_func_t) _NS_id_is_equal,
+  (NSHT_retain_func_t) _NS_id_retain,
+  (NSHT_release_func_t) _NS_id_release,
+  (NSHT_describe_func_t) _NS_id_describe
+};
+
+/** For sets of pointers with transfer of ownership upon insertion. */
+const NSHashTableCallBacks NSOwnedPointerHashCallBacks = 
+{
+  (NSHT_hash_func_t) _NS_owned_void_p_hash,
+  (NSHT_isEqual_func_t) _NS_owned_void_p_is_equal,
+  (NSHT_retain_func_t) _NS_owned_void_p_retain,
+  (NSHT_release_func_t) _NS_owned_void_p_release,
+  (NSHT_describe_func_t) _NS_owned_void_p_describe
+};
+
+/** For sets of pointers to structs when the first field of the
+ * struct is the size of an int. */
+const NSHashTableCallBacks NSPointerToStructHashCallBacks = 
+{
+  (NSHT_hash_func_t) _NS_int_p_hash,
+  (NSHT_isEqual_func_t) _NS_int_p_is_equal,
+  (NSHT_retain_func_t) _NS_int_p_retain,
+  (NSHT_release_func_t) _NS_int_p_release,
+  (NSHT_describe_func_t) _NS_int_p_describe
+};
+
