@@ -1,7 +1,7 @@
 /** Implementation of NSInvocation for GNUStep
-   Copyright (C) 1998 Free Software Foundation, Inc.
+   Copyright (C) 1998,2003 Free Software Foundation, Inc.
    
-   Written:     Richard Frith-Macdonald <richard@brainstorm.co.uk>
+   Author:     Richard Frith-Macdonald <richard@brainstorm.co.uk>
    Date: August 1998
    Based on code by: Andrew Kachites McCallum <mccallum@gnu.ai.mit.edu>
    
@@ -22,14 +22,14 @@
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA.
 
    <title>NSInvocation class reference</title>
-   $Date: 2002/02/20 05:08:46 $ $Revision: 1.46 $
+   $Date: 2003/08/24 23:07:40 $ $Revision: 1.60 $
    */ 
 
-#include <Foundation/NSException.h>
-#include <Foundation/NSCoder.h>
-#include <Foundation/NSInvocation.h>
-#include <base/GSInvocation.h>
-#include <config.h>
+#include "Foundation/NSException.h"
+#include "Foundation/NSCoder.h"
+#include "Foundation/NSInvocation.h"
+#include "GSInvocation.h"
+#include "config.h"
 #include <mframe.h>
 #if defined(USE_LIBFFI)
 #include "cifframe.h"
@@ -37,23 +37,66 @@
 #include "callframe.h"
 #endif
 
-
 static Class   NSInvocation_abstract_class;
 static Class   NSInvocation_concrete_class;
 
+@interface GSInvocationProxy
+{
+@public
+  Class		isa;
+  id		target;
+  NSInvocation	*invocation;
+}
++ (id) _newWithTarget: (id)t;
+- (NSInvocation*) _invocation;
+- (void) forwardInvocation: (NSInvocation*)anInvocation;
+- (NSMethodSignature*) methodSignatureForSelector: (SEL)aSelector;
+@end
+@interface GSMessageProxy : GSInvocationProxy
+@end
+
+
+
+/**
+ * <p>The NSInvocation class implements a mechanism of constructing
+ * messages (as NSInvocation instances), sending these to other
+ * objects, and handling the returned values.
+ * </p>
+ * <p>An NSInvocation object may contain a target object to which a
+ * message can be sent, or may send the message to an arbitrary object.<br />
+ * Each message consists of a selector for that method and an argument
+ * list.  Once the message has been sent, the invocation will contain
+ * a return value whose contents may be copied out of it.
+ * </p>
+ * <p>The target, selector, and arguments of an instance be constructed
+ * dynamically, providing a great deal of power/flexibility.
+ * </p>
+ * <p>The sending of the message to the target object (using the -invoke
+ * or -invokeWithTarget: method) can be done at any time, but a standard
+ * use of this is by the [NSObject-forwardInvocation:] method which is
+ * called whenever a method is not implemented by the class of the
+ * object to which it was sent.
+ * </p>
+ * <p>Related to the class are two convenience macros ... NS_MESSAGE()
+ * and NS_INVOCATION() to allow easy construction of invocations with
+ * all the arguments set up.
+ * </p>
+ */
 @implementation NSInvocation
 
 #ifdef USE_LIBFFI
 static inline void
 _get_arg(NSInvocation *inv, int index, void *buffer)
 {
-  cifframe_get_arg((cifframe_t *)inv->_cframe, index, buffer);
+  cifframe_get_arg((cifframe_t *)inv->_cframe, index, buffer,
+		   inv->_info[index+1].size);
 }
 
 static inline void
 _set_arg(NSInvocation *inv, int index, void *buffer)
 {
-  cifframe_set_arg((cifframe_t *)inv->_cframe, index, buffer);
+  cifframe_set_arg((cifframe_t *)inv->_cframe, index, buffer,
+		   inv->_info[index+1].size);
 }
 
 static inline void *
@@ -131,6 +174,13 @@ _arg_addr(NSInvocation *inv, int index)
     }
 }
 
+/**
+ * Returns an invocation instance which can be used to send messages to
+ * a target object using the described signature.<br />
+ * You must set the target and selector (using -setTarget: and -setSelector:)
+ * before you attempt to use the invocation.<br />
+ * Raises an NSInvalidArgumentException if the signature is nil.
+ */
 + (NSInvocation*) invocationWithMethodSignature: (NSMethodSignature*)_signature
 {
   return AUTORELEASE([[NSInvocation_concrete_class alloc]
@@ -145,7 +195,7 @@ _arg_addr(NSInvocation *inv, int index)
       _argsRetained = NO;
       if (_cframe && _sig)
 	{
-	  int	i;
+	  unsigned int	i;
 
 	  for (i = 3; i <= _numArgs; i++)
 	    {
@@ -166,10 +216,14 @@ _arg_addr(NSInvocation *inv, int index)
 	    }
 	}
     }
+
+
+  CLEAR_RETURN_VALUE_IF_OBJECT;
+  
 #if	defined(USE_LIBFFI)
   if (_cframe)
     {
-      cifframe_free((cifframe_t *)_cframe);
+      NSZoneFree(NSDefaultMallocZone(), _cframe);
       _retval = 0;	// Part of _cframe
     }
 #elif defined(USE_FFCALL)
@@ -192,10 +246,12 @@ _arg_addr(NSInvocation *inv, int index)
   [super dealloc];
 }
 
-/*
- *      Accessing message elements.
+/**
+ * Copies the argument identified by index into the memory location specified
+ * by the buffer argument.<br />
+ * An index of zero is the target object, an index of one is the selector,
+ * so the actual method arguments start at index 2.
  */
-
 - (void) getArgument: (void*)buffer
 	     atIndex: (int)index
 {
@@ -218,6 +274,11 @@ _arg_addr(NSInvocation *inv, int index)
     }		
 }
 
+/**
+ * Copies the invocations return value to the location pointed to by buffer
+ * if a return value has been set (see the -setReturnValue: method).<br />
+ * If there isn't a return value then this method raises an exception.
+ */
 - (void) getReturnValue: (void*)buffer
 {
   const char	*type;
@@ -241,11 +302,25 @@ _arg_addr(NSInvocation *inv, int index)
     }
 }
 
+/**
+ * Returns the selector of the invocation (the argument at index 1)
+ */
 - (SEL) selector
 {
   return _selector;
 }
 
+/**
+ * Sets the argument identified by index from the memory location specified
+ * by the buffer argument.<br />
+ * Using an index of 0 is equivalent to calling -setTarget: and using an
+ * argument of 1 is equivalent to -setSelector:<br />
+ * Proper arguments start at index 2.<br />
+ * NB. Unlike -setTarget: and -setSelector: the value of buffer must be
+ * <em>a pointer to</em> the argument to be set in the invocation.<br />
+ * If -retainArguments was called, then any object argument set in the
+ * receiver is retained by it.
+ */
 - (void) setArgument: (void*)buffer
 	     atIndex: (int)index
 {
@@ -312,12 +387,17 @@ _arg_addr(NSInvocation *inv, int index)
     }		
 }
 
+/**
+ * Sets the return value of the invocation to the item that buffer points to.
+ */
 - (void) setReturnValue: (void*)buffer
 {
   const char	*type;
 
   type = _info[0].type;
 
+  CLEAR_RETURN_VALUE_IF_OBJECT;
+  
   if (*type != _C_VOID)
     {
       int	length = _info[0].size;
@@ -328,14 +408,23 @@ _arg_addr(NSInvocation *inv, int index)
 #endif
       memcpy(_retval, buffer, length);
     }
+
+  RETAIN_RETURN_VALUE;
   _validReturn = YES;
 }
 
+/**
+ * Sets the selector for the invocation.
+ */
 - (void) setSelector: (SEL)aSelector
 {
   _selector = aSelector;
 }
 
+/**
+ * Sets the target object for the invocation.<br />
+ * If -retainArguments was called, then the target is retained.
+ */
 - (void) setTarget: (id)anObject
 {
   if (_argsRetained)
@@ -345,20 +434,27 @@ _arg_addr(NSInvocation *inv, int index)
   _target = anObject;
 }
 
+/**
+ * Returns the target object of the invocation.
+ */
 - (id) target
 {
   return _target;
 }
 
-/*
- *      Managing arguments.
+/**
+ * Returns a flag to indicate whether object arguments of the invocation
+ * (including its target) are retained by the invocation.
  */
-
 - (BOOL) argumentsRetained
 {
   return _argsRetained;
 }
 
+/**
+ * Instructs the invocation to retain its object arguments (including the
+ * target).  The default is not to retain them.
+ */
 - (void) retainArguments
 {
   if (_argsRetained)
@@ -367,7 +463,7 @@ _arg_addr(NSInvocation *inv, int index)
     }
   else
     {
-      int	i;
+      unsigned int	i;
 
       _argsRetained = YES;
       IF_NO_GC(RETAIN(_target));
@@ -408,15 +504,17 @@ _arg_addr(NSInvocation *inv, int index)
     }		
 }
 
-/*
- *      Dispatching an Invocation.
+/**
+ * Sends the message encapsulated in the invocation to its target.
  */
-
 - (void) invoke
 {
   [self invokeWithTarget: _target];
 }
 
+/**
+ * Sends the message encapsulated in the invocation to anObject.
+ */
 - (void) invokeWithTarget: (id)anObject
 {
   id		old_target;
@@ -424,11 +522,16 @@ _arg_addr(NSInvocation *inv, int index)
   IMP		imp;
   int		stack_argsize;
 
+
+  CLEAR_RETURN_VALUE_IF_OBJECT;
+  _validReturn = NO;
+
   /*
    *	A message to a nil object returns nil.
    */
   if (anObject == nil)
     {
+      _validReturn = YES;
       memset(_retval, '\0', _info[0].size);	/* Clear return value */
       return;
     }
@@ -455,18 +558,18 @@ _arg_addr(NSInvocation *inv, int index)
       s.receiver = _target;
 #endif
       if (GSObjCIsInstance(_target))
-	s.class = class_get_super_class(GSObjCClass(_target));
+	s.class = GSObjCSuper(GSObjCClass(_target));
       else
-	s.class = class_get_super_class((Class)_target);
+	s.class = GSObjCSuper((Class)_target);
       imp = objc_msg_lookup_super(&s, _selector);
     }
   else
     {
-      imp = method_get_imp(object_is_instance(_target) ?
-	class_get_instance_method(
+      imp = method_get_imp(GSObjCIsInstance(_target) ?
+	GSGetInstanceMethod(
 		    ((struct objc_class*)_target)->class_pointer, _selector)
-	: class_get_class_method(
-		    ((struct objc_class*)_target)->class_pointer, _selector));
+	: GSGetClassMethod(
+		    ((struct objc_class*)_target), _selector));
 	/*
 	 * If fast lookup failed, we may be forwarding or something ...
 	 */
@@ -478,28 +581,20 @@ _arg_addr(NSInvocation *inv, int index)
 
   stack_argsize = [_sig frameLength];
 
-#ifdef USE_LIBFFI
-  ffi_call(&((cifframe_t *)_cframe)->cif, (void(*)(void))imp, _retval,
-	   ((cifframe_t *)_cframe)->values);
-  if (_info[0].size)
-    {
-      cifframe_decode_return(_info[0].type, _retval);
-    }
-#else
   returned = __builtin_apply((void(*)(void))imp,
     (arglist_t)_cframe, stack_argsize);
   if (_info[0].size)
     {
       mframe_decode_return(_info[0].type, _retval, returned);
     }
-#endif
+
+  RETAIN_RETURN_VALUE;
   _validReturn = YES;
 }
 
-/*
- *      Getting the method _signature.
+/**
+ * Returns the method signature of the invocation.
  */
-
 - (NSMethodSignature*) methodSignature
 {
   return _sig;
@@ -513,20 +608,20 @@ _arg_addr(NSInvocation *inv, int index)
    */
   char buffer[1024];
 
-  sprintf (buffer, "<%s %p selector: %s target: %s>", \
-                (char*)object_get_class_name(self), \
+  snprintf (buffer, 1024, "<%s %p selector: %s target: %s>", \
+                GSClassNameFromObject(self), \
                 self, \
-                _selector ? [NSStringFromSelector(_selector) cString] : "nil", \
-                _target ? [NSStringFromClass([_target class]) cString] : "nil" \
+                _selector ? GSNameFromSelector(_selector) : "nil", \
+                _target ?   GSNameFromClass([_target class]) : "nil" \
                 );
 
-  return [NSString stringWithCString:buffer];
+  return [NSString stringWithCString: buffer];
 }
 
 - (void) encodeWithCoder: (NSCoder*)aCoder
 {
   const char	*types = [_sig methodType];
-  int		i;
+  unsigned int	i;
 
   [aCoder encodeValueOfObjCType: @encode(char*)
 			     at: &types];
@@ -575,8 +670,8 @@ _arg_addr(NSInvocation *inv, int index)
   NSMethodSignature	*newSig;
   const char		*types;
   void			*datum;
-  int			i;
-
+  unsigned int		i;
+  
   [aCoder decodeValueOfObjCType: @encode(char*) at: &types];
   newSig = [NSMethodSignature signatureWithObjCTypes: types];
   NSZoneFree(NSDefaultMallocZone(), (void*)types);
@@ -598,7 +693,7 @@ _arg_addr(NSInvocation *inv, int index)
         const char      *t = _info[i].type;
         if (*t == _C_STRUCT_B || *t == _C_UNION_B || *t == _C_ARY_B)
           {
-	    *(void**)datum = _fastMallocBuffer(_info[i].size);
+	    *(void**)datum = GSAutoreleasedBuffer(_info[i].size);
             datum = *(void**)datum;
           }
       }
@@ -611,25 +706,40 @@ _arg_addr(NSInvocation *inv, int index)
     {
       [aCoder decodeValueOfObjCType: @encode(BOOL) at: &_validReturn];
       if (_validReturn)
-	{
-	  [aCoder decodeValueOfObjCType: _info[0].type at: _retval];
-	}
+        {
+          [aCoder decodeValueOfObjCType: _info[0].type at: _retval];
+        }
     }
   return self;
 }
 
 @end
 
+/**
+ * Provides some minor extensions and some utility methods to aid
+ * integration of NSInvocation with the ObjectiveC runtime.
+ */
 @implementation NSInvocation (GNUstep)
 
+/**
+ * Internal use.<br />
+ * Initialises the receiver with a known selector and argument list
+ * as supplied to the forward:: method by the ObjectiveC runtime
+ * when it is unable to locate an implementation for mthe selector
+ * in a class.
+ */
 - (id) initWithArgframe: (arglist_t)frame selector: (SEL)aSelector
 {
   [self subclassResponsibility: _cmd];
   return nil;
 }
 
-/*
- *	This is the de_signated initialiser.
+/** <init /><override-subclass />
+ * Initialised an invocation instance which can be used to send messages to
+ * a target object using aSignature.<br />
+ * You must set the target and selector (using -setTarget: and -setSelector:)
+ * before you attempt to use the invocation.<br />
+ * Raises an NSInvalidArgumentException if aSignature is nil.
  */
 - (id) initWithMethodSignature: (NSMethodSignature*)aSignature
 {
@@ -637,21 +747,57 @@ _arg_addr(NSInvocation *inv, int index)
   return nil;
 }
 
+/**
+ * Tries to produce a method signature based on aSelector and uses that to
+ * initialise self by calling the -initWithMethodSignature: method.<br />
+ * If the argument type of aSelector cannot be determined, this releases self
+ * and returns nil.
+ */
 - (id) initWithSelector: (SEL)aSelector
 {
-  return [self initWithArgframe: 0 selector: aSelector];
+  const char *types;
+  NSMethodSignature *newSig;
+
+  types = sel_get_type(aSelector);
+  if (types == 0)
+    {
+      types = sel_get_type(sel_get_any_typed_uid(GSNameFromSelector(aSelector)));
+    }
+  if (types == 0)
+    {
+      NSLog(@"Couldn't find encoding type for selector %s.",
+	   GSNameFromSelector(aSelector));
+      RELEASE(self);
+      return nil;
+    }
+  newSig = [NSMethodSignature signatureWithObjCTypes: types];
+  return [self initWithMethodSignature: newSig];
 }
 
+/**
+ * Initialises the reciever with the specified target, selector, and
+ * a variable number of arguments.
+ */
 - (id) initWithTarget: anObject selector: (SEL)aSelector, ...
 {
-  va_list	ap;
+  va_list	     ap;
+  NSMethodSignature *newSig;
 
-  self = [self initWithArgframe: 0 selector: aSelector];
+  if (anObject)
+    {
+      newSig = [anObject methodSignatureForSelector: aSelector];
+      self = [self initWithMethodSignature: newSig];
+    }
+  else
+    {
+      self = [self initWithSelector: aSelector];
+    }
   if (self)
     {
-      int	i;
+      unsigned int	i;
 
       [self setTarget: anObject];
+      [self setSelector: aSelector];
       va_start (ap, aSelector);
       for (i = 3; i <= _numArgs; i++)
 	{
@@ -659,9 +805,6 @@ _arg_addr(NSInvocation *inv, int index)
 	  unsigned	size = _info[i].size;
 	  void		*datum;
 
-#ifdef USE_LIBFFI
-	  size = ((cifframe_t *)_cframe)->args[i-1]->size;
-#endif
 	  datum = _arg_addr(self, i-1);
 
 #define CASE_TYPE(_C,_T) case _C: *(_T*)datum = va_arg (ap, _T); break
@@ -738,22 +881,58 @@ _arg_addr(NSInvocation *inv, int index)
   return self;
 }
 
+/**
+ * Internal use.<br />
+ * Provides a return frame that the ObjectiveC runtime can use to
+ * return the result of an invocation to a calling function.
+ */
 - (void*) returnFrame: (arglist_t)argFrame
 {
   [self subclassResponsibility: _cmd];
   return NULL;
 }
 
+/**
+ * Returns the status of the flag set by -setSendsToSuper:
+ */
 - (BOOL) sendsToSuper
 {
   return _sendToSuper;
 }
 
+/**
+ * Sets the flag to tell the invocation that it should actually invoke a
+ * method in the superclass of the target rather than the method of the
+ * target itsself.<br />
+ * This extension permits an invocation to act like a regular method
+ * call sent to <em>super</em> in the method of a class.
+ */
 - (void) setSendsToSuper: (BOOL)flag
 {
   _sendToSuper = flag;
 }
+@end
 
+/**
+ * These methods are for internal use only ... not public API<br />
+ * They are used by the NS_INVOCATION() and NS_MESSAGE() macros to help
+ * create invocations.
+ */
+@implementation NSInvocation (MacroSetup)
++ (id) _newProxyForInvocation: (id)target
+{
+  return [GSInvocationProxy _newWithTarget: target];
+}
++ (id) _newProxyForMessage: (id)target
+{
+  return [GSMessageProxy _newWithTarget: target];
+}
++ (NSInvocation*) _returnInvocationAndDestroyProxy: (id)proxy
+{
+  NSInvocation	*inv = [proxy _invocation];
+  NSDeallocateObject(proxy);
+  return inv;
+}
 @end
 
 @implementation NSInvocation (BackwardCompatibility)
@@ -769,22 +948,7 @@ _arg_addr(NSInvocation *inv, int index)
 
 - (id) initWithArgframe: (arglist_t)frame selector: (SEL)aSelector
 {
-  const char		*types;
-  NSMethodSignature	*newSig;
-
-  types = sel_get_type(aSelector);
-  if (types == 0)
-    {
-      types = sel_get_type(sel_get_any_typed_uid(sel_get_name(aSelector)));
-    }
-  if (types == 0)
-    {
-      [NSException raise: NSInvalidArgumentException
-		  format: @"Couldn't find encoding type for selector %s.",
-			 sel_get_name(aSelector)];
-    }
-  newSig = [NSMethodSignature signatureWithObjCTypes: types];
-  self = [self initWithMethodSignature: newSig];
+  self = [self initWithSelector: aSelector];
   if (self)
     {
       [self setSelector: aSelector];
@@ -793,7 +957,7 @@ _arg_addr(NSInvocation *inv, int index)
        */
       if (frame)
 	{
-	  int	i;
+	  unsigned int	i;
 
 	  mframe_get_arg(frame, &_info[1], &_target);
 	  for (i = 1; i <= _numArgs; i++)
@@ -810,6 +974,11 @@ _arg_addr(NSInvocation *inv, int index)
  */
 - (id) initWithMethodSignature: (NSMethodSignature*)aSignature
 {
+  if (aSignature == nil)
+    {
+      RELEASE(self);
+      return nil;
+    }
   _sig = RETAIN(aSignature);
   _numArgs = [aSignature numberOfArguments];
   _info = [aSignature methodInfo];
@@ -826,3 +995,47 @@ _arg_addr(NSInvocation *inv, int index)
   return mframe_handle_return(_info[0].type, _retval, argFrame);
 }
 @end
+
+
+@implementation	GSInvocationProxy
++ (id) _newWithTarget: (id)t
+{
+  GSInvocationProxy	*o;
+  o = (GSInvocationProxy*) NSAllocateObject(self, 0, NSDefaultMallocZone());
+  o->target = RETAIN(t);
+  return o;
+}
+- (NSInvocation*) _invocation
+{
+  return invocation;
+}
+- (retval_t) forward: (SEL)aSel : (arglist_t)argFrame
+{
+  NSInvocation	*inv;
+
+  if (aSel == 0)
+    [NSException raise: NSInvalidArgumentException
+		format: @"%@ null selector given", NSStringFromSelector(_cmd)];
+
+  inv = AUTORELEASE([[NSInvocation alloc] initWithArgframe: argFrame
+						  selector: aSel]);
+  [self forwardInvocation: inv];
+  return [inv returnFrame: argFrame];
+}
+- (void) forwardInvocation: (NSInvocation*)anInvocation
+{
+  invocation = anInvocation;
+}
+- (NSMethodSignature*) methodSignatureForSelector: (SEL)aSelector
+{
+  return [target methodSignatureForSelector: aSelector];
+}
+@end
+
+@implementation	GSMessageProxy
+- (NSInvocation*) _invocation
+{
+  [invocation setTarget: target];
+  return invocation;
+}
+@end
