@@ -1,8 +1,9 @@
 /** Mutual exclusion locking classes
-   Copyright (C) 1996 Free Software Foundation, Inc.
+   Copyright (C) 1996,2003 Free Software Foundation, Inc.
 
    Author:  Scott Christley <scottc@net-community.com>
    Created: 1996
+   Author:  Richard Frith-Macdonald <rfm@gnu.org>
    
    This file is part of the GNUstep Objective-C Library.
 
@@ -21,18 +22,81 @@
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA.
 
    <title>NSLock class reference</title>
-   $Date: 2002/02/20 06:42:05 $ $Revision: 1.19 $
+   $Date: 2004/02/17 12:55:01 $ $Revision: 1.28 $
 */ 
 
-#include <config.h>
+#include "config.h"
 #include <errno.h>
-#if HAVE_UNISTD_H
+#ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif
-#include <base/preface.h>
-#include <Foundation/NSLock.h>
-#include <Foundation/NSException.h>
-#include <Foundation/NSDebug.h>
+#include "GNUstepBase/preface.h"
+#include "Foundation/NSLock.h"
+#include "Foundation/NSException.h"
+#include "Foundation/NSDebug.h"
+#include "Foundation/NSThread.h"
+#ifdef NeXT_RUNTIME
+#include "thr-mach.h"
+#endif
+
+#define _MUTEX     ((objc_mutex_t)_mutex)
+#define _CONDITION ((objc_condition_t)_condition)
+
+extern void		GSSleepUntilIntervalSinceReferenceDate(NSTimeInterval);
+extern NSTimeInterval	GSTimeNow();
+
+typedef struct {
+  NSTimeInterval	end;
+  NSTimeInterval	i0;
+  NSTimeInterval	i1;
+  NSTimeInterval	max;
+} GSSleepInfo;
+
+static void GSSleepInit(NSDate *limit, GSSleepInfo *context)
+{
+  context->end = [limit timeIntervalSinceReferenceDate];
+  context->i0 = 0.0;
+  context->i1 = 0.0001;		// Initial pause interval.
+  context->max = 0.25;		// Maximum pause interval.
+}
+
+/**
+ * <p>Using a pointer to a context structure initialised using GSSleepInit()
+ * we either pause for a while and return YES or, if the limit date
+ * has passed, return NO.
+ * </p>
+ * <p>The pause intervals start off very small, but rapidly increase
+ * (following a fibonacci sequence) up to a maximum value.
+ * </p>
+ * <p>We use the GSSleepUntilIntervalSinceReferenceDate() function to
+ * avoid objc runtime messaging overheads and overheads of creating and
+ * destroying temporary date objects.
+ * </p>
+ */
+static BOOL GSSleepOrFail(GSSleepInfo *context)
+{
+  NSTimeInterval	when = GSTimeNow();
+  NSTimeInterval	tmp;
+
+  if (when >= context->end)
+    {
+      return NO;
+    }
+  tmp = context->i0 + context->i1;
+  context->i0 = context->i1;
+  context->i1 = tmp;
+  if (tmp > context->max)
+    {
+      tmp = context->max;
+    }
+  when += tmp;
+  if (when > context->end)
+    {
+      when = context->end;
+    }
+  GSSleepUntilIntervalSinceReferenceDate(when);
+  return YES;		// Paused.
+}
 
 // Exceptions
 
@@ -67,6 +131,16 @@ NSString *NSRecursiveLockException = @"NSRecursiveLockException";
 // NSLock class
 // Simplest lock for protecting critical sections of code
 
+/**
+ * An NSLock is used in multi-threaded applications to protect critical
+ * pieces of code. While one thread holds a lock within a piece of code,
+ * another thread cannot execute that code until the first thread has
+ * given up it's hold on the lock. The limitation of NSLock is that
+ * you can only lock an
+ * NSLock once and it must be unlocked before it can be aquired again.<br />
+ * Other lock classes, notably NSRecursiveLock, have
+ * different restrictions.
+ */
 @implementation NSLock
 
 // Designated initializer
@@ -99,65 +173,72 @@ NSString *NSRecursiveLockException = @"NSRecursiveLockException";
     {
       // Ask the runtime to deallocate the mutex
       // If there are outstanding locks then it will block
-      if (objc_mutex_deallocate(_mutex) == -1)
+      if (objc_mutex_deallocate(_MUTEX) == -1)
 	{
 	  NSWarnMLog(@"objc_mutex_deallocate() failed");
 	}
+      _mutex = 0;
     }
 }
 
-// Try to acquire the lock
-// Does not block
+/**
+ * Attempts to aquire a lock, but returns immediately if the lock
+ * cannot be aquired. It returns YES if the lock is aquired. It returns
+ * NO if the lock cannot be aquired or if the current thread already has
+ * the lock.
+ */
 - (BOOL) tryLock
 {
   /* Return NO if we're already locked */
-  if (_mutex->owner == objc_thread_id())
+  if (_MUTEX->owner == objc_thread_id())
     {	
       return NO;
     }
 
   // Ask the runtime to acquire a lock on the mutex
-  if (objc_mutex_trylock(_mutex) == -1)
+  if (objc_mutex_trylock(_MUTEX) == -1)
     {
       return NO;
     }
   return YES;
 }
 
-- (BOOL) lockBeforeDate: (NSDate *)limit
+/**
+ * Attempts to aquire a lock before the date limit passes. It returns YES
+ * if it can. It returns NO if it cannot, or if the current thread already
+ * has the lock (but it waits until the time limit is up before returning
+ * NO).
+ */
+- (BOOL) lockBeforeDate: (NSDate*)limit
 {
-  int x;
+  int		x;
+  GSSleepInfo	ctxt;
+
+  GSSleepInit(limit, &ctxt);
 
   /* This is really the behavior of OpenStep, if the current thread has
      the lock, we just block until the time limit is up. Very odd */
-  while (_mutex->owner == objc_thread_id()
-	 || (x = objc_mutex_trylock(_mutex)) == -1)
+  while (_MUTEX->owner == objc_thread_id()
+    || (x = objc_mutex_trylock(_MUTEX)) == -1)
     {
-      NSDate *current = [NSDate date];
-      NSComparisonResult compare;
-      
-      compare = [current compare: limit];
-      if (compare == NSOrderedSame || compare == NSOrderedDescending)
+      if (GSSleepOrFail(&ctxt) == NO)
 	{
 	  return NO;
 	}
-      /*
-       * This should probably be more accurate like usleep(250)
-       * but usleep is known to NOT be thread safe under all architectures.
-       */
-      sleep(1);
     }
   return YES;
 }
 
-// NSLocking protocol
+/**
+ * Attempts to aquire a lock, and waits until it can do so.
+ */
 - (void) lock
 {
-  CHECK_RECURSIVE_LOCK(_mutex);
+  CHECK_RECURSIVE_LOCK(_MUTEX);
 
   // Ask the runtime to acquire a lock on the mutex
   // This will block
-  if (objc_mutex_lock(_mutex) == -1)
+  if (objc_mutex_lock(_MUTEX) == -1)
     {
       [NSException raise: NSLockException
         format: @"failed to lock mutex"];
@@ -168,7 +249,7 @@ NSString *NSRecursiveLockException = @"NSRecursiveLockException";
 - (void) unlock
 {
   // Ask the runtime to release a lock on the mutex
-  if (objc_mutex_unlock(_mutex) == -1)
+  if (objc_mutex_unlock(_MUTEX) == -1)
     {
       [NSException raise: NSLockException
 		  format: @"unlock: failed to unlock mutex"];
@@ -228,7 +309,7 @@ NSString *NSRecursiveLockException = @"NSRecursiveLockException";
   if (_condition != 0)
     {
       // Ask the runtime to deallocate the condition
-      if (objc_condition_deallocate(_condition) == -1)
+      if (objc_condition_deallocate(_CONDITION) == -1)
 	{
 	  NSWarnMLog(@"objc_condition_deallocate() failed");
 	}
@@ -237,7 +318,7 @@ NSString *NSRecursiveLockException = @"NSRecursiveLockException";
     {
       // Ask the runtime to deallocate the mutex
       // If there are outstanding locks then it will block
-      if (objc_mutex_deallocate(_mutex) == -1)
+      if (objc_mutex_deallocate(_MUTEX) == -1)
 	{
 	  NSWarnMLog(@"objc_mutex_deallocate() failed");
 	}
@@ -253,9 +334,9 @@ NSString *NSRecursiveLockException = @"NSRecursiveLockException";
 // Acquiring and release the lock
 - (void) lockWhenCondition: (int)value
 {
-  CHECK_RECURSIVE_CONDITION_LOCK(_mutex);
+  CHECK_RECURSIVE_CONDITION_LOCK(_MUTEX);
 
-  if (objc_mutex_lock(_mutex) == -1)
+  if (objc_mutex_lock(_MUTEX) == -1)
     {
       [NSException raise: NSConditionLockException
         format: @"lockWhenCondition: failed to lock mutex"];
@@ -264,7 +345,7 @@ NSString *NSRecursiveLockException = @"NSRecursiveLockException";
 
   while (_condition_value != value)
     {
-      if (objc_condition_wait(_condition, _mutex) == -1)
+      if (objc_condition_wait(_CONDITION, _MUTEX) == -1)
         {
           [NSException raise: NSConditionLockException
             format: @"objc_condition_wait failed"];
@@ -278,7 +359,7 @@ NSString *NSRecursiveLockException = @"NSRecursiveLockException";
   int depth;
 
   // First check to make sure we have the lock
-  depth = objc_mutex_trylock(_mutex);
+  depth = objc_mutex_trylock(_MUTEX);
 
   // Another thread has the lock so abort
   if (depth == -1)
@@ -301,7 +382,7 @@ NSString *NSRecursiveLockException = @"NSRecursiveLockException";
   _condition_value = value;
 
   // wake up blocked threads
-  if (objc_condition_broadcast(_condition) == -1)
+  if (objc_condition_broadcast(_CONDITION) == -1)
     {
       [NSException raise: NSConditionLockException
         format: @"unlockWithCondition: objc_condition_broadcast failed"];
@@ -309,8 +390,8 @@ NSString *NSRecursiveLockException = @"NSRecursiveLockException";
     }
 
   // and unlock twice
-  if ((objc_mutex_unlock(_mutex) == -1)
-      || (objc_mutex_unlock(_mutex) == -1))
+  if ((objc_mutex_unlock(_MUTEX) == -1)
+    || (objc_mutex_unlock(_MUTEX) == -1))
     {
       [NSException raise: NSConditionLockException
         format: @"unlockWithCondition: failed to unlock mutex"];
@@ -320,10 +401,10 @@ NSString *NSRecursiveLockException = @"NSRecursiveLockException";
 
 - (BOOL) tryLock
 {
-  CHECK_RECURSIVE_CONDITION_LOCK(_mutex);
+  CHECK_RECURSIVE_CONDITION_LOCK(_MUTEX);
 
   // Ask the runtime to acquire a lock on the mutex
-  if (objc_mutex_trylock(_mutex) == -1)
+  if (objc_mutex_trylock(_MUTEX) == -1)
     return NO;
   else
     return YES;
@@ -351,23 +432,18 @@ NSString *NSRecursiveLockException = @"NSRecursiveLockException";
 // Acquiring the lock with a date condition
 - (BOOL) lockBeforeDate: (NSDate*)limit
 { 
-  CHECK_RECURSIVE_CONDITION_LOCK(_mutex);
+  GSSleepInfo	ctxt;
 
-  while (objc_mutex_trylock(_mutex) == -1)
+  CHECK_RECURSIVE_CONDITION_LOCK(_MUTEX);
+
+  GSSleepInit(limit, &ctxt);
+
+  while (objc_mutex_trylock(_MUTEX) == -1)
     {
-      NSDate *current = [NSDate date];
-      NSComparisonResult compare;
-      
-      compare = [current compare: limit];
-      if (compare == NSOrderedSame || compare == NSOrderedDescending)
+      if (GSSleepOrFail(&ctxt) == NO)
 	{
 	  return NO;
 	}
-      /*
-       * This should probably be more accurate like usleep(250)
-       * but usleep is known to NOT be thread safe under all architectures.
-       */
-      sleep(1);
     }
   return YES;
 }
@@ -377,15 +453,47 @@ NSString *NSRecursiveLockException = @"NSRecursiveLockException";
                 beforeDate: (NSDate*)limitDate
 {
 #ifndef HAVE_OBJC_CONDITION_TIMEDWAIT
-  [self notImplemented: _cmd];
+  GSSleepInfo	ctxt;
+
+  CHECK_RECURSIVE_CONDITION_LOCK(_MUTEX);
+
+  GSSleepInit(limitDate, &ctxt);
+
+  do
+    {
+      if (_condition_value == condition_to_meet)
+	{
+	  while (objc_mutex_trylock(_MUTEX) == -1)
+	    {
+	      if (GSSleepOrFail(&ctxt) == NO)
+		{
+		  return NO;
+		}
+	    }
+	  if (_condition_value == condition_to_meet)
+	    {
+	      return YES;
+	    }
+	  if (objc_mutex_unlock(_MUTEX) == -1)
+	    {
+	      [NSException raise: NSConditionLockException
+			   format: @"%s failed to unlock mutex",
+			   GSNameFromSelector(_cmd)];
+	      /* NOT REACHED */
+	    }
+	} 
+    }
+  while (GSSleepOrFail(&ctxt) == YES);
+
   return NO;
+
 #else
   NSTimeInterval atimeinterval;
   struct timespec endtime;
   
-  CHECK_RECURSIVE_CONDITION_LOCK(_mutex);
+  CHECK_RECURSIVE_CONDITION_LOCK(_MUTEX);
   
-  if (-1 == objc_mutex_lock(_mutex))
+  if (-1 == objc_mutex_lock(_MUTEX))
     [NSException raise: NSConditionLockException 
 		 format: @"lockWhenCondition: failed to lock mutex"];
 	
@@ -399,7 +507,7 @@ NSString *NSRecursiveLockException = @"NSRecursiveLockException";
   
   while (_condition_value != condition_to_meet)
     {
-      switch (objc_condition_timedwait(_condition, _mutex, &endtime))
+      switch (objc_condition_timedwait(_CONDITION, _MUTEX, &endtime))
 	{
 	  case 0: 
 	    break;
@@ -423,11 +531,11 @@ NSString *NSRecursiveLockException = @"NSRecursiveLockException";
 // These methods ignore the condition
 - (void) lock
 {
-  CHECK_RECURSIVE_CONDITION_LOCK(_mutex);
+  CHECK_RECURSIVE_CONDITION_LOCK(_MUTEX);
 
   // Ask the runtime to acquire a lock on the mutex
   // This will block
-  if (objc_mutex_lock(_mutex) == -1)
+  if (objc_mutex_lock(_MUTEX) == -1)
     {
       [NSException raise: NSConditionLockException
         format: @"lock: failed to lock mutex"];
@@ -438,7 +546,7 @@ NSString *NSRecursiveLockException = @"NSRecursiveLockException";
 - (void) unlock
 {
   // wake up blocked threads
-  if (objc_condition_broadcast(_condition) == -1)
+  if (objc_condition_broadcast(_CONDITION) == -1)
     {
       [NSException raise: NSConditionLockException
         format: @"unlockWithCondition: objc_condition_broadcast failed"];
@@ -446,7 +554,7 @@ NSString *NSRecursiveLockException = @"NSRecursiveLockException";
     }
 
   // Ask the runtime to release a lock on the mutex
-  if (objc_mutex_unlock(_mutex) == -1)
+  if (objc_mutex_unlock(_MUTEX) == -1)
     {
       [NSException raise: NSConditionLockException
         format: @"unlock: failed to unlock mutex"];
@@ -457,16 +565,18 @@ NSString *NSRecursiveLockException = @"NSRecursiveLockException";
 @end
 
 
-// NSRecursiveLock
-// Allows the lock to be recursively acquired by the same thread
-//
-// If the same thread locks the mutex (n) times then that same 
-// thread must also unlock it (n) times before another thread 
-// can acquire the lock.
 
+/**
+ * See NSLock for more information about what a lock is. A recursive
+ * lock extends NSLock in that you can lock a recursive lock multiple
+ * times. Each lock must be balanced by a cooresponding unlock, and the
+ * lock is not released for another thread to aquire until the last
+ * unlock call is made (corresponding to the first lock message).
+ */
 @implementation NSRecursiveLock
 
-// Designated initializer
+/** <init />
+ */
 - (id) init
 {
   self = [super init];
@@ -496,41 +606,44 @@ NSString *NSRecursiveLockException = @"NSRecursiveLockException";
     {
       // Ask the runtime to deallocate the mutex
       // If there are outstanding locks then it will block
-      if (objc_mutex_deallocate(_mutex) == -1)
+      if (objc_mutex_deallocate(_MUTEX) == -1)
 	{
 	  NSWarnMLog(@"objc_mutex_deallocate() failed");
 	}
+      _mutex = 0;
     }
 }
 
-// Try to acquire the lock
-// Does not block
+/**
+ * Attempts to aquire a lock, but returns NO immediately if the lock
+ * cannot be aquired. It returns YES if the lock is aquired. Can be
+ * called multiple times to make nested locks.
+ */
 - (BOOL) tryLock
 {
   // Ask the runtime to acquire a lock on the mutex
-  if (objc_mutex_trylock(_mutex) == -1)
+  if (objc_mutex_trylock(_MUTEX) == -1)
     return NO;
   else
     return YES;
 }
 
-- (BOOL) lockBeforeDate: (NSDate *)limit
+/**
+ * Attempts to aquire a lock before the date limit passes. It returns
+ * YES if it can. It returns NO if it cannot
+ * (but it waits until the time limit is up before returning NO).
+ */
+- (BOOL) lockBeforeDate: (NSDate*)limit
 {
-  while (objc_mutex_trylock(_mutex) == -1)
+  GSSleepInfo	ctxt;
+
+  GSSleepInit(limit, &ctxt);
+  while (objc_mutex_trylock(_MUTEX) == -1)
     {
-      NSDate *current = [NSDate date];
-      NSComparisonResult compare;
-      
-      compare = [current compare: limit];
-      if (compare == NSOrderedSame || compare == NSOrderedDescending)
+      if (GSSleepOrFail(&ctxt) == NO)
 	{
 	  return NO;
 	}
-      /*
-       * This should probably be more accurate like usleep(250)
-       * but usleep is known to NOT be thread safe under all architectures.
-       */
-      sleep(1);
     }
   return YES;
 }
@@ -540,7 +653,7 @@ NSString *NSRecursiveLockException = @"NSRecursiveLockException";
 {
   // Ask the runtime to acquire a lock on the mutex
   // This will block
-  if (objc_mutex_lock(_mutex) == -1)
+  if (objc_mutex_lock(_MUTEX) == -1)
     {
       [NSException raise: NSRecursiveLockException
         format: @"lock: failed to lock mutex"];
@@ -551,7 +664,7 @@ NSString *NSRecursiveLockException = @"NSRecursiveLockException";
 - (void) unlock
 {
   // Ask the runtime to release a lock on the mutex
-  if (objc_mutex_unlock(_mutex) == -1)
+  if (objc_mutex_unlock(_MUTEX) == -1)
     {
       [NSException raise: NSRecursiveLockException
         format: @"unlock: failed to unlock mutex"];
