@@ -24,19 +24,19 @@
  * Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA.
  *
  * <title>NSMapTable class reference</title>
- * $Date: 2002/03/08 07:11:10 $ $Revision: 1.28 $
+ * $Date: 2003/07/31 23:49:31 $ $Revision: 1.35 $
  */
 
 /**** Included Headers *******************************************************/
 
-#include <config.h>
-#include <Foundation/NSObject.h>
-#include <Foundation/NSString.h>
-#include <Foundation/NSArray.h>
-#include <Foundation/NSException.h>
-#include <Foundation/NSZone.h>
-#include <Foundation/NSMapTable.h>
-#include <Foundation/NSDebug.h>
+#include "config.h"
+#include "Foundation/NSObject.h"
+#include "Foundation/NSString.h"
+#include "Foundation/NSArray.h"
+#include "Foundation/NSException.h"
+#include "Foundation/NSZone.h"
+#include "Foundation/NSMapTable.h"
+#include "Foundation/NSDebug.h"
 #include "NSCallBacks.h"
 
 
@@ -66,7 +66,7 @@ typedef struct {
  (M->extra.v.retain)((NSMapTable*)M, X.ptr)
 #define	GSI_MAP_ENUMERATOR	NSMapEnumerator
 
-#include <base/GSIMap.h>
+#include "GNUstepBase/GSIMap.h"
 
 /**** Function Implementations ****/
 
@@ -99,6 +99,7 @@ NSAllMapTableKeys(NSMapTable *table)
     {
       [keyArray addObject: key];
     }
+  NSEndMapTableEnumeration(&enumerator);
   return keyArray;
 }
 
@@ -131,6 +132,7 @@ NSAllMapTableValues(NSMapTable *table)
     {
       [valueArray addObject: value];
     }
+  NSEndMapTableEnumeration(&enumerator);
   return valueArray;
 }
 
@@ -169,29 +171,35 @@ NSCompareMapTables(NSMapTable *table1, NSMapTable *table2)
     }
   else
     {
-      GSIMapNode	n = t1->firstNode;
+      NSMapEnumerator enumerator = GSIMapEnumeratorForMap((GSIMapTable)t1);
+      GSIMapNode n;
 
-      while (n != 0)
-	{
-	  if (GSIMapNodeForKey(t2, n->key) == 0)
-	    {
-	      return NO;
-	    }
-	  n = n->nextInMap;
-	}
+      while ((n = GSIMapEnumeratorNextNode(&enumerator)) != 0)
+        {
+          if (GSIMapNodeForKey(t2, n->key) == 0)
+            {
+	      GSIMapEndEnumerator((GSIMapEnumerator)&enumerator);
+              return NO;
+            }
+        }
+      GSIMapEndEnumerator((GSIMapEnumerator)&enumerator);
       return YES;
     }
 }
 
 /**
- * Copy the supplied map table creating the new table in the specified zone.
+ * Copy the supplied map table.<br />
+ * Returns a map table, space for which is allocated in zone, which
+ * has (newly retained) copies of table's keys and values.  As always,
+ * if zone is 0, then NSDefaultMallocZone() is used.
  */
 NSMapTable *
 NSCopyMapTableWithZone(NSMapTable *table, NSZone *zone)
 {
   GSIMapTable	t;
   GSIMapNode	n;
-
+  NSMapEnumerator enumerator;
+  
   if (table == 0)
     {
       NSWarnFLog(@"Nul table argument supplied");
@@ -202,18 +210,18 @@ NSCopyMapTableWithZone(NSMapTable *table, NSZone *zone)
   GSIMapInitWithZoneAndCapacity(t, zone, ((GSIMapTable)table)->nodeCount);
   t->extra.k = ((GSIMapTable)table)->extra.k;
   t->extra.v = ((GSIMapTable)table)->extra.v;
-  n = ((GSIMapTable)table)->firstNode;
-  while (n != 0)
+  enumerator = GSIMapEnumeratorForMap((GSIMapTable)table);
+  while ((n = GSIMapEnumeratorNextNode(&enumerator)) != 0)
     {
       GSIMapAddPair(t, n->key, n->value);
-      n = n->nextInMap;
     }
+  GSIMapEndEnumerator((GSIMapEnumerator)&enumerator);
 
   return (NSMapTable*)t;
 }
 
 /**
- * Returns the number of keys in the table.
+ * Returns the number of key/value pairs in the table.
  */
 unsigned int
 NSCountMapTable(NSMapTable *table)
@@ -228,7 +236,10 @@ NSCountMapTable(NSMapTable *table)
 
 /**
  * Create a new map table by calling NSCreateMapTableWithZone() using
- * NSDefaultMallocZone().
+ * NSDefaultMallocZone().<br />
+ * Returns a (pointer to) an NSMapTable space for which is allocated
+ * in the default zone.  If capacity is small or 0, then the returned
+ * table has a reasonable capacity. 
  */
 NSMapTable *
 NSCreateMapTable(
@@ -243,9 +254,11 @@ NSCreateMapTable(
 /**
  * Create a new map table using the supplied callbacks structures.
  * If any functions in the callback structures are null the default
- * values are used ... as for non-owned pointers.
+ * values are used ... as for non-owned pointers.<br />
+ * Of course, if you send 0 for zone, then the map table will be
+ * created in NSDefaultMallocZone().<br />
  * The table will be created with the specified capacity ... ie ready
- * to hold at lest that many items.
+ * to hold at least that many items.
  */
 NSMapTable *
 NSCreateMapTableWithZone(
@@ -284,7 +297,7 @@ NSCreateMapTableWithZone(
 
 /**
  * Function to be called when finished with the enumerator.
- * Not required in GNUstep ... just provided for MacOS-X compatibility.
+ * This permits memory used by the enumerator to be released!
  */
 void
 NSEndMapTableEnumeration(NSMapEnumerator *enumerator)
@@ -292,7 +305,9 @@ NSEndMapTableEnumeration(NSMapEnumerator *enumerator)
   if (enumerator == 0)
     {
       NSWarnFLog(@"Nul enumerator argument supplied");
+      return;
     }
+  GSIMapEndEnumerator((GSIMapEnumerator)enumerator);
 }
 
 /**
@@ -313,7 +328,10 @@ NSEnumerateMapTable(NSMapTable *table)
 }
 
 /**
- * Destroy the map table and relase its contents.
+ * Destroy the map table and relase its contents.<br />
+ * Releases all the keys and values of table (using the key and
+ * value callbacks specified at the time of table's creation),
+ * and then proceeds to deallocate the space allocated for table itself.
  */
 void
 NSFreeMapTable(NSMapTable *table)
@@ -332,13 +350,13 @@ NSFreeMapTable(NSMapTable *table)
 }
 
 /**
- * Returns the value for the specified key, or a nul pointer if the
+ * Returns the value for the specified key, or a null pointer if the
  * key is not found in the table.
  */
 void *
 NSMapGet(NSMapTable *table, const void *key)
 {
-  GSIMapNode	n = GSIMapNodeForKey((GSIMapTable)table, (GSIMapKey)key);
+  GSIMapNode	n;
 
   if (table == 0)
     {
@@ -360,7 +378,7 @@ NSMapGet(NSMapTable *table, const void *key)
  * Adds the key and value to table.<br />
  * If an equal key is already in table, replaces its mapped value
  * with the new one, without changing the key itsself.<br />
- * If key is equal to the notAKeyMarker field of the tables
+ * If key is equal to the notAKeyMarker field of the table's
  * NSMapTableKeyCallBacks, raises an NSInvalidArgumentException.
  */
 void
@@ -372,7 +390,7 @@ NSMapInsert(NSMapTable *table, const void *key, const void *value)
   if (table == 0)
     {
       [NSException raise: NSInvalidArgumentException
-		  format: @"Attempt to place key-value in nul table"];
+		  format: @"Attempt to place key-value in null table"];
     }
   if (key == t->extra.k.notAKeyMarker)
     {
@@ -398,7 +416,7 @@ NSMapInsert(NSMapTable *table, const void *key, const void *value)
  * Adds the key and value to table and returns nul.<br />
  * If an equal key is already in table, returns the old key
  * instead of adding the new key-value pair.<br />
- * If key is equal to the notAKeyMarker field of the tables
+ * If key is equal to the notAKeyMarker field of the table's
  * NSMapTableKeyCallBacks, raises an NSInvalidArgumentException.
  */
 void *
@@ -410,7 +428,7 @@ NSMapInsertIfAbsent(NSMapTable *table, const void *key, const void *value)
   if (table == 0)
     {
       [NSException raise: NSInvalidArgumentException
-		  format: @"Attempt to place key-value in nul table"];
+		  format: @"Attempt to place key-value in null table"];
     }
   if (key == t->extra.k.notAKeyMarker)
     {
@@ -432,7 +450,7 @@ NSMapInsertIfAbsent(NSMapTable *table, const void *key, const void *value)
 /**
  * Adds the key and value to table and returns nul.<br />
  * If an equal key is already in table, raises an NSInvalidArgumentException.
- * <br />If key is equal to the notAKeyMarker field of the tables
+ * <br />If key is equal to the notAKeyMarker field of the table's
  * NSMapTableKeyCallBacks, raises an NSInvalidArgumentException.
  */
 void
@@ -444,7 +462,7 @@ NSMapInsertKnownAbsent(NSMapTable *table, const void *key, const void *value)
   if (table == 0)
     {
       [NSException raise: NSInvalidArgumentException
-		  format: @"Attempt to place key-value in nul table"];
+		  format: @"Attempt to place key-value in null table"];
     }
   if (key == t->extra.k.notAKeyMarker)
     {
@@ -467,7 +485,7 @@ NSMapInsertKnownAbsent(NSMapTable *table, const void *key, const void *value)
  * Returns a flag to say whether the table contains the specified key.
  * Returns the original key and the value it maps to.<br />
  * The GNUstep implementation checks originalKey and value to see if
- * they are nul pointers, and only updates them if non-null.
+ * they are null pointers, and only updates them if non-null.
  */
 BOOL
 NSMapMember(NSMapTable *table, const void *key,
@@ -500,7 +518,8 @@ NSMapMember(NSMapTable *table, const void *key,
 }
 
 /**
- * Remove the specified key from the table.
+ * Remove the specified key from the table (if present).<br />
+ * Causes the key and its associated value to be released.
  */
 void
 NSMapRemove(NSMapTable *table, const void *key)
@@ -516,8 +535,10 @@ NSMapRemove(NSMapTable *table, const void *key)
 /**
  * Step through the map table ... return the next key-value pair and
  * return YES, or hit the end of the table and return NO.<br />
+ * The enumerator parameter is a value supplied by NSEnumerateMapTable()
+ * and must be destroyed using NSEndMapTableEnumeration().<br />
  * The GNUstep implementation permits either key or value to be a
- * nul pointer, and refrains from attempting to return the appropriate
+ * null pointer, and refrains from attempting to return the appropriate
  * result in that case.
  */
 BOOL
@@ -560,7 +581,8 @@ NSNextMapEnumeratorPair(NSMapEnumerator *enumerator,
 }
 
 /**
- * Empty the map table, but preserve its capacity.
+ * Empty the map table (releasing every key and value),
+ * but preserve its capacity.
  */
 void
 NSResetMapTable(NSMapTable *table)
@@ -608,7 +630,124 @@ NSStringFromMapTable(NSMapTable *table)
 	(t->extra.k.describe)(table, key),
 	(t->extra.v.describe)(table, value)];
     }
+  NSEndMapTableEnumeration(&enumerator);
   return string;
 }
 
+
+
+
+/* These are to increase readabilty locally. */
+typedef unsigned int (*NSMT_hash_func_t)(NSMapTable *, const void *);
+typedef BOOL (*NSMT_is_equal_func_t)(NSMapTable *, const void *, const void *);
+typedef void (*NSMT_retain_func_t)(NSMapTable *, const void *);
+typedef void (*NSMT_release_func_t)(NSMapTable *, void *);
+typedef NSString *(*NSMT_describe_func_t)(NSMapTable *, const void *);
+
+
+/** For keys that are pointer-sized or smaller quantities. */
+const NSMapTableKeyCallBacks NSIntMapKeyCallBacks = 
+{
+  (NSMT_hash_func_t) _NS_int_hash,
+  (NSMT_is_equal_func_t) _NS_int_is_equal,
+  (NSMT_retain_func_t) _NS_int_retain,
+  (NSMT_release_func_t) _NS_int_release,
+  (NSMT_describe_func_t) _NS_int_describe,
+  NSNotAnIntMapKey
+};
+
+/** For keys that are pointers not freed. */
+const NSMapTableKeyCallBacks NSNonOwnedPointerMapKeyCallBacks = 
+{
+  (NSMT_hash_func_t) _NS_non_owned_void_p_hash,
+  (NSMT_is_equal_func_t) _NS_non_owned_void_p_is_equal,
+  (NSMT_retain_func_t) _NS_non_owned_void_p_retain,
+  (NSMT_release_func_t) _NS_non_owned_void_p_release,
+  (NSMT_describe_func_t) _NS_non_owned_void_p_describe,
+  NSNotAPointerMapKey
+};
+
+/** For keys that are pointers not freed, or 0. */
+const NSMapTableKeyCallBacks NSNonOwnedPointerOrNullMapKeyCallBacks = 
+{
+  (NSMT_hash_func_t) _NS_non_owned_void_p_hash,
+  (NSMT_is_equal_func_t) _NS_non_owned_void_p_is_equal,
+  (NSMT_retain_func_t) _NS_non_owned_void_p_retain,
+  (NSMT_release_func_t) _NS_non_owned_void_p_release,
+  (NSMT_describe_func_t) _NS_non_owned_void_p_describe,
+  NSNotAPointerMapKey
+};
+
+/** For sets of objects without retaining and releasing. */
+const NSMapTableKeyCallBacks NSNonRetainedObjectMapKeyCallBacks = 
+{
+  (NSMT_hash_func_t) _NS_non_retained_id_hash,
+  (NSMT_is_equal_func_t) _NS_non_retained_id_is_equal,
+  (NSMT_retain_func_t) _NS_non_retained_id_retain,
+  (NSMT_release_func_t) _NS_non_retained_id_release,
+  (NSMT_describe_func_t) _NS_non_retained_id_describe,
+  NSNotAPointerMapKey
+};
+
+/** For keys that are objects. */
+const NSMapTableKeyCallBacks NSObjectMapKeyCallBacks = 
+{
+  (NSMT_hash_func_t) _NS_id_hash,
+  (NSMT_is_equal_func_t) _NS_id_is_equal,
+  (NSMT_retain_func_t) _NS_id_retain,
+  (NSMT_release_func_t) _NS_id_release,
+  (NSMT_describe_func_t) _NS_id_describe,
+  NSNotAPointerMapKey
+};
+
+/** For keys that are pointers with transfer of ownership upon insertion. */
+const NSMapTableKeyCallBacks NSOwnedPointerMapKeyCallBacks = 
+{
+  (NSMT_hash_func_t) _NS_owned_void_p_hash,
+  (NSMT_is_equal_func_t) _NS_owned_void_p_is_equal,
+  (NSMT_retain_func_t) _NS_owned_void_p_retain,
+  (NSMT_release_func_t) _NS_owned_void_p_release,
+  (NSMT_describe_func_t) _NS_owned_void_p_describe,
+  NSNotAPointerMapKey
+};
+
+/** For values that are pointer-sized integer quantities. */
+const NSMapTableValueCallBacks NSIntMapValueCallBacks = 
+{
+  (NSMT_retain_func_t) _NS_int_retain,
+  (NSMT_release_func_t) _NS_int_release,
+  (NSMT_describe_func_t) _NS_int_describe
+};
+
+/** For values that are pointers not freed. */
+const NSMapTableValueCallBacks NSNonOwnedPointerMapValueCallBacks = 
+{
+  (NSMT_retain_func_t) _NS_non_owned_void_p_retain,
+  (NSMT_release_func_t) _NS_non_owned_void_p_release,
+  (NSMT_describe_func_t) _NS_non_owned_void_p_describe
+};
+
+/** For sets of objects without retaining and releasing. */
+const NSMapTableValueCallBacks NSNonRetainedObjectMapValueCallBacks = 
+{
+  (NSMT_retain_func_t) _NS_non_retained_id_retain,
+  (NSMT_release_func_t) _NS_non_retained_id_release,
+  (NSMT_describe_func_t) _NS_non_retained_id_describe
+};
+
+/** For values that are objects. */
+const NSMapTableValueCallBacks NSObjectMapValueCallBacks = 
+{
+  (NSMT_retain_func_t) _NS_id_retain,
+  (NSMT_release_func_t) _NS_id_release,
+  (NSMT_describe_func_t) _NS_id_describe
+};
+
+/** For values that are pointers with transfer of ownership upon insertion. */
+const NSMapTableValueCallBacks NSOwnedPointerMapValueCallBacks = 
+{
+  (NSMT_retain_func_t) _NS_owned_void_p_retain,
+  (NSMT_release_func_t) _NS_owned_void_p_release,
+  (NSMT_describe_func_t) _NS_owned_void_p_describe
+};
 
