@@ -23,16 +23,16 @@
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA.
 
    <title>NSMethodSignature class reference</title>
-   $Date: 2001/12/18 16:54:15 $ $Revision: 1.26 $
+   $Date: 2003/07/31 23:49:31 $ $Revision: 1.30 $
    */ 
 
-#include <config.h>
-#include <base/preface.h>
+#include "config.h"
+#include "GNUstepBase/preface.h"
 #include <mframe.h>
 
-#include <Foundation/NSMethodSignature.h>
-#include <Foundation/NSException.h>
-#include <Foundation/NSString.h>
+#include "Foundation/NSMethodSignature.h"
+#include "Foundation/NSException.h"
+#include "Foundation/NSString.h"
 
 
 @implementation NSMethodSignature
@@ -126,6 +126,41 @@
   [super dealloc];
 }
 
+- (BOOL)isEqual:(id)other
+{
+  BOOL isEqual = YES;
+  if (other == nil)
+    {
+      return NO;
+    }
+  if (((NSMethodSignature *)other)->isa != isa)
+    {
+      return NO;
+    }
+  isEqual = ([self numberOfArguments] == [other numberOfArguments] 
+    && [self frameLength] == [other frameLength] 
+    && *[self methodReturnType] == *[other methodReturnType] 
+    && [self isOneway] == [other isOneway]);
+  if(isEqual == NO)
+    {
+      return NO;
+    }
+  else
+    {
+      int i, n;
+      n = [self numberOfArguments];
+      for(i = 0; i < n; i++)
+        {
+          if ( (*[self getArgumentTypeAtIndex:i] 
+		== *[other getArgumentTypeAtIndex:i]) == NO)
+            {
+              return NO;
+            }
+        }
+    }
+  return isEqual;
+}
+
 @end
 
 @implementation NSMethodSignature(GNU)
@@ -134,7 +169,7 @@
   if (_info == 0)
     {
       const char	*types = _methodTypes;
-      int		i;
+      unsigned int	i;
 
       _info = NSZoneMalloc(NSDefaultMallocZone(),
 	sizeof(NSArgumentInfo)*(_numArgs+1));
