@@ -1,4 +1,5 @@
 /** Implementation of NSNotificationCenter for GNUstep
+  notification = (id)NSAllocateObject(concrete, 0, NSDefaultMallocZone());
    Copyright (C) 1999 Free Software Foundation, Inc.
 
    Written by:  Richard Frith-Macdonald <richard@brainstorm.co.uk>
@@ -25,18 +26,94 @@
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA.
 
    <title>NSNotificationCenter class reference</title>
-   $Date: 2002/02/01 10:41:00 $ $Revision: 1.32 $
+   $Date: 2004/02/04 00:14:00 $ $Revision: 1.42 $
 */
 
 #include "config.h"
-#include <Foundation/NSNotification.h>
-#include <Foundation/NSException.h>
-#include <Foundation/NSLock.h>
-#include <Foundation/NSThread.h>
+#include "Foundation/NSNotification.h"
+#include "Foundation/NSException.h"
+#include "Foundation/NSLock.h"
+#include "Foundation/NSThread.h"
+#include "GNUstepBase/GSLock.h"
+
+
+/**
+ * Concrete class implementing NSNotification
+ */
+@interface	GSNotification : NSNotification
+{
+@public
+  NSString	*_name;
+  id		_object;
+  NSDictionary	*_info;
+}
+@end
+
+@implementation GSNotification
+
+static Class concrete = 0;
+
++ (void) initialize
+{
+  if (concrete == 0)
+    {
+      concrete = [GSNotification class];
+    }
+}
+
++ (NSNotification*) notificationWithName: (NSString*)name
+				  object: (id)object
+			        userInfo: (NSDictionary*)info
+{
+  GSNotification	*n;
+
+  n = (GSNotification*)NSAllocateObject(self, 0, NSDefaultMallocZone());
+  n->_name = [name copyWithZone: GSObjCZone(self)];
+  n->_object = TEST_RETAIN(object);
+  n->_info = TEST_RETAIN(info);
+  return AUTORELEASE(n);
+}
+
+- (id) copyWithZone: (NSZone*)zone
+{
+  GSNotification	*n;
+
+  if (NSShouldRetainWithZone (self, zone))
+    {
+      return [self retain];
+    }
+  n = (GSNotification*)NSAllocateObject(concrete, 0, NSDefaultMallocZone());
+  n->_name = [_name copyWithZone: GSObjCZone(self)];
+  n->_object = TEST_RETAIN(_object);
+  n->_info = TEST_RETAIN(_info);
+  return n;
+}
+
+- (void) dealloc
+{
+  RELEASE(_name);
+  TEST_RELEASE(_object);
+  TEST_RELEASE(_info);
+  [super dealloc];
+}
+
+- (NSString*) name
+{
+  return _name;
+}
+
+- (id) object
+{
+  return _object;
+}
+
+- (NSDictionary*) userInfo
+{
+  return _info;
+}
+
+@end
 
-typedef struct {
-  @defs(NSNotification)
-} NotificationStruct;
 
 /*
  * Garbage collection considerations -
@@ -115,7 +192,7 @@ static void obsFree(Observation *o);
 #define GSI_ARRAY_RELEASE(A, X)   obsFree(X.ext)
 #define GSI_ARRAY_RETAIN(A, X)    obsRetain(X.ext)
 
-#include <base/GSIArray.h>
+#include "GNUstepBase/GSIArray.h"
 
 #define GSI_MAP_RETAIN_KEY(M, X)  
 #define GSI_MAP_RELEASE_KEY(M, X) ({if ((((gsaddr)X.obj) & 1) == 0) \
@@ -130,7 +207,9 @@ static void obsFree(Observation *o);
 #define GSI_MAP_VEXTRA Observation*
 #define	GSI_MAP_EXTRA	void*
 
-#include <base/GSIMap.h>
+#include "GNUstepBase/GSIMap.h"
+
+@class	GSLazyRecursiveLock;
 
 /*
  * An NC table is used to keep track of memory allocated to store
@@ -161,9 +240,7 @@ typedef struct NCTbl {
   GSIMapTable		named;		/* Getting named messages only.	*/
   GSIArray		array;		/* Temp store during posting.	*/
   unsigned		lockCount;	/* Count recursive operations.	*/
-  NSRecursiveLock	*_lock;		/* Lock out other threads.	*/
-  IMP			lImp;
-  IMP			uImp;
+  GSLazyRecursiveLock	*_lock;		/* Lock out other threads.	*/
   BOOL			lockingDisabled;
   BOOL			immutableInPost;
 
@@ -171,8 +248,8 @@ typedef struct NCTbl {
   Observation	**chunks;
   unsigned	numChunks;
   GSIMapTable	cache[CACHESIZE];
-  short		chunkIndex;
-  short		cacheIndex;
+  unsigned short	chunkIndex;
+  unsigned short	cacheIndex;
 } NCTable;
 
 #define	TABLE		((NCTable*)_table)
@@ -316,7 +393,7 @@ static void endNCTable(NCTable *t)
   TEST_RELEASE(t->_lock);
 }
 
-static NCTable *newNCTable()
+static NCTable *newNCTable(void)
 {
   NCTable	*t;
 
@@ -334,21 +411,23 @@ static NCTable *newNCTable()
   t->array = NSZoneMalloc(NSDefaultMallocZone(), sizeof(GSIArray_t));
   GSIArrayInitWithZoneAndCapacity(t->array, NSDefaultMallocZone(), 16);
 
+  // t->_lock = [GSLazyRecursiveLock new];
+  t->_lock = [NSRecursiveLock new];
   return t;
 }
 
 static inline void lockNCTable(NCTable* t)
 {
-  if (t->_lock != nil && t->lockingDisabled == NO)
-    (*t->lImp)(t->_lock, @selector(lock));
+  if (t->lockingDisabled == NO)
+    [t->_lock lock];
   t->lockCount++;
 }
 
 static inline void unlockNCTable(NCTable* t)
 {
   t->lockCount--;
-  if (t->_lock != nil && t->lockingDisabled == NO)
-    (*t->uImp)(t->_lock, @selector(unlock));
+  if (t->lockingDisabled == NO)
+    [t->_lock unlock];
 }
 
 static void obsFree(Observation *o)
@@ -485,6 +564,10 @@ static NSNotificationCenter *default_center = nil;
 {
   if (self == [NSNotificationCenter class])
     {
+      if (concrete == 0)
+	{
+	  concrete = [GSNotification class];
+	}
       /*
        * Do alloc and init separately so the default center can refer to
        * the 'default_center' variable during initialisation.
@@ -502,42 +585,12 @@ static NSNotificationCenter *default_center = nil;
 
 /* Initializing. */
 
-- (void) _becomeThreaded: (NSNotification*)notification
-{
-  unsigned	count;
-
-  TABLE->_lock = [NSRecursiveLock new];
-  TABLE->lImp = [TABLE->_lock methodForSelector: @selector(lock)];
-  TABLE->uImp = [TABLE->_lock methodForSelector: @selector(unlock)];
-  count = LOCKCOUNT;
-  /*
-   * If we start locking inside a method that would normally have been
-   * locked, we must lock the lock enough times so that when we leave
-   * the method the number of unlocks will match.
-   */
-  while (count-- > 0)
-    {
-      (*TABLE->lImp)(TABLE->_lock, @selector(lock));
-    }
-}
-
 - (id) init
 {
-  [super init];
-  TABLE = newNCTable();
-  if ([NSThread isMultiThreaded])
+  if ((self = [super init]) != nil)
     {
-      [self _becomeThreaded: nil];
+      _table = newNCTable();
     }
-  else
-    {
-      [[NSNotificationCenter defaultCenter]
-	addObserver: self
-	   selector: @selector(_becomeThreaded:)
-	       name: NSWillBecomeMultiThreadedNotification
-	     object: nil];
-    }
-
   return self;
 }
 
@@ -854,35 +907,31 @@ static NSNotificationCenter *default_center = nil;
 }
 
 
-/*
- * Post NOTIFICATION to all the observers that match its NAME and OBJECT.
- *
- * For performance reasons, we don't wrap an exception handler round every
- * message sent to an observer.  This means that, if one observer raises
- * an exception, later observers in the lists will not get the notification.
+/**
+ * Private method to perform the actual posting of a notification.
+ * Release the notification before returning, or before we raise
+ * any exception ... to avoid leaks.
  */
-- (void) postNotification: (NSNotification*)notification
+- (void) _postAndRelease: (NSNotification*)notification 
 {
-  NSString	*n_name;
-  id		n_object;
   Observation	*o;
   unsigned	count;
   volatile GSIArray	a;
   unsigned	arrayBase;
+  NSString	*name = [notification name];
+  id		object;
 
-  if (notification == nil)
-    [NSException raise: NSInvalidArgumentException
-		format: @"Tried to post a nil notification."];
-
-  n_name = ((NotificationStruct*)notification)->_name;
-  n_object = ((NotificationStruct*)notification)->_object;
-  if (n_object != nil)
-    n_object = CHEATGC(n_object);
-
-  if (n_name == nil)
-    [NSException raise: NSInvalidArgumentException
-		format: @"Tried to post a notification with no name."];
-
+  if (name == nil)
+    {
+      RELEASE(notification);
+      [NSException raise: NSInvalidArgumentException
+		  format: @"Tried to post a notification with no name."];
+    }
+  object = [notification object];
+  if (object != nil)
+    {
+      object = CHEATGC(object);
+    }
   lockNCTable(TABLE);
 
   a = ARRAY;
@@ -922,9 +971,9 @@ static NSNotificationCenter *default_center = nil;
 	   * Post the notification to all the observers that specified OBJECT,
 	   * but didn't specify NAME.
 	   */
-	  if (n_object)
+	  if (object)
 	    {
-	      n = GSIMapNodeForSimpleKey(NAMELESS, (GSIMapKey)n_object);
+	      n = GSIMapNodeForSimpleKey(NAMELESS, (GSIMapKey)object);
 	      if (n != 0)
 		{
 		  o = n->value.ext;
@@ -941,19 +990,23 @@ static NSNotificationCenter *default_center = nil;
 	   * observers with a non-nil OBJECT that doesn't match the
 	   * notification's OBJECT).
 	   */
-	  if (n_name)
+	  if (name)
 	    {
-	      n = GSIMapNodeForKey(NAMED, (GSIMapKey)n_name);
+	      n = GSIMapNodeForKey(NAMED, (GSIMapKey)name);
 	      if (n)
-		m = (GSIMapTable)n->value.ptr;
+		{
+		  m = (GSIMapTable)n->value.ptr;
+		}
 	      else
-		m = 0;
+		{
+		  m = 0;
+		}
 	      if (m != 0)
 		{
 		  /*
 		   * First, observers with a matching object.
 		   */
-		  n = GSIMapNodeForSimpleKey(m, (GSIMapKey)n_object);
+		  n = GSIMapNodeForSimpleKey(m, (GSIMapKey)object);
 		  if (n != 0)
 		    {
 		      o = n->value.ext;
@@ -965,7 +1018,7 @@ static NSNotificationCenter *default_center = nil;
 			}
 		    }
 
-		  if (n_object != nil)
+		  if (object != nil)
 		    {
 		      /*
 		       * Now observers with a nil object.
@@ -1008,9 +1061,9 @@ static NSNotificationCenter *default_center = nil;
 	   * Post the notification to all the observers that specified OBJECT,
 	   * but didn't specify NAME.
 	   */
-	  if (n_object)
+	  if (object)
 	    {
-	      n = GSIMapNodeForSimpleKey(NAMELESS, (GSIMapKey)n_object);
+	      n = GSIMapNodeForSimpleKey(NAMELESS, (GSIMapKey)object);
 	      if (n != 0)
 		{
 		  o = n->value.ext;
@@ -1035,19 +1088,23 @@ static NSNotificationCenter *default_center = nil;
 	   * observers with a non-nil OBJECT that doesn't match the
 	   * notification's OBJECT).
 	   */
-	  if (n_name)
+	  if (name)
 	    {
-	      n = GSIMapNodeForKey(NAMED, (GSIMapKey)n_name);
+	      n = GSIMapNodeForKey(NAMED, (GSIMapKey)name);
 	      if (n)
-		m = (GSIMapTable)n->value.ptr;
+		{
+		  m = (GSIMapTable)n->value.ptr;
+		}
 	      else
-		m = 0;
+		{
+		  m = 0;
+		}
 	      if (m != 0)
 		{
 		  /*
 		   * First, observers with a matching object.
 		   */
-		  n = GSIMapNodeForSimpleKey(m, (GSIMapKey)n_object);
+		  n = GSIMapNodeForSimpleKey(m, (GSIMapKey)object);
 		  if (n != 0)
 		    {
 		      o = n->value.ext;
@@ -1058,7 +1115,7 @@ static NSNotificationCenter *default_center = nil;
 			}
 		    }
 
-		  if (n_object != nil)
+		  if (object != nil)
 		    {
 		      /*
 		       * Now observers with a nil object.
@@ -1097,28 +1154,60 @@ static NSNotificationCenter *default_center = nil;
       GSIArrayRemoveItemsFromIndex(ARRAY, arrayBase);
       unlockNCTable(TABLE);
 
+      RELEASE(notification);
       [localException raise];
     }
   NS_ENDHANDLER
 #endif
 
   unlockNCTable(TABLE);
+  RELEASE(notification);
 }
 
+
+/**
+ * Posts notification to all the observers that match its NAME and OBJECT.<br />
+ * The GNUstep implementation calls -postNotificationName:object:userInfo: to
+ * perform the actual posting.
+ */
+- (void) postNotification: (NSNotification*)notification
+{
+  if (notification == nil)
+    {
+      [NSException raise: NSInvalidArgumentException
+		  format: @"Tried to post a nil notification."];
+    }
+  [self _postAndRelease: RETAIN(notification)];
+}
+
+/**
+ * Creates and posts a notification using the
+ * -postNotificationName:object:userInfo: passing a nil user info argument.
+ */
 - (void) postNotificationName: (NSString*)name 
 		       object: (id)object
 {
-  [self postNotification: [NSNotification notificationWithName: name
-							object: object]];
+  [self postNotificationName: name object: object userInfo: nil];
 }
 
+/**
+ * The preferred method for posting a notification.
+ * <br />
+ * For performance reasons, we don't wrap an exception handler round every
+ * message sent to an observer.  This means that, if one observer raises
+ * an exception, later observers in the lists will not get the notification.
+ */
 - (void) postNotificationName: (NSString*)name 
 		       object: (id)object
 		     userInfo: (NSDictionary*)info
 {
-  [self postNotification: [NSNotification notificationWithName: name
-							object: object
-						      userInfo: info]];
+  GSNotification	*notification;
+
+  notification = (id)NSAllocateObject(concrete, 0, NSDefaultMallocZone());
+  name = notification->_name = [name copyWithZone: GSObjCZone(self)];
+  object = notification->_object = TEST_RETAIN(object);
+  notification->_info = TEST_RETAIN(info);
+  [self _postAndRelease: notification];
 }
 
 @end
