@@ -22,7 +22,7 @@
    or in connection with the use or performance of this software.
 
    <title>NSNotificationQueue class reference</title>
-   $Date: 2002/01/28 15:16:33 $ $Revision: 1.14 $
+   $Date: 2003/07/31 23:49:31 $ $Revision: 1.19 $
 */
 
 /* Implementation for NSNotificationQueue for GNUStep
@@ -49,15 +49,15 @@
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA.
    */
 
-#include <config.h>
-#include <base/preface.h>
-#include <Foundation/NSRunLoop.h>
-#include <Foundation/NSNotificationQueue.h>
-#include <Foundation/NSNotification.h>
-#include <Foundation/NSDictionary.h>
-#include <Foundation/NSArray.h>
-#include <Foundation/NSString.h>
-#include <Foundation/NSThread.h>
+#include "config.h"
+#include "GNUstepBase/preface.h"
+#include "Foundation/NSRunLoop.h"
+#include "Foundation/NSNotificationQueue.h"
+#include "Foundation/NSNotification.h"
+#include "Foundation/NSDictionary.h"
+#include "Foundation/NSArray.h"
+#include "Foundation/NSString.h"
+#include "Foundation/NSThread.h"
 
 /* NotificationQueueList by Richard Frith-Macdonald
    These objects are used to maintain lists of NSNotificationQueue objects.
@@ -83,7 +83,7 @@ typedef struct {
 @end
 
 static NotificationQueueList*
-currentList()
+currentList(void)
 {
   NotificationQueueList	*list;
   NSMutableDictionary	*d;
@@ -249,6 +249,11 @@ add_to_queue(NSNotificationQueueList *queue, NSNotification *notification,
   NSNotificationQueueRegistration	*item;
 
   item = NSZoneCalloc(_zone, 1, sizeof(NSNotificationQueueRegistration));
+  if (item == 0)
+    {
+      [NSException raise: NSMallocException
+      		  format: @"Unable to add to notification queue"];
+    }
       
   item->notification = RETAIN(notification);
   item->name = [notification name];
@@ -274,6 +279,11 @@ add_to_queue(NSNotificationQueueList *queue, NSNotification *notification,
  * NSNotificationQueue class implementation
  */
 
+@interface NSNotificationQueue (Private)
+- (void) _postNotification: (NSNotification*)notification
+		  forModes: (NSArray*)modes;
+@end
+
 @implementation NSNotificationQueue
 
 + (NSNotificationQueue*) defaultQueue
@@ -307,12 +317,17 @@ add_to_queue(NSNotificationQueueList *queue, NSNotification *notification,
   _center = RETAIN(notificationCenter);
   _asapQueue = NSZoneCalloc(_zone, 1, sizeof(NSNotificationQueueList));
   _idleQueue = NSZoneCalloc(_zone, 1, sizeof(NSNotificationQueueList));
-
-  /*
-   * insert in global queue list
-   */
-  [NotificationQueueList registerQueue: self];
-
+  if (_asapQueue == 0 || _idleQueue == 0)
+    {
+      DESTROY(self);
+    }
+  else
+    {
+      /*
+       * insert in global queue list
+       */
+      [NotificationQueueList registerQueue: self];
+    }
   return self;
 }
 
@@ -432,19 +447,6 @@ add_to_queue(NSNotificationQueueList *queue, NSNotification *notification,
     }
 }
 
-- (void) postNotification: (NSNotification*)notification
-		 forModes: (NSArray*)modes
-{
-  NSString	*mode = [[NSRunLoop currentRunLoop] currentMode];
-
-  // check to see if run loop is in a valid mode
-  if (mode == nil || modes == nil
-    || [modes indexOfObject: mode] != NSNotFound)
-    {
-      [_center postNotification: notification];
-    }
-}
-
 - (void) enqueueNotification: (NSNotification*)notification
 		postingStyle: (NSPostingStyle)postingStyle	
 {
@@ -468,7 +470,7 @@ add_to_queue(NSNotificationQueueList *queue, NSNotification *notification,
   switch (postingStyle)
     {
       case NSPostNow: 
-	[self postNotification: notification forModes: modes];
+	[self _postNotification: notification forModes: modes];
 	break;
       case NSPostASAP: 
 	add_to_queue(_asapQueue, notification, modes, _zone);
@@ -481,6 +483,23 @@ add_to_queue(NSNotificationQueueList *queue, NSNotification *notification,
 
 @end
 
+@implementation NSNotificationQueue (Private)
+
+- (void) _postNotification: (NSNotification*)notification
+		  forModes: (NSArray*)modes
+{
+  NSString	*mode = [[NSRunLoop currentRunLoop] currentMode];
+
+  // check to see if run loop is in a valid mode
+  if (mode == nil || modes == nil
+    || [modes indexOfObject: mode] != NSNotFound)
+    {
+      [_center postNotification: notification];
+    }
+}
+
+@end
+
 /*
  *	The following code handles sending of queued notifications by
  *	NSRunLoop.
@@ -500,7 +519,7 @@ static inline void notifyASAP(NSNotificationQueue *q)
       NSArray				*modes = item->modes;
 
       remove_from_queue_no_release(list, item);
-      [q postNotification: notification forModes: modes];
+      [q _postNotification: notification forModes: modes];
       RELEASE(notification);
       RELEASE(modes);
       NSZoneFree(((accessQueue)q)->_zone, item);
@@ -535,7 +554,7 @@ static inline void notifyIdle(NSNotificationQueue *q)
       NSArray				*modes = item->modes;
 
       remove_from_queue_no_release(list, item);
-      [q postNotification: notification forModes: modes];
+      [q _postNotification: notification forModes: modes];
       RELEASE(notification);
       RELEASE(modes);
       NSZoneFree(((accessQueue)q)->_zone, item);
