@@ -21,14 +21,14 @@
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA.
 
    <title>NSNull class reference</title>
-   $Date: 2001/12/18 16:54:15 $ $Revision: 1.4 $
+   $Date: 2003/06/07 01:24:41 $ $Revision: 1.6 $
    */ 
 
-/*
+#include "Foundation/NSNull.h"
+
+/**
  * An object to use as a placeholder - in collections for instance.
  */
-#include <Foundation/NSNull.h>
-
 @implementation	NSNull
 
 static NSNull	*null = 0;
@@ -51,6 +51,10 @@ static NSNull	*null = 0;
     }
 }
 
+/**
+ * Return an object that can be used as a placeholder in a collection.
+ * This method always returns the same object.
+ */
 + (NSNull*) null
 {
   return null;
