@@ -24,21 +24,49 @@
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA.
 
    <title>NSNumber class reference</title>
-   $Date: 2001/12/18 16:54:15 $ $Revision: 1.39 $
+   $Date: 2004/01/31 16:28:35 $ $Revision: 1.50 $
 */
 
 #include <string.h>
-#include <config.h>
-#include <base/preface.h>
-#include <Foundation/NSException.h>
-#include <Foundation/NSString.h>
-#include <Foundation/NSNotification.h>
-#include <Foundation/NSConcreteNumber.h>
-#include <Foundation/NSMapTable.h>
-#include <Foundation/NSThread.h>
-#include <Foundation/NSCoder.h>
-#include <Foundation/NSPortCoder.h>
-#include <Foundation/NSObjCRuntime.h>
+#include "config.h"
+#include "GNUstepBase/preface.h"
+#include "Foundation/NSException.h"
+#include "Foundation/NSString.h"
+#include "Foundation/NSNotification.h"
+#include "Foundation/NSMapTable.h"
+#include "Foundation/NSThread.h"
+#include "Foundation/NSCoder.h"
+#include "Foundation/NSPortCoder.h"
+#include "Foundation/NSObjCRuntime.h"
+
+#include "NSConcreteNumber.h"
+
+@interface GSCachedBool : NSBoolNumber
+@end
+@interface GSCachedInt : NSIntNumber
+@end
+@implementation GSCachedBool
+- (id) copyWithZone: (NSZone*)zone
+{
+  return RETAIN(self);
+}
+- (void) dealloc
+{
+  [NSException raise: NSGenericException
+	      format: @"Attempt to deallocate bool number owned by cache"];
+}
+@end
+@implementation GSCachedInt
+- (id) copyWithZone: (NSZone*)zone
+{
+  return RETAIN(self);
+}
+- (void) dealloc
+{
+  [NSException raise: NSGenericException
+	      format: @"Attempt to deallocate int number owned by cache"];
+}
+@end
 
 @implementation NSNumber
 
@@ -203,12 +231,12 @@ static Class	doubleNumberClass;
       /*
        * cache bool values.
        */
-      boolN = (NSNumber*)NSAllocateObject(boolNumberClass, 0,
+      boolN = (NSNumber*)NSAllocateObject([GSCachedBool class], 0,
 	NSDefaultMallocZone());
       boolean = NO;
       boolN = [boolN initWithBytes: &boolean objCType: NULL];
 
-      boolY = (NSNumber*)NSAllocateObject(boolNumberClass, 0,
+      boolY = (NSNumber*)NSAllocateObject([GSCachedBool class], 0,
 	NSDefaultMallocZone());
       boolean = YES;
       boolY = [boolY initWithBytes: &boolean objCType: NULL];
@@ -220,7 +248,7 @@ static Class	doubleNumberClass;
 	{
 	  NSNumber	*num;
 
-	  num = (NSNumber*)NSAllocateObject(intNumberClass, 0,
+	  num = (NSNumber*)NSAllocateObject([GSCachedInt class], 0,
 	    NSDefaultMallocZone());
 	  num = [num initWithBytes: &integer objCType: NULL];
 	  smallIntegers[integer + GS_SMALL] = num;
@@ -298,192 +326,305 @@ static Class	doubleNumberClass;
 
 + (NSNumber*) numberWithBool: (BOOL)value
 {
-  if (value == YES)
+  // if class is NSNumber, replace by appropriate object
+  if (self == abstractClass)
     {
-      return boolY;
+      if (value == NO)
+        {
+          return boolN;
+        }
+      else
+        {
+          return boolY;
+        }
     }
-  else
+  else // alloc class and init with object intWithXX method
     {
-      return boolN;
+      return AUTORELEASE([[self allocWithZone: NSDefaultMallocZone()]
+                           initWithBool: value]);
     }
 }
 
 + (NSNumber*) numberWithChar: (signed char)value
 {
-  NSNumber	*theObj;
+  NSNumber	*theObj = nil;
 
-  if (value <= GS_SMALL && value >= -GS_SMALL)
+  // if class is NSNumber, replace by appropriate object
+  if (self == abstractClass)
     {
-      return smallIntegers[value + GS_SMALL];
+      if (value <= GS_SMALL && value >= -GS_SMALL)
+        {
+          return smallIntegers[value + GS_SMALL];
+        }
+      theObj = (NSNumber*)NSAllocateObject(charNumberClass, 0,
+                                           NSDefaultMallocZone());
+      theObj = [theObj initWithBytes: &value objCType: NULL];
     }
-  theObj = (NSNumber*)NSAllocateObject(charNumberClass, 0,
-    NSDefaultMallocZone());
-  theObj = [theObj initWithBytes: &value objCType: NULL];
+  else // alloc class and init with object intWithXX method
+    {
+      theObj = [[self allocWithZone: NSDefaultMallocZone()]
+                 initWithChar: value];
+    }
+
   return AUTORELEASE(theObj);
 }
 
 + (NSNumber*) numberWithDouble: (double)value
 {
-  NSNumber	*theObj;
+  NSNumber	*theObj = nil;
+
+  // if class is NSNumber, replace by appropriate object
+  if (self == abstractClass)
+    {
+      theObj = (NSNumber*)NSAllocateObject(doubleNumberClass, 0,
+                                           NSDefaultMallocZone());
+      theObj = [theObj initWithBytes: &value objCType: NULL];
+    }
+  else // alloc class and init with object intWithXX method
+    {
+      theObj = [[self allocWithZone: NSDefaultMallocZone()]
+                 initWithDouble: value];
+    }
 
-  theObj = (NSNumber*)NSAllocateObject(doubleNumberClass, 0,
-    NSDefaultMallocZone());
-  theObj = [theObj initWithBytes: &value objCType: NULL];
   return AUTORELEASE(theObj);
 }
 
 + (NSNumber*) numberWithFloat: (float)value
 {
-  NSNumber	*theObj;
+  NSNumber	*theObj = nil;
+
+  // if class is NSNumber, replace by appropriate object
+  if (self == abstractClass)
+    {
+      theObj = (NSNumber*)NSAllocateObject(floatNumberClass, 0,
+                                           NSDefaultMallocZone());
+      theObj = [theObj initWithBytes: &value objCType: NULL];
+    }
+  else // alloc class and init with object intWithXX method
+    {
+      theObj = [[self allocWithZone: NSDefaultMallocZone()]
+                 initWithFloat: value];
+    }
 
-  theObj = (NSNumber*)NSAllocateObject(floatNumberClass, 0,
-    NSDefaultMallocZone());
-  theObj = [theObj initWithBytes: &value objCType: NULL];
   return AUTORELEASE(theObj);
 }
 
 + (NSNumber*) numberWithInt: (signed int)value
 {
-  NSNumber	*theObj;
+  NSNumber	*theObj = nil;
 
-  if (value <= GS_SMALL && value >= -GS_SMALL)
+  // if class is NSNumber, replace by appropriate object
+  if (self == abstractClass)
     {
-      return smallIntegers[value + GS_SMALL];
+      if (value <= GS_SMALL && value >= -GS_SMALL)
+        {
+          return smallIntegers[value + GS_SMALL];
+        }
+      theObj = (NSNumber*)NSAllocateObject(intNumberClass, 0,
+                                           NSDefaultMallocZone());
+      theObj = [theObj initWithBytes: &value objCType: NULL];
     }
-  theObj = (NSNumber*)NSAllocateObject(intNumberClass, 0,
-    NSDefaultMallocZone());
-  theObj = [theObj initWithBytes: &value objCType: NULL];
+  else // alloc class and init with object intWithXX method
+    {
+      theObj = [[self allocWithZone: NSDefaultMallocZone()]
+                 initWithInt: value];
+    }
+
   return AUTORELEASE(theObj);
 }
 
 + (NSNumber*) numberWithLong: (signed long)value
 {
-  NSNumber	*theObj;
+  NSNumber	*theObj = nil;
 
-  if (value <= GS_SMALL && value >= -GS_SMALL)
+  // if class is NSNumber, replace by appropriate object
+  if (self == abstractClass)
     {
-      return smallIntegers[value + GS_SMALL];
+      if (value <= GS_SMALL && value >= -GS_SMALL)
+        {
+          return smallIntegers[value + GS_SMALL];
+        }
+      theObj = (NSNumber*)NSAllocateObject(longNumberClass, 0,
+                                           NSDefaultMallocZone());
+      theObj = [theObj initWithBytes: &value objCType: NULL];
     }
-  theObj = (NSNumber*)NSAllocateObject(longNumberClass, 0,
-    NSDefaultMallocZone());
-  theObj = [theObj initWithBytes: &value objCType: NULL];
+  else // alloc class and init with object intWithXX method
+    {
+      theObj = [[self allocWithZone: NSDefaultMallocZone()]
+                 initWithLong: value];
+    }
+
   return AUTORELEASE(theObj);
 }
 
 + (NSNumber*) numberWithLongLong: (signed long long)value
 {
-  NSNumber	*theObj;
+  NSNumber	*theObj = nil;
 
-  if (value <= GS_SMALL && value >= -GS_SMALL)
+  // if class is NSNumber, replace by appropriate object
+  if (self == abstractClass)
     {
-      return smallIntegers[value + GS_SMALL];
+      if (value <= GS_SMALL && value >= -GS_SMALL)
+        {
+          return smallIntegers[value + GS_SMALL];
+        }
+      theObj = (NSNumber*)NSAllocateObject(longLongNumberClass, 0,
+                                           NSDefaultMallocZone());
+      theObj = [theObj initWithBytes: &value objCType: NULL];
     }
-  theObj = (NSNumber*)NSAllocateObject(longLongNumberClass, 0,
-    NSDefaultMallocZone());
-  theObj = [theObj initWithBytes: &value objCType: NULL];
+  else // alloc class and init with object intWithXX method
+    {
+      theObj = [[self allocWithZone: NSDefaultMallocZone()]
+                 initWithLongLong: value];
+    }
+
   return AUTORELEASE(theObj);
 }
 
 + (NSNumber*) numberWithShort: (signed short)value
 {
-  NSNumber	*theObj;
+  NSNumber	*theObj = nil;
 
-  if (value <= GS_SMALL && value >= -GS_SMALL)
+  // if class is NSNumber, replace by appropriate object
+  if (self == abstractClass)
     {
-      return smallIntegers[value + GS_SMALL];
+      if (value <= GS_SMALL && value >= -GS_SMALL)
+        {
+          return smallIntegers[value + GS_SMALL];
+        }
+      theObj = (NSNumber*)NSAllocateObject(shortNumberClass, 0,
+                                           NSDefaultMallocZone());
+      theObj = [theObj initWithBytes: &value objCType: NULL];
     }
-  theObj = (NSNumber*)NSAllocateObject(shortNumberClass, 0,
-    NSDefaultMallocZone());
-  theObj = [theObj initWithBytes: &value objCType: NULL];
+  else // alloc class and init with object intWithXX method
+    {
+      theObj = [[self allocWithZone: NSDefaultMallocZone()]
+                 initWithShort: value];
+    }
+
   return AUTORELEASE(theObj);
 }
 
 + (NSNumber*) numberWithUnsignedChar: (unsigned char)value
 {
-  NSNumber	*theObj;
+  NSNumber	*theObj = nil;
 
-  if (value <= GS_SMALL)
+  // if class is NSNumber, replace by appropriate object
+  if (self == abstractClass)
     {
-      return smallIntegers[value + GS_SMALL];
+      if (value <= GS_SMALL)
+        {
+          return smallIntegers[value + GS_SMALL];
+        }
+      theObj = (NSNumber*)NSAllocateObject(uCharNumberClass, 0,
+                                           NSDefaultMallocZone());
+      theObj = [theObj initWithBytes: &value objCType: NULL];
     }
-  theObj = (NSNumber*)NSAllocateObject(uCharNumberClass, 0,
-    NSDefaultMallocZone());
-  theObj = [theObj initWithBytes: &value objCType: NULL];
+  else // alloc class and init with object intWithXX method
+    {
+      theObj = [[self allocWithZone: NSDefaultMallocZone()]
+                 initWithUnsignedChar: value];
+    }
+
   return AUTORELEASE(theObj);
 }
 
 + (NSNumber*) numberWithUnsignedInt: (unsigned int)value
 {
-  NSNumber	*theObj;
+  NSNumber	*theObj = nil;
 
-  if (value <= GS_SMALL)
+  // if class is NSNumber, replace by appropriate object
+  if (self == abstractClass)
     {
-      return smallIntegers[value + GS_SMALL];
+      if (value <= GS_SMALL)
+        {
+          return smallIntegers[value + GS_SMALL];
+        }
+      theObj = (NSNumber*)NSAllocateObject(uIntNumberClass, 0,
+                                           NSDefaultMallocZone());
+      theObj = [theObj initWithBytes: &value objCType: NULL];
     }
-  theObj = (NSNumber*)NSAllocateObject(uIntNumberClass, 0,
-    NSDefaultMallocZone());
-  theObj = [theObj initWithBytes: &value objCType: NULL];
+  else // alloc class and init with object intWithXX method
+    {
+      theObj = [[self allocWithZone: NSDefaultMallocZone()]
+                 initWithUnsignedInt: value];
+    }
+
   return AUTORELEASE(theObj);
 }
 
 + (NSNumber*) numberWithUnsignedLong: (unsigned long)value
 {
-  NSNumber	*theObj;
+  NSNumber	*theObj = nil;
 
-  if (value <= GS_SMALL)
+  // if class is NSNumber, replace by appropriate object
+  if (self == abstractClass)
     {
-      return smallIntegers[value + GS_SMALL];
+      if (value <= GS_SMALL)
+        {
+          return smallIntegers[value + GS_SMALL];
+        }
+      theObj = (NSNumber*)NSAllocateObject(uLongNumberClass, 0,
+                                           NSDefaultMallocZone());
+      theObj = [theObj initWithBytes: &value objCType: NULL];
     }
-  theObj = (NSNumber*)NSAllocateObject(uLongNumberClass, 0,
-    NSDefaultMallocZone());
-  theObj = [theObj initWithBytes: &value objCType: NULL];
+  else // alloc class and init with object intWithXX method
+    {
+      theObj = [[self allocWithZone: NSDefaultMallocZone()]
+                 initWithUnsignedLong: value];
+    }
+
   return AUTORELEASE(theObj);
 }
 
 + (NSNumber*) numberWithUnsignedLongLong: (unsigned long long)value
 {
-  NSNumber	*theObj;
+  NSNumber	*theObj = nil;
 
-  if (value <= GS_SMALL)
+  // if class is NSNumber, replace by appropriate object
+  if (self == abstractClass)
     {
-      return smallIntegers[value + GS_SMALL];
+      if (value <= GS_SMALL)
+        {
+          return smallIntegers[value + GS_SMALL];
+        }
+      theObj = (NSNumber*)NSAllocateObject(uLongLongNumberClass, 0,
+                                           NSDefaultMallocZone());
+      theObj = [theObj initWithBytes: &value objCType: NULL];
     }
-  theObj = (NSNumber*)NSAllocateObject(uLongLongNumberClass, 0,
-    NSDefaultMallocZone());
-  theObj = [theObj initWithBytes: &value objCType: NULL];
+  else // alloc class and init with object intWithXX method
+    {
+      theObj = [[self allocWithZone: NSDefaultMallocZone()]
+                 initWithUnsignedLongLong: value];
+    }
+
   return AUTORELEASE(theObj);
 }
 
 + (NSNumber*) numberWithUnsignedShort: (unsigned short)value
 {
-  NSNumber	*theObj;
+  NSNumber	*theObj = nil;
 
-  if (value <= GS_SMALL)
+  // if class is NSNumber, replace by appropriate object
+  if (self == abstractClass)
     {
-      return smallIntegers[value + GS_SMALL];
+      if (value <= GS_SMALL)
+        {
+          return smallIntegers[value + GS_SMALL];
+        }
+      theObj = (NSNumber*)NSAllocateObject(uShortNumberClass, 0,
+                                           NSDefaultMallocZone());
+      theObj = [theObj initWithBytes: &value objCType: NULL];
     }
-  theObj = (NSNumber*)NSAllocateObject(uShortNumberClass, 0,
-    NSDefaultMallocZone());
-  theObj = [theObj initWithBytes: &value objCType: NULL];
+  else // alloc class and init with object intWithXX method
+    {
+      theObj = [[self allocWithZone: NSDefaultMallocZone()]
+                 initWithUnsignedShort: value];
+    }
+
   return AUTORELEASE(theObj);
 }
 
-+ (NSValue*) valueFromString: (NSString*)string
-{
-  /* FIXME: implement this better */
-  const char *str;
-
-  str = [string cString];
-  if (strchr(str, '.') >= 0 || strchr(str, 'e') >= 0 
-      || strchr(str, 'E') >= 0)
-    return [NSNumber numberWithDouble: atof(str)];
-  else if (strchr(str, '-') >= 0)
-    return [NSNumber numberWithInt: atoi(str)];
-  else
-    return [NSNumber numberWithUnsignedInt: atoi(str)];
-  return [NSNumber numberWithInt: 0];
-}
-
 /*
  * A moderately sane default init method - a zero value integer.
  */
@@ -495,13 +636,13 @@ static Class	doubleNumberClass;
 - (id) initWithBool: (BOOL)value
 {
   NSDeallocateObject(self);
-  if (value == YES)
+  if (value == NO)
     {
-      self = boolY;
+      self = boolN;
     }
   else
     {
-      self = boolN;
+      self = boolY;
     }
   return RETAIN(self);
 }
@@ -662,11 +803,50 @@ static Class	doubleNumberClass;
     return NSCopyObject(self, 0, zone);
 }
 
+/**
+ * Returns the string representation of this number using a non-localised
+ * conversion (decimal point is '.' irrespective of the locale).
+ */
 - (NSString*) description
 {
   return [self descriptionWithLocale: nil];
 }
 
+/**
+ * <p>
+ *   Produces a string representation of the number.  For a boolean
+ *   this will be either 'true' or 'false'.  For other numbers the
+ *   format is produced using the initWithFormat:locale:... method
+ *   of NSString, and the format depends on the type of number as
+ *   follows -
+ * </p>
+ * <deflist>
+ *   <term>char</term>
+ *   <desc>%i</desc>
+ *   <term> short</term>
+ *   <desc>%hi</desc>
+ *   <term> int</term>
+ *   <desc>%i</desc>
+ *   <term> long</term>
+ *   <desc>%li</desc>
+ *   <term> long long</term>
+ *   <desc>%lli</desc>
+ *   <term> unsigned char</term>
+ *   <desc>%u</desc>
+ *   <term> unsigned short</term>
+ *   <desc>%hu</desc>
+ *   <term> unsigned int</term>
+ *   <desc>%u</desc>
+ *   <term> unsigned long</term>
+ *   <desc>%lu</desc>
+ *   <term> unsigned long long</term>
+ *   <desc>%llu</desc>
+ *   <term> float</term>
+ *   <desc>%0.7g</desc>
+ *   <term> double</term>
+ *   <desc>%0.16g</desc>
+ * </deflist>
+ */
 - (NSString*) descriptionWithLocale: (NSDictionary*)locale
 {
   NSString	*result = nil;
@@ -683,7 +863,7 @@ static Class	doubleNumberClass;
       switch (info->typeLevel)
 	{
 	  case 0:
-	    return [self boolValue] ? @"true" : @"false";
+	    return [self boolValue] ? @"YES" : @"NO";
 	    break;
 
 	  case 1:
@@ -771,98 +951,98 @@ static Class	doubleNumberClass;
 	      BOOL	oData;
 
 	      (*(info->getValue))(self, @selector(getValue:), &oData);
-	      return oData;
+	      return (oData == 0) ? NO : YES;
 	    }
 	  case 1:
 	    {
 	      signed char	oData;
 
 	      (*(info->getValue))(self, @selector(getValue:), &oData);
-	      return oData;
+	      return (oData == 0) ? NO : YES;
 	    }
 	  case 2:
 	    {
 	      unsigned char	oData;
 
 	      (*(info->getValue))(self, @selector(getValue:), &oData);
-	      return oData;
+	      return (oData == 0) ? NO : YES;
 	    }
 	  case 3:
 	    {
 	      signed short	oData;
 
 	      (*(info->getValue))(self, @selector(getValue:), &oData);
-	      return oData;
+	      return (oData == 0) ? NO : YES;
 	    }
 	  case 4:
 	    {
 	      unsigned short	oData;
 
 	      (*(info->getValue))(self, @selector(getValue:), &oData);
-	      return oData;
+	      return (oData == 0) ? NO : YES;
 	    }
 	  case 5:
 	    {
 	      signed int	oData;
 
 	      (*(info->getValue))(self, @selector(getValue:), &oData);
-	      return oData;
+	      return (oData == 0) ? NO : YES;
 	    }
 	  case 6:
 	    {
 	      unsigned int	oData;
 
 	      (*(info->getValue))(self, @selector(getValue:), &oData);
-	      return oData;
+	      return (oData == 0) ? NO : YES;
 	    }
 	  case 7:
 	    {
 	      signed long	oData;
 
 	      (*(info->getValue))(self, @selector(getValue:), &oData);
-	      return oData;
+	      return (oData == 0) ? NO : YES;
 	    }
 	  case 8:
 	    {
 	      unsigned long	oData;
 
 	      (*(info->getValue))(self, @selector(getValue:), &oData);
-	      return oData;
+	      return (oData == 0) ? NO : YES;
 	    }
 	  case 9:
 	    {
 	      signed long long	oData;
 
 	      (*(info->getValue))(self, @selector(getValue:), &oData);
-	      return oData;
+	      return (oData == 0) ? NO : YES;
 	    }
 	  case 10:
 	    {
 	      unsigned long long	oData;
 
 	      (*(info->getValue))(self, @selector(getValue:), &oData);
-	      return oData;
+	      return (oData == 0) ? NO : YES;
 	    }
 	  case 11:
 	    {
 	      float	oData;
 
 	      (*(info->getValue))(self, @selector(getValue:), &oData);
-	      return oData;
+	      return (oData == 0) ? NO : YES;
 	    }
 	  case 12:
 	    {
 	      double	oData;
 
 	      (*(info->getValue))(self, @selector(getValue:), &oData);
-	      return oData;
+	      return (oData == 0) ? NO : YES;
 	    }
 	  default:
 	    [NSException raise: NSInvalidArgumentException
 			format: @"unknown number type value for get"];
 	}
     }
-  return 0;
+  return NO;
 }
 
 - (signed char) charValue
