@@ -23,14 +23,15 @@
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA.
 
    <title>NSNumberFormatter class reference</title>
-   $Date: 2001/12/18 16:54:15 $ $Revision: 1.8 $
+   $Date: 2003/06/12 03:29:35 $ $Revision: 1.10 $
    */
 
-#include <Foundation/NSDictionary.h>
-#include <Foundation/NSString.h>
-#include <Foundation/NSAttributedString.h>
-#include <Foundation/NSDecimalNumber.h>
-#include <Foundation/NSNumberFormatter.h>
+#include "Foundation/NSDictionary.h"
+#include "Foundation/NSException.h"
+#include "Foundation/NSString.h"
+#include "Foundation/NSAttributedString.h"
+#include "Foundation/NSDecimalNumber.h"
+#include "Foundation/NSNumberFormatter.h"
 
 
 @implementation NSNumberFormatter
@@ -148,7 +149,23 @@
 	      forString: (NSString*)string
        errorDescription: (NSString**)error
 {
-  return NO;	// FIXME
+  if (anObject == NULL)
+    {
+      [NSException raise: NSInvalidArgumentException
+		   format: @"NULL Object pointer to NSNumberFormatter"];
+      /* NO RETURN. */
+      return NO;
+    }
+
+  /* FIXME: This is just a quick hack implementation.  */
+  NSLog(@"NSNumberFormatter-getObjectValue:forString:... not implemented");
+  if (_positiveFormat == nil &&  _negativeFormat == nil)
+    {
+      /* Just assume nothing else has been setup and do a simple conversion. */
+      *anObject = [NSDecimalNumber decimalNumberWithString: string]; 
+      return YES;
+    }
+  return NO;
 }
 
 - (BOOL) hasThousandSeparators
