@@ -21,18 +21,18 @@
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA.
 
    <title>NSPage class reference</title>
-   $Date: 2002/02/20 06:42:05 $ $Revision: 1.23 $
+   $Date: 2003/07/31 23:49:31 $ $Revision: 1.30 $
    */ 
 
-#include <config.h>
-#include <base/preface.h>
-#include <Foundation/NSObjCRuntime.h>
-#include <Foundation/NSZone.h>
+#include "config.h"
+#include "GNUstepBase/preface.h"
+#include "Foundation/NSObjCRuntime.h"
+#include "Foundation/NSZone.h"
 #include <string.h>
 #ifdef __WIN32__
 #include <malloc.h>
 #endif
-#if HAVE_UNISTD_H
+#ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif
 #include <stdio.h>
@@ -68,11 +68,18 @@ getpagesize(void)
 #define getpagesize vm_page_size
 #endif
 
+#ifdef __BEOS__
+#include <kernel/OS.h>
+#define getpagesize()  B_PAGE_SIZE
+#endif
+
 /* Cache the size of a memory page here, so we don't have to make the
    getpagesize() system call repeatedly. */
 static unsigned ns_page_size = 0;
 
-/* Return the number of bytes in a memory page. */
+/**
+ * Return the number of bytes in a memory page.
+ */
 unsigned
 NSPageSize (void)
 {
@@ -81,7 +88,9 @@ NSPageSize (void)
   return ns_page_size;
 }
 
-/* Return log base 2 of the number of bytes in a memory page. */
+/**
+ * Return log base 2 of the number of bytes in a memory page.
+ */
 unsigned
 NSLogPageSize (void)
 {
@@ -93,8 +102,10 @@ NSLogPageSize (void)
   return log;
 }
 
-/* Round BYTES down to the nearest multiple of the memory page size,
-   and return it. */
+/**
+ * Round bytes down to the nearest multiple of the memory page size,
+ * and return it.
+ */
 unsigned
 NSRoundDownToMultipleOfPageSize (unsigned bytes)
 {
@@ -103,8 +114,10 @@ NSRoundDownToMultipleOfPageSize (unsigned bytes)
   return (bytes / a) * a;
 }
 
-/* Round BYTES up to the nearest multiple of the memory page size,
-   and return it. */
+/**
+ * Round bytes up to the nearest multiple of the memory page size,
+ * and return it.
+ */
 unsigned
 NSRoundUpToMultipleOfPageSize (unsigned bytes)
 {
@@ -117,6 +130,9 @@ NSRoundUpToMultipleOfPageSize (unsigned bytes)
 #include	<sys/sysinfo.h>
 #endif
 
+/**
+ * Return the number of bytes of real (physical) memory available.
+ */
 unsigned
 NSRealMemoryAvailable ()
 {
@@ -126,12 +142,27 @@ NSRealMemoryAvailable ()
   if ((sysinfo(&info)) != 0)
     return 0;
   return (unsigned) info.freeram;
+#elif defined(__MINGW__)
+  MEMORYSTATUS memory;
+  
+  GlobalMemoryStatus(&memory);
+  return (unsigned)memory.dwAvailPhys;
+#elif defined(__BEOS__)
+  system_info info;
+  
+  if (get_system_info(&info) != B_OK)
+    return 0;
+  return (unsigned)(info.max_pages - info.used_pages) * B_PAGE_SIZE;
 #else
   fprintf (stderr, "NSRealMemoryAvailable() not implemented.\n");
   return 0;
 #endif
 }
 
+/**
+ * Allocate memory for this process and return a pointer to it (or a nul
+ * pointer on failure).  
+ */
 void *
 NSAllocateMemoryPages (unsigned bytes)
 {
@@ -143,7 +174,7 @@ NSAllocateMemoryPages (unsigned bytes)
     return NULL;
   return where;
 #else
-#if HAVE_VALLOC
+#ifdef HAVE_VALLOC
   where = valloc (bytes);
 #else
   where = malloc (bytes);
@@ -155,6 +186,10 @@ NSAllocateMemoryPages (unsigned bytes)
 #endif
 }
 
+/**
+ * Deallocate memory which was previously allocated using the
+ * NSAllocateMemoryPages() function.
+ */
 void
 NSDeallocateMemoryPages (void *ptr, unsigned bytes)
 {
@@ -165,29 +200,19 @@ NSDeallocateMemoryPages (void *ptr, unsigned bytes)
 #endif
 }
 
+/**
+ * Perform an efficient large scale copy of data from src to dest.
+ * The value bytes specifies the length of the data copied.
+ */
 void
-NSCopyMemoryPages (const void *source, void *dest, unsigned bytes)
+NSCopyMemoryPages (const void *src, void *dest, unsigned bytes)
 {
 #if __mach__
   kern_return_t r;
-  r = vm_copy (mach_task_self(), source, bytes, dest);
+  r = vm_copy (mach_task_self(), src, bytes, dest);
   NSParameterAssert (r == KERN_SUCCESS);
 #else
-  memcpy (dest, source, bytes);
+  memcpy (dest, src, bytes);
 #endif
 }
 
-
-
-
-
-
-
-
-
-
-
-
-
-
-
