@@ -20,23 +20,34 @@
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA.
 
-   <title>NSPipe class reference</title>
-   $Date: 2002/02/20 06:42:05 $ $Revision: 1.13 $
+   $Date: 2003/07/31 23:49:31 $ $Revision: 1.19 $
    */
 
-#include <config.h>
-#include <base/preface.h>
-#include <Foundation/NSObject.h>
-#include <Foundation/NSFileHandle.h>
-#include <Foundation/NSDebug.h>
-#if HAVE_UNISTD_H
+#include "config.h"
+#include "GNUstepBase/preface.h"
+#include "Foundation/NSObject.h"
+#include "Foundation/NSFileHandle.h"
+#include "Foundation/NSDebug.h"
+#ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif
 
+/**
+ * <p>The NSPipe provides an encapsulation of the UNIX concept of pipe.<br />
+ * With NSPipe, it is possible to redirect the standard input or
+ * standard output.
+ * </p>
+ * <p>The file handles created by NSPipe are automatically closed when they
+ * are no longer in use (ie when the NSPipe instance is deallocated), so you
+ * don't need to close them explicitly.
+ * </p>
+ */
 @implementation NSPipe
 
-// Allocating and Initializing a FileHandle Object
-
+/**
+ * Returns a newly allocated and initialized NSPipe object that has been
+ * sent an autorelease message.
+ */
 + (id) pipe
 {
   return AUTORELEASE([[self alloc] init]);
@@ -59,8 +70,10 @@
 
       if (pipe(p) == 0)
         {
-          readHandle = [[NSFileHandle alloc] initWithFileDescriptor: p[0]];
-          writeHandle = [[NSFileHandle alloc] initWithFileDescriptor: p[1]];
+          readHandle = [[NSFileHandle alloc] initWithFileDescriptor: p[0]
+						     closeOnDealloc: YES];
+          writeHandle = [[NSFileHandle alloc] initWithFileDescriptor: p[1]
+						      closeOnDealloc: YES];
         }
       else
 	{
@@ -72,19 +85,27 @@
 
       if (CreatePipe(&readh, &writeh, NULL, 0) != 0)
         {
-          readHandle = [[NSFileHandle alloc] initWithNativeHandle: readh];
-          writeHandle = [[NSFileHandle alloc] initWithNativeHandle: writeh];
+          readHandle = [[NSFileHandle alloc] initWithNativeHandle: readh
+						   closeOnDealloc: YES];
+          writeHandle = [[NSFileHandle alloc] initWithNativeHandle: writeh
+						    closeOnDealloc: YES];
         }
 #endif
     }
   return self;
 }
 
+/**
+ * Returns the file handle for reading from the pipe.
+ */
 - (id) fileHandleForReading
 {
   return readHandle;
 }
 
+/**
+ * Returns the file handle for writing to the pipe.
+ */
 - (id) fileHandleForWriting
 {
   return writeHandle;
