@@ -22,7 +22,7 @@
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA.
 
    <title>NSProcessInfo class reference</title>
-   $Date: 2002/03/13 16:49:56 $ $Revision: 1.73 $
+   $Date: 2004/01/11 04:49:10 $ $Revision: 1.98 $
 */ 
 
 /*************************************************************************
@@ -40,9 +40,8 @@
  * purpose to override the autorelease, retain, and release methods.
  * To Do      : 
  * 1) To test the class on more platforms;
- * 2) To change the format of the string renurned by globallyUniqueString;
  * Bugs       : Not known
- * Last update: 22-jul-1999
+ * Last update: 07-aug-2002
  * History    : 06-aug-1995    - Birth and the first beta version (v. 0.5);
  *              08-aug-1995    - V. 0.6 (tested on NS, SunOS, Solaris, OSF/1
  *              The use of the environ global var was changed to more 
@@ -54,9 +53,9 @@
  * - To the NEXTSTEP/GNUStep community
  *************************************************************************/
 
-#include <config.h>
-#include <base/preface.h>
-#if HAVE_UNISTD_H
+#include "config.h"
+#include "GNUstepBase/preface.h"
+#ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif
 
@@ -67,16 +66,33 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-#include <GSConfig.h>
-#include <Foundation/NSString.h>
-#include <Foundation/NSArray.h>
-#include <Foundation/NSSet.h>
-#include <Foundation/NSDictionary.h>
-#include <Foundation/NSDate.h>
-#include <Foundation/NSException.h>
-#include <Foundation/NSProcessInfo.h>
-#include <Foundation/NSAutoreleasePool.h>
-#include <Foundation/NSHost.h>
+#include <sys/file.h>
+#ifdef HAVE_SYS_FCNTL_H
+#include <sys/fcntl.h>
+#endif
+
+#ifdef HAVE_LIBKVM
+#include <kvm.h>
+#include <fcntl.h>
+#include <sys/param.h>
+#include <sys/sysctl.h>
+#endif /* HAVE_LIBKVM */
+
+#include "GSConfig.h"
+#include "Foundation/NSString.h"
+#include "Foundation/NSArray.h"
+#include "Foundation/NSBundle.h"
+#include "Foundation/NSSet.h"
+#include "Foundation/NSDictionary.h"
+#include "Foundation/NSDate.h"
+#include "Foundation/NSException.h"
+#include "Foundation/NSProcessInfo.h"
+#include "Foundation/NSAutoreleasePool.h"
+#include "Foundation/NSHost.h"
+#include "Foundation/NSLock.h"
+#include "GNUstepBase/GSCategories.h"
+
+#include "GSPrivate.h"
 
 /* This error message should be called only if the private main function
  * was not executed successfully. This may happen ONLY if another library
@@ -142,6 +158,8 @@ static NSProcessInfo	*_gnu_sharedProcessInfoObject = nil;
 // Host name of the CPU executing the process
 static NSString		*_gnu_hostName = nil;   
 
+static char		*_gnu_arg_zero = 0;
+
 // Current process name
 static NSString		*_gnu_processName = nil;
 
@@ -161,31 +179,80 @@ static NSMutableSet	*_debug_set = nil;
 void 
 _gnu_process_args(int argc, char *argv[], char *env[])
 {
-  NSAutoreleasePool	*arp = [NSAutoreleasePool new];
+  CREATE_AUTORELEASE_POOL(arp);
   int i;
 
+  if (_gnu_arg_zero != 0)
+    {
+      free(_gnu_arg_zero);
+    }
+
+  if (argv != 0 && argv[0] != 0)
+    {
+      _gnu_arg_zero = (char*)malloc(strlen(argv[0]) + 1);
+      strcpy(_gnu_arg_zero, argv[0]);
+    }
+  else
+    {
+#ifdef __MINGW__
+      char	*buffer;
+      int	buffer_size = 0;
+      int	needed_size = 0;
+
+      while (needed_size == buffer_size)
+	{
+          buffer_size = buffer_size + 256;
+          buffer = (char*)malloc(buffer_size);
+          needed_size = GetModuleFileNameA(NULL, buffer, buffer_size);
+          if (needed_size < buffer_size)
+	    {
+              _gnu_arg_zero = buffer;
+	    }
+          else
+	    {
+              free(buffer);
+	    }
+	}
+#else      
+      fprintf(stderr, "Error: for some reason, argv not properly set up " 
+	      "during GNUstep base initialization\n");
+      abort();
+#endif      
+    }
+
   /* Getting the process name */
-  _gnu_processName = [[NSString stringWithCString: argv[0]] lastPathComponent];
+  IF_NO_GC(RELEASE(_gnu_processName));
+  _gnu_processName
+    = [[NSString stringWithCString: _gnu_arg_zero] lastPathComponent];
   IF_NO_GC(RETAIN(_gnu_processName));
 
 
   /* Copy the argument list */
   {
+    NSString            *str;
     NSMutableSet	*mySet;
     id			obj_argv[argc];
-    int			added = 0;
+    int			added = 1;
 
     mySet = [NSMutableSet new];
-    for (i = 0; i < argc; i++) 
+
+    /* Copy the zero'th argument to the argument list */
+    str = [NSString stringWithCString: _gnu_arg_zero];
+    obj_argv[0] = str;
+    
+    for (i = 1; i < argc; i++) 
       {
-	NSString	*str = [NSString stringWithCString: argv[i]];
+	str = [NSString stringWithCString: argv[i]];
 
 	if ([str hasPrefix: @"--GNU-Debug="])
 	  [mySet addObject: [str substringFromIndex: 12]];
 	else
           obj_argv[added++] = str;
       }
+
+    IF_NO_GC(RELEASE(_gnu_arguments));
     _gnu_arguments = [[NSArray alloc] initWithObjects: obj_argv count: added];
+    IF_NO_GC(RELEASE(_debug_set));
     _debug_set = mySet;
   }
 	
@@ -239,31 +306,111 @@ _gnu_process_args(int argc, char *argv[], char *env[])
 #endif
 	i++;
       }
+    IF_NO_GC(RELEASE(_gnu_environment));
     _gnu_environment = [[NSDictionary alloc] initWithObjects: values
 						     forKeys: keys];
-    [keys release];
-    [values release];
+    IF_NO_GC(RELEASE(keys));
+    IF_NO_GC(RELEASE(values));
   }
-  [arp release];
+  IF_NO_GC(RELEASE(arp));
 }
 
-#if !GS_FAKE_MAIN && (defined(HAVE_PROCFS) && defined(HAVE_LOAD_METHOD))
+#if !GS_FAKE_MAIN && ((defined(HAVE_PROCFS)  || defined(HAVE_LIBKVM)) && (defined(HAVE_LOAD_METHOD)))
 /*
  * We have to save program arguments and environment before main () is
  * executed, because main () could modify their values before we get a
  * chance to read them 
  */
-static int	_gnu_noobjc_argc;
-static char	**_gnu_noobjc_argv;
-static char	**_gnu_noobjc_env;
+static int	_gnu_noobjc_argc = 0;
+static char	**_gnu_noobjc_argv = NULL;
+static char	**_gnu_noobjc_env = NULL;
 
 /*
  * The +load method (an extension of the GNU compiler) is invoked
  * before main and +initialize (for this class) is executed.  This is
- * guaranteed if +load contains only pure C code, as we have here. 
+ * guaranteed if +load contains only pure C code, as we have here. The
+ * code in here either uses libkvm if available, or else procfs.
  */
 + (void) load 
 {
+#ifdef HAVE_LIBKVM
+  /*
+   * Use the kvm library to open the kernel and read the environment and
+   * arguments. As we are not running as root we cannot open the memory
+   * device and thus we fake it using /dev/null. This is allowed under
+   * FreeBSD, but may fail on other operating systems which check the
+   * file type. The kvm calls used are those which are supposedly backward
+   * compatible with Solaris rather than being FreeBSD specific
+   */
+  kvm_t *kptr = NULL;
+  struct kinfo_proc *proc_ptr = NULL;
+  int nprocs, i, count;
+  char **vectors;
+
+  /* open the kernel */
+  kptr = kvm_open(NULL, "/dev/null", NULL, O_RDONLY, "NSProcessInfo");
+  if (!kptr)
+    {
+      fprintf(stderr, "Error: Your system appears to provide libkvm, but the kernel open fails\n");
+      fprintf(stderr, "Try to reconfigure gnustep-base with --enable-fake-main. to work\n");
+      fprintf(stderr, "around this problem.");
+      abort();
+    }
+
+  /* find the process */
+  proc_ptr = kvm_getprocs(kptr, KERN_PROC_PID, getpid(), &nprocs);
+  if (!proc_ptr || (nprocs != 1))
+    {
+      fprintf(stderr, "Error: libkvm cannot find the current process\n");
+      abort();
+    }
+
+  /* get the environment vectors */
+  vectors = kvm_getenvv(kptr, proc_ptr, 0);
+  if (!vectors)
+    {
+      fprintf(stderr, "Error: libkvm does not return an environment for the current process\n");
+      abort();
+    }
+
+  /* copy the environment strings */
+  for(count = 0; vectors[count]; count++)
+    ;
+  _gnu_noobjc_env = (char**)malloc(sizeof(char*) * (count + 1));
+  if (!_gnu_noobjc_env)
+    goto malloc_error;
+  for (i = 0; i < count; i++)
+    {
+      _gnu_noobjc_env[i] = (char *)strdup(vectors[i]);
+      if (!_gnu_noobjc_env[i])
+	goto malloc_error;
+    }
+  _gnu_noobjc_env[i] = NULL;
+
+  /* get the argument vectors */
+  vectors = kvm_getargv(kptr, proc_ptr, 0);
+  if (!vectors)
+    {
+      fprintf(stderr, "Error: libkvm does not return arguments for the current process\n");
+      abort();
+    }
+
+  /* copy the argument strings */
+  for(_gnu_noobjc_argc = 0; vectors[_gnu_noobjc_argc]; _gnu_noobjc_argc++)
+    ;
+  _gnu_noobjc_argv = (char**)malloc(sizeof(char*) * (_gnu_noobjc_argc + 1));
+  if (!_gnu_noobjc_argv)
+    goto malloc_error;
+  for (i = 0; i < _gnu_noobjc_argc; i++)
+    {
+      _gnu_noobjc_argv[i] = (char *)strdup(vectors[i]);
+      if (!_gnu_noobjc_argv[i])
+	goto malloc_error;
+    }
+  _gnu_noobjc_argv[i] = NULL;
+
+  return;
+#else /* !HAVE_LIBKVM (i.e. HAVE_PROCFS).  */
   /*
    * Now we have the problem of reading program arguments and
    * environment.  We take the environment from extern char **environ, and
@@ -445,6 +592,7 @@ static char	**_gnu_noobjc_env;
    */
   abort();
 #endif /* HAVE_PROGRAM_INVOCATION_NAME */
+#endif /* !HAVE_LIBKVM (e.g. HAVE_PROCFS) */
  malloc_error: 
   fprintf(stderr, "malloc() error when starting gnustep-base.\n");
   fprintf(stderr, "Free some memory and then re-run the program.\n");
@@ -485,7 +633,7 @@ _gnu_noobjc_free_vars(void)
       _gnu_noobjc_free_vars();
     }
 }
-#else /* !HAVE_PROCFS || !HAVE_LOAD_METHOD */
+#else /*! HAVE_PROCFS !HAVE_LOAD_METHOD !HAVE_LIBKVM */
 
 #ifdef __MINGW__
 /* For WindowsAPI Library, we know the global variables (argc, etc) */
@@ -494,6 +642,16 @@ _gnu_noobjc_free_vars(void)
   if (self == [NSProcessInfo class])
     _gnu_process_args(__argc, __argv, _environ);
 }
+#elif defined(__BEOS__)
+
+extern int __libc_argc;
+extern char **__libc_argv;
++ (void) initialize
+{
+  if (self == [NSProcessInfo class])
+    _gnu_process_args(__libc_argc, __libc_argv, environ);
+}
+
 
 #else
 #ifndef GS_PASS_ARGUMENTS
@@ -533,9 +691,9 @@ int main(int argc, char *argv[], char *env[])
 
 #endif /* HAS_LOAD_METHOD && HAS_PROCFS */ 
 
-/*************************************************************************
- *** Getting an NSProcessInfo Object
- *************************************************************************/
+/**
+ * Returns the shared NSProcessInfo object for the current process.
+ */
 + (NSProcessInfo *) processInfo
 {
   // Check if the main() function was successfully called
@@ -554,19 +712,65 @@ int main(int argc, char *argv[], char *env[])
   return _gnu_sharedProcessInfoObject;
 }
 
-/*************************************************************************
- *** Returning Process Information
- *************************************************************************/
+/**
+ * Returns an array containing the arguments supplied to start this
+ * process.  NB. In GNUstep, any arguments of the form --GNU-Debug=...
+ * are <em>not</em> included in this array ... they are part of the
+ * debug mechanism, and are hidden so that setting debug variables
+ * will not effect the normal operation of the program.
+ */
 - (NSArray *) arguments
 {
   return _gnu_arguments;
 }
 
+/**
+ * Returns a dictionary giving the environment variables which were
+ * provided for the process to use.
+ */
 - (NSDictionary *) environment
 {
   return _gnu_environment;
 }
 
+/**
+ * Returns a string which may be used as a globally unique identifier.<br />
+ * The string contains the host name, the process ID, a timestamp and a
+ * counter.<br />
+ * The first three values identify the process in which the string is
+ * generated, while the fourth ensures that multiple strings generated
+ * within the same process are unique.
+ */
+- (NSString *) globallyUniqueString
+{
+  static unsigned long	counter = 0;
+  int			count;
+  static NSString	*host = nil;
+  static int		pid;
+  static unsigned long	start;
+
+  [gnustep_global_lock lock];
+  if (host == nil)
+    {
+      extern NSTimeInterval GSTimeNow(void);
+
+      pid = [self processIdentifier];
+      start = (unsigned long)GSTimeNow();
+      host = [[self hostName] stringByReplacingString: @"." withString: @"_"];
+      RETAIN(host);
+    }
+  count = counter++;
+  [gnustep_global_lock unlock];
+
+  // $$$ The format of the string is not specified by the OpenStep 
+  // specification.
+  return [NSString stringWithFormat: @"%@_%x_%lx_%lx",
+    host, pid, start, count];
+}
+
+/**
+ * Returns the name of the machine on which this process is running.
+ */
 - (NSString *) hostName
 {
   if (!_gnu_hostName) 
@@ -576,6 +780,102 @@ int main(int argc, char *argv[], char *env[])
   return _gnu_hostName;
 }
 
+/**
+ * Return a number representing the operating system type.<br />
+ * The known types are listed in the header file, but not all of the
+ * listed types are actually implemented ... some are present for
+ * MacOS-X compatibility only.<br />
+ * <list>
+ * <item>NSWindowsNTOperatingSystem - used for windows NT, 2000, XP</item>
+ * <item>NSWindows95OperatingSystem - probably never to be implemented</item>
+ * <item>NSSolarisOperatingSystem - not yet recognised</item>
+ * <item>NSHPUXOperatingSystem - not implemented</item>
+ * <item>NSMACHOperatingSystem - perhaps the HURD in future?</item>
+ * <item>NSSunOSOperatingSystem - probably never to be implemented</item>
+ * <item>NSOSF1OperatingSystem - probably never to be implemented</item>
+ * <item>NSGNULinuxOperatingSystem - the GNUstep 'standard'</item>
+ * <item>NSBSDOperatingSystem - BSD derived operating systems</item>
+ * </list>
+ */
+- (unsigned int) operatingSystem
+{
+  static unsigned int	os = 0;
+
+  if (os == 0)
+    {
+      NSString	*n = [self operatingSystemName];
+
+      if ([n isEqualToString: @"linux-gnu"] == YES)
+        {
+	  os = NSGNULinuxOperatingSystem;
+	}
+      else if ([n isEqualToString: @"mingw"] == YES)
+        {
+	  os = NSWindowsNTOperatingSystem;
+	}
+      else if ([n isEqualToString: @"cygwin"] == YES)
+        {
+	  os = NSWindowsNTOperatingSystem;
+	}
+      else if ([n hasPrefix: @"bsd"] == YES)
+        {
+	  os = NSBSDOperatingSystem;
+	}
+      else if ([n hasPrefix: @"freebsd"] == YES)
+        {
+	  os = NSBSDOperatingSystem;
+	}
+      else if ([n hasPrefix: @"netbsd"] == YES)
+        {
+	  os = NSBSDOperatingSystem;
+	}
+      else if ([n hasPrefix: @"openbsd"] == YES)
+        {
+	  os = NSBSDOperatingSystem;
+	}
+      else if ([n isEqualToString: @"beos"] == YES)
+	{
+          os = NSBeOperatingSystem;
+        }
+      else if ([n hasPrefix: @"darwin"] == YES)
+	{
+          os = NSMACHOperatingSystem;
+        }
+      else if ([n hasPrefix: @"solaris"] == YES)
+	{
+          os = NSSolarisOperatingSystem;
+        }
+      else if ([n hasPrefix: @"hpux"] == YES)
+	{
+          os = NSHPUXOperatingSystem;
+        }
+      else
+        {
+	  NSLog(@"Unable to determine O/S ... assuming GNU/Linux");
+	  os = NSGNULinuxOperatingSystem;
+	}
+    }
+  return os;
+}
+
+/**
+ * Returns the name of the operating system in use.
+ */
+- (NSString*) operatingSystemName
+{
+  static NSString	*os = nil;
+
+  if (os == nil)
+    {
+      os = [[NSBundle _gnustep_target_os] copy];
+    }
+  return os;
+}
+
+/**
+ * Returns the process identifier number which identifies this process
+ * on this machine.
+ */
 - (int) processIdentifier
 {
   int	pid;
@@ -588,28 +888,16 @@ int main(int argc, char *argv[], char *env[])
   return pid;
 }
 
+/**
+ * Returns the process name for this process. This may have been set using
+ * the -setProcessName: method, or may be the default process name (the
+ * file name of the binary being executed).
+ */
 - (NSString *) processName
 {
   return _gnu_processName;
 }
 
-- (NSString *) globallyUniqueString
-{
-  int	pid;
-
-#if defined(__MINGW__)
-  pid = (int)GetCurrentProcessId();
-#else
-  pid = (int)getpid();
-#endif
-
-  // $$$ The format of the string is not specified by the OpenStep 
-  // specification. It could be useful to change this format after
-  // NeXTSTEP release 4.0 comes out.
-  return [NSString stringWithFormat: @"%@:%d:[%@]",
-    [self hostName], pid, [NSDate date]];
-}
-
 /**
  * Change the name of the current process to newName.
  */
@@ -625,24 +913,44 @@ int main(int argc, char *argv[], char *env[])
 @end
 
 @implementation	NSProcessInfo (GNUstep)
+
+static BOOL	debugTemporarilyDisabled = NO;
+
 /**
  * Fallback method. The developer must call this method to initialize
  * the NSProcessInfo system if none of the system-specific hacks to
- * auto initiailise it are working.
+ * auto-initialize it are working.
  */
 + (void) initializeWithArguments: (char**)argv
                            count: (int)argc
                      environment: (char**)env
 {
-  if (!_gnu_processName && !_gnu_arguments && !_gnu_environment)
+  [gnustep_global_lock lock];
+  _gnu_process_args(argc, argv, env);
+  [gnustep_global_lock unlock];
+}
+
+/**
+ * Returns a indication of whether debug logging is enabled.
+ * This returns YES unless a call to -setDebugLoggingEnabled: has
+ * been used to turn logging off.
+ */
+- (BOOL) debugLoggingEnabled
+{
+  if (debugTemporarilyDisabled == YES)
     {
-      _gnu_process_args(argc, argv, env);
+      return NO;
+    }
+  else
+    {
+      return YES;
     }
 }
 
 /**
  * This method returns a set of debug levels set using the
- * --GNU-Debug=... command line option.<br />
+ * --GNU-Debug=... command line option and/or the GNU-Debug
+ * user default.<br />
  * You can modify this set to change the debug logging under
  * your programs control ... but such modifications are not
  * thread-safe.
@@ -652,17 +960,61 @@ int main(int argc, char *argv[], char *env[])
   return _debug_set;
 }
 
+/**
+ * This method permits you to turn all debug logging on or off
+ * without modifying the set of debug levels in use.
+ */
+- (void) setDebugLoggingEnabled: (BOOL)flag
+{
+  if (flag == NO)
+    {
+      debugTemporarilyDisabled = YES;
+    }
+  else
+    {
+      debugTemporarilyDisabled = NO;
+    }
+}
+
+/**
+ * Set the file to which NSLog output should be directed.<br />
+ * Returns YES on success, NO on failure.<br />
+ * By default logging goes to standard error.
+ */
+- (BOOL) setLogFile: (NSString*)path
+{
+  extern int	_NSLogDescriptor;
+  int		desc;
+
+  desc = open([path fileSystemRepresentation], O_RDWR|O_CREAT|O_APPEND, 0644);
+  if (desc >= 0)
+    {
+      if (_NSLogDescriptor >= 0 && _NSLogDescriptor != 2)
+	{
+	  close(_NSLogDescriptor);
+	}
+      _NSLogDescriptor = desc;
+      return YES;
+    }
+  return NO;
+}
 @end
 
 /**
- * Function for rapid testing to see if a debug level is set.
- * This is used by the debugging macros.
+ * Function for rapid testing to see if a debug level is set.<br />
+ * This is used by the debugging macros.<br />
+ * If debug logging has been turned off, this returns NO even if
+ * the specified level exists in the set of debug levels.
  */
 BOOL GSDebugSet(NSString *level)
 {
   static IMP debugImp = 0;
   static SEL debugSel;
 
+  if (debugTemporarilyDisabled == YES)
+    {
+      return NO;
+    }
   if (debugImp == 0)
     {
       debugSel = @selector(member:);
@@ -679,3 +1031,46 @@ BOOL GSDebugSet(NSString *level)
   return YES;
 }
 
+
+BOOL
+GSEnvironmentFlag(const char *name, BOOL def)
+{
+  const char	*c = getenv(name);
+  BOOL		a = def;
+
+  if (c != 0)
+    {
+      a = NO;
+      if ((c[0] == 'y' || c[0] == 'Y') && (c[1] == 'e' || c[1] == 'E')
+	&& (c[2] == 's' || c[2] == 'S') && c[3] == 0)
+	{
+	  a = YES;
+	}
+      else if ((c[0] == 't' || c[0] == 'T') && (c[1] == 'r' || c[1] == 'R')
+	&& (c[2] == 'u' || c[2] == 'U') && (c[3] == 'e' || c[3] == 'E')
+	&& c[4] == 0)
+	{
+	  a = YES;
+	}
+      else if (isdigit(c[0]) && c[0] != '0')
+	{
+	  a = YES;
+	}
+    }
+  return a;
+}
+
+/**
+ * Used by NSException uncaught exception handler - must not call any
+ * methods/functions which might cause a recursive exception.
+ */ 
+const char*
+GSArgZero(void)
+{
+  if (_gnu_arg_zero == 0)
+    return "";
+  else
+    return _gnu_arg_zero;
+}
+
+
