@@ -21,15 +21,15 @@
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA.
 
    <title>NSProtocolChecker class reference</title>
-   $Date: 2001/12/18 16:54:15 $ $Revision: 1.10 $
+   $Date: 2004/02/04 00:14:00 $ $Revision: 1.13 $
    */ 
 
 #include "config.h"
-#include <base/preface.h>
-#include <Foundation/NSProtocolChecker.h>
-#include <Foundation/NSException.h>
-#include <Foundation/NSInvocation.h>
-#include <Foundation/NSMethodSignature.h>
+#include "GNUstepBase/preface.h"
+#include "Foundation/NSProtocolChecker.h"
+#include "Foundation/NSException.h"
+#include "Foundation/NSInvocation.h"
+#include "Foundation/NSMethodSignature.h"
 
 @implementation NSProtocolChecker
 
@@ -75,7 +75,7 @@
     {
       if (((id)buffer) == _myTarget)
    	{
-	  ((id)buffer) = self;
+	  buffer = self;
 	  [anInvocation setReturnValue: buffer];
    	}
     }
