@@ -21,56 +21,109 @@
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA.
 
    <title>NSProxy class reference</title>
-   $Date: 2001/12/18 16:54:15 $ $Revision: 1.19 $
+   $Date: 2003/08/24 23:07:41 $ $Revision: 1.27 $
    */
 
-#include <config.h>
-#include <base/preface.h>
-#include <Foundation/NSInvocation.h>
-#include <Foundation/NSProxy.h>
-#include <Foundation/NSMethodSignature.h>
-#include <Foundation/NSAutoreleasePool.h>
-#include <Foundation/NSException.h>
-#include <Foundation/NSObjCRuntime.h>
-#include "limits.h"
+#include "config.h"
+#include "GNUstepBase/preface.h"
+#include "Foundation/NSInvocation.h"
+#include "Foundation/NSProxy.h"
+#include "Foundation/NSMethodSignature.h"
+#include "Foundation/NSAutoreleasePool.h"
+#include "Foundation/NSException.h"
+#include "Foundation/NSObjCRuntime.h"
+#include "Foundation/NSDistantObject.h"
+#include "Foundation/NSPortCoder.h"
+
+#ifdef	HAVE_LIMITS_H
+/* For UINT_MAX */
+#include <limits.h>
+#endif
+
+@class	NSDistantObject;
 
 #ifndef NeXT_RUNTIME
 extern BOOL __objc_responds_to(id, SEL);
 #endif
 
+/**
+ * The NSProxy class provides a basic implementation of a class whose
+ * instances are used to <em>stand in</em> for other objects.<br />
+ * The class provides the most basic methods of NSObject, and expects
+ * messages for other methods to be forwarded to the <em>real</em>
+ * object represented by the proxy.  You must subclass NSProxy to
+ * implement -forwardInvocation: to these <em>real</em> objects.
+ */
 @implementation NSProxy
 
+/**
+ * Allocates and returns an NSProxy instance in the default zone.
+ */
 + (id) alloc
 {
   return [self allocWithZone: NSDefaultMallocZone()];
 }
 
+/**
+ * Allocates and returns an NSProxy instance in the specified zone z.
+ */
 + (id) allocWithZone: (NSZone*)z
 {
   NSProxy*	ob = (NSProxy*) NSAllocateObject(self, 0, z);
   return ob;
 }
 
+/**
+ * Returns the receiver
+ */
 + (id) autorelease
 {
   return self;
 }
 
+/**
+ * Returns the receiver
+ */
 + (Class) class
 {
   return self;
 }
 
+/**
+ * Returns a string describing the receiver.
+ */
 + (NSString*) description
 {
-  return [NSString stringWithFormat: @"<%s>", object_get_class_name(self)];
+  return [NSString stringWithFormat: @"<%s>", GSClassNameFromObject(self)];
 }
 
+/**
+ * Returns NO ... the NSProxy class cannot be an instance of any class.
+ */
++ (BOOL) isKindOfClass: (Class)aClass
+{
+  return NO;
+}
+
+/**
+ * Returns YES if aClass is identical to the receiver, NO otherwise.
+ */
++ (BOOL) isMemberOfClass: (Class)aClass
+{
+  return(self == aClass);
+}
+
+/**
+ * A dummy method ...
+ */
 + (void) load
 {
   /* Do nothing	*/
 }
 
+/**
+ * Returns the method signature for the specified selector.
+ */
 + (NSMethodSignature*) methodSignatureForSelector: (SEL)aSelector
 {
   struct objc_method	*mth;
@@ -79,7 +132,7 @@ extern BOOL __objc_responds_to(id, SEL);
     {
       return nil;
     }
-  mth = class_get_class_method(GSObjCClass(self), aSelector);
+  mth = GSGetClassMethod(self, aSelector);
   if (mth != 0)
     {
       const char	*types = mth->method_types;
@@ -92,11 +145,17 @@ extern BOOL __objc_responds_to(id, SEL);
   return nil;
 }
 
+/**
+ * A dummy method to ensure that the class can safely be held in containers.
+ */
 + (void) release
 {
   /* Do nothing	*/
 }
 
+/**
+ * Returns YES if the receiver responds to aSelector, NO otherwise.
+ */
 + (BOOL) respondsToSelector: (SEL)aSelector
 {
   if (__objc_responds_to(self, aSelector))
@@ -105,16 +164,33 @@ extern BOOL __objc_responds_to(id, SEL);
     return NO;
 }
 
+/**
+ * Returns the receiver.
+ */
 + (id) retain
 {
   return self;
 }
 
-+ (Class) superclass
+/**
+ * Returns the maximum unsigned integer value.
+ */
++ (unsigned int) retainCount
 {
-  return class_get_super_class (self);
+  return UINT_MAX;
 }
 
+/**
+ * Returns the superclass of the receiver
+ */
++ (Class) superclass
+{
+  return GSObjCSuper(self);
+}
+
+/**
+ * Adds the receiver to the current autorelease pool and returns self.
+ */
 - (id) autorelease
 {
 #if	GS_WITH_GC == 0
@@ -123,30 +199,62 @@ extern BOOL __objc_responds_to(id, SEL);
   return self;
 }
 
+/**
+ * Dummy method ... returns the receiver.
+ */
+- (id) awakeAfterUsingCoder: (NSCoder*)aDecoder
+{
+  return self;
+}
+
+/**
+ * Returns the class of the receiver.
+ */
 - (Class) class
 {
   return object_get_class(self);
 }
 
+/**
+ * Calls the -forwardInvocation: method to determine if the 'real' object
+ * referred to by the proxy conforms to aProtocol.  Returns the result.<br />
+ * NB. The default operation of -forwardInvocation: is to raise an exception.
+ */
 - (BOOL) conformsToProtocol: (Protocol*)aProtocol
 {
-  [NSException raise: NSGenericException
-    format: @"subclass %s should override %s", object_get_class_name(self),
-    sel_get_name(_cmd)];
-  return NO;
+  NSMethodSignature	*sig;
+  NSInvocation		*inv;
+  BOOL			ret;
+
+  sig = [self methodSignatureForSelector: _cmd];
+  inv = [NSInvocation invocationWithMethodSignature: sig];
+  [inv setSelector: _cmd];
+  [inv setArgument: &aProtocol atIndex: 2];
+  [self forwardInvocation: inv];
+  [inv getReturnValue: &ret];
+  return ret;
 }
 
+/**
+ * Frees the memory used by the receiver.
+ */
 - (void) dealloc
 {
   NSDeallocateObject((NSObject*)self);
 }
 
+/**
+ * Returns a text descrioption of the receiver.
+ */
 - (NSString*) description
 {
   return [NSString stringWithFormat: @"<%s %lx>",
-	object_get_class_name(self), (unsigned long)self];
+	GSClassNameFromObject(self), (unsigned long)self];
 }
 
+/**
+ * Calls the -forwardInvocation: method and returns the result.
+ */
 - (retval_t) forward:(SEL)aSel :(arglist_t)argFrame
 {
   NSInvocation *inv;
@@ -157,61 +265,88 @@ extern BOOL __objc_responds_to(id, SEL);
   return [inv returnFrame: argFrame];
 }
 
+/** <override-subclass />
+ * Raises an NSInvalidArgumentException
+ */
 - (void) forwardInvocation: (NSInvocation*)anInvocation
 {
   [NSException raise: NSInvalidArgumentException
 	      format: @"NSProxy should not implement '%s'",
-				sel_get_name(_cmd)];
+				GSNameFromSelector(_cmd)];
 }
 
+/**
+ * Returns the address of the receiver ... so it can be stored in a dictionary.
+ */
 - (unsigned int) hash
 {
   return (unsigned int)self;
 }
 
+/** <init /> <override-subclass />
+ * Initialises the receiver and returns the resulting instance.
+ */
 - (id) init
 {
   [NSException raise: NSGenericException
-    format: @"subclass %s should override %s", object_get_class_name(self),
-    sel_get_name(_cmd)];
+    format: @"subclass %s should override %s", GSClassNameFromObject(self),
+    GSNameFromSelector(_cmd)];
   return self;
 }
 
+/**
+ * Tests for pointer equality with anObject
+ */
 - (BOOL) isEqual: (id)anObject
 {
   return (self == anObject);
 }
 
-+ (BOOL) isKindOfClass: (Class)aClass
-{
-  return NO;
-}
-
+/**
+ * Calls the -forwardInvocation: method to determine if the 'real' object
+ * referred to by the proxy is an instance of the specified class.
+ * Returns the result.<br />
+ * NB. The default operation of -forwardInvocation: is to raise an exception.
+ */
 - (BOOL) isKindOfClass: (Class)aClass
 {
-  Class class = self->isa;
+  NSMethodSignature	*sig;
+  NSInvocation		*inv;
+  BOOL			ret;
 
-  while (class != nil)
-    {
-      if (class == aClass)
-	{
-	  return YES;
-	}
-      class = class_get_super_class(class);
-    }
-  return NO;
-}
-
-+ (BOOL) isMemberOfClass: (Class)aClass
-{
-  return(self == aClass);
+  sig = [self methodSignatureForSelector: _cmd];
+  inv = [NSInvocation invocationWithMethodSignature: sig];
+  [inv setSelector: _cmd];
+  [inv setArgument: &aClass atIndex: 2];
+  [self forwardInvocation: inv];
+  [inv getReturnValue: &ret];
+  return ret;
 }
 
+/**
+ * Calls the -forwardInvocation: method to determine if the 'real' object
+ * referred to by the proxy is an instance of the specified class.
+ * Returns the result.<br />
+ * NB. The default operation of -forwardInvocation: is to raise an exception.
+ */
 - (BOOL) isMemberOfClass: (Class)aClass
 {
-  return(self->isa == aClass);
+  NSMethodSignature	*sig;
+  NSInvocation		*inv;
+  BOOL			ret;
+
+  sig = [self methodSignatureForSelector: _cmd];
+  inv = [NSInvocation invocationWithMethodSignature: sig];
+  [inv setSelector: _cmd];
+  [inv setArgument: &aClass atIndex: 2];
+  [self forwardInvocation: inv];
+  [inv getReturnValue: &ret];
+  return ret;
 }
 
+/**
+ * Returns YES
+ */
 - (BOOL) isProxy
 {
   return YES;
@@ -220,11 +355,11 @@ extern BOOL __objc_responds_to(id, SEL);
 - (id) notImplemented: (SEL)aSel
 {
   [NSException raise: NSGenericException
-	      format: @"NSProxy notImplemented %s", sel_get_name(aSel)];
+	      format: @"NSProxy notImplemented %s", GSNameFromSelector(aSel)];
   return self;
 }
 
-/*
+/**
  * If we respond to the method directly, create and return a method
  * signature.  Otherwise raise an exception.
  */
@@ -236,7 +371,7 @@ extern BOOL __objc_responds_to(id, SEL);
     {
       return nil;
     }
-  mth = class_get_instance_method(GSObjCClass(self), aSelector);
+  mth = GSGetInstanceMethod(GSObjCClass(self), aSelector);
   if (mth != 0)
     {
       const char	*types = mth->method_types;
@@ -259,7 +394,7 @@ extern BOOL __objc_responds_to(id, SEL);
     {
       [NSException raise: NSGenericException
 		  format: @"invalid selector passed to %s",
-				sel_get_name(_cmd)];
+				GSNameFromSelector(_cmd)];
       return nil;
     }
   return (*msg)(self, aSelector);
@@ -274,7 +409,7 @@ extern BOOL __objc_responds_to(id, SEL);
     {
       [NSException raise: NSGenericException
 		  format: @"invalid selector passed to %s",
-				sel_get_name(_cmd)];
+				GSNameFromSelector(_cmd)];
       return nil;
     }
   return (*msg)(self, aSelector, anObject);
@@ -290,23 +425,55 @@ extern BOOL __objc_responds_to(id, SEL);
     {
       [NSException raise: NSGenericException
 		  format: @"invalid selector passed to %s",
-				sel_get_name(_cmd)];
+				GSNameFromSelector(_cmd)];
       return nil;
     }
   return (*msg)(self, aSelector, anObject, anotherObject);
 }
 
+/**
+ * Decrement the retain count for the receiver ... deallocate if it would
+ * become negative.
+ */
 - (void) release
 {
 #if	GS_WITH_GC == 0
-  if (_retain_count-- == 0)
+  if (_retain_count == 0)
     {
       [self dealloc];
     }
+  else
+    {
+      _retain_count--;
+    }
 #endif
 }
 
-/*
+/**
+ * Returns the actual object to be encoded for sending over the
+ * network on a Distributed Objects connection.
+ */
+- (id) replacementObjectForPortCoder: (NSPortCoder*)aCoder
+{
+  static Class	proxyClass = 0;
+  static IMP	proxyImp = 0;
+
+  if (proxyImp == 0)
+    {
+      proxyClass = [NSDistantObject class];
+      /*
+       * use get_imp() because NSDistantObject doesn't implement
+       * methodForSelector:
+       */
+      proxyImp = get_imp(GSObjCClass((id)proxyClass),
+	@selector(proxyWithLocal:connection:));
+    }
+
+  return (*proxyImp)(proxyClass, @selector(proxyWithLocal:connection:),
+    self, [aCoder connection]);
+}
+
+/**
  * If we respond to the method directly, return YES, otherwise
  * forward this request to the object we are acting as a proxy for.
  */
@@ -336,6 +503,9 @@ extern BOOL __objc_responds_to(id, SEL);
     }
 }
 
+/**
+ * Increment the retain count for the receiver.
+ */
 - (id) retain
 {
 #if	GS_WITH_GC == 0
@@ -344,26 +514,33 @@ extern BOOL __objc_responds_to(id, SEL);
   return self;
 }
 
+/**
+ * Return the retain count for the receiver.
+ */
 - (unsigned int) retainCount
 {
   return _retain_count + 1;
 }
 
-+ (unsigned) retainCount
-{
-  return UINT_MAX;
-}
-
+/**
+ * Returns the receiver.
+ */
 - (id) self
 {
   return self;
 }
 
+/**
+ * Returns the superclass of the receivers class.
+ */
 - (Class) superclass
 {
   return object_get_super_class(self);
 }
 
+/**
+ * Returns the zone in which the receiver was allocated.
+ */
 - (NSZone*) zone
 {
   return NSZoneFromPointer(self);
