@@ -21,16 +21,16 @@
  * Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA.
 
    <title>NSRange class reference</title>
-   $Date: 2002/02/13 22:25:38 $ $Revision: 1.14 $
+   $Date: 2003/06/07 01:24:41 $ $Revision: 1.16 $
  */ 
 
-#include <config.h>
+#include "config.h"
 
 #define	IN_NSRANGE_M 1
-#include <Foundation/NSException.h>
-#include <Foundation/NSString.h>
-#include <Foundation/NSRange.h>
-#include <Foundation/NSScanner.h>
+#include "Foundation/NSException.h"
+#include "Foundation/NSString.h"
+#include "Foundation/NSRange.h"
+#include "Foundation/NSScanner.h"
 
 @class	NSString;
 
@@ -44,7 +44,7 @@ static BOOL	(*scanStringImp)(NSScanner*, SEL, NSString*, NSString**);
 static id 	(*scannerImp)(Class, SEL, NSString*);
 
 static inline void
-setupCache()
+setupCache(void)
 {
   if (NSStringClass == 0)
     {
