@@ -25,32 +25,32 @@
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA.
 
    <title>NSRunLoop class reference</title>
-   $Date: 2002/03/07 16:49:32 $ $Revision: 1.85 $
+   $Date: 2003/10/28 11:10:23 $ $Revision: 1.111 $
 */
 
-#include <config.h>
-#include <base/preface.h>
-#include <Foundation/NSMapTable.h>
-#include <Foundation/NSDate.h>
-#include <Foundation/NSValue.h>
-#include <Foundation/NSAutoreleasePool.h>
-#include <Foundation/NSPort.h>
-#include <Foundation/NSTimer.h>
-#include <Foundation/NSNotificationQueue.h>
-#include <Foundation/NSRunLoop.h>
-#include <Foundation/NSThread.h>
-#include <Foundation/NSDebug.h>
+#include "config.h"
+#include "GNUstepBase/preface.h"
+#include "Foundation/NSMapTable.h"
+#include "Foundation/NSDate.h"
+#include "Foundation/NSValue.h"
+#include "Foundation/NSAutoreleasePool.h"
+#include "Foundation/NSPort.h"
+#include "Foundation/NSTimer.h"
+#include "Foundation/NSNotificationQueue.h"
+#include "Foundation/NSRunLoop.h"
+#include "Foundation/NSThread.h"
+#include "Foundation/NSDebug.h"
 
-#if HAVE_SYS_TYPES_H
+#ifdef HAVE_SYS_TYPES_H
 #include <sys/types.h>
 #endif
-#if HAVE_SYS_TIME_H
+#ifdef HAVE_SYS_TIME_H
 #include <sys/time.h>
 #endif
-#if HAVE_POLL_H
+#ifdef HAVE_POLL_H
 #include <poll.h>
 #endif
-#if HAVE_UNISTD_H
+#ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif
 #include <time.h>
@@ -58,11 +58,13 @@
 #include <string.h>		/* for memset() */
 
 
+NSString * const NSDefaultRunLoopMode = @"NSDefaultRunLoopMode";
+
 static NSDate	*theFuture = nil;
 
 extern BOOL	GSCheckTasks();
 
-#if	HAVE_POLL
+#ifdef	HAVE_POLL
 typedef struct {
   int		limit;
   short		*index;
@@ -251,6 +253,7 @@ static inline BOOL timerInvalidated(NSTimer* timer)
 		 target: (id)target
 	       argument: (id)argument
 		  delay: (NSTimeInterval)delay;
+- (void) invalidate;
 @end
 
 @implementation GSTimedPerformer
@@ -274,10 +277,7 @@ static inline BOOL timerInvalidated(NSTimer* timer)
 
 - (void) gcFinalize
 {
-  if (timer != nil)
-    {
-      [timer invalidate];
-    }
+  [self invalidate];
 }
 
 - (id) initWithSelector: (SEL)aSelector
@@ -292,14 +292,25 @@ static inline BOOL timerInvalidated(NSTimer* timer)
       target = RETAIN(aTarget);
       argument = RETAIN(anArgument);
       timer = [[NSTimer allocWithZone: NSDefaultMallocZone()]
-	initWithTimeInterval: delay
-	  targetOrInvocation: self
-		    selector: @selector(fire)
-		    userInfo: nil
-		     repeats: NO];
+	initWithFireDate: nil
+		interval: delay
+		  target: self
+		selector: @selector(fire)
+		userInfo: nil
+		 repeats: NO];
     }
   return self;
 }
+
+- (void) invalidate
+{
+  if (timer != nil)
+    {
+      [timer invalidate];
+      DESTROY(timer);
+    }
+}
+
 @end
 
 
@@ -318,7 +329,7 @@ static inline BOOL timerInvalidated(NSTimer* timer)
 #define GSI_ARRAY_RETAIN(A, X)	
 #endif
 
-#include <base/GSIArray.h>
+#include "GNUstepBase/GSIArray.h"
 
 static NSComparisonResult aSort(GSIArrayItem i0, GSIArrayItem i1)
 {
@@ -333,7 +344,7 @@ static IMP	wRelImp;
 static IMP	wRetImp;
 
 static void
-wRelease(NSMapTable* t, const void* w)
+wRelease(NSMapTable* t, void* w)
 {
   (*wRelImp)((id)w, wRelSel);
 }
@@ -344,7 +355,7 @@ wRetain(NSMapTable* t, const void* w)
   (*wRetImp)((id)w, wRetSel);
 }
 
-const NSMapTableValueCallBacks WatcherMapValueCallBacks = 
+static const NSMapTableValueCallBacks WatcherMapValueCallBacks = 
 {
   wRetain,
   wRelease,
@@ -354,25 +365,6 @@ const NSMapTableValueCallBacks WatcherMapValueCallBacks =
 #define	WatcherMapValueCallBacks	NSOwnedPointerMapValueCallBacks 
 #endif
 
-static void
-aRetain(NSMapTable* t, const void* a)
-{
-}
-
-static void
-aRelease(NSMapTable* t, const void* a)
-{
-  GSIArrayEmpty((GSIArray)a);
-  NSZoneFree(((GSIArray)a)->zone, (void*)a);
-}
-
-const NSMapTableValueCallBacks ArrayMapValueCallBacks = 
-{
-  aRetain,
-  aRelease,
-  0
-};
-
 
 
 /**
@@ -396,9 +388,9 @@ const NSMapTableValueCallBacks ArrayMapValueCallBacks =
   NSMapTable	*_wfdMap;
   int		fairStart;	// For trying to ensure fair handling.
   BOOL		completed;	// To mark operation as completed.
-#if	HAVE_POLL
-  int		pollfds_capacity;
-  int		pollfds_count;
+#ifdef	HAVE_POLL
+  unsigned int	pollfds_capacity;
+  unsigned int	pollfds_count;
   struct pollfd	*pollfds;
 #endif
 }
@@ -419,10 +411,19 @@ const NSMapTableValueCallBacks ArrayMapValueCallBacks =
   NSZoneFree(timers->zone, (void*)timers);
   GSIArrayEmpty(watchers);
   NSZoneFree(watchers->zone, (void*)watchers);
-  NSFreeMapTable(_efdMap);
-  NSFreeMapTable(_rfdMap);
-  NSFreeMapTable(_wfdMap);
-#if	HAVE_POLL
+  if (_efdMap != 0)
+    {
+      NSFreeMapTable(_efdMap);
+    }
+  if (_rfdMap != 0)
+    {
+      NSFreeMapTable(_rfdMap);
+    }
+  if (_wfdMap != 0)
+    {
+      NSFreeMapTable(_wfdMap);
+    }
+#ifdef	HAVE_POLL
   if (pollfds != 0)
     {
       objc_free(pollfds);
@@ -463,7 +464,7 @@ const NSMapTableValueCallBacks ArrayMapValueCallBacks =
 }
 
 /**
- * Mark this poll conext as having completed, so that if we are
+ * Mark this poll context as having completed, so that if we are
  * executing a re-entrant poll, the enclosing poll operations
  * know they can stop what they are doing because an inner
  * operation has done the job.
@@ -506,7 +507,7 @@ const NSMapTableValueCallBacks ArrayMapValueCallBacks =
   return self;
 }
 
-#if	HAVE_POLL
+#ifdef	HAVE_POLL
 
 static void setPollfd(int fd, int event, GSRunLoopCtxt *ctxt)
 {
@@ -562,10 +563,10 @@ static void setPollfd(int fd, int event, GSRunLoopCtxt *ctxt)
 - (BOOL) pollUntil: (int)milliseconds within: (NSArray*)contexts
 {
   int		poll_return;
-  int		fdEnd = 0;	/* Number of descriptors being monitored. */
+  int		fdEnd;	/* Number of descriptors being monitored. */
   int		fdIndex;
   int		fdFinish;
-  unsigned	i;
+  unsigned int	i;
 
   i = GSIArrayCount(watchers);
 
@@ -603,7 +604,6 @@ static void setPollfd(int fd, int event, GSRunLoopCtxt *ctxt)
       info = GSIArrayItemAtIndex(watchers, i).obj;
       if (info->_invalidated == YES)
 	{
-	  GSIArrayRemoveItemAtIndex(watchers, i);
 	  continue;
 	}
 
@@ -613,21 +613,18 @@ static void setPollfd(int fd, int event, GSRunLoopCtxt *ctxt)
 	    fd = (int)info->data;
 	    setPollfd(fd, POLLPRI, self);
 	    NSMapInsert(_efdMap, (void*)fd, info);
-	    fdEnd++;
 	    break;
 
 	  case ET_RDESC: 
 	    fd = (int)info->data;
 	    setPollfd(fd, POLLIN, self);
 	    NSMapInsert(_rfdMap, (void*)fd, info);
-	    fdEnd++;
 	    break;
 
 	  case ET_WDESC: 
 	    fd = (int)info->data;
 	    setPollfd(fd, POLLOUT, self);
 	    NSMapInsert(_wfdMap, (void*)fd, info);
-	    fdEnd++;
 	    break;
 
 	  case ET_RPORT: 
@@ -659,7 +656,6 @@ static void setPollfd(int fd, int event, GSRunLoopCtxt *ctxt)
 		    setPollfd(fd, POLLIN, self);
 		    NSMapInsert(_rfdMap, 
 		      (void*)port_fd_array[port_fd_count], info);
-		    fdEnd++;
 		  }
 	      }
 	    break;
@@ -677,21 +673,25 @@ static void setPollfd(int fd, int event, GSRunLoopCtxt *ctxt)
       milliseconds = 0;
     }
 
-if (0) {
-  int i;
+#if 0
+{
+  unsigned int i;
   fprintf(stderr, "poll %d %d:", milliseconds, pollfds_count);
   for (i = 0; i < pollfds_count; i++)
     fprintf(stderr, " %d,%x", pollfds[i].fd, pollfds[i].events);
   fprintf(stderr, "\n");
 }
+#endif
   poll_return = poll (pollfds, pollfds_count, milliseconds);
-if (0) {
-  int i;
+#if 0
+{
+  unsigned int i;
   fprintf(stderr, "ret %d %d:", poll_return, pollfds_count);
   for (i = 0; i < pollfds_count; i++)
     fprintf(stderr, " %d,%x", pollfds[i].fd, pollfds[i].revents);
   fprintf(stderr, "\n");
 }
+#endif
 
   NSDebugMLLog(@"NSRunLoop", @"poll returned %d\n", poll_return);
 
@@ -702,13 +702,11 @@ if (0) {
 	  GSCheckTasks();
 	  poll_return = 0;
 	}
-#ifdef __MINGW__
       else if (errno == 0)
 	{
-	  /* MinGW often returns an errno == 0. Not sure why */
+	  /* Some systems returns an errno == 0. Not sure why */
 	  poll_return = 0;
 	}
-#endif
       else
 	{
 	  /* Some exceptional condition happened. */
@@ -737,6 +735,7 @@ if (0) {
    * inputs where multiple inputs are in use.  Note - fairStart can be
    * modified while we are in the loop (by recursive calls).
    */
+  fdEnd = pollfds_count;
   if (++fairStart >= fdEnd)
     {
       fairStart = 0;
@@ -765,17 +764,9 @@ if (0) {
 	   */
 	  if (pollfds[fdIndex].revents & (POLLPRI|POLLERR|POLLHUP|POLLNVAL))
 	    {
-	      watcher = (GSRunLoopWatcher*)NSMapGet(_efdMap,
-		(void*)fd);
+	      watcher = (GSRunLoopWatcher*)NSMapGet(_efdMap, (void*)fd);
 	      if (watcher != nil && watcher->_invalidated == NO)
 		{
-		  /*
-		   * The watcher is still valid - so call its
-		   * receivers event handling method.
-		   */
-		  (*watcher->handleEvent)(watcher->receiver,
-		    eventSel, watcher->data, watcher->type,
-		    (void*)(gsaddr)fd, mode);
 		  i = [contexts count];
 		  while (i-- > 0)
 		    {
@@ -783,6 +774,13 @@ if (0) {
 
 		      if (c != self) [c endEvent: (void*)fd type: ET_EDESC];
 		    }
+		  /*
+		   * The watcher is still valid - so call its
+		   * receivers event handling method.
+		   */
+		  (*watcher->handleEvent)(watcher->receiver,
+		    eventSel, watcher->data, watcher->type,
+		    (void*)(gsaddr)fd, mode);
 		}
 	      GSNotifyASAP();
 	      if (completed == YES)
@@ -793,17 +791,9 @@ if (0) {
 	    }
 	  if (pollfds[fdIndex].revents & (POLLOUT|POLLERR|POLLHUP|POLLNVAL))
 	    {
-	      watcher = (GSRunLoopWatcher*)NSMapGet(_wfdMap,
-		(void*)fd);
+	      watcher = (GSRunLoopWatcher*)NSMapGet(_wfdMap, (void*)fd);
 	      if (watcher != nil && watcher->_invalidated == NO)
 		{
-		  /*
-		   * The watcher is still valid - so call its
-		   * receivers event handling method.
-		   */
-		  (*watcher->handleEvent)(watcher->receiver,
-		    eventSel, watcher->data, watcher->type,
-		    (void*)(gsaddr)fd, mode);
 		  i = [contexts count];
 		  while (i-- > 0)
 		    {
@@ -811,6 +801,13 @@ if (0) {
 
 		      if (c != self) [c endEvent: (void*)fd type: ET_WDESC];
 		    }
+		  /*
+		   * The watcher is still valid - so call its
+		   * receivers event handling method.
+		   */
+		  (*watcher->handleEvent)(watcher->receiver,
+		    eventSel, watcher->data, watcher->type,
+		    (void*)(gsaddr)fd, mode);
 		}
 	      GSNotifyASAP();
 	      if (completed == YES)
@@ -821,17 +818,9 @@ if (0) {
 	    }
 	  if (pollfds[fdIndex].revents & (POLLIN|POLLERR|POLLHUP|POLLNVAL))
 	    {
-	      watcher = (GSRunLoopWatcher*)NSMapGet(_rfdMap,
-		(void*)fd);
+	      watcher = (GSRunLoopWatcher*)NSMapGet(_rfdMap, (void*)fd);
 	      if (watcher != nil && watcher->_invalidated == NO)
 		{
-		  /*
-		   * The watcher is still valid - so call its
-		   * receivers event handling method.
-		   */
-		  (*watcher->handleEvent)(watcher->receiver,
-		    eventSel, watcher->data, watcher->type,
-		    (void*)(gsaddr)fd, mode);
 		  i = [contexts count];
 		  while (i-- > 0)
 		    {
@@ -839,6 +828,13 @@ if (0) {
 
 		      if (c != self) [c endEvent: (void*)fd type: ET_RDESC];
 		    }
+		  /*
+		   * The watcher is still valid - so call its
+		   * receivers event handling method.
+		   */
+		  (*watcher->handleEvent)(watcher->receiver,
+		    eventSel, watcher->data, watcher->type,
+		    (void*)(gsaddr)fd, mode);
 		}
 	      GSNotifyASAP();
 	      if (completed == YES)
@@ -899,6 +895,8 @@ if (0) {
     }
   else 
     {
+      timeout.tv_sec = -1;
+      timeout.tv_usec = -1;
       select_timeout = NULL;
     }
 
@@ -984,7 +982,7 @@ if (0) {
 		  {
 		    fd = port_fd_array[port_fd_count];
 		    FD_SET (port_fd_array[port_fd_count], &read_fds);
-		    if (fd > fdEnd)
+	            if (fd > fdEnd)
 		      fdEnd = fd;
 		    NSMapInsert(_rfdMap, 
 		      (void*)port_fd_array[port_fd_count], info);
@@ -1009,6 +1007,8 @@ if (0) {
       select_timeout = &timeout;
     }
 
+  // NSDebugMLLog(@"NSRunLoop", @"select timeout %d,%d", timeout.tv_sec, timeout.tv_usec);
+
   select_return = select (fdEnd, &read_fds, &write_fds,
     &exception_fds, select_timeout);
 
@@ -1021,13 +1021,11 @@ if (0) {
 	  GSCheckTasks();
 	  select_return = 0;
 	}
-#ifdef __MINGW__
       else if (errno == 0)
 	{
-	  /* MinGW often returns an errno == 0. Not sure why */
-	    select_return = 0;
+	  /* Some systems return an errno == 0. Not sure why */
+	  select_return = 0;
 	}
-#endif
       else
 	{
 	  /* Some exceptional condition happened. */
@@ -1072,17 +1070,9 @@ if (0) {
 	{
 	  GSRunLoopWatcher	*watcher;
 
-	  watcher = (GSRunLoopWatcher*)NSMapGet(_efdMap,
-	    (void*)fdIndex);
+	  watcher = (GSRunLoopWatcher*)NSMapGet(_efdMap, (void*)fdIndex);
 	  if (watcher != nil && watcher->_invalidated == NO)
 	    {
-	      /*
-	       * The watcher is still valid - so call its receivers
-	       * event handling method.
-	       */
-	      (*watcher->handleEvent)(watcher->receiver,
-		eventSel, watcher->data, watcher->type,
-		(void*)(gsaddr)fdIndex, mode);
 	      i = [contexts count];
 	      while (i-- > 0)
 		{
@@ -1090,6 +1080,13 @@ if (0) {
 
 		  if (c != self) [c endEvent: (void*)fdIndex type: ET_EDESC];
 		}
+	      /*
+	       * The watcher is still valid - so call its receivers
+	       * event handling method.
+	       */
+	      (*watcher->handleEvent)(watcher->receiver,
+		eventSel, watcher->data, watcher->type,
+		(void*)(gsaddr)fdIndex, mode);
 	    }
 	  GSNotifyASAP();
 	  if (completed == YES)
@@ -1102,16 +1099,9 @@ if (0) {
 	{
 	  GSRunLoopWatcher	*watcher;
 
-	  watcher = NSMapGet(_wfdMap, (void*)fdIndex);
+	  watcher = (GSRunLoopWatcher*)NSMapGet(_wfdMap, (void*)fdIndex);
 	  if (watcher != nil && watcher->_invalidated == NO)
 	    {
-	      /*
-	       * The watcher is still valid - so call its receivers
-	       * event handling method.
-	       */
-	      (*watcher->handleEvent)(watcher->receiver,
-		eventSel, watcher->data, watcher->type,
-		(void*)(gsaddr)fdIndex, mode);
 	      i = [contexts count];
 	      while (i-- > 0)
 		{
@@ -1119,6 +1109,13 @@ if (0) {
 
 		  if (c != self) [c endEvent: (void*)fdIndex type: ET_WDESC];
 		}
+	      /*
+	       * The watcher is still valid - so call its receivers
+	       * event handling method.
+	       */
+	      (*watcher->handleEvent)(watcher->receiver,
+		eventSel, watcher->data, watcher->type,
+		(void*)(gsaddr)fdIndex, mode);
 	    }
 	  GSNotifyASAP();
 	  if (completed == YES)
@@ -1134,13 +1131,6 @@ if (0) {
 	  watcher = (GSRunLoopWatcher*)NSMapGet(_rfdMap, (void*)fdIndex);
 	  if (watcher != nil && watcher->_invalidated == NO)
 	    {
-	      /*
-	       * The watcher is still valid - so call its receivers
-	       * event handling method.
-	       */
-	      (*watcher->handleEvent)(watcher->receiver,
-		eventSel, watcher->data, watcher->type,
-		(void*)(gsaddr)fdIndex, mode);
 	      i = [contexts count];
 	      while (i-- > 0)
 		{
@@ -1148,6 +1138,13 @@ if (0) {
 
 		  if (c != self) [c endEvent: (void*)fdIndex type: ET_RDESC];
 		}
+	      /*
+	       * The watcher is still valid - so call its receivers
+	       * event handling method.
+	       */
+	      (*watcher->handleEvent)(watcher->receiver,
+		eventSel, watcher->data, watcher->type,
+		(void*)(gsaddr)fdIndex, mode);
 	    }
 	  GSNotifyASAP();
 	  if (completed == YES)
@@ -1180,6 +1177,42 @@ if (0) {
 
 @implementation NSObject (TimedPerformers)
 
+/**
+ * Cancels any perform operations set up for the specified target
+ * in the current run loop.
+ */
++ (void) cancelPreviousPerformRequestsWithTarget: (id)target
+{
+  NSMutableArray	*perf = [[NSRunLoop currentRunLoop] _timedPerformers];
+  unsigned		count = [perf count];
+
+  if (count > 0)
+    {
+      GSTimedPerformer	*array[count];
+
+      IF_NO_GC(RETAIN(target));
+      [perf getObjects: array];
+      while (count-- > 0)
+	{
+	  GSTimedPerformer	*p = array[count];
+
+	  if (p->target == target)
+	    {
+	      [p invalidate];
+	      [perf removeObjectAtIndex: count];
+	    }
+	}
+      RELEASE(target);
+    }
+}
+
+/**
+ * Cancels any perform operations set up for the specified target
+ * in the current loop, but only if the value of aSelector and argument
+ * with which the performs were set up match those supplied.<br />
+ * Matching of the argument may be either by pointer equality or by
+ * use of the [NSObject-isEqual:] method.
+ */
 + (void) cancelPreviousPerformRequestsWithTarget: (id)target
 					selector: (SEL)aSelector
 					  object: (id)arg
@@ -1199,8 +1232,9 @@ if (0) {
 	  GSTimedPerformer	*p = array[count];
 
 	  if (p->target == target && sel_eq(p->selector, aSelector)
-	    && [p->argument isEqual: arg])
+	    && (p->argument == arg || [p->argument isEqual: arg]))
 	    {
+	      [p invalidate];
 	      [perf removeObjectAtIndex: count];
 	    }
 	}
@@ -1312,10 +1346,14 @@ if (0) {
 	  item->_date = RETAIN(d);
 	}
       else
-	item->_date = RETAIN(theFuture);
+	{
+	  item->_date = RETAIN(theFuture);
+	}
     }
   else
-    item->_date = RETAIN(theFuture);
+    {
+      item->_date = RETAIN(theFuture);
+    }
   GSIArrayInsertSorted(watchers, (GSIArrayItem)item, aSort);
 }
 
@@ -1369,6 +1407,7 @@ if (0) {
 		    }
 		}
 	    }
+	  NSEndMapTableEnumeration(&enumerator);
 
 	  /*
 	   * Finally, fire the requests.
@@ -1490,7 +1529,7 @@ if (0) {
 
   info = [self _getWatcher: data type: type forMode: mode];
 
-  if (info && info->receiver == (id)watcher)
+  if (info != nil && (id)info->receiver == (id)watcher)
     {
       /* Increment usage count for this watcher. */
       info->count++;
@@ -1578,22 +1617,9 @@ if (0) {
 
 + (NSRunLoop*) currentRunLoop
 {
-  static NSString	*key = @"NSRunLoopThreadKey";
-  NSMutableDictionary	*d;
-  NSRunLoop		*r;
+  extern NSRunLoop	*GSRunLoopForThread();
 
-  d = GSCurrentThreadDictionary();
-  r = [d objectForKey: key];
-  if (r == nil)
-    {
-      if (d != nil)
-	{
-	  r = [self new];
-	  [d setObject: r forKey: key];
-	  RELEASE(r);
-	}
-    }
-  return r;
+  return GSRunLoopForThread(nil);
 }
 
 /* This is the designated initializer. */
@@ -1606,7 +1632,7 @@ if (0) {
       _contextMap = NSCreateMapTable (NSNonRetainedObjectMapKeyCallBacks,
 					 NSObjectMapValueCallBacks, 0);
       _timedPerformers = [[NSMutableArray alloc] initWithCapacity: 8];
-#if	HAVE_POLL
+#ifdef	HAVE_POLL
       _extra = objc_malloc(sizeof(pollextra));
       memset(_extra, '\0', sizeof(pollextra));
 #endif
@@ -1622,7 +1648,7 @@ if (0) {
 
 - (void) gcFinalize
 {
-#if	HAVE_POLL
+#ifdef	HAVE_POLL
   if (_extra != 0)
     {
       pollextra	*e = (pollextra*)_extra;
@@ -1650,8 +1676,10 @@ if (0) {
 }
 
 
-/* Adding timers.  They are removed when they are invalid. */
-
+/**
+ * Adds a timer to the loop in the specified mode.<br />
+ * Timers are removed automatically when they are invalid.<br />
+ */
 - (void) addTimer: (NSTimer*)timer
 	  forMode: (NSString*)mode
 {
@@ -1671,11 +1699,13 @@ if (0) {
 
 
 /**
- * Fire appropriate timers and determine the earliest time that anything
- * watched for becomes useless.
+ * Fires timers whose fire date has passed, and checks timers and limit dates
+ * for input sources, determining the earliest time that anything watched for
+ * becomes useless.  Returns that date/time.
  */
 - (NSDate*) limitDateForMode: (NSString*)mode
 {
+  extern NSTimer	*GSHousekeeper();
   GSRunLoopCtxt		*context = NSMapGet(_contextMap, mode);
   NSDate		*when = nil;
 
@@ -1792,6 +1822,32 @@ if (0) {
 		  min_watcher = nil;
 		}
 	    }
+
+	  /*
+	   * If there is nothing being watched, and no valid timers
+	   * other than the housekeeper, we set min_timer to nil so
+	   * that the housekeeper timer does not keep the runloop
+	   * active.  It's a special case set up in NSThread.m
+	   */
+	  if (min_watcher == nil && min_timer != nil
+	    && min_timer == GSHousekeeper())
+	    {
+	      unsigned count = GSIArrayCount(timers);
+
+	      while (count-- > 1)
+		{
+		  NSTimer	*tmp = GSIArrayItemAtIndex(timers, 0).obj;
+		  if (timerInvalidated(tmp) == YES)
+		    {
+		      GSIArrayRemoveItemAtIndex(timers, count);
+		    }
+		}
+	      if (GSIArrayCount(timers) == 1)
+		{
+		  min_timer = nil;
+		}
+	    }
+
 	  _currentMode = savedMode;
 	}
       NS_HANDLER
@@ -1834,7 +1890,7 @@ if (0) {
 }
 
 /**
- * Listen to input sources.<br />
+ * Listen for events from input sources.<br />
  * If limit_date is nil or in the past, then don't wait;
  * just poll inputs and return,
  * otherwise block until input is available or until the
@@ -1871,36 +1927,11 @@ if (0) {
 	  NSDebugMLLog(@"NSRunLoop", @"no inputs in mode %@", mode);
 	  GSNotifyASAP();
 	  GSNotifyIdle();
-	  ti = [limit_date timeIntervalSinceNow];
 	  /*
 	   * Pause for as long as possible (up to the limit date)
 	   */
-	  if (ti > 0.0)
-	    {
-#if	defined(HAVE_USLEEP)
-	      if (ti >= INT_MAX / 1000000)
-		{
-		  ti = INT_MAX;
-		}
-	      else
-		{
-		  ti *= 1000000;
-		}
-	      usleep (ti);
-#elif	defined(__MINGW__)
-	      if (ti >= INT_MAX / 1000)
-		{
-		  ti = INT_MAX;
-		}
-	      else
-		{
-		  ti *= 1000;
-		}
-	      Sleep (ti);
-#else
-	      sleep (ti);
-#endif
-	    }
+	  [NSThread sleepUntilDate: limit_date];
+	  ti = [limit_date timeIntervalSinceNow];
 	  GSCheckTasks();
 	  if (context != nil)
 	    {
@@ -1913,12 +1944,13 @@ if (0) {
 	}
 
       /* Find out how much time we should wait, and set SELECT_TIMEOUT. */
-      if (!limit_date)
+      if (limit_date == nil
+       || (ti = [limit_date timeIntervalSinceNow]) <= 0.0)
 	{
 	  /* Don't wait at all. */
 	  timeout_ms = 0;
 	}
-      else if ((ti = [limit_date timeIntervalSinceNow]) > 0.0)
+      else
 	{
 	  /* Wait until the LIMIT_DATE. */
 	  NSDebugMLLog(@"NSRunLoop", @"accept I/P before %f (sec from now %f)", 
@@ -1932,24 +1964,6 @@ if (0) {
 	      timeout_ms = ti * 1000;
 	    }
 	}
-      else if (ti <= 0.0)
-	{
-	  /* The LIMIT_DATE has already past; return immediately without
-	     polling any inputs. */
-	  GSCheckTasks();
-	  [self _checkPerformers: context];
-	  GSNotifyASAP();
-	  NSDebugMLLog(@"NSRunLoop", @"limit date past, returning");
-	  _currentMode = savedMode;
-	  RELEASE(arp);
-	  NS_VOIDRETURN;
-	}
-      else
-	{
-	  /* Wait forever. */
-	  NSDebugMLLog(@"NSRunLoop", @"accept input waiting forever");
-	  timeout_ms = -1;
-	}
 
       if ([_contextStack indexOfObjectIdenticalTo: context] == NSNotFound)
 	{
@@ -1985,27 +1999,18 @@ if (0) {
 
 /**
  * Calls -acceptInputForMode:beforeDate: to run the loop once.<br />
- * If the limit dates for all of mode's input sources have passed,
- * returns NO without running the loop, otherwise returns YES.
+ * The specified date may be nil ... in which case the loop runs
+ * until the limit date of the first input or timeout.<br />
+ * If the specified date is in the past, runs the loop once only, to
+ * handle any events already available.<br />
+ * If there are no input sources in mode, returns NO without running the loop,
+ * otherwise returns YES.
  */
 - (BOOL) runMode: (NSString*)mode beforeDate: (NSDate*)date
 {
   id	d;
 
-  NSAssert(mode && date, NSInvalidArgumentException);
-  /* If date has already passed, simply return. */
-  if ([date timeIntervalSinceNow] < 0)
-    {
-      NSDebugMLLog(@"NSRunLoop", @"run mode with date already past");
-      /*
-       * Notify if any tasks have completed.
-       */
-      if (GSCheckTasks() == YES)
-	{
-	  GSNotifyASAP();
-	}
-      return NO;
-    }
+  NSAssert(mode != nil, NSInvalidArgumentException);
 
   /* Find out how long we can wait before first limit date. */
   d = [self limitDateForMode: mode];
@@ -2027,7 +2032,10 @@ if (0) {
    * Retain the date in case the firing of a timer (or some other event)
    * releases it.
    */
-  d = [d earlierDate: date];
+  if (date != nil)
+    {
+      d = [d earlierDate: date];
+    }
   IF_NO_GC(RETAIN(d));
 
   /* Wait, listening to our input sources. */
@@ -2066,12 +2074,53 @@ if (0) {
 - (void) addPort: (NSPort*)port
          forMode: (NSString*)mode
 {
-  return [self addEvent: (void*)port
-		   type: ET_RPORT
-		watcher: (id<RunLoopEvents>)port
-		forMode: (NSString*)mode];
+  [self addEvent: (void*)port
+	    type: ET_RPORT
+	 watcher: (id<RunLoopEvents>)port
+	 forMode: (NSString*)mode];
 }
 
+/**
+ * Cancels any perform operations set up for the specified target
+ * in the receiver.
+ */
+- (void) cancelPerformSelectorsWithTarget: (id) target
+{
+  NSMapEnumerator	enumerator;
+  GSRunLoopCtxt		*context;
+  void			*mode;
+
+  enumerator = NSEnumerateMapTable(_contextMap);
+
+  while (NSNextMapEnumeratorPair(&enumerator, &mode, (void**)&context))
+    {
+      if (context != nil)
+	{
+	  GSIArray	performers = context->performers;
+	  unsigned	count = GSIArrayCount(performers);
+
+	  while (count--)
+	    {
+	      GSRunLoopPerformer	*p;
+
+	      p = GSIArrayItemAtIndex(performers, count).obj;
+	      if (p->target == target)
+		{
+		  GSIArrayRemoveItemAtIndex(performers, count);
+		}
+	    }
+	}
+    }
+  NSEndMapTableEnumeration(&enumerator);
+}
+
+/**
+ * Cancels any perform operations set up for the specified target
+ * in the receiver, but only if the value of aSelector and argument
+ * with which the performs were set up match those supplied.<br />
+ * Matching of the argument may be either by pointer equality or by
+ * use of the [NSObject-isEqual:] method.
+ */
 - (void) cancelPerformSelector: (SEL)aSelector
 			target: (id) target
 		      argument: (id) argument
@@ -2095,13 +2144,14 @@ if (0) {
 
 	      p = GSIArrayItemAtIndex(performers, count).obj;
 	      if (p->target == target && sel_eq(p->selector, aSelector)
-		&& p->argument == argument)
+		&& (p->argument == argument || [p->argument isEqual: argument]))
 		{
 		  GSIArrayRemoveItemAtIndex(performers, count);
 		}
 	    }
 	}
     }
+  NSEndMapTableEnumeration(&enumerator);
 }
 
 - (void) configureAsServer
@@ -2109,6 +2159,15 @@ if (0) {
 /* Nothing to do here */
 }
 
+/**
+ * Sets up sending of aSelector to target with argument.<br />
+ * The selector is sent before the next runloop iteration (unless
+ * cancelled before then).<br />
+ * The target and argument objects are <em>not</em> retained.<br />
+ * The order value is used to determine the order in which messages
+ * are sent if multiple messages have been set up. Messages with a lower
+ * order value are sent first.
+ */
 - (void) performSelector: (SEL)aSelector
 		  target: (id)target
 		argument: (id)argument
@@ -2152,7 +2211,7 @@ if (0) {
 	      GSRunLoopPerformer	*p;
 
 	      p = GSIArrayItemAtIndex(performers, i).obj;
-	      if (p->order <= order)
+	      if (p->order > order)
 		{
 		  GSIArrayInsertItem(performers, (GSIArrayItem)item, i);
 		  break;
@@ -2170,7 +2229,7 @@ if (0) {
 - (void) removePort: (NSPort*)port
             forMode: (NSString*)mode
 {
-  return [self removeEvent: (void*)port type: ET_RPORT forMode: mode all: NO];
+  [self removeEvent: (void*)port type: ET_RPORT forMode: mode all: NO];
 }
 
 @end
