@@ -23,15 +23,20 @@
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA.
 
    <title>NSScanner class reference</title>
-   $Date: 2002/03/13 09:58:43 $ $Revision: 1.36 $
+   $Date: 2004/02/29 02:57:22 $ $Revision: 1.46 $
 */ 
 
-#include <config.h>
-#include <base/Unicode.h>
-#include <Foundation/NSScanner.h>
-#include <Foundation/NSException.h>
-#include <Foundation/NSObjCRuntime.h>
-#include <Foundation/NSUserDefaults.h>
+#include "config.h"
+#include "GNUstepBase/Unicode.h"
+#include "Foundation/NSScanner.h"
+#include "Foundation/NSException.h"
+#include "Foundation/NSObjCRuntime.h"
+#include "Foundation/NSUserDefaults.h"
+
+/* We need to define _GNU_SOURCE on systems (SuSE) to get LONG_LONG_MAX.  */
+#ifndef _GNU_SOURCE
+#define _GNU_SOURCE
+#endif
 #include <float.h>
 #include <limits.h>
 #include <math.h>
@@ -45,6 +50,20 @@
 #define ULONG_LONG_MAX ULLONG_MAX
 #endif
 
+/**
+ * <p>
+ *   The NSScanner class cluster (currently a single class in GNUstep)
+ *   provides a mechanism to parse the contents of a string into number
+ *   and string values by making a sequence of scan operations to
+ *   step through the string retrieving successive items.
+ * </p>
+ * <p>
+ *   You can tell the scanner whether its scanning is supposed to be
+ *   case sensitive or not, and you can specify a set of characters
+ *   to be skipped before each scanning operation (by default,
+ *   whitespace and newlines).
+ * </p>
+ */
 @implementation NSScanner
 
 @class	GSCString;
@@ -103,8 +122,9 @@ typedef struct {
     }
 }
 
-/*
- * Create and return a scanner that scans aString.
+/**
+ * Create and return a scanner that scans aString.<br />
+ * Uses -initWithString: and with no locale set.
  */
 + (id) scannerWithString: (NSString *)aString
 {
@@ -112,6 +132,11 @@ typedef struct {
     initWithString: aString]);
 }
 
+/**
+ * Returns an NSScanner instance set up to scan aString
+ * (using -initWithString: and with a locale set the default locale
+ * (using -setLocale:
+ */
 + (id) localizedScannerWithString: (NSString*)aString
 {
   NSScanner		*scanner = [self scannerWithString: aString];
@@ -123,9 +148,14 @@ typedef struct {
   return scanner;
 }
 
-/*
- * Initialize a a newly-allocated scanner to scan aString.
- * Returns self.
+/**
+ * Initialises the scanner to scan aString.  The GNUstep
+ * implementation may make an internal copy of the original
+ * string - so it is not safe to assume that if you modify a
+ * mutable string that you initialised a scanner with, the changes
+ * will be visible to the scanner.
+ * <br/>
+ * Returns the scanner object.
  */
 - (id) initWithString: (NSString *)aString
 {
@@ -200,9 +230,10 @@ typedef struct {
   [super dealloc];
 }
 
-/*
- * Returns YES if no more characters remain to be scanned.
- * Returns YES if all characters remaining to be scanned are to be skipped.
+/**
+ * Returns YES if no more characters remain to be scanned.<br />
+ * Returns YES if all characters remaining to be scanned
+ * are to be skipped.<br />
  * Returns NO if there are characters left to scan.
  */
 - (BOOL) isAtEnd
@@ -288,8 +319,15 @@ typedef struct {
   return YES;
 }
 
-/*
- * Scan an int into value.
+/**
+ * After initial skipping (if any), this method scans a integer value,
+ * placing it in <em>intValue</em> if that is not null.
+ * <br/>
+ * Returns YES if anything is scanned, NO otherwise.
+ * <br/>
+ * On overflow, INT_MAX or INT_MIN is put into <em>intValue</em>
+ * <br/>
+ * Scans past any excess digits
  */
 - (BOOL) scanInt: (int*)value
 {
@@ -306,7 +344,7 @@ typedef struct {
  * Internal version used by scanRadixUnsignedInt: and scanHexInt: .
  */
 - (BOOL) scanUnsignedInt_: (unsigned int *)value
-		    radix: (int)radix
+		    radix: (unsigned int)radix
 		gotDigits: (BOOL)gotDigits
 {
   unsigned int	num = 0;
@@ -381,12 +419,22 @@ typedef struct {
   return YES;
 }
 
-/*
- * Scan an unsigned int of the given radix into value.
+/**
+ * After initial skipping (if any), this method scans an unsigned
+ * integer value placing it in <em>intValue</em> if that is not null.
+ * If the number begins with "0x" or "0X" it is treated as hexadecimal,
+ * otherwise if the number begins with "0" it is treated as octal,
+ * otherwise the number is treated as decimal.
+ * <br/>
+ * Returns YES if anything is scanned, NO otherwise.
+ * <br/>
+ * On overflow, INT_MAX or INT_MIN is put into <em>intValue</em>
+ * <br/>
+ * Scans past any excess digits
  */
 - (BOOL) scanRadixUnsignedInt: (unsigned int *)value
 {
-  int		radix;
+  unsigned int	radix;
   BOOL		gotDigits = NO;
   unsigned int	saveScanLocation = _scanLocation;
 
@@ -423,8 +471,16 @@ typedef struct {
   return NO;
 }
 
-/*
- * Scan a hexadecimal unsigned integer into value.
+/**
+ * After initial skipping (if any), this method scans a hexadecimal
+ * integer value (optionally prefixed by "0x" or "0X"),
+ * placing it in <em>intValue</em> if that is not null.
+ * <br/>
+ * Returns YES if anything is scanned, NO otherwise.
+ * <br/>
+ * On overflow, INT_MAX or INT_MIN is put into <em>intValue</em>
+ * <br/>
+ * Scans past any excess digits
  */
 - (BOOL) scanHexInt: (unsigned int *)value
 {
@@ -464,9 +520,17 @@ typedef struct {
   return NO;
 }
 
-/*
- * Scan a long long int into value.
- * Same as scanInt, except with different variable types and limits.
+/**
+ * After initial skipping (if any), this method scans a long
+ * decimal integer value placing it in <em>longLongValue</em> if that
+ * is not null.
+ * <br/>
+ * Returns YES if anything is scanned, NO otherwise.
+ * <br/>
+ * On overflow, LONG_LONG_MAX or LONG_LONG_MIN is put into
+ * <em>longLongValue</em>
+ * <br/>
+ * Scans past any excess digits
  */
 - (BOOL) scanLongLong: (long long *)value
 {
@@ -552,19 +616,25 @@ typedef struct {
 #endif /* defined(LONG_LONG_MAX) */
 }
 
+/**
+ * Not implemented.
+ */
 - (BOOL) scanDecimal: (NSDecimal*)value
 {
   [self notImplemented:_cmd];			/* FIXME */
   return NO;
 }
 
-/*
- * Scan a double into value.
- * Returns YES if a valid floating-point expression was scanned. 
- * Returns NO otherwise.
- * On overflow, HUGE_VAL or -HUGE_VAL is put into value and YES is returned.
- * On underflow, 0.0 is put into value and YES is returned.
- * Based on the strtod code from the GNU C library.
+/**
+ * After initial skipping (if any), this method scans a double value,
+ * placing it in <em>doubleValue</em> if that is not null.
+ * Returns YES if anything is scanned, NO otherwise.
+ * <br/>
+ * On overflow, HUGE_VAL or - HUGE_VAL is put into <em>doubleValue</em>
+ * <br/>
+ * On underflow, 0.0 is put into <em>doubleValue</em>
+ * <br/>
+ * Scans past any excess digits
  */
 - (BOOL) scanDouble: (double *)value
 {
@@ -680,12 +750,16 @@ typedef struct {
   return YES;
 }
 
-/*
- * Scan a float into value.
- * Returns YES if a valid floating-point expression was scanned. 
- * Returns NO otherwise.
- * On overflow, HUGE_VAL or -HUGE_VAL is put into value and YES is returned.
- * On underflow, 0.0 is put into value and YES is returned.
+/**
+ * After initial skipping (if any), this method scans a float value,
+ * placing it in <em>floatValue</em> if that is not null.
+ * Returns YES if anything is scanned, NO otherwise.
+ * <br/>
+ * On overflow, HUGE_VAL or - HUGE_VAL is put into <em>floatValue</em>
+ * <br/>
+ * On underflow, 0.0 is put into <em>floatValue</em>
+ * <br/>
+ * Scans past any excess digits
  */
 - (BOOL) scanFloat: (float*)value
 {
@@ -701,12 +775,13 @@ typedef struct {
   return NO;
 }
     
-/*
- * Scan as long as characters from aSet are encountered.
- * Returns YES if any characters were scanned.
- * Returns NO if no characters were scanned.
- * If value is non-NULL, and any characters were scanned, a string
- * containing the scanned characters is returned by reference in value.
+/**
+ * After initial skipping (if any), this method scans any characters
+ * from aSet, terminating when a character not in the set
+ * is found.<br />
+ * Returns YES if any character is scanned, NO otherwise.<br />
+ * If value is not null, any character scanned are
+ * stored in a string returned in this location.
  */
 - (BOOL) scanCharactersFromSet: (NSCharacterSet *)aSet 
 		    intoString: (NSString **)value
@@ -760,12 +835,12 @@ typedef struct {
   return NO;
 }
 
-/*
- * Scan until a character from aSet is encountered.
- * Returns YES if any characters were scanned.
- * Returns NO if no characters were scanned.
- * If value is non-NULL, and any characters were scanned, a string
- * containing the scanned characters is returned by reference in value.
+/**
+ * After initial skipping (if any), this method scans characters until
+ * it finds one in <em>set</em>.  The scanned characters are placed in
+ * <em>stringValue</em> if that is not null.
+ * <br/>
+ * Returns YES if anything is scanned, NO otherwise.
  */
 - (BOOL) scanUpToCharactersFromSet: (NSCharacterSet *)aSet 
 		        intoString: (NSString **)value
@@ -819,13 +894,10 @@ typedef struct {
   return YES;
 }
 
-/*
- * Scans for aString.
- * Returns YES if the characters at the scan location match aString.
- * Returns NO if the characters at the scan location do not match aString.
- * If the characters at the scan location match aString.
- * If value is non-NULL, and the characters at the scan location match aString,
- * a string containing the matching string is returned by reference in value.
+/**
+ * After initial skipping (if any), this method scans for string
+ * and places the characters found in value if that is not null.<br/>
+ * Returns YES if anything is scanned, NO otherwise.
  */
 - (BOOL) scanString: (NSString *)string intoString: (NSString **)value
 {
@@ -851,12 +923,33 @@ typedef struct {
   return YES;
 }
 
-/*
- * Scans the string until aString is encountered..
- * Returns YES if any characters were scanned.
- * Returns NO if no characters were scanned.
- * If value is non-NULL, and any characters were scanned, a string
- * containing the scanned characters is returned by reference in value.
+/**
+ * <p>After initial skipping (if any), this method scans characters until
+ * it finds string.  The scanned characters are placed in
+ * value if that is not null.  If string is not found, all the characters
+ * up to the end of the scanned string will be returned.
+ * </p>
+ * Returns YES if anything is scanned, NO otherwise.<br />
+ * <p>NB. If the current scanner location points to a copy of string, or
+ * points to skippable characters immediately before a copy of string
+ * then this method returns NO since it finds no characters to store
+ * in value before it finds string.
+ * </p>
+ * <p>To count the occurrances of string, this should be used in
+ * conjunction with the -scanString:intoString: method.
+ * </p>
+ * <example>
+ * NSString *ch = @"["; 
+ * unsigned total = 0; 
+ *
+ * [scanner scanUpToString: ch intoString: NULL]; 
+ * while ([scanner scanString: ch intoString: NULL] == YES)
+ *  { 
+ *    total++;
+ *    [scanner scanUpToString: ch intoString: NULL]; 
+ *  } 
+ * NSLog(@"total %d", total); 
+ * </example>
  */
 - (BOOL) scanUpToString: (NSString *)string 
 	     intoString: (NSString **)value
@@ -884,7 +977,7 @@ typedef struct {
   return YES;
 }
 
-/*
+/**
  * Returns the string being scanned.
  */
 - (NSString *) string
@@ -892,18 +985,21 @@ typedef struct {
   return _string;
 }
 
-/*
- * Returns the character index at which the scanner
- * will begin the next scanning operation.
+/**
+ * Returns the current position that the scanner has reached in
+ * scanning the string.  This is the position at which the next scan
+ * operation will begin.
  */
 - (unsigned) scanLocation
 {
   return _scanLocation;
 }
 
-/*
- * Set the character location at which the scanner
- * will begin the next scanning operation to anIndex.
+/**
+ * This method sets the location in the scanned string at which the
+ * next scan operation begins.
+ * Raises an NSRangeException if index is beyond the end of the
+ * scanned string.
  */
 - (void) setScanLocation: (unsigned int)anIndex
 {
@@ -914,37 +1010,53 @@ typedef struct {
 		format: @"Attempt to set scan location beyond end of string"];
 }
 
-/*
- * Returns YES if the scanner makes a distinction
- * between upper and lower case characters.
+/**
+ * If the scanner is set to be case-sensitive in its scanning of
+ * the string (other than characters to be skipped), this method
+ * returns YES, otherwise it returns NO.
+ * <br/>
+ * The default is for a scanner to <em>not</em> be case sensitive.
  */
 - (BOOL) caseSensitive
 {
   return _caseSensitive;
 }
 
-/*
- * If flag is YES the scanner will consider upper and lower case
- * to be the same during scanning.  If flag is NO the scanner will
- * not make a distinction between upper and lower case characters.
+/**
+ * Sets the case sensitivity of the scanner.
+ * <br/>
+ * Case sensitivity governs matching of characters being scanned,
+ * but does not effect the characters in the set to be skipped.
+ * <br/>
+ * The default is for a scanner to <em>not</em> be case sensitive.
  */
 - (void) setCaseSensitive: (BOOL)flag
 {
   _caseSensitive = flag;
 }
 
-/*
- * Return a character set object containing the characters the scanner
- * will ignore when searching for the next element to be scanned.
+/**
+ * Returns a set of characters containing those characters that the
+ * scanner ignores when starting any scan operation.  Once a character
+ * not in this set has been encountered during an operation, skipping
+ * is finished, and any further characters from this set that are
+ * found are scanned normally.
+ * <br/>
+ * The default for this is the whitespaceAndNewlineCharacterSet.
  */
 - (NSCharacterSet *) charactersToBeSkipped
 {
   return _charactersToBeSkipped;
 }
 
-/*
- * Set the characters to be ignored when the scanner
- * searches for the next element to be scanned.
+/**
+ * Sets the set of characters that the scanner will skip over at the
+ * start of each scanning operation to be <em>skipSet</em>.
+ * Skipping is performed by literal character matchins - the case
+ * sensitivity of the scanner does not effect it.
+ * If this is set to nil, no skipping is done.
+ * <br/>
+ * The default for this is the whitespaceAndNewlineCharacterSet.
  */
 - (void) setCharactersToBeSkipped: (NSCharacterSet *)aSet
 {
@@ -953,18 +1065,20 @@ typedef struct {
     [_charactersToBeSkipped methodForSelector: memSel];
 }
 
-/*
- * Returns a dictionary object containing the locale
- * information used by the scanner.
+/**
+ * Returns the locale set for the scanner, or nil if no locale has
+ * been set.  A scanner uses it's locale to alter the way it handles
+ * scanning - it uses the NSDecimalSeparator value for scanning
+ * numbers.
  */
 - (NSDictionary *) locale
 {
   return _locale;
 }
 
-/*
- * Set the dictionary containing the locale
- * information used by the scanner to localeDictionary.
+/**
+ * This method sets the locale used by the scanner to <em>aLocale</em>.
+ * The locale may be set to nil.
  */
 - (void) setLocale: (NSDictionary *)localeDictionary
 {
@@ -1004,3 +1118,181 @@ typedef struct {
 }
 
 @end
+
+/*
+ * Some utilities
+ */
+BOOL
+GSScanInt(unichar *buf, unsigned length, int *result)
+{
+  unsigned int num = 0;
+  const unsigned int limit = UINT_MAX / 10;
+  BOOL negative = NO;
+  BOOL overflow = NO;
+  BOOL got_digits = NO;
+  unsigned int pos = 0;
+
+  /* Check for sign */
+  if (pos < length)
+    {
+      switch (buf[pos])
+	{
+	  case '+': 
+	    pos++;
+	    break;
+	  case '-': 
+	    negative = YES;
+	    pos++;
+	    break;
+	}
+    }
+
+  /* Process digits */
+  while (pos < length)
+    {
+      unichar digit = buf[pos];
+
+      if ((digit < '0') || (digit > '9'))
+	break;
+      if (!overflow)
+	{
+	  if (num >= limit)
+	    overflow = YES;
+	  else
+	    num = num * 10 + (digit - '0');
+	}
+      pos++;
+      got_digits = YES;
+    }
+
+  /* Save result */
+  if (!got_digits)
+    {
+      return NO;
+    }
+  if (result)
+    {
+      if (overflow
+	|| (num > (negative ? (unsigned int)INT_MIN : (unsigned int)INT_MAX)))
+	*result = negative ? INT_MIN: INT_MAX;
+      else if (negative)
+	*result = -num;
+      else
+	*result = num;
+    }
+  return YES;
+}
+
+/**
+ * Scan in a double value in the standard locale ('.' as decimal point).<br />
+ * Return YES on success, NO on failure.<br />
+ * The value pointed to by result is unmodified on failure.<br />
+ * No value is returned in result if it is a null pointer.
+ */
+BOOL
+GSScanDouble(unichar *buf, unsigned length, double *result)
+{
+  unichar	c = 0;
+  double	num = 0.0;
+  long int	exponent = 0;
+  BOOL		negative = NO;
+  BOOL		got_dot = NO;
+  BOOL		got_digit = NO;
+  unsigned	pos = 0;
+
+  /* Skip whitespace */
+  while (pos < length && isspace((int)buf[pos]))
+    {
+      pos++;
+    }
+
+  /* Check for sign */
+  if (pos < length)
+    {
+      switch (buf[pos])
+	{
+	  case '+': 
+	    pos++;
+	    break;
+	  case '-': 
+	    negative = YES;
+	    pos++;
+	    break;
+	}
+    }
+
+    /* Process number */
+  while (pos < length)
+    {
+      c = buf[pos];
+      if ((c >= '0') && (c <= '9'))
+	{
+	  /* Ensure that the number being accumulated will not overflow. */
+	  if (num >= (DBL_MAX / 10.000000001))
+	    {
+	      ++exponent;
+	    }
+	  else
+	    {
+	      num = (num * 10.0) + (c - '0');
+	      got_digit = YES;
+	    }
+            /* Keep track of the number of digits after the decimal point.
+	       If we just divided by 10 here, we would lose precision. */
+	  if (got_dot)
+	    {
+	      --exponent;
+	    }
+        }
+      else if (!got_dot && (c == '.'))
+	{
+	  /* Note that we have found the decimal point. */
+	  got_dot = YES;
+        }
+      else
+	{
+	  /* Any other character terminates the number. */
+	  break;
+        }
+      pos++;
+    }
+  if (!got_digit)
+    {
+      return NO;
+    }
+
+  /* Check for trailing exponent */
+  if ((pos < length) && ((c == 'e') || (c == 'E')))
+    {
+      int expval;
+
+      pos++;
+      if (GSScanInt(&buf[pos], length - pos, &expval) == YES)
+	{
+      /* Check for exponent overflow */
+	if (num)
+	  {
+	    if ((exponent > 0) && (expval > (LONG_MAX - exponent)))
+	      exponent = LONG_MAX;
+	    else if ((exponent < 0) && (expval < (LONG_MIN - exponent)))
+	      exponent = LONG_MIN;
+	    else
+	      exponent += expval;
+	  }
+	}
+      else
+	{
+	  return NO;
+	}
+    }
+  if (result)
+    {
+      if (num && exponent)
+	num *= pow(10.0, (double) exponent);
+      if (negative)
+	*result = -num;
+      else
+	*result = num;
+    }
+  return YES;
+}
