@@ -21,22 +21,23 @@
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA.
 
    <title>NSSerializer class reference</title>
-   $Date: 2002/03/13 09:58:43 $ $Revision: 1.44 $
+   $Date: 2003/11/05 13:52:18 $ $Revision: 1.56 $
    */
 
-#include <config.h>
-#include <base/preface.h>
-#include <Foundation/NSData.h>
-#include <Foundation/NSDictionary.h>
-#include <Foundation/NSArray.h>
-#include <Foundation/NSString.h>
-#include <Foundation/NSException.h>
-#include <Foundation/NSProxy.h>
-#include <Foundation/NSLock.h>
-#include <Foundation/NSSet.h>
-#include <Foundation/NSThread.h>
-#include <Foundation/NSNotificationQueue.h>
-#include <Foundation/NSObjCRuntime.h>
+#include "config.h"
+#include "GNUstepBase/preface.h"
+#include "Foundation/NSData.h"
+#include "Foundation/NSDictionary.h"
+#include "Foundation/NSArray.h"
+#include "Foundation/NSString.h"
+#include "Foundation/NSException.h"
+#include "Foundation/NSProxy.h"
+#include "Foundation/NSLock.h"
+#include "Foundation/NSSet.h"
+#include "Foundation/NSThread.h"
+#include "Foundation/NSNotificationQueue.h"
+#include "Foundation/NSObjCRuntime.h"
+#include "Foundation/NSValue.h"
 
 #include "GSPrivate.h"
 
@@ -44,9 +45,11 @@
 @class	GSMutableDictionary;
 @class	NSDataMalloc;
 @class	GSInlineArray;
-@class  GSMutableArray;
+@class	GSMutableArray;
 @class	GSCString;
+@class	GSCBufferString;
 @class	GSUnicodeString;
+@class	GSUnicodeBufferString;
 @class	GSMutableString;
 
 /*
@@ -60,7 +63,7 @@
 #define	GSI_MAP_EQUAL(M, X,Y)	[(X).obj isEqualToString: (Y).obj]
 #define	GSI_MAP_NOCLEAN	1
 
-#include <base/GSIMap.h>
+#include "GNUstepBase/GSIMap.h"
 
 /*
  *	Setup for inline operation of string arrays.
@@ -69,7 +72,7 @@
 #define	GSI_ARRAY_RELEASE(A, X)	
 #define	GSI_ARRAY_TYPES	GSUNION_OBJ
 
-#include <base/GSIArray.h>
+#include "GNUstepBase/GSIArray.h"
 
 /*
  *	Define constants for data types and variables to hold them.
@@ -82,6 +85,8 @@
 #define ST_DICT		5
 #define ST_MDICT	6
 #define ST_DATA		7
+#define ST_DATE		8
+#define ST_NUMBER	9
 
 static char	st_xref = (char)ST_XREF;
 static char	st_cstring = (char)ST_CSTRING;
@@ -91,20 +96,29 @@ static char	st_marray = (char)ST_MARRAY;
 static char	st_dict = (char)ST_DICT;
 static char	st_mdict = (char)ST_MDICT;
 static char	st_data = (char)ST_DATA;
+static char	st_date = (char)ST_DATE;
+static char	st_number = (char)ST_NUMBER;
 
 
 
 /*
- *	Variables to cache class information.
+ * Variables to cache class information. These are used to check how
+ * an instance should be serialized.
  */
 static Class	ArrayClass = 0;
 static Class	MutableArrayClass = 0;
 static Class	DataClass = 0;
+static Class	DateClass = 0;
 static Class	DictionaryClass = 0;
 static Class	MutableDictionaryClass = 0;
 static Class	CStringClass = 0;
 static Class	MStringClass = 0;
 static Class	StringClass = 0;
+static Class	NumberClass = 0;
+
+#if NEED_WORD_ALIGNMENT
+static unsigned	gsu32Align;
+#endif
 
 typedef struct {
   @defs(GSString)
@@ -112,11 +126,11 @@ typedef struct {
 
 typedef struct {
   NSMutableData	*data;
-  void		(*appImp)();		// Append to data.
-  void*		(*datImp)();		// Bytes pointer.
-  unsigned int	(*lenImp)();		// Length of data.
-  void		(*serImp)();		// Serialize integer.
-  void		(*setImp)();		// Set length of data.
+  void		(*appImp)(NSData*,SEL,const void*,unsigned);
+  void*		(*datImp)(NSMutableData*,SEL);		// Bytes pointer.
+  unsigned int	(*lenImp)(NSData*,SEL);			// Length of data.
+  void		(*serImp)(NSMutableData*,SEL,int);	// Serialize integer.
+  void		(*setImp)(NSMutableData*,SEL,unsigned);	// Set length of data.
   unsigned	count;			// String counter.
   GSIMapTable_t	map;			// For uniquing.
   BOOL		shouldUnique;		// Do we do uniquing?
@@ -135,11 +149,11 @@ initSerializerInfo(_NSSerializerInfo* info, NSMutableData *d, BOOL u)
 
   c = GSObjCClass(d);
   info->data = d; 
-  info->appImp = (void (*)())get_imp(c, appSel);
-  info->datImp = (void* (*)())get_imp(c, datSel);
-  info->lenImp = (unsigned int (*)())get_imp(c, lenSel);
-  info->serImp = (void (*)())get_imp(c, serSel);
-  info->setImp = (void (*)())get_imp(c, setSel);
+  info->appImp = (void (*)(NSData*,SEL,const void*,unsigned))get_imp(c, appSel);
+  info->datImp = (void* (*)(NSMutableData*,SEL))get_imp(c, datSel);
+  info->lenImp = (unsigned int (*)(NSData*,SEL))get_imp(c, lenSel);
+  info->serImp = (void (*)(NSMutableData*,SEL,int))get_imp(c, serSel);
+  info->setImp = (void (*)(NSMutableData*,SEL,unsigned))get_imp(c, setSel);
   info->shouldUnique = u;
   (*info->appImp)(d, appSel, &info->shouldUnique, 1);
   if (u)
@@ -168,8 +182,15 @@ serializeToInfo(id object, _NSSerializerInfo* info)
 				[object description]];
     }
   c = GSObjCClass(object);
-  if (GSObjCIsKindOf(c, CStringClass)
-    || (c == MStringClass && ((ivars)object)->_flags.wide == 0))
+
+  if ((GSObjCIsKindOf(c, CStringClass)
+       || (c == MStringClass && ((ivars)object)->_flags.wide == 0))
+      /*
+      We can only save it as a c-string if it only contains ASCII characters.
+      Other characters might be decoded incorrectly when deserialized since
+      the c-string encoding might be different then.
+      */
+      && [object canBeConvertedToEncoding: NSASCIIStringEncoding])
     {
       GSIMapNode	node;
 
@@ -222,7 +243,7 @@ serializeToInfo(id object, _NSSerializerInfo* info)
 	   * word boundary, so we work with an aligned buffer
 	   * and use memcmpy()
 	   */
- 	  if ((dlen % __alignof__(gsu32)) != 0)
+ 	  if ((dlen % gsu32Align) != 0)
 	    {
 	      unichar buffer[slen];
 	      [object getCharacters: buffer];
@@ -296,6 +317,24 @@ serializeToInfo(id object, _NSSerializerInfo* info)
       (*info->serImp)(info->data, serSel, [object length]);
       (*info->appImp)(info->data, appSel, [object bytes], [object length]);
     }
+  else if (GSObjCIsKindOf(c, DateClass))
+    {
+      NSTimeInterval	ti = [object timeIntervalSinceReferenceDate];
+
+      (*info->appImp)(info->data, appSel, &st_date, 1);
+      [info->data serializeDataAt: &ti
+		       ofObjCType: @encode(NSTimeInterval)
+			  context: nil];
+    }
+  else if (GSObjCIsKindOf(c, NumberClass))
+    {
+      double	d = [object doubleValue];
+
+      (*info->appImp)(info->data, appSel, &st_number, 1);
+      [info->data serializeDataAt: &d
+		       ofObjCType: @encode(double)
+			  context: nil];
+    }
   else
     {
       [NSException raise: NSInvalidArgumentException
@@ -314,6 +353,9 @@ static BOOL	shouldBeCompact = NO;
 {
   if (self == [NSSerializer class])
     {
+#if NEED_WORD_ALIGNMENT
+      gsu32Align = objc_alignof_type(@encode(gsu32));
+#endif
       appSel = @selector(appendBytes:length:);
       datSel = @selector(mutableBytes);
       lenSel = @selector(length);
@@ -322,6 +364,8 @@ static BOOL	shouldBeCompact = NO;
       ArrayClass = [NSArray class];
       MutableArrayClass = [NSMutableArray class];
       DataClass = [NSData class];
+      DateClass = [NSDate class];
+      NumberClass = [NSNumber class];
       DictionaryClass = [NSDictionary class];
       MutableDictionaryClass = [NSMutableDictionary class];
       StringClass = [NSString class];
@@ -330,6 +374,10 @@ static BOOL	shouldBeCompact = NO;
     }
 }
 
+/**
+ * Converts the supplied propertyList into a serialized representation
+ * in the returned NSData object.
+ */
 + (NSData*) serializePropertyList: (id)propertyList
 {
   _NSSerializerInfo	info;
@@ -343,6 +391,10 @@ static BOOL	shouldBeCompact = NO;
   return info.data;
 }
 
+/**
+ * Converts the supplied propertyList into a serialized representation
+ * stored in d.
+ */
 + (void) serializePropertyList: (id)propertyList
 		      intoData: (NSMutableData*)d
 {
@@ -379,14 +431,14 @@ static BOOL	shouldBeCompact = NO;
 
 
 
-/*
- *	Variables to cache class information.
- */
 static BOOL	uniquing = NO;	/* Make incoming strings unique	*/
-static Class	IACls = 0;	/* Immutable Array	*/
+
+/*
+ * Variables to cache class information. These are used to create instances
+ * when deserializing.
+ */
 static Class	MACls = 0;	/* Mutable Array	*/
 static Class	DCls = 0;	/* Data			*/
-static Class	IDCls = 0;	/* Immutable Dictionary	*/
 static Class	MDCls = 0;	/* Mutable Dictionary	*/
 static Class	USCls = 0;	/* Unicode String	*/
 static Class	CSCls = 0;	/* C String 		*/
@@ -406,29 +458,57 @@ static SEL deiSel;
 static SEL csInitSel;
 static SEL usInitSel;
 static SEL dInitSel;
-static SEL iaInitSel;
 static SEL maInitSel;
-static SEL idInitSel;
 static SEL mdInitSel;
+static SEL maAddSel;
+static SEL mdSetSel;
 static IMP csInitImp;
 static IMP usInitImp;
 static IMP dInitImp;
-static IMP iaInitImp;
 static IMP maInitImp;
-static IMP idInitImp;
 static IMP mdInitImp;
+static IMP maAddImp;
+static IMP mdSetImp;
 
-static void
+static BOOL
 initDeserializerInfo(_NSDeserializerInfo* info, NSData *d, unsigned *c, BOOL m)
 {
+  unsigned char	u;
+
   info->data = d;
   info->cursor = c;
   info->mutable = m;
   info->debImp = (void (*)())[d methodForSelector: debSel];
   info->deiImp = (unsigned int (*)())[d methodForSelector: deiSel];
-  (*info->debImp)(d, debSel, &info->didUnique, 1, c);
+  (*info->debImp)(d, debSel, &u, 1, c);
+  if (u == 0 || u == 1)
+    {
+      info->didUnique = u;		// Old (current) format
+    }
+  else
+    {
+      if (u == 'G')
+	{
+	  const unsigned char	*b = [d bytes];
+	  unsigned int		l = [d length];
+
+	  if (*c + 11 < l && memcmp(&b[*c-1], "GNUstepSer", 10) == 0)
+	    {
+	      *c += 9; 
+	      (*info->debImp)(d, debSel, &u, 1, c);
+	      NSLog(@"Serialised data version %d not supported ..."
+		@" try another version of GNUstep");
+	      return NO;
+	    }
+	}
+      NSLog(@"Bad serialised data");
+      return NO;
+    }
   if (info->didUnique)
-    GSIArrayInitWithZoneAndCapacity(&info->array, NSDefaultMallocZone(), 16);
+    {
+      GSIArrayInitWithZoneAndCapacity(&info->array, NSDefaultMallocZone(), 16);
+    }
+  return YES;
 }
 
 static void
@@ -445,12 +525,12 @@ deserializeFromInfo(_NSDeserializerInfo* info)
   unsigned int	size;
 
   (*info->debImp)(info->data, debSel, &code, 1, info->cursor);
-  size = (*info->deiImp)(info->data, deiSel, info->cursor);
 
   switch (code)
     {
       case ST_XREF:
 	{
+	  size = (*info->deiImp)(info->data, deiSel, info->cursor);
 	  return RETAIN(GSIArrayItemAtIndex(&info->array, size).obj);
 	}
 
@@ -459,6 +539,7 @@ deserializeFromInfo(_NSDeserializerInfo* info)
 	  GSCString	*s;
 	  char		*b;
 	
+	  size = (*info->deiImp)(info->data, deiSel, info->cursor);
 	  b = NSZoneMalloc(NSDefaultMallocZone(), size);
 	  (*info->debImp)(info->data, debSel, b, size, info->cursor);
 	  s = (GSCString*)NSAllocateObject(CSCls, 0, NSDefaultMallocZone());
@@ -485,6 +566,7 @@ deserializeFromInfo(_NSDeserializerInfo* info)
 	  unichar	*b;
 	  unsigned	i;
 	
+	  size = (*info->deiImp)(info->data, deiSel, info->cursor);
 	  b = NSZoneMalloc(NSDefaultMallocZone(), size*sizeof(unichar));
 	  (*info->debImp)(info->data, debSel, b, size*sizeof(unichar),
 	    info->cursor);
@@ -534,99 +616,80 @@ deserializeFromInfo(_NSDeserializerInfo* info)
 
       case ST_ARRAY:
       case ST_MARRAY:
+	size = (*info->deiImp)(info->data, deiSel, info->cursor);
 	{
-	  id		objects[size];
 	  id		a;
-	  unsigned	i;
 
-	  for (i = 0; i < size; i++)
+	  a = NSAllocateObject(MACls, 0, NSDefaultMallocZone());
+	  a = (*maInitImp)(a, maInitSel, size);
+	  if (a != nil)
 	    {
-	      objects[i] = deserializeFromInfo(info);
-	      if (objects[i] == nil)
+	      unsigned	i;
+
+	      for (i = 0; i < size; i++)
 		{
-#if	!GS_WITH_GC
-		  while (i > 0)
+		  id	o = deserializeFromInfo(info);
+
+		  if (o == nil)
 		    {
-		      [objects[--i] release];
+		      RELEASE(a);
+		      return nil;
 		    }
-#endif
-		  objc_free(objects);
-		  return nil;
+		  (*maAddImp)(a, maAddSel, o);
+		  RELEASE(o);
+		}
+	      if (code != ST_MARRAY && info->mutable == NO)
+		{
+		  [a makeImmutableCopyOnFail: NO];
 		}
 	    }
-	  if (code == ST_MARRAY || info->mutable)
-	    {
-	      a = NSAllocateObject(MACls, 0, NSDefaultMallocZone());
-	      a = (*maInitImp)(a, maInitSel, objects, size);
-	    }
-	  else
-	    {
-	      a = NSAllocateObject(IACls, sizeof(id)*size,
-		NSDefaultMallocZone());
-	      a = (*iaInitImp)(a, iaInitSel, objects, size);
-	    }
-#if	!GS_WITH_GC
-	  for (i = 0; i < size; i++)
-	    {
-	      [objects[i] release];
-	    }
-#endif
 	  return a;
 	}
 
       case ST_DICT:
       case ST_MDICT:
+	size = (*info->deiImp)(info->data, deiSel, info->cursor);
 	{
-	  id	keys[size];
-	  id	objects[size];
-	  id	d;
-	  int	i;
+	  id		d;
 
-	  for (i = 0; i < size; i++)
+	  d = NSAllocateObject(MDCls, 0, NSDefaultMallocZone());
+	  d = (*mdInitImp)(d, mdInitSel, size);
+	  if (d != nil)
 	    {
-	      keys[i] = deserializeFromInfo(info);
-	      if (keys[i] == nil)
+	      unsigned int	i;
+
+	      for (i = 0; i < size; i++)
 		{
-#if	!GS_WITH_GC
-		  while (i > 0)
+		  id	k = deserializeFromInfo(info);
+
+		  if (k == nil)
 		    {
-		      [keys[--i] release];
-		      [objects[i] release];
+		      RELEASE(d);
+		      return nil;
+		    }
+		  else
+		    {
+		      id	o = deserializeFromInfo(info);
+
+		      if (o == nil)
+			{
+			  RELEASE(k);
+			  RELEASE(d);
+			  return nil;
+			}
+		      else
+			{
+			  (*mdSetImp)(d, mdSetSel, o, k);
+			  RELEASE(k);
+			  RELEASE(o);
+			}
 		    }
-#endif
-		  return nil;
 		}
-	      objects[i] = deserializeFromInfo(info);
-	      if (objects[i] == nil)
+	      if (code != ST_MDICT && info->mutable == NO)
 		{
-#if	!GS_WITH_GC
-		  [keys[i] release];
-		  while (i > 0)
-		    {
-		      [keys[--i] release];
-		      [objects[i] release];
-		    }
-#endif
-		  return nil;
+		  [d makeImmutableCopyOnFail: NO];
 		}
 	    }
-	  if (code == ST_MDICT || info->mutable)
-	    {
-	      d = NSAllocateObject(MDCls, 0, NSDefaultMallocZone());
-	      d = (*mdInitImp)(d, mdInitSel, objects, keys, size);
-	    }
-	  else
-	    {
-	      d = NSAllocateObject(IDCls, 0, NSDefaultMallocZone());
-	      d = (*idInitImp)(d, idInitSel, objects, keys, size);
-	    }
-#if	!GS_WITH_GC
-	  for (i = 0; i < size; i++)
-	    {
-	      [keys[i] release];
-	      [objects[i] release];
-	    }
-#endif
 	  return d;
 	}
 
@@ -634,6 +697,7 @@ deserializeFromInfo(_NSDeserializerInfo* info)
 	{
 	  NSData	*d;
 
+	  size = (*info->deiImp)(info->data, deiSel, info->cursor);
 	  d = (NSData*)NSAllocateObject(DCls, 0, NSDefaultMallocZone());
 	  if (size > 0)
 	    {
@@ -649,6 +713,28 @@ deserializeFromInfo(_NSDeserializerInfo* info)
 	  return d;
 	}
 
+      case ST_DATE:
+	{
+	  NSTimeInterval	ti;
+
+	  [info->data deserializeDataAt: &ti
+			     ofObjCType: @encode(NSTimeInterval)
+			       atCursor: info->cursor
+				context: nil];
+	  return [[NSDate alloc] initWithTimeIntervalSinceReferenceDate: ti];
+	}
+
+      case ST_NUMBER:
+	{
+	  double	d;
+
+	  [info->data deserializeDataAt: &d
+			     ofObjCType: @encode(double)
+			       atCursor: info->cursor
+				context: nil];
+	  return [[NSNumber alloc] initWithDouble: d];
+	}
+
       default:
 	return nil;
     }
@@ -674,8 +760,15 @@ deserializeFromInfo(_NSDeserializerInfo* info)
   _NSDeserializerProxy	*proxy;
 
   proxy = (_NSDeserializerProxy*)NSAllocateObject(self,0,NSDefaultMallocZone());
-  initDeserializerInfo(&proxy->info, RETAIN(d), c, m);
-  return AUTORELEASE(proxy);
+  if (initDeserializerInfo(&proxy->info, RETAIN(d), c, m) == YES)
+    {
+      return AUTORELEASE(proxy);
+    }
+  else
+    {
+      DESTROY(proxy);
+      return nil;
+    }
 }
 
 - (void) dealloc
@@ -730,27 +823,35 @@ deserializeFromInfo(_NSDeserializerInfo* info)
       csInitSel = @selector(initWithCStringNoCopy:length:freeWhenDone:);
       usInitSel = @selector(initWithCharactersNoCopy:length:freeWhenDone:);
       dInitSel = @selector(initWithBytesNoCopy:length:);
-      iaInitSel = @selector(initWithObjects:count:);
-      maInitSel = @selector(initWithObjects:count:);
-      idInitSel = @selector(initWithObjects:forKeys:count:);
-      mdInitSel = @selector(initWithObjects:forKeys:count:);
-      IACls = [GSInlineArray class];
+      maInitSel = @selector(initWithCapacity:);
+      mdInitSel = @selector(initWithCapacity:);
+      maAddSel = @selector(addObject:);
+      mdSetSel = @selector(setObject:forKey:);
       MACls = [GSMutableArray class];
       DCls = [NSDataMalloc class];
-      IDCls = [GSDictionary class];
       MDCls = [GSMutableDictionary class];
-      USCls = [GSUnicodeString class];
-      CSCls = [GSCString class];
+      USCls = [GSUnicodeBufferString class];
+      CSCls = [GSCBufferString class];
       csInitImp = [CSCls instanceMethodForSelector: csInitSel];
       usInitImp = [USCls instanceMethodForSelector: usInitSel];
       dInitImp = [DCls instanceMethodForSelector: dInitSel];
-      iaInitImp = [IACls instanceMethodForSelector: iaInitSel];
       maInitImp = [MACls instanceMethodForSelector: maInitSel];
-      idInitImp = [IDCls instanceMethodForSelector: idInitSel];
       mdInitImp = [MDCls instanceMethodForSelector: mdInitSel];
+      maAddImp = [MACls instanceMethodForSelector: maAddSel];
+      mdSetImp = [MDCls instanceMethodForSelector: mdSetSel];
     }
 }
 
+/**
+ * <p>Deserializes the property list stored in data at the offset specified
+ * by the value pointed to by cursor.  Upon completion the value in cursor
+ * is updated to refer to a position immediately after the end of the
+ * deserialized sequence.
+ * </p>
+ * <p>The flag is used to specify whether container objects in the
+ * deserialized list should be instances of mutable or immutable classes.
+ * </p>
+ */
 + (id) deserializePropertyListFromData: (NSData*)data
                               atCursor: (unsigned int*)cursor
                      mutableContainers: (BOOL)flag
@@ -763,12 +864,26 @@ deserializeFromInfo(_NSDeserializerInfo* info)
       return nil;
     }
   NSAssert(cursor != 0, NSInvalidArgumentException);
-  initDeserializerInfo(&info, data, cursor, flag);
-  o = deserializeFromInfo(&info);
-  endDeserializerInfo(&info);
-  return AUTORELEASE(o);
+  if (initDeserializerInfo(&info, data, cursor, flag) == YES)
+    {
+      o = deserializeFromInfo(&info);
+      endDeserializerInfo(&info);
+      return AUTORELEASE(o);
+    }
+  else
+    {
+      return nil;
+    }
 }
 
+/**
+ * <p>Deserializes the property list stored in data to produce a single
+ * property list object (dictionary, array, string, data, number or date).
+ * </p>
+ * <p>The flag is used to specify whether container objects in the
+ * deserialized list should be instances of mutable or immutable classes.
+ * </p>
+ */
 + (id) deserializePropertyListFromData: (NSData*)data
                      mutableContainers: (BOOL)flag
 {
@@ -780,12 +895,33 @@ deserializeFromInfo(_NSDeserializerInfo* info)
     {
       return nil;
     }
-  initDeserializerInfo(&info, data, &cursor, flag);
-  o = deserializeFromInfo(&info);
-  endDeserializerInfo(&info);
-  return AUTORELEASE(o);
+  if (initDeserializerInfo(&info, data, &cursor, flag) == YES)
+    {
+      o = deserializeFromInfo(&info);
+      endDeserializerInfo(&info);
+      return AUTORELEASE(o);
+    }
+  else
+    {
+      return nil;
+    }
 }
 
+/**
+ * <p>Deserializes the property list stored in data at the offset specified
+ * by the value pointed to by cursor.  Upon completion the value in cursor
+ * is updated to refer to a position immediately after the end of the
+ * deserialized sequence.
+ * </p>
+ * <p>The flag is used to specify whether container objects in the
+ * deserialized list should be instances of mutable or immutable classes.
+ * </p>
+ * <p>The length is used to determine  whether lazy deserialization is done,
+ * if the data is longer than this value, a proxy is returned rather than
+ * the actual property list, and the real deserialization can be done
+ * later.
+ * </p>
+ */
 + (id) deserializePropertyListLazilyFromData: (NSData*)data
                                     atCursor: (unsigned*)cursor
                                       length: (unsigned)length
@@ -801,10 +937,16 @@ deserializeFromInfo(_NSDeserializerInfo* info)
       _NSDeserializerInfo   info;
       id    o;
 
-      initDeserializerInfo(&info, data, cursor, flag);
-      o = deserializeFromInfo(&info);
-      endDeserializerInfo(&info);
-      return AUTORELEASE(o);
+      if (initDeserializerInfo(&info, data, cursor, flag) == YES)
+	{
+	  o = deserializeFromInfo(&info);
+	  endDeserializerInfo(&info);
+	  return AUTORELEASE(o);
+	}
+      else
+	{
+	  return nil;
+	}
     }
   else
     {
@@ -816,6 +958,12 @@ deserializeFromInfo(_NSDeserializerInfo* info)
 @end
 
 @implementation NSDeserializer (GNUstep)
+/**
+ * This method turns on/off uniquing of strings as they are
+ * deserialized from data objects.  The uniquing mechanism
+ * employs the GNUstep-specific functions documented with
+ * the NSCountedSet class.
+ */
 + (void) uniquing: (BOOL)flag
 {
   if (flag == YES)
