@@ -21,31 +21,33 @@
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA.
 
    <title>NSTask class reference</title>
-   $Date: 2002/02/24 06:14:31 $ $Revision: 1.47 $
+   $Date: 2003/11/23 06:23:09 $ $Revision: 1.58 $
    */
 
-#include <config.h>
-#include <base/preface.h>
-#include <Foundation/NSObject.h>
-#include <Foundation/NSBundle.h>
-#include <Foundation/NSData.h>
-#include <Foundation/NSDate.h>
-#include <Foundation/NSString.h>
-#include <Foundation/NSException.h>
-#include <Foundation/NSFileHandle.h>
-#include <Foundation/NSFileManager.h>
-#include <Foundation/NSMapTable.h>
-#include <Foundation/NSProcessInfo.h>
-#include <Foundation/NSRunLoop.h>
-#include <Foundation/NSNotification.h>
-#include <Foundation/NSNotificationQueue.h>
-#include <Foundation/NSTask.h>
-#include <Foundation/NSTimer.h>
-#include <Foundation/NSLock.h>
-#include <Foundation/NSDebug.h>
+#include "config.h"
+#include "GNUstepBase/preface.h"
+#include "Foundation/NSObject.h"
+#include "Foundation/NSBundle.h"
+#include "Foundation/NSCharacterSet.h"
+#include "Foundation/NSData.h"
+#include "Foundation/NSDate.h"
+#include "Foundation/NSEnumerator.h"
+#include "Foundation/NSString.h"
+#include "Foundation/NSException.h"
+#include "Foundation/NSFileHandle.h"
+#include "Foundation/NSFileManager.h"
+#include "Foundation/NSMapTable.h"
+#include "Foundation/NSProcessInfo.h"
+#include "Foundation/NSRunLoop.h"
+#include "Foundation/NSNotification.h"
+#include "Foundation/NSNotificationQueue.h"
+#include "Foundation/NSTask.h"
+#include "Foundation/NSTimer.h"
+#include "Foundation/NSLock.h"
+#include "Foundation/NSDebug.h"
 
 #include <string.h>
-#if HAVE_UNISTD_H
+#ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif
 #include <sys/types.h>
@@ -53,29 +55,29 @@
 #include <fcntl.h>
 #endif
 
-#if HAVE_WINDOWS_H
+#ifdef HAVE_WINDOWS_H
 #  include <windows.h>
 #endif
 
-#if	HAVE_SYS_SIGNAL_H
+#ifdef	HAVE_SYS_SIGNAL_H
 #include <sys/signal.h>
 #endif
-#if	HAVE_SIGNAL_H
+#ifdef	HAVE_SIGNAL_H
 #include <signal.h>
 #endif
-#if	HAVE_SYS_FILE_H
+#ifdef	HAVE_SYS_FILE_H
 #include <sys/file.h>
 #endif
-#if	HAVE_SYS_FCNTL_H
+#ifdef	HAVE_SYS_FCNTL_H
 #include <sys/fcntl.h>
 #endif
-#if	HAVE_SYS_IOCTL_H
+#ifdef	HAVE_SYS_IOCTL_H
 #include <sys/ioctl.h>
 #endif
-#if	HAVE_SYS_WAIT_H
+#ifdef	HAVE_SYS_WAIT_H
 #include <sys/wait.h>
 #endif
-#if	HAVE_SYS_PARAM_H
+#ifdef	HAVE_SYS_PARAM_H
 #include <sys/param.h>
 #endif
 
@@ -83,7 +85,7 @@
  *	If we are on a streams based system, we need to include stropts.h
  *	for definitions needed to set up slave pseudo-terminal stream.
  */
-#if	HAVE_SYS_STROPTS_H
+#ifdef	HAVE_SYS_STROPTS_H
 #include <sys/stropts.h>
 #endif
 
@@ -115,7 +117,9 @@ static void handleSignal(int sig)
 #ifdef __MINGW__
 @interface NSConcreteWindowsTask : NSTask
 {
-  PROCESS_INFORMATION proc_info;
+@public
+  PROCESS_INFORMATION	procInfo;
+  HANDLE		wThread;
 }
 @end
 #define NSConcreteTask NSConcreteWindowsTask
@@ -137,7 +141,7 @@ pty_master(char* name, int len)
    *	If we have grantpt(), assume we are using sysv-style pseudo-terminals,
    *	otherwise assume bsd style.
    */
-#if	HAVE_GRANTPT
+#ifdef	HAVE_GRANTPT
   master = open("/dev/ptmx", O_RDWR);
   if (master >= 0)
     {
@@ -192,8 +196,8 @@ pty_slave(const char* name)
   int	slave;
 
   slave = open(name, O_RDWR);
-#if	HAVE_SYS_STROPTS_H
-#if	HAVE_PTS_STREAM_MODULES
+#ifdef	HAVE_SYS_STROPTS_H
+#ifdef	HAVE_PTS_STREAM_MODULES
   if (slave >= 0 && isastream(slave))
     {
       if (ioctl(slave, I_PUSH, "ptem") < 0)
@@ -499,13 +503,13 @@ pty_slave(const char* name)
  */
 - (void) setStandardError: (id)hdl
 {
-  NSAssert([hdl isKindOfClass: [NSFileHandle class]] ||
-	   [hdl isKindOfClass: [NSPipe class]], NSInvalidArgumentException);
   if (_hasLaunched)
     {
       [NSException raise: NSInvalidArgumentException
                   format: @"NSTask - task has been launched"];
     }
+  NSAssert(hdl != nil && ([hdl isKindOfClass: [NSFileHandle class]] ||
+    [hdl isKindOfClass: [NSPipe class]]), NSInvalidArgumentException);
   ASSIGN(_standardError, hdl);
 }
 
@@ -521,13 +525,13 @@ pty_slave(const char* name)
  */
 - (void) setStandardInput: (id)hdl
 {
-  NSAssert([hdl isKindOfClass: [NSFileHandle class]] ||
-	   [hdl isKindOfClass: [NSPipe class]], NSInvalidArgumentException);
   if (_hasLaunched)
     {
       [NSException raise: NSInvalidArgumentException
                   format: @"NSTask - task has been launched"];
     }
+  NSAssert(hdl != nil && ([hdl isKindOfClass: [NSFileHandle class]] ||
+    [hdl isKindOfClass: [NSPipe class]]), NSInvalidArgumentException);
   ASSIGN(_standardInput, hdl);
 }
 
@@ -543,13 +547,13 @@ pty_slave(const char* name)
  */
 - (void) setStandardOutput: (id)hdl
 {
-  NSAssert([hdl isKindOfClass: [NSFileHandle class]] ||
-	   [hdl isKindOfClass: [NSPipe class]], NSInvalidArgumentException);
   if (_hasLaunched)
     {
       [NSException raise: NSInvalidArgumentException
                   format: @"NSTask - task has been launched"];
     }
+  NSAssert(hdl != nil && ([hdl isKindOfClass: [NSFileHandle class]] ||
+    [hdl isKindOfClass: [NSPipe class]]), NSInvalidArgumentException);
   ASSIGN(_standardOutput, hdl);
 }
 
@@ -606,9 +610,9 @@ pty_slave(const char* name)
     }
 #ifndef __MINGW__
 #ifdef	HAVE_KILLPG
-  killpg(_taskId, SIGTERM);
+  killpg(_taskId, SIGSTOP);
 #else
-  kill(-_taskId, SIGTERM);
+  kill(-_taskId, SIGSTOP);
 #endif
 #endif
   return YES;
@@ -737,13 +741,31 @@ pty_slave(const char* name)
   full_path = [arch_path stringByAppendingPathComponent: libs];
 
   lpath = [full_path stringByAppendingPathComponent: prog];
+#ifdef	__MINGW__
+  if ([mgr isExecutableFileAtPath: lpath] == NO
+    && [mgr isExecutableFileAtPath:
+    (lpath = [lpath stringByAppendingPathExtension: @"exe"])] == NO)
+#else
   if ([mgr isExecutableFileAtPath: lpath] == NO)
+#endif
     {
       lpath = [arch_path stringByAppendingPathComponent: prog];
+#ifdef	__MINGW__
+      if ([mgr isExecutableFileAtPath: lpath] == NO
+	&& [mgr isExecutableFileAtPath:
+	(lpath = [lpath stringByAppendingPathExtension: @"exe"])] == NO)
+#else
       if ([mgr isExecutableFileAtPath: lpath] == NO)
+#endif
 	{
 	  lpath = [base_path stringByAppendingPathComponent: prog];
+#ifdef	__MINGW__
+	  if ([mgr isExecutableFileAtPath: lpath] == NO
+	    && [mgr isExecutableFileAtPath:
+	    (lpath = [lpath stringByAppendingPathExtension: @"exe"])] == NO)
+#else
 	  if ([mgr isExecutableFileAtPath: lpath] == NO)
+#endif
 	    {
 	      /*
 	       * Last resort - if the launch path was simply a program name
@@ -754,9 +776,19 @@ pty_slave(const char* name)
 		{
 		  lpath = [NSBundle _absolutePathOfExecutable: prog];
 		}
-	      if (lpath != nil && [mgr isExecutableFileAtPath: lpath] == NO)
+	      if (lpath != nil)
 		{
-		  lpath = nil;
+#ifdef	__MINGW__
+		  if ([mgr isExecutableFileAtPath: lpath] == NO
+		    && [mgr isExecutableFileAtPath:
+		    (lpath = [lpath stringByAppendingPathExtension: @"exe"])]
+		    == NO)
+#else
+		  if ([mgr isExecutableFileAtPath: lpath] == NO)
+#endif
+		    {
+		      lpath = nil;
+		    }
 		}
 	    }
 	}
@@ -877,17 +909,51 @@ pty_slave(const char* name)
 BOOL
 GSCheckTasks()
 {
-  /* FIXME: Implement */
-  return YES;
+  BOOL	found = NO;
+
+  if (hadChildSignal == YES)
+    {
+      NSArray	*a;
+      unsigned	c;
+
+      hadChildSignal = NO;
+      [tasksLock lock];
+      a = NSAllMapTableValues(activeTasks);
+      [tasksLock unlock];
+      c = [a count];
+      while (c-- > 0)
+	{
+	  NSConcreteWindowsTask	*t = [a objectAtIndex: c];
+	  DWORD			eCode;
+
+	  if (GetExitCodeProcess(t->procInfo.hProcess, &eCode) != 0)
+	    {
+	      if (eCode != STILL_ACTIVE)
+		{
+		  [t _terminatedChild: eCode];
+		  found = YES;
+		}
+	    }
+	}
+    }
+  return found;
 }
 
 - (void) gcFinalize
 {
   [super gcFinalize];
-  if (proc_info.hProcess != NULL)
-    CloseHandle(proc_info.hProcess);
-  if (proc_info.hThread != NULL)
-    CloseHandle(proc_info.hThread);
+  if (wThread != NULL)
+    {
+      CloseHandle(wThread);
+    }
+  if (procInfo.hProcess != NULL)
+    {
+      CloseHandle(procInfo.hProcess);
+    }
+  if (procInfo.hThread != NULL)
+    {
+      CloseHandle(procInfo.hThread);
+    }
 }
 
 - (void) interrupt
@@ -907,11 +973,98 @@ GSCheckTasks()
     }
 
   _hasTerminated = YES;
-  TerminateProcess(proc_info.hProcess, 10);
+  TerminateProcess(procInfo.hProcess, 10);
+}
+
+/*
+ * Wait for child process completion.
+ */
+static DWORD WINAPI _threadFunction(LPVOID t)
+{
+  DWORD	milliseconds = 60000;
+  int	taskId = [(NSTask*)t processIdentifier]; 
+
+  for (;;)
+    {
+      NSConcreteWindowsTask	*task;
+
+      [tasksLock lock];
+      task = (NSConcreteWindowsTask*)NSMapGet(activeTasks, (void*)taskId);
+      [tasksLock unlock];
+      if (t == nil)
+	{
+	  return 0;	// Task gone away.
+	}
+      switch (WaitForSingleObject(task->procInfo.hProcess, milliseconds))
+	{
+	  case WAIT_OBJECT_0:
+	    handleSignal(0);	// Signal child process state change.
+	    return 0;
+
+	  case WAIT_TIMEOUT:
+	    break;		// Timeout ... retry
+
+	  default:
+	    return 0;		// Error ... stop watching.
+	}
+    }
+}
+
+
+static NSString*
+endSlashesDoubledFromString(NSString *aString)
+{
+  int			i = [aString length] - 2;
+  NSMutableString	*returnString;
+  
+  if (![aString hasSuffix:@"\\"])
+    {
+      return aString;
+    }
+  returnString = [NSMutableString stringWithFormat: @"%@\\", aString];
+  while ([aString characterAtIndex: i] == '\\' && i >= 0)
+    {
+      [returnString appendString:@"\\"];
+      i--;
+    }
+
+  return returnString;
+}
+
+static NSString*
+quotedFromString(NSString *aString)
+{
+  NSString		*resultString;
+  NSMutableArray	*components;
+  int			i;
+  
+  /* First split on "'s */
+  components = [NSMutableArray arrayWithArray:
+    [aString componentsSeparatedByString: @"\""]];
+
+  /* Iterate over all but the last component and double slashes if needed */
+  i = [components count];
+  while (i-- > 0)
+    {
+      [components replaceObjectAtIndex: i withObject:
+	endSlashesDoubledFromString([components objectAtIndex: i])];
+    }
+
+  /* Join them again with \" as separator */
+  resultString = [components componentsJoinedByString: @"\\\""];
+
+  /* Put in in "'s if it contains spaces */
+  if ([resultString rangeOfCharacterFromSet:
+    [NSCharacterSet whitespaceCharacterSet]].length > 0)
+    {
+      resultString = [NSString stringWithFormat: @"\"%@\"", resultString];
+    }
+  return resultString;
 }
 
 - (void) launch
 {
+  DWORD		tid;
   STARTUPINFO	start_info;
   NSString      *lpath;
   NSString      *arg;
@@ -919,20 +1072,25 @@ GSCheckTasks()
   NSMutableString *args;
   char		*c_args;
   int		result;
-
+  const char	*executable;
+  
   if (_hasLaunched)
     {
       return;
     }
 
   lpath = [self _fullLaunchPath];
-  args = [lpath mutableCopy];
+  executable = [lpath fileSystemRepresentation];
+
+  args = [[NSMutableString alloc] initWithString:
+    quotedFromString([NSString stringWithCString: executable])];
   arg_enum = [[self arguments] objectEnumerator];
   while ((arg = [arg_enum nextObject]))
     {
       [args appendString: @" "];
-      [args appendString: arg];
+      [args appendString: quotedFromString(arg)];
     }
+  
   c_args = NSZoneMalloc(NSDefaultMallocZone(), [args cStringLength]+1);
   [args getCString: c_args];
 
@@ -943,7 +1101,7 @@ GSCheckTasks()
   start_info.hStdOutput = GetStdHandle(STD_OUTPUT_HANDLE);
   start_info.hStdError  = GetStdHandle(STD_ERROR_HANDLE);
 
-  result = CreateProcess([lpath fileSystemRepresentation],
+  result = CreateProcess(executable,
 			 c_args,
 			 NULL,      /* proc attrs */
 			 NULL,      /* thread attrs */
@@ -952,53 +1110,44 @@ GSCheckTasks()
 			 NULL,      /* env block */
 			 [[self currentDirectoryPath] fileSystemRepresentation],
 			 &start_info,
-			 &proc_info);
+			 &procInfo);
   NSZoneFree(NSDefaultMallocZone(), c_args);
   if (result == 0)
     {
-      NSLog(@"Error launching task: %@", lpath);
+      NSLog(@"Error launching task: %s", executable);
       return;
     }
-  _taskId = proc_info.dwProcessId;
+  
+  _taskId = procInfo.dwProcessId;
   _hasLaunched = YES;
   ASSIGN(_launchPath, lpath);	// Actual path used.
 
   [tasksLock lock];
   NSMapInsert(activeTasks, (void*)_taskId, (void*)self);
   [tasksLock unlock];
+/*
+ * Create thread to watch for termination of process.
+ */
+  wThread = CreateThread(NULL, 0, _threadFunction, (LPVOID)self, 0, &tid);
 }
 
 - (void) _collectChild
 {
   if (_hasCollected == NO)
     {
-      /* FIXME: Implement */
+      DWORD		eCode;
+
+      if (GetExitCodeProcess(procInfo.hProcess, &eCode) == 0)
+	{
+	  NSLog(@"Error getting exit code for process %d", _taskId);
+	}
+      else if (eCode != STILL_ACTIVE)
+	{
+	  [self _terminatedChild: eCode];
+	}
     }
 }
 
-- (int) terminationStatus
-{
-  DWORD	exit_code;
-  int	result;
-
-  [super terminationStatus];
-  result = GetExitCodeProcess(proc_info.hProcess, &exit_code);
-  _terminationStatus = exit_code;
-  if (result == 0)
-    {
-      NSLog(@"Error getting exit code");
-      return -1;
-    }
-  return exit_code;
-}
-
-- (void) waitUntilExit
-{
-  DWORD result;
-
-  result = WaitForSingleObject(proc_info.hProcess, INFINITE);
-}
-  
 @end
 
 #else /* !MINGW */
@@ -1159,7 +1308,7 @@ GSCheckTasks()
       /*
        * Make sure task is run in it's own process group.
        */
-#if     HAVE_SETPGRP
+#ifdef     HAVE_SETPGRP
 #ifdef	SETPGRP_VOID
       setpgrp();
 #else
@@ -1171,7 +1320,7 @@ GSCheckTasks()
 #else
       pid = (int)getpid();
 #endif
-#if     HAVE_SETPGID
+#ifdef     HAVE_SETPGID
       setpgid(pid, pid);
 #endif
 #endif
@@ -1186,7 +1335,7 @@ GSCheckTasks()
 	      exit(1);			/* Failed to open slave!	*/
 	    }
 
-#if	HAVE_SETSID
+#ifdef	HAVE_SETSID
 	  i = setsid();
 #endif
 #ifdef	TIOCNOTTY
@@ -1311,7 +1460,7 @@ GSCheckTasks()
       if (result < 0)
         {
           NSLog(@"waitpid %d, result %d, error %s",
-                _taskId, result, GSLastErrorStr(errno));
+	    _taskId, result, GSLastErrorStr(errno));
           [self _terminatedChild: -1];
         }
       else if (result == _taskId || (result > 0 && errno == 0))
@@ -1320,7 +1469,7 @@ GSCheckTasks()
 	    {
 #ifdef  WAITDEBUG
               NSLog(@"waitpid %d, termination status = %d",
-                        _taskId, _terminationStatus);
+		_taskId, _terminationStatus);
 #endif
               [self _terminatedChild: WEXITSTATUS(_terminationStatus)];
 	    }
@@ -1328,20 +1477,24 @@ GSCheckTasks()
 	    {
 #ifdef  WAITDEBUG
               NSLog(@"waitpid %d, termination status = %d",
-                        _taskId, _terminationStatus);
+		_taskId, _terminationStatus);
 #endif
               [self _terminatedChild: WTERMSIG(_terminationStatus)];
 	    }
 #ifdef  WAITDEBUG
           else
-            NSLog(@"waitpid %d, event status = %d",
-                        _taskId, _terminationStatus);
+	    {
+	      NSLog(@"waitpid %d, event status = %d",
+		_taskId, _terminationStatus);
+	    }
 #endif
 	}
 #ifdef  WAITDEBUG
       else
-        NSLog(@"waitpid %d, result %d, error %s",
-                _taskId, result, GSLastErrorStr(errno));
+	{
+	  NSLog(@"waitpid %d, result %d, error %s",
+	    _taskId, result, GSLastErrorStr(errno));
+	}
 #endif
     }
 }
