@@ -21,15 +21,15 @@
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA.
 
    <title>NSUnarchiver class reference</title>
-   $Date: 2002/02/26 13:55:47 $ $Revision: 1.46 $
+   $Date: 2003/07/31 23:49:31 $ $Revision: 1.53 $
    */
 
-#include <config.h>
+#include "config.h"
 #include <string.h>
-#include <Foundation/NSObjCRuntime.h>
-#include <Foundation/NSZone.h>
-#include <Foundation/NSException.h>
-#include <Foundation/NSByteOrder.h>
+#include "Foundation/NSObjCRuntime.h"
+#include "Foundation/NSZone.h"
+#include "Foundation/NSException.h"
+#include "Foundation/NSByteOrder.h"
 
 /*
  *	Setup for inline operation of arrays.
@@ -38,17 +38,17 @@
 #define	GSI_ARRAY_RELEASE(A, X)	
 #define	GSI_ARRAY_TYPES	GSUNION_OBJ|GSUNION_SEL|GSUNION_PTR
 
-#include <base/GSIArray.h>
+#include "GNUstepBase/GSIArray.h"
 
 #define	_IN_NSUNARCHIVER_M
-#include <Foundation/NSArchiver.h>
+#include "Foundation/NSArchiver.h"
 #undef	_IN_NSUNARCHIVER_M
 
-#include <Foundation/NSAutoreleasePool.h>
-#include <Foundation/NSCoder.h>
-#include <Foundation/NSData.h>
-#include <Foundation/NSUtilities.h>
-#include <Foundation/NSString.h>
+#include "Foundation/NSAutoreleasePool.h"
+#include "Foundation/NSCoder.h"
+#include "Foundation/NSData.h"
+#include "Foundation/NSUtilities.h"
+#include "Foundation/NSString.h"
 
 static const char*
 typeToName1(char type)
@@ -292,6 +292,52 @@ mapClassName(NSUnarchiverObjectInfo *info)
 @implementation	NSUnarchiverObjectInfo
 @end
 
+/**
+ * This class reconstructs objects from an archive.<br />
+ * <strong>Re-using the archiver</strong><br />
+ * <p>
+ *   The -resetUnarchiverWithData:atIndex: method lets you re-use
+ *   the archive to decode a new data object or, in conjunction
+ *   with the 'cursor' method (which reports the current decoding
+ *   position in the archive), decode a second archive that exists
+ *   in the data object after the first one.
+ * </p>
+ * <strong>Subclassing with different input format.</strong><br />
+ * NSUnarchiver normally reads directly from an NSData object using
+ * the methods -
+ * <deflist>
+ *   <term>-deserializeTypeTag:andCrossRef:atCursor:</term>
+ *   <desc>
+ *     to decode type tags for data items, the tag is the
+ *     first byte of the character encoding string for the
+ *     data type (as provided by '@encode(xxx)'), possibly
+ *     with the top bit set to indicate that what follows is
+ *     a crossreference to an item already encoded.<br />
+ *     Also decode a crossreference number either to identify the
+ *     following item, or to refer to a previously encoded item.
+ *     Objects, Classes, Selectors, CStrings and Pointer items
+ *     have crossreference encoding, other types do not.
+ *   </desc>
+ *   <term>[NSData-deserializeDataAt:ofObjCType:atCursor:context:]</term>
+ *   <desc>
+ *     to decode all other information.
+ *   </desc>
+ * </deflist>
+ * <p>
+ *   And uses other NSData methods to read the archive header information
+ *   from within the method:
+ *   [-deserializeHeaderAt:version:classes:objects:pointers:]
+ *   to read a fixed size header including archiver version
+ *   (obtained by [self systemVersion]) and crossreference
+ *   table sizes.
+ * </p>
+ * <p>
+ *   To subclass NSUnarchiver, you must implement your own versions of the
+ *   four methods above, and override the 'directDataAccess' method to
+ *   return NO so that the archiver knows to use your serialization
+ *   methods rather than those in the NSData object.
+ * </p>
+ */
 @implementation NSUnarchiver
 
 @class NSDataMalloc;
@@ -409,9 +455,9 @@ static Class NSDataMallocClass;
 			 count: (unsigned)expected
 			    at: (void*)buf
 {
-  int		i;
+  unsigned int	i;
   int		offset = 0;
-  int		size = objc_sizeof_type(type);
+  unsigned int	size = (unsigned int)objc_sizeof_type(type);
   unsigned char	info;
   unsigned	count;
 
@@ -700,29 +746,19 @@ static Class NSDataMallocClass;
 
       case _GSC_STRUCT_B:
 	{
-	  int offset = 0;
+	  struct objc_struct_layout layout;
 
 	  typeCheck(*type, _GSC_STRUCT_B);
-	  while (*type != _C_STRUCT_E && *type++ != '='); /* skip "<name>=" */
-	  for (;;)
+	  objc_layout_structure (type, &layout);
+	  while (objc_layout_structure_next_member (&layout))
 	    {
-	      (*dValImp)(self, dValSel, type, (char*)address + offset);
-	      offset += objc_sizeof_type(type);
-	      type = objc_skip_typespec(type);
-	      if (*type == _C_STRUCT_E)
-		{
-		  break;
-		}
-	      else
-		{
-		  int	align = objc_alignof_type(type);
-		  int	rem = offset % align;
+	      int		offset;
+	      int		align;
+	      const char	*ftype;
 
-		  if (rem != 0)
-		    {
-		      offset += align - rem;
-		    }
-		}
+	      objc_layout_structure_get_info (&layout, &offset, &align, &ftype);
+
+	      (*dValImp)(self, dValSel, ftype, (char*)address + offset);
 	    }
 	  return;
 	}
@@ -762,7 +798,7 @@ static Class NSDataMallocClass;
 	       *	add it to the crossref map.
 	       */
 	      size = objc_sizeof_type(++type);
-	      *(void**)address = _fastMallocBuffer(size);
+	      *(void**)address = GSAutoreleasedBuffer(size);
 	      GSIArrayAddItem(ptrMap, (GSIArrayItem)*(void**)address);
 
 	      /*
