@@ -21,31 +21,32 @@
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA.
 
    <title>NSUser class reference</title>
-   $Date: 2002/03/13 09:58:43 $ $Revision: 1.63 $
+   $Date: 2004/02/24 03:57:41 $ $Revision: 1.85 $
    */ 
 
-#include <config.h>
-#include <base/preface.h>
-#include <Foundation/NSObjCRuntime.h>
-#include <Foundation/NSString.h>
-#include <Foundation/NSPathUtilities.h>
-#include <Foundation/NSException.h>
-#include <Foundation/NSArray.h>
-#include <Foundation/NSDictionary.h>
-#include <Foundation/NSFileManager.h>
-#include <Foundation/NSProcessInfo.h>
-#include <Foundation/NSString.h>
-#include <Foundation/NSValue.h>
-#include <Foundation/NSLock.h>
-#include <Foundation/NSUserDefaults.h>
+#include "config.h"
+#include "GNUstepBase/preface.h"
+#include "Foundation/NSObjCRuntime.h"
+#include "Foundation/NSString.h"
+#include "Foundation/NSPathUtilities.h"
+#include "Foundation/NSException.h"
+#include "Foundation/NSArray.h"
+#include "Foundation/NSDictionary.h"
+#include "Foundation/NSFileManager.h"
+#include "Foundation/NSProcessInfo.h"
+#include "Foundation/NSString.h"
+#include "Foundation/NSValue.h"
+#include "Foundation/NSLock.h"
+#include "Foundation/NSUserDefaults.h"
+#include "GNUstepBase/GSCategories.h"
 
 #include "GSPrivate.h"
 
 #include <stdlib.h>		// for getenv()
-#if HAVE_UNISTD_H
+#ifdef HAVE_UNISTD_H
 #include <unistd.h>		// for getlogin()
 #endif
-#if	HAVE_PWD_H
+#ifdef	HAVE_PWD_H
 #include <pwd.h>		// for getpwnam()
 #endif
 #include <sys/types.h>
@@ -64,6 +65,30 @@ static NSString	*gnustep_system_root = nil;  /* GNUSTEP_SYSTEM_ROOT */
 static void	setupPathNames();
 static NSString	*userDirectory(NSString *name, BOOL defaults);
 
+static NSString*
+ImportPath(NSString *s, const char *c)
+{
+  static NSFileManager	*mgr = nil;
+  const char		*ptr = c;
+  unsigned		len;
+
+  if (mgr == nil)
+    {
+      mgr = [NSFileManager defaultManager];
+      RETAIN(mgr);
+    }
+  if (ptr == 0)
+    {
+      if (s == nil)
+	{
+	  return nil;
+	}
+      ptr = [s cString];
+    }
+  len = strlen(ptr);
+  return [mgr stringWithFileSystemRepresentation: ptr length: len]; 
+}
+
 /**
  * Sets the user name for this process.  This method is supplied to enable
  * setuid programs to run properly as the user indicated by their effective
@@ -79,31 +104,26 @@ GSSetUserName(NSString* name)
     {
       NSUserName();	// Ensure we know the old user name.
     }
-  if ([theUserName isEqualToString: name] == NO)
-    {
-      /*
-       * We can destroy the cached user path so that next time
-       * anything wants it, it will be regenerated.
-       */
-      DESTROY(gnustep_user_root);
-      /*
-       * Next we can set up the new user name, and reset the user defaults
-       * system so that standard user defaults will be those of the new
-       * user.
-       */
-      ASSIGN(theUserName, name);
-      [NSUserDefaults resetStandardUserDefaults];
-    }
+  /*
+   * We can destroy the cached user path so that next time
+   * anything wants it, it will be regenerated.
+   */
+  DESTROY(gnustep_user_root);
+  /*
+   * Next we can set up the new user name, and reset the user defaults
+   * system so that standard user defaults will be those of the new
+   * user.
+   */
+  ASSIGN(theUserName, name);
+  [NSUserDefaults resetStandardUserDefaults];
 }
 
 /**
- * Return the caller's login name as an NSString object.
- * The 'LOGNAME' environment variable is our primary source, but we use
- * other system-dependent sources if LOGNAME is not set.  This function
- * is intended to return the name under which the user logged in rather
- * than the name associated with their numeric user ID (though the two
- * are usually the same).  If you have a setuid program and want to
- * change the user to reflect the uid, use GSSetUserName()
+ * Return the caller's login name as an NSString object.<br />
+ * Under ms-windows, the 'LOGNAME' environment variable is used as the
+ * user name.<br />
+ * Under unix-like systems, the name associated with the current
+ * effective user ID is used.
  */
 /* NOTE FOR DEVELOPERS.
  * If you change the behavior of this method you must also change
@@ -112,60 +132,47 @@ GSSetUserName(NSString* name)
 NSString *
 NSUserName(void)
 {
+#if defined(__WIN32__)
   if (theUserName == nil)
     {
       const char *loginName = 0;
-#if defined(__WIN32__)
       /* The GetUserName function returns the current user name */
       char buf[1024];
       DWORD n = 1024;
 
-      if (GetEnvironmentVariable("LOGNAME", buf, 1024))
+      if (GetEnvironmentVariable("LOGNAME", buf, 1024) != 0 && buf[0] != '\0')
 	loginName = buf;
-      else if (GetUserName(buf, &n))
+      else if (GetUserName(buf, &n) != 0 && buf[0] != '\0')
 	loginName = buf;
-#else
-      loginName = getenv("LOGNAME");
-#if	HAVE_GETPWNAM
-      /*
-       * Check that LOGNAME contained legal name.
-       */
-      if (loginName != 0 && getpwnam(loginName) == 0)
-	{
-	  loginName = 0;
-	}
-#endif	/* HAVE_GETPWNAM */
-#if	HAVE_GETLOGIN
-      /*
-       * Try getlogin() if LOGNAME environmentm variable didn't work.
-       */
-      if (loginName == 0)
-	{
-	  loginName = getlogin();
-	}
-#endif	/* HAVE_GETLOGIN */
-#if HAVE_GETPWUID
-      /*
-       * Try getting the name of the effective user as a last resort.
-       */
-      if (loginName == 0)
-	{
-#if HAVE_GETEUID
-	  int uid = geteuid();
-#else
-	  int uid = getuid();
-#endif /* HAVE_GETEUID */
-	  struct passwd *pwent = getpwuid (uid);
-	  loginName = pwent->pw_name;
-	}
-#endif /* HAVE_GETPWUID */
-#endif
       if (loginName)
 	theUserName = [[NSString alloc] initWithCString: loginName];
       else
 	[NSException raise: NSInternalInconsistencyException
 		    format: @"Unable to determine current user name"];
     }
+#else
+  static int	olduid = 0;
+#ifdef HAVE_GETEUID
+  int uid = geteuid();
+#else
+  int uid = getuid();
+#endif /* HAVE_GETEUID */
+
+  if (theUserName == nil || uid != olduid)
+    {
+      const char *loginName = 0;
+#ifdef HAVE_GETPWUID
+      struct passwd *pwent = getpwuid (uid);
+      loginName = pwent->pw_name;
+#endif /* HAVE_GETPWUID */
+      olduid = uid;
+      if (loginName)
+	theUserName = [[NSString alloc] initWithCString: loginName];
+      else
+	[NSException raise: NSInternalInconsistencyException
+		    format: @"Unable to determine current user name"];
+    }
+#endif
   return theUserName;
 }
 
@@ -219,36 +226,51 @@ GSStringFromWin32EnvironmentVariable(const char * envVar)
 NSString *
 NSHomeDirectoryForUser(NSString *loginName)
 {
-  NSString	*s;
+  NSString	*s = nil;
 #if !defined(__MINGW__)
   struct passwd *pw;
 
   [gnustep_global_lock lock];
   pw = getpwnam ([loginName cString]);
-  if (pw == 0)
-    {
-      NSLog(@"Unable to locate home directory for '%@'", loginName);
-      s = nil;
-    }
-  else
+  if (pw != 0)
     {
       s = [NSString stringWithCString: pw->pw_dir];
     }
   [gnustep_global_lock unlock];
-  return s;
 #else
-  /* Then environment variable HOMEPATH holds the home directory
-     for the user on Windows NT; Win95 has no concept of home. */
-  [gnustep_global_lock lock];
-  s = GSStringFromWin32EnvironmentVariable("HOMEPATH");
-  if (s != nil)
+  if ([loginName isEqual: NSUserName()] == YES)
     {
-      s = [GSStringFromWin32EnvironmentVariable("HOMEDRIVE")
-        stringByAppendingString: s];
+      [gnustep_global_lock lock];
+      /*
+       * The environment variable HOMEPATH holds the home directory
+       * for the user on Windows NT; Win95 has no concept of home.
+       * For OPENSTEP compatibility (and because USERPROFILE is usually
+       * unusable because it contains spaces), we use HOMEPATH in
+       * preference to USERPROFILE.
+       */
+      s = GSStringFromWin32EnvironmentVariable("HOMEPATH");
+      if (s != nil && ([s length] < 2 || [s characterAtIndex: 1] != ':'))
+	{
+	  s = [GSStringFromWin32EnvironmentVariable("HOMEDRIVE")
+	    stringByAppendingString: s];
+	}
+      if (s == nil)
+	{
+	  /* The environment variable USERPROFILE holds the home directory
+	     for the user on more modern versions of windoze. */
+	  s = GSStringFromWin32EnvironmentVariable("USERPROFILE");
+	}
+      [gnustep_global_lock unlock];
+    }
+  if ([s length] == 0 && [loginName length] != 1)
+    {
+      s = nil;
+      NSLog(@"NSHomeDirectoryForUser(%@) failed", loginName);
     }
-  [gnustep_global_lock unlock];
-  return s;
 #endif
+  s = ImportPath(s, 0);
+// NSLog(@"Home for %@ is %@", loginName, s);
+  return s;
 }
 
 /**
@@ -258,7 +280,7 @@ NSHomeDirectoryForUser(NSString *loginName)
 NSString *
 NSFullUserName(void)
 {
-#if HAVE_PWD_H
+#ifdef HAVE_PWD_H
   struct passwd	*pw;
 
   pw = getpwnam([NSUserName() cString]);
@@ -269,134 +291,150 @@ NSFullUserName(void)
 #endif
 }
 
+static BOOL
+setupSystemRoot(NSDictionary *env)
+{
+  BOOL	warned = NO;
+	  
+  if (gnustep_system_root == nil)
+    {
+      /* Any of the following might be nil */
+      gnustep_system_root = [env objectForKey: @"GNUSTEP_SYSTEM_ROOT"];
+      gnustep_system_root = ImportPath(gnustep_system_root, 0);
+      TEST_RETAIN (gnustep_system_root);
+      if (gnustep_system_root == nil)
+	{
+	  /*
+	   * This is pretty important as we need it to load
+	   * character sets, language settings and similar
+	   * resources.  Use fprintf to avoid recursive calls.
+	   */
+	  warned = YES;
+	  gnustep_system_root
+	    = ImportPath(nil, stringify(GNUSTEP_INSTALL_PREFIX));
+	  RETAIN(gnustep_system_root);
+	  fprintf (stderr, 
+	    "Warning - GNUSTEP_SYSTEM_ROOT is not set "
+		    "- using %s\n", [gnustep_system_root lossyCString]);
+	}
+    }
+  return warned;
+}
+
+static BOOL
+setupLocalRoot(NSDictionary *env, BOOL warned)
+{
+  if (gnustep_local_root == nil)
+    {
+      gnustep_local_root = [env objectForKey: @"GNUSTEP_LOCAL_ROOT"];
+      gnustep_local_root = ImportPath(gnustep_local_root, 0);
+      TEST_RETAIN (gnustep_local_root);
+      if (gnustep_local_root == nil)
+	{
+	  gnustep_local_root
+	    = ImportPath(nil, stringify(GNUSTEP_LOCAL_ROOT));
+	  if ([gnustep_local_root length] == 0)
+	    {
+	      gnustep_local_root = nil;
+	    }
+	  else
+	    {
+	      RETAIN(gnustep_local_root);
+	    }
+	}
+      if (gnustep_local_root == nil)
+	{
+	  if ([[gnustep_system_root lastPathComponent] isEqual:
+	    @"System"] == YES)
+	    {
+	      gnustep_local_root = [[gnustep_system_root
+		stringByDeletingLastPathComponent]
+		stringByAppendingPathComponent: @"Local"];
+	      TEST_RETAIN (gnustep_local_root);
+	    }
+	  else
+	    {
+	      gnustep_local_root = @"/usr/GNUstep/Local";
+	    }
+#ifndef	NDEBUG
+	  if (warned == NO)
+	    {
+	      warned = YES;
+	      fprintf (stderr, 
+		"Warning - GNUSTEP_LOCAL_ROOT is not set "
+		"- using %s\n", [gnustep_local_root lossyCString]);
+	    }
+#endif
+	}
+    }
+  return warned;
+}
+
+static BOOL
+setupNetworkRoot(NSDictionary *env, BOOL warned)
+{
+  if (gnustep_network_root == nil)
+    {
+      gnustep_network_root = [env objectForKey: @"GNUSTEP_NETWORK_ROOT"];
+      gnustep_network_root = ImportPath(gnustep_network_root, 0);
+      TEST_RETAIN (gnustep_network_root);
+      if (gnustep_network_root == nil)
+	{
+	  gnustep_network_root
+	    = ImportPath(nil, stringify(GNUSTEP_NETWORK_ROOT));
+	  if ([gnustep_network_root length] == 0)
+	    {
+	      gnustep_network_root = nil;
+	    }
+	  else
+	    {
+	      RETAIN(gnustep_network_root);
+	    }
+	}
+      if (gnustep_network_root == nil)
+	{
+	  if ([[gnustep_system_root lastPathComponent] isEqual:
+	    @"System"] == YES)
+	    {
+	      gnustep_network_root = [[gnustep_system_root
+		stringByDeletingLastPathComponent]
+		stringByAppendingPathComponent: @"Network"];
+	      TEST_RETAIN (gnustep_network_root);
+	    }
+	  else
+	    {
+	      gnustep_network_root = @"/usr/GNUstep/Network";
+	    }
+#ifndef	NDEBUG
+	  if (warned == NO)
+	    {
+	      warned = YES;
+	      fprintf (stderr, 
+		"Warning - GNUSTEP_NETWORK_ROOT is not set "
+		"- using %s\n", [gnustep_network_root lossyCString]);
+	    }
+#endif
+	}
+    }
+  return warned;
+}
+
+
 static void
 setupPathNames()
 {
-#if defined (__MINGW32__)
-  NSString *systemDrive = GSStringFromWin32EnvironmentVariable("SystemDrive");
-#endif
   if (gnustep_user_root == nil)
     {
       NS_DURING
 	{
 	  BOOL	warned = NO;
-	  NSDictionary	*env;
+	  NSDictionary	*env = [[NSProcessInfo processInfo] environment];
 	  
 	  [gnustep_global_lock lock];
 
-	  /* Double-Locking Pattern */
-	  if (gnustep_system_root == nil)
-	    {
-	      env = [[NSProcessInfo processInfo] environment];
-	      /* Any of the following might be nil */
-	      gnustep_system_root = [env objectForKey: @"GNUSTEP_SYSTEM_ROOT"];
-	      TEST_RETAIN (gnustep_system_root);
-	      if (gnustep_system_root == nil)
-		{
-		  /*
-		   * This is pretty important as we need it to load
-		   * character sets, language settings and similar
-		   * resources.  Use fprintf to avoid recursive calls.
-		   */
-		  warned = YES;
-		  gnustep_system_root = [NSString stringWithCString:
-					  stringify(GNUSTEP_INSTALL_PREFIX)];
-#if defined (__MINGW32__)
-                  gnustep_system_root = [systemDrive stringByAppendingString:
-                                          gnustep_system_root];
-#endif
-
-		  RETAIN(gnustep_system_root);
-		  fprintf (stderr, 
-		    "Warning - GNUSTEP_SYSTEM_ROOT is not set "
-		    "- using %s\n", [gnustep_system_root lossyCString]);
-		}
-	    }
-	  if (gnustep_local_root == nil)
-	    {
-	      gnustep_local_root = [env objectForKey: @"GNUSTEP_LOCAL_ROOT"];
-	      TEST_RETAIN (gnustep_local_root);
-	      if (gnustep_local_root == nil)
-		{
-		  gnustep_local_root = [NSString stringWithCString:
-					  stringify(GNUSTEP_LOCAL_ROOT)];
-#if defined (__MINGW32__)
-                  gnustep_local_root = [systemDrive stringByAppendingString:
-                                         gnustep_local_root];
-#endif
-		  if ([gnustep_local_root length] == 0)
-		    gnustep_local_root = nil;
-		  else
-		    RETAIN(gnustep_local_root);
-		}
-	      if (gnustep_local_root == nil)
-		{
-		  if ([[gnustep_system_root lastPathComponent] isEqual:
-		    @"System"] == YES)
-		    {
-		      gnustep_local_root = [[gnustep_system_root
-			stringByDeletingLastPathComponent]
-			stringByAppendingPathComponent: @"Local"];
-		      TEST_RETAIN (gnustep_local_root);
-		    }
-		  else
-		    {
-		      gnustep_local_root = @"/usr/GNUstep/Local";
-		    }
-#ifndef	NDEBUG
-		  if (warned == NO)
-		    {
-		      warned = YES;
-		      fprintf (stderr, 
-			"Warning - GNUSTEP_LOCAL_ROOT is not set "
-			"- using %s\n", [gnustep_local_root lossyCString]);
-		    }
-#endif
-		}
-	    }
-	  if (gnustep_network_root == nil)
-	    {
-	      gnustep_network_root = [env objectForKey: 
-		@"GNUSTEP_NETWORK_ROOT"];
-	      TEST_RETAIN (gnustep_network_root);
-	      if (gnustep_network_root == nil)
-		{
-		  gnustep_network_root = [NSString stringWithCString:
-					  stringify(GNUSTEP_NETWORK_ROOT)];
-#if defined (__MINGW32__)
-                  gnustep_network_root = [systemDrive stringByAppendingString:
-                                           gnustep_network_root];
-#endif
-		  if ([gnustep_network_root length] == 0)
-		    gnustep_network_root = nil;
-		  else
-		    RETAIN(gnustep_network_root);
-		}
-	      if (gnustep_network_root == nil)
-		{
-		  if ([[gnustep_system_root lastPathComponent] isEqual:
-		    @"System"] == YES)
-		    {
-		      gnustep_network_root = [[gnustep_system_root
-			stringByDeletingLastPathComponent]
-			stringByAppendingPathComponent: @"Network"];
-		      TEST_RETAIN (gnustep_network_root);
-		    }
-		  else
-		    {
-		      gnustep_network_root = @"/usr/GNUstep/Network";
-		    }
-#ifndef	NDEBUG
-		  if (warned == NO)
-		    {
-		      warned = YES;
-		      fprintf (stderr, 
-			"Warning - GNUSTEP_NETWORK_ROOT is not set "
-			"- using %s\n", [gnustep_network_root lossyCString]);
-		    }
-#endif
-		}
-	    }
+	  warned = setupSystemRoot(env);
+	  warned = setupLocalRoot(env, warned);
+	  warned = setupNetworkRoot(env, warned);
 	  if (gnustep_user_root == nil)
 	    {
 	      gnustep_user_root = [userDirectory(NSUserName(), NO) copy];
@@ -449,42 +487,31 @@ userDirectory(NSString *name, BOOL defaults)
   NSString	*file;
   NSString	*user = nil;
   NSString	*defs = nil;
+  BOOL		forceD = NO;
+  BOOL		forceU = NO;
+  NSDictionary	*attributes;
 
   NSCAssert([name length] > 0, NSInvalidArgumentException);
 
-  if (defaults == YES)
-    {
-#ifdef	FORCE_DEFAULTS_ROOT
-      return [NSString stringWithCString: stringify(FORCE_DEFAULTS_ROOT)];
-#endif
-    }
-  else
-    {
-#ifdef	FORCE_USER_ROOT
-      return [NSString stringWithCString: stringify(FORCE_USER_ROOT)];
-#endif
-    }
-
   home = NSHomeDirectoryForUser(name);
-  file = [home stringByAppendingPathComponent: @".GNUsteprc"];
   manager = [NSFileManager defaultManager];
-  /*
-   * If there is no per-user  startup file, look in the one in the
-   * configured system root directory.  NB. Don't use the environment
-   * variable ... that could have been changed!
-   */
-  if ([manager isReadableFileAtPath: file] == NO)
-    {
-      NSString	*root;
 
-      root = [NSString stringWithCString: stringify(GNUSTEP_INSTALL_PREFIX)];
-#if defined (__MINGW32__)
-      root = [GSStringFromWin32EnvironmentVariable("SystemDrive")
-	stringByAppendingString: gnustep_system_root];
-#endif
-      file = [root stringByAppendingPathComponent: @".GNUsteprc"];
+  if (gnustep_system_root == nil)
+    {
+      NSDictionary	*env = [[NSProcessInfo processInfo] environment];
+	  
+      [gnustep_global_lock lock];
+      setupSystemRoot(env);
+      [gnustep_global_lock unlock];
     }
-  if ([manager isReadableFileAtPath: file] == YES)
+  file = [gnustep_system_root stringByAppendingPathComponent: @".GNUsteprc"];
+  attributes = [manager fileAttributesAtPath: file traverseLink: YES];
+  if (([attributes filePosixPermissions] & 022) != 0)
+    {
+      fprintf(stderr, "The file '%s' is writable by someone other than"
+	" its owner.\nIgnoring it.\n", [file fileSystemRepresentation]);
+    }
+  else if ([manager isReadableFileAtPath: file] == YES)
     {
       NSArray	*lines;
       unsigned	count;
@@ -496,33 +523,119 @@ userDirectory(NSString *name, BOOL defaults)
 	{
 	  NSRange	r;
 	  NSString	*line;
+	  NSString	*key;
+	  NSString	*val;
 
 	  line = [[lines objectAtIndex: count] stringByTrimmingSpaces];
 	  r = [line rangeOfString: @"="];
 	  if (r.length == 1)
 	    {
-	      NSString	*key = [line substringToIndex: r.location];
-	      NSString	*val = [line substringFromIndex: NSMaxRange(r)];
+	      key = [line substringToIndex: r.location];
+	      val = [line substringFromIndex: NSMaxRange(r)];
 
 	      key = [key stringByTrimmingSpaces];
 	      val = [val stringByTrimmingSpaces];
-	      if ([key isEqualToString: @"GNUSTEP_USER_ROOT"] == YES)
+	    }
+	  else
+	    {
+	      key = [line stringByTrimmingSpaces];
+	      val = nil;
+	    }
+
+	  if ([key isEqualToString: @"GNUSTEP_USER_ROOT"] == YES)
+	    {
+	      if ([val length] > 0 && [val characterAtIndex: 0] == '~')
 		{
-		  if ([val length] > 0 && [val characterAtIndex: 0] == '~')
-		    {
-		      val = [home stringByAppendingString:
-			[val substringFromIndex: 1]];
-		    }
-		  user = val;
+		  val = [home stringByAppendingString:
+		    [val substringFromIndex: 1]];
 		}
-	      else if ([key isEqualToString: @"GNUSTEP_DEFAULTS_ROOT"] == YES)
+	      user = val;
+	    }
+	  else if ([key isEqualToString: @"GNUSTEP_DEFAULTS_ROOT"] == YES)
+	    {
+	      if ([val length] > 0 && [val characterAtIndex: 0] == '~')
 		{
-		  if ([val length] > 0 && [val characterAtIndex: 0] == '~')
+		  val = [home stringByAppendingString:
+		    [val substringFromIndex: 1]];
+		}
+	      defs = val;
+	    }
+	  else if ([key isEqualToString: @"FORCE_USER_ROOT"] == YES)
+	    {
+	      forceU = YES;
+	    }
+	  else if ([key isEqualToString: @"FORCE_DEFAULTS_ROOT"] == YES)
+	    {
+	      forceD = YES;
+	    }
+	}
+    }
+
+  if (forceD == NO || defs == nil || forceU == NO || user == nil)
+    {
+      file = [home stringByAppendingPathComponent: @".GNUsteprc"];
+
+      attributes = [manager fileAttributesAtPath: file traverseLink: YES];
+      if (([attributes filePosixPermissions] & 022) != 0)
+	{
+	  fprintf(stderr, "The file '%s' is writable by someone other than"
+	    " its owner.\nIgnoring it.\n", [file fileSystemRepresentation]);
+	}
+#ifndef	__MINGW__
+/* FIXME ... need to get mingw working */
+      else if (attributes != nil
+	&& [[attributes fileOwnerAccountName] isEqual: NSUserName()] == NO)
+	{
+	  fprintf(stderr, "The file '%s' is not owned by the current user."
+	    "\nIgnoring it.\n", [file fileSystemRepresentation]);
+	}
+#endif
+      else if ([manager isReadableFileAtPath: file] == YES)
+	{
+	  NSArray	*lines;
+	  unsigned	count;
+
+	  file = [NSString stringWithContentsOfFile: file];
+	  lines = [file componentsSeparatedByString: @"\n"];
+	  count = [lines count];
+	  while (count-- > 0)
+	    {
+	      NSRange	r;
+	      NSString	*line;
+
+	      line = [[lines objectAtIndex: count] stringByTrimmingSpaces];
+	      r = [line rangeOfString: @"="];
+	      if (r.length == 1)
+		{
+		  NSString	*key = [line substringToIndex: r.location];
+		  NSString	*val = [line substringFromIndex: NSMaxRange(r)];
+
+		  key = [key stringByTrimmingSpaces];
+		  val = [val stringByTrimmingSpaces];
+		  if ([key isEqualToString: @"GNUSTEP_USER_ROOT"] == YES)
 		    {
-		      val = [home stringByAppendingString:
-			[val substringFromIndex: 1]];
+		      if ([val length] > 0 && [val characterAtIndex: 0] == '~')
+			{
+			  val = [home stringByAppendingString:
+			    [val substringFromIndex: 1]];
+			}
+		      if (user == nil || forceU == NO)
+			{
+			  user = val;
+			}
+		    }
+		  else if ([key isEqualToString: @"GNUSTEP_DEFAULTS_ROOT"])
+		    {
+		      if ([val length] > 0 && [val characterAtIndex: 0] == '~')
+			{
+			  val = [home stringByAppendingString:
+			    [val substringFromIndex: 1]];
+			}
+		      if (defs == nil || forceD == NO)
+			{
+			  defs = val;
+			}
 		    }
-		  defs = val;
 		}
 	    }
 	}
@@ -552,7 +665,7 @@ userDirectory(NSString *name, BOOL defaults)
       path = [home stringByAppendingPathComponent: @"GNUstep"];
     }
 
-  return path;
+  return ImportPath(path, 0);
 }
 
 /** Returns an array of strings which contain paths that should be in
@@ -574,17 +687,31 @@ GSStandardPathPrefixes(void)
   prefixes = [env objectForKey: @"GNUSTEP_PATHPREFIX_LIST"];
   if (prefixes != nil)
     {
+      unsigned	c;
+
 #if	defined(__WIN32__)
       prefixArray = [prefixes componentsSeparatedByString: @";"];
 #else
       prefixArray = [prefixes componentsSeparatedByString: @":"];
 #endif
-      if ([prefixArray count] <= 1)
+      if ((c = [prefixArray count]) <= 1)
 	{
 	  /* This probably means there was some parsing error, but who
 	     knows. Play it safe though... */
 	  prefixArray = nil;
 	}
+      else
+	{
+	  NSString	*a[c];
+	  unsigned	i;
+
+	  [prefixArray getObjects: a];
+	  for (i = 0; i < c; i++)
+	    {
+	      a[c] = ImportPath(a[c], 0);
+	    }
+	  prefixArray = [NSArray arrayWithObjects: a count: c];
+	}
     }
   if (prefixes == nil)
     {
@@ -660,7 +787,7 @@ NSTemporaryDirectory(void)
 
   if (GetTempPath(1024, buffer))
     {
-      baseTempDirName = [NSString stringWithCString: buffer];
+      baseTempDirName = ImportPath(nil, buffer);
     }
 #endif
 
@@ -679,8 +806,12 @@ NSTemporaryDirectory(void)
 	  baseTempDirName = [env objectForKey: @"TMP"];
 	  if (baseTempDirName == nil)
 	    {
-#if	defined(__WIN32__)
-	      baseTempDirName = @"C:\\";
+#if	defined(__MINGW__)
+#ifdef  __CYGWIN__
+	      baseTempDirName = @"/cygdrive/c/";
+#else
+	      baseTempDirName = @"/c/";
+#endif
 #else
 	      baseTempDirName = @"/tmp";
 #endif
@@ -753,11 +884,21 @@ NSOpenStepRootDirectory(void)
   root = [[[NSProcessInfo processInfo] environment]
     objectForKey: @"GNUSTEP_ROOT"];
   if (root == nil)
+    {
 #if	defined(__MINGW__)
-    root = @"C:\\";
+#ifdef  __CYGWIN__
+      root = @"/cygdrive/c/";
 #else
-    root = @"/";
+      root = @"/c/";
 #endif
+#else
+      root = @"/";
+#endif
+    }
+  else
+    {
+      root = ImportPath(root, 0);
+    }
   return root;
 }
 
@@ -774,9 +915,10 @@ NSSearchPathForDirectoriesInDomains(NSSearchPathDirectory directoryKey,
   NSString		*demosDir = @"Demos";
   NSString		*devDir = @"Developer";
   NSString		*libraryDir = @"Library";
-  NSString		*libsDir = @"Libraries";
+  NSString		*libsDir = @"Library/Libraries";
   NSString		*toolsDir = @"Tools";
-  NSString		*docDir = @"Documentation";
+  NSString		*docDir = @"Library/Documentation";
+  NSString		*supportDir = @"Library/ApplicationSupport";
   NSMutableArray	*paths = [NSMutableArray new];
   NSString		*path;
   unsigned		i;
@@ -811,26 +953,26 @@ if (domainMask & mask) \
   if (directoryKey == NSApplicationDirectory
     || directoryKey == NSAllApplicationsDirectory)
     {
-      ADD_PATH (NSUserDomainMask, gnustep_user_root, appsDir);
-      ADD_PATH (NSLocalDomainMask, gnustep_local_root, appsDir);
-      ADD_PATH (NSNetworkDomainMask, gnustep_network_root, appsDir);
-      ADD_PATH (NSSystemDomainMask, gnustep_system_root, appsDir);
+      ADD_PATH(NSUserDomainMask, gnustep_user_root, appsDir);
+      ADD_PATH(NSLocalDomainMask, gnustep_local_root, appsDir);
+      ADD_PATH(NSNetworkDomainMask, gnustep_network_root, appsDir);
+      ADD_PATH(NSSystemDomainMask, gnustep_system_root, appsDir);
     }
   if (directoryKey == NSDemoApplicationDirectory
     || directoryKey == NSAllApplicationsDirectory);
     {
       NSString *devDemosDir = [devDir stringByAppendingPathComponent: demosDir];
-      ADD_PATH (NSSystemDomainMask, gnustep_system_root, devDemosDir);
+      ADD_PATH(NSSystemDomainMask, gnustep_system_root, devDemosDir);
     }
   if (directoryKey == NSDeveloperApplicationDirectory
     || directoryKey == NSAllApplicationsDirectory)
     {
       NSString *devAppsDir = [devDir stringByAppendingPathComponent: appsDir];
 
-      ADD_PATH (NSUserDomainMask, gnustep_local_root, devAppsDir);
-      ADD_PATH (NSLocalDomainMask, gnustep_local_root, devAppsDir);
-      ADD_PATH (NSNetworkDomainMask, gnustep_network_root, devAppsDir);
-      ADD_PATH (NSSystemDomainMask, gnustep_system_root, devAppsDir);
+      ADD_PATH(NSUserDomainMask, gnustep_user_root, devAppsDir);
+      ADD_PATH(NSLocalDomainMask, gnustep_local_root, devAppsDir);
+      ADD_PATH(NSNetworkDomainMask, gnustep_network_root, devAppsDir);
+      ADD_PATH(NSSystemDomainMask, gnustep_system_root, devAppsDir);
     }
   if (directoryKey == NSAdminApplicationDirectory
     || directoryKey == NSAllApplicationsDirectory)
@@ -838,24 +980,24 @@ if (domainMask & mask) \
       NSString *devAdminDir = [devDir stringByAppendingPathComponent: adminDir];
 
       /* FIXME - NSUserDomainMask ? - users have no Administrator directory */
-      ADD_PATH (NSLocalDomainMask, gnustep_local_root, devAdminDir);
-      ADD_PATH (NSNetworkDomainMask, gnustep_network_root, devAdminDir);
-      ADD_PATH (NSSystemDomainMask, gnustep_system_root, devAdminDir);
+      ADD_PATH(NSLocalDomainMask, gnustep_local_root, devAdminDir);
+      ADD_PATH(NSNetworkDomainMask, gnustep_network_root, devAdminDir);
+      ADD_PATH(NSSystemDomainMask, gnustep_system_root, devAdminDir);
     }
   if (directoryKey == NSLibraryDirectory
     || directoryKey == NSAllLibrariesDirectory)
     {
-      ADD_PATH (NSUserDomainMask, gnustep_user_root, libraryDir);
-      ADD_PATH (NSLocalDomainMask, gnustep_local_root, libraryDir);
-      ADD_PATH (NSNetworkDomainMask, gnustep_network_root, libraryDir);
-      ADD_PATH (NSSystemDomainMask, gnustep_system_root, libraryDir);
+      ADD_PATH(NSUserDomainMask, gnustep_user_root, libraryDir);
+      ADD_PATH(NSLocalDomainMask, gnustep_local_root, libraryDir);
+      ADD_PATH(NSNetworkDomainMask, gnustep_network_root, libraryDir);
+      ADD_PATH(NSSystemDomainMask, gnustep_system_root, libraryDir);
     }
   if (directoryKey == NSDeveloperDirectory)
     {
-      ADD_PATH (NSUserDomainMask, gnustep_local_root, devDir);
-      ADD_PATH (NSLocalDomainMask, gnustep_local_root, devDir);
-      ADD_PATH (NSNetworkDomainMask, gnustep_network_root, devDir);
-      ADD_PATH (NSSystemDomainMask, gnustep_system_root, devDir);
+      ADD_PATH(NSUserDomainMask, gnustep_local_root, devDir);
+      ADD_PATH(NSLocalDomainMask, gnustep_local_root, devDir);
+      ADD_PATH(NSNetworkDomainMask, gnustep_network_root, devDir);
+      ADD_PATH(NSSystemDomainMask, gnustep_system_root, devDir);
     }
   if (directoryKey == NSUserDirectory)
     {
@@ -866,25 +1008,33 @@ if (domainMask & mask) \
     }
   if (directoryKey == NSDocumentationDirectory)
     {
-      ADD_PATH (NSUserDomainMask, gnustep_user_root, docDir);
-      ADD_PATH (NSLocalDomainMask, gnustep_local_root, docDir);
-      ADD_PATH (NSNetworkDomainMask, gnustep_network_root, docDir);
-      ADD_PATH (NSSystemDomainMask, gnustep_system_root, docDir);
+      ADD_PATH(NSUserDomainMask, gnustep_user_root, docDir);
+      ADD_PATH(NSLocalDomainMask, gnustep_local_root, docDir);
+      ADD_PATH(NSNetworkDomainMask, gnustep_network_root, docDir);
+      ADD_PATH(NSSystemDomainMask, gnustep_system_root, docDir);
     }
   if (directoryKey == GSLibrariesDirectory)
     {
-      ADD_PATH (NSUserDomainMask, gnustep_user_root, libsDir);
-      ADD_PATH (NSLocalDomainMask, gnustep_local_root, libsDir);
-      ADD_PATH (NSNetworkDomainMask, gnustep_network_root, libsDir);
-      ADD_PATH (NSSystemDomainMask, gnustep_system_root, libsDir);
+      ADD_PATH(NSUserDomainMask, gnustep_user_root, libsDir);
+      ADD_PATH(NSLocalDomainMask, gnustep_local_root, libsDir);
+      ADD_PATH(NSNetworkDomainMask, gnustep_network_root, libsDir);
+      ADD_PATH(NSSystemDomainMask, gnustep_system_root, libsDir);
     }
   if (directoryKey == GSToolsDirectory)
     {
-      ADD_PATH (NSUserDomainMask, gnustep_user_root, toolsDir);
-      ADD_PATH (NSLocalDomainMask, gnustep_local_root, toolsDir);
-      ADD_PATH (NSNetworkDomainMask, gnustep_network_root, toolsDir);
-      ADD_PATH (NSSystemDomainMask, gnustep_system_root, toolsDir);
+      ADD_PATH(NSUserDomainMask, gnustep_user_root, toolsDir);
+      ADD_PATH(NSLocalDomainMask, gnustep_local_root, toolsDir);
+      ADD_PATH(NSNetworkDomainMask, gnustep_network_root, toolsDir);
+      ADD_PATH(NSSystemDomainMask, gnustep_system_root, toolsDir);
     }
+  if (directoryKey == GSApplicationSupportDirectory)
+    {
+      ADD_PATH(NSUserDomainMask, gnustep_user_root, supportDir);
+      ADD_PATH(NSLocalDomainMask, gnustep_local_root, supportDir);
+      ADD_PATH(NSNetworkDomainMask, gnustep_network_root, supportDir);
+      ADD_PATH(NSSystemDomainMask, gnustep_system_root, supportDir);
+    }
+
 #undef ADD_PATH
 
   fm = [NSFileManager defaultManager];
