@@ -23,32 +23,39 @@
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA.
 
    <title>NSUserDefaults class reference</title>
-   $Date: 2002/03/13 09:58:43 $ $Revision: 1.77 $
+   $Date: 2003/11/10 06:40:24 $ $Revision: 1.106 $
 */
 
-#include <config.h>
-#include <base/preface.h>
+#include "config.h"
+#include "GNUstepBase/preface.h"
 #include <sys/stat.h>
 #include <sys/types.h>
 #include <errno.h>
 
-#include <Foundation/NSUserDefaults.h>
-#include <Foundation/NSFileManager.h>
-#include <Foundation/NSPathUtilities.h>
-#include <Foundation/NSDictionary.h>
-#include <Foundation/NSArray.h>
-#include <Foundation/NSDate.h>
-#include <Foundation/NSUtilities.h>
-#include <Foundation/NSArchiver.h>
-#include <Foundation/NSException.h>
-#include <Foundation/NSNotification.h>
-#include <Foundation/NSTimer.h>
-#include <Foundation/NSProcessInfo.h>
-#include <Foundation/NSRunLoop.h>
-#include <Foundation/NSBundle.h>
-#include <Foundation/NSValue.h>
-#include <Foundation/NSLock.h>
-#include <base/GSLocale.h>
+#include "Foundation/NSUserDefaults.h"
+#include "Foundation/NSArchiver.h"
+#include "Foundation/NSArray.h"
+#include "Foundation/NSBundle.h"
+#include "Foundation/NSDate.h"
+#include "Foundation/NSDictionary.h"
+#include "Foundation/NSDistributedLock.h"
+#include "Foundation/NSException.h"
+#include "Foundation/NSFileManager.h"
+#include "Foundation/NSLock.h"
+#include "Foundation/NSNotification.h"
+#include "Foundation/NSPathUtilities.h"
+#include "Foundation/NSProcessInfo.h"
+#include "Foundation/NSRunLoop.h"
+#include "Foundation/NSSet.h"
+#include "Foundation/NSThread.h"
+#include "Foundation/NSTimer.h"
+#include "Foundation/NSUtilities.h"
+#include "Foundation/NSValue.h"
+#include "Foundation/NSDebug.h"
+#include "GNUstepBase/GSLocale.h"
+#ifdef HAVE_LOCALE_H
+#include <locale.h>
+#endif
 
 #include "GSPrivate.h"
 
@@ -64,7 +71,9 @@ static SEL	addSel;
 
 static Class	NSArrayClass;
 static Class	NSDataClass;
+static Class	NSDateClass;
 static Class	NSDictionaryClass;
+static Class	NSNumberClass;
 static Class	NSMutableDictionaryClass;
 static Class	NSStringClass;
 
@@ -82,6 +91,27 @@ static void updateCache(NSUserDefaults *self)
 {
   if (self == sharedDefaults)
     {
+      NSArray	*debug;
+
+      /**
+       * If there is an array NSUserDefault called GNU-Debug,
+       * we add its contents to the set of active debug levels.
+       */
+      debug = [self arrayForKey: @"GNU-Debug"];
+      if (debug != nil)
+        {
+	  unsigned	c = [debug count];
+	  NSMutableSet	*s;
+
+	  s = [[NSProcessInfo processInfo] debugSet];
+	  while (c-- > 0)
+	    {
+	      NSString	*level = [debug objectAtIndex: c];
+
+	      [s addObject: level];
+	    }
+	}
+
       flags[GSMacOSXCompatible]
 	= [self boolForKey: @"GSMacOSXCompatible"];
       flags[GSOldStyleGeometry]
@@ -100,7 +130,6 @@ static void updateCache(NSUserDefaults *self)
 - (void) __createStandardSearchList;
 - (NSDictionary*) __createArgumentDictionary;
 - (void) __changePersistentDomain: (NSString*)domainName;
-- (void) __timerTicked: (NSTimer*)tim;
 @end
 
 /**
@@ -196,7 +225,9 @@ static BOOL setSharedDefaults = NO;	/* Flag to prevent infinite recursion */
        */
       NSArrayClass = [NSArray class];
       NSDataClass = [NSData class];
+      NSDateClass = [NSDate class];
       NSDictionaryClass = [NSDictionary class];
+      NSNumberClass = [NSNumber class];
       NSMutableDictionaryClass = [NSMutableDictionary class];
       NSStringClass = [NSString class];
       classLock = [NSRecursiveLock new];
@@ -217,6 +248,7 @@ static BOOL setSharedDefaults = NO;	/* Flag to prevent infinite recursion */
     {
       NSDictionary	*regDefs;
 
+      [sharedDefaults synchronize];	// Ensure changes are written.
       regDefs = RETAIN([sharedDefaults->_tempDomains
 	objectForKey: NSRegistrationDomain]);
       setSharedDefaults = NO;
@@ -389,9 +421,11 @@ static BOOL setSharedDefaults = NO;	/* Flag to prevent infinite recursion */
     {
       NSString *path;
       NSDictionary *dict;
-      path = [NSBundle pathForGNUstepResource: lang
-		                       ofType: nil
-		                  inDirectory: @"Resources/Languages"];
+      NSBundle *gbundle;
+      gbundle = [NSBundle bundleForLibrary: @"gnustep-base"];
+      path = [gbundle pathForResource: lang
+		               ofType: nil
+		          inDirectory: @"Languages"];
       dict = nil;
       if (path)
 	dict = [NSDictionary dictionaryWithContentsOfFile: path];
@@ -402,19 +436,31 @@ static BOOL setSharedDefaults = NO;	/* Flag to prevent infinite recursion */
 	}
       else if (added_locale == NO)
 	{
-	  NSString *locale = GSSetLocale(nil);
+	  NSString	*locale = nil;
+
+#ifdef HAVE_LOCALE_H
+#ifdef LC_MESSAGES
+	  locale = GSSetLocale(LC_MESSAGES, nil);
+#endif
+#endif
 	  if (locale == nil)
-	    continue;
+	    {
+	      continue;
+	    }
 	  /* See if we can get the dictionary from i18n functions.
 	     Note that we get the dict from the current locale regardless
 	     of what 'lang' is, since it should match anyway. */
 	  /* Also, I don't think that the i18n routines can handle more than
 	     one locale, but tell me if I'm wrong... */
 	  if (GSLanguageFromLocale(locale))
-	    lang = GSLanguageFromLocale(locale);
+	    {
+	      lang = GSLanguageFromLocale(locale);
+	    }
 	  dict = GSDomainFromDefaultLocale();
-	  if (dict)
-	    [sharedDefaults setVolatileDomain: dict forName: lang];
+	  if (dict != nil)
+	    {
+	      [sharedDefaults setVolatileDomain: dict forName: lang];
+	    }
 	  added_locale = YES;
 	}
     }
@@ -439,8 +485,13 @@ static BOOL setSharedDefaults = NO;	/* Flag to prevent infinite recursion */
 + (NSArray*) userLanguages
 {
   NSArray	*currLang = nil;
-  NSString	*locale;
+  NSString	*locale = nil;
 
+#ifdef HAVE_LOCALE_H
+#ifdef LC_MESSAGES
+  locale = GSSetLocale(LC_MESSAGES, nil);
+#endif
+#endif
   [classLock lock];
   if (userLanguages != nil)
     {
@@ -449,7 +500,6 @@ static BOOL setSharedDefaults = NO;	/* Flag to prevent infinite recursion */
       return AUTORELEASE(userLanguages);
     }
   userLanguages = RETAIN([NSMutableArray arrayWithCapacity: 5]);
-  locale = GSSetLocale(@"");
   if (sharedDefaults == nil)
     {
       /* Create our own defaults to get "NSLanguages" since sharedDefaults
@@ -473,7 +523,8 @@ static BOOL setSharedDefaults = NO;	/* Flag to prevent infinite recursion */
 	  [tempDefaults setSearchList: sList];
 	  RELEASE(sList);
 	  currLang = [tempDefaults stringArrayForKey: @"NSLanguages"];
-	  AUTORELEASE(tempDefaults);
+	  AUTORELEASE(RETAIN(currLang));
+	  RELEASE(tempDefaults);
 	}
     }
   else
@@ -481,12 +532,12 @@ static BOOL setSharedDefaults = NO;	/* Flag to prevent infinite recursion */
       currLang
 	= [[self standardUserDefaults] stringArrayForKey: @"NSLanguages"];
     }
-  if (currLang == nil && locale != 0 && GSLanguageFromLocale(locale))
+  if (currLang == nil && locale != nil && GSLanguageFromLocale(locale))
     {
       currLang = [NSArray arrayWithObject: GSLanguageFromLocale(locale)];
     }
 #ifdef __MINGW__
-  if (currLang == nil && locale != 0)
+  if (currLang == nil && locale != nil)
     {
       /* Check for language as the first part of the locale string */
       NSRange under = [locale rangeOfString: @"_"];
@@ -549,7 +600,6 @@ static BOOL setSharedDefaults = NO;	/* Flag to prevent infinite recursion */
   [[self standardUserDefaults]
     setPersistentDomain: globDict forName: NSGlobalDomain];
   RELEASE(globDict);
-  return;
 }
 
 /*************************************************************************
@@ -592,7 +642,7 @@ static NSString	*pathForUser(NSString *user)
 #endif
   attr = [NSDictionary dictionaryWithObjectsAndKeys:
     NSUserName(), NSFileOwnerAccountName,
-    [NSNumber numberWithUnsignedLong: desired], NSFilePosixPermissions,
+    [NSNumberClass numberWithUnsignedLong: desired], NSFilePosixPermissions,
     nil];
 
   if ([mgr fileExistsAtPath: home isDirectory: &isDir] == NO)
@@ -694,6 +744,9 @@ static NSString	*pathForUser(NSString *user)
  */
 - (id) initWithContentsOfFile: (NSString*)path
 {
+  NSFileManager	*mgr = [NSFileManager defaultManager];
+  BOOL		flag;
+
   self = [super init];
 
   /*
@@ -704,14 +757,40 @@ static NSString	*pathForUser(NSString *user)
       processName = RETAIN([[NSProcessInfo processInfo] processName]);
     }
 
-  if (path != nil && [path isEqual: @""] == NO)
+  if (path == nil || [path isEqual: @""] == YES)
     {
-      _defaultsDatabase = [path copy];
+      path = pathForUser(NSUserName());
+    }
+  path = [path stringByStandardizingPath];
+  _defaultsDatabase = [path copy];
+  path = [path stringByDeletingLastPathComponent];
+  if ([mgr isWritableFileAtPath: path] == NO)
+    {
+      NSWarnMLog(@"Path '%@' is not writable - making user defaults for '%@' "
+	@" read-only\n", path, _defaultsDatabase);
+    }
+  else if ([mgr fileExistsAtPath: path isDirectory: &flag] == NO && flag == NO)
+    {
+      NSWarnMLog(@"Path '%@' is not an accessible directory - making user "
+	@"defaults for '%@' read-only\n", path, _defaultsDatabase);
+    }
+  else if ([mgr fileExistsAtPath: _defaultsDatabase] == YES
+    && [mgr isReadableFileAtPath: _defaultsDatabase] == NO)
+    {
+      NSWarnMLog(@"Path '%@' is not readable - making user defaults blank\n",
+	_defaultsDatabase);
     }
   else
     {
-      _defaultsDatabase = [pathForUser(NSUserName()) copy];
+      /*
+       * Only create the file lock if we can update the file ...
+       * if we can't the absence of the lock tells us we must be
+       * in read-only mode.
+       */
+      _fileLock = [[NSDistributedLock alloc] initWithPath:
+	[_defaultsDatabase stringByAppendingPathExtension: @"lck"]];
     }
+  _lock = [NSRecursiveLock new];
 
   // Create an empty search list
   _searchList = [[NSMutableArray alloc] initWithCapacity: 10];
@@ -720,25 +799,8 @@ static NSString	*pathForUser(NSString *user)
   _persDomains = [[NSMutableDictionaryClass alloc] initWithCapacity: 10];
   if ([self synchronize] == NO)
     {
-      NSRunLoop	*runLoop = [NSRunLoop currentRunLoop];
-      BOOL	done = NO;
-      int	attempts;
-
-      // Retry for a couple of seconds in case we are locked out.
-      for (attempts = 0; done == NO && attempts < 10; attempts++)
-	{
-	  [runLoop runMode: NSDefaultRunLoopMode
-		beforeDate: [NSDate dateWithTimeIntervalSinceNow: 0.2]];
-	  if ([self synchronize] == YES)
-	    {
-	      done = YES;
-	    }
-        }
-      if (done == NO)
-	{
-          DESTROY(self);
-          return self;
-        }
+      DESTROY(self);
+      return self;
     }
 
   // Check and if not existent add the Application and the Global domains
@@ -765,20 +827,24 @@ static NSString	*pathForUser(NSString *user)
     setObject: [NSMutableDictionaryClass dictionaryWithCapacity: 10]
     forKey: NSRegistrationDomain];
 
-  _lock = [NSRecursiveLock new];
+  [[NSNotificationCenter defaultCenter] addObserver: self
+           selector: @selector(synchronize)
+               name: @"GSHousekeeping"
+             object: nil];
+
   return self;
 }
 
 - (void) dealloc
 {
-  if (_tickingTimer)
-    [_tickingTimer invalidate];
+  [[NSNotificationCenter defaultCenter] removeObserver: self];
   RELEASE(_lastSync);
   RELEASE(_searchList);
   RELEASE(_persDomains);
   RELEASE(_tempDomains);
   RELEASE(_changedDomains);
   RELEASE(_dictionaryRep);
+  RELEASE(_fileLock);
   RELEASE(_lock);
   [super dealloc];
 }
@@ -811,14 +877,23 @@ static NSString	*pathForUser(NSString *user)
 
 /**
  * Looks up a value for a specified default using -objectForKey:
- * and checks that it is a boolean.  Returns NO if it is not.
+ * and returns its boolean representation.<br />
+ * Returns NO if it is not a boolean.<br />
+ * The text 'yes' or 'true' or any non zero numeric value is considered
+ * to be a boolean YES.  Other string values are NO.<br />
+ * NB. This differs slightly from the documented behavior for MacOS-X
+ * (August 2002) in that the GNUstep version accepts the string 'TRUE'
+ * as equivalent to 'YES'.
  */
 - (BOOL) boolForKey: (NSString*)defaultName
 {
-  id	obj = [self stringForKey: defaultName];
+  id	obj = [self objectForKey: defaultName];
 
-  if (obj != nil)
-    return [obj boolValue];
+  if (obj != nil && ([obj isKindOfClass: NSStringClass]
+    || [obj isKindOfClass: NSNumberClass]))
+    {
+      return [obj boolValue];
+    }
   return NO;
 }
 
@@ -844,7 +919,9 @@ static NSString	*pathForUser(NSString *user)
   id	obj = [self objectForKey: defaultName];
 
   if (obj != nil && [obj isKindOfClass: NSDictionaryClass])
-    return obj;
+    {
+      return obj;
+    }
   return nil;
 }
 
@@ -854,10 +931,13 @@ static NSString	*pathForUser(NSString *user)
  */
 - (float) floatForKey: (NSString*)defaultName
 {
-  id	obj = [self stringForKey: defaultName];
+  id	obj = [self objectForKey: defaultName];
 
-  if (obj != nil)
-    return [obj floatValue];
+  if (obj != nil && ([obj isKindOfClass: NSStringClass]
+    || [obj isKindOfClass: NSNumberClass]))
+    {
+      return [obj floatValue];
+    }
   return 0.0;
 }
 
@@ -867,10 +947,13 @@ static NSString	*pathForUser(NSString *user)
  */
 - (int) integerForKey: (NSString*)defaultName
 {
-  id	obj = [self stringForKey: defaultName];
+  id	obj = [self objectForKey: defaultName];
 
-  if (obj != nil)
-    return [obj intValue];
+  if (obj != nil && ([obj isKindOfClass: NSStringClass]
+    || [obj isKindOfClass: NSNumberClass]))
+    {
+      return [obj intValue];
+    }
   return 0;
 }
 
@@ -940,19 +1023,18 @@ static NSString	*pathForUser(NSString *user)
       [self __changePersistentDomain: processName];
     }
   [_lock unlock];
-  return;
 }
 
 /**
- * Sets a boolean value for defaultName in the application domain.
- * <br />Calls -setObject:forKey: to make the change.
+ * Sets a boolean value for defaultName in the application domain.<br />
+ * The boolean value is stored as a string - either YES or NO.
+ * Calls -setObject:forKey: to make the change.
  */
 - (void) setBool: (BOOL)value forKey: (NSString*)defaultName
 {
-  id	obj = (value)?@"YES": @"NO";
+  NSNumber	*n = [NSNumberClass numberWithBool: value];
 
-  [self setObject: obj forKey: defaultName];
-  return;
+  [self setObject: n forKey: defaultName];
 }
 
 /**
@@ -961,11 +1043,9 @@ static NSString	*pathForUser(NSString *user)
  */
 - (void) setFloat: (float)value forKey: (NSString*)defaultName
 {
-  char	buf[32];
+  NSNumber	*n = [NSNumberClass numberWithFloat: value];
 
-  sprintf(buf,"%g",value);
-  [self setObject: [NSStringClass stringWithCString: buf] forKey: defaultName];
-  return;
+  [self setObject: n forKey: defaultName];
 }
 
 /**
@@ -974,43 +1054,113 @@ static NSString	*pathForUser(NSString *user)
  */
 - (void) setInteger: (int)value forKey: (NSString*)defaultName
 {
-  char	buf[32];
+  NSNumber	*n = [NSNumberClass numberWithInt: value];
 
-  sprintf(buf,"%d",value);
-  [self setObject: [NSStringClass stringWithCString: buf] forKey: defaultName];
-  return;
+  [self setObject: n forKey: defaultName];
+}
+
+static BOOL isPlistObject(id o)
+{
+  if ([o isKindOfClass: NSStringClass] == YES)
+    {
+      return YES;
+    }
+  if ([o isKindOfClass: NSDataClass] == YES)
+    {
+      return YES;
+    }
+  if ([o isKindOfClass: NSDateClass] == YES)
+    {
+      return YES;
+    }
+  if ([o isKindOfClass: NSNumberClass] == YES)
+    {
+      return YES;
+    }
+  if ([o isKindOfClass: NSArrayClass] == YES)
+    {
+      NSEnumerator	*e = [o objectEnumerator];
+      id		tmp;
+
+      while ((tmp = [e nextObject]) != nil)
+	{
+	  if (isPlistObject(tmp) == NO)
+	    {
+	      return NO;
+	    }
+	}
+      return YES;
+    }
+  if ([o isKindOfClass: NSDictionaryClass] == YES)
+    {
+      NSEnumerator	*e = [o keyEnumerator];
+      id		tmp;
+
+      while ((tmp = [e nextObject]) != nil)
+	{
+	  if (isPlistObject(tmp) == NO)
+	    {
+	      return NO;
+	    }
+	  tmp = [o objectForKey: tmp];
+	  if (isPlistObject(tmp) == NO)
+	    {
+	      return NO;
+	    }
+	}
+      return YES;
+    }
+  return NO;
 }
 
 /**
- * Sets an object value for defaultName in the application domain.
- * <br />Causes a NSUserDefaultsDidChangeNotification to be posted
+ * Sets an object value for defaultName in the application domain.<br />
+ * The defaultName must be a non-empty string.<br />
+ * The value must be an instance of one of the [NSString-propertyList]
+ * classes.<br />
+ * <p>Causes a NSUserDefaultsDidChangeNotification to be posted
  * if this is the first change to a persistent-domain since the
  * last -synchronize.
+ * </p>
  */
 - (void) setObject: (id)value forKey: (NSString*)defaultName
 {
-  if (value && defaultName && ([defaultName length] > 0))
-    {
-      NSMutableDictionary	*dict;
-      id			obj;
+  NSMutableDictionary	*dict;
+  id			obj;
 
-      [_lock lock];
-      obj = [_persDomains objectForKey: processName];
-      if ([obj isKindOfClass: NSMutableDictionaryClass] == YES)
-	{
-	  dict = obj;
-	}
-      else
-	{
-	  dict = [obj mutableCopy];
-	  [_persDomains setObject: dict forKey: processName];
-	  RELEASE(dict);
-	}
-      [dict setObject: value forKey: defaultName];
-      [self __changePersistentDomain: processName];
-      [_lock unlock];
+  if ([defaultName isKindOfClass: [NSString class]] == NO
+    || [defaultName length] == 0)
+    {
+      [NSException raise: NSInvalidArgumentException
+	format: @"attempt to set object with bad key (%@)", defaultName];
     }
-  return;
+  if (value == nil)
+    {
+      [NSException raise: NSInvalidArgumentException
+	format: @"attempt to set nil object for key (%@)", defaultName];
+    }
+  if (isPlistObject(value) == NO)
+    {
+      [NSException raise: NSInvalidArgumentException
+	format: @"attempt to set non property list object for key (%@)",
+	defaultName];
+    }
+
+  [_lock lock];
+  obj = [_persDomains objectForKey: processName];
+  if ([obj isKindOfClass: NSMutableDictionaryClass] == YES)
+    {
+      dict = obj;
+    }
+  else
+    {
+      dict = [obj mutableCopy];
+      [_persDomains setObject: dict forKey: processName];
+      RELEASE(dict);
+    }
+  [dict setObject: value forKey: defaultName];
+  [self __changePersistentDomain: processName];
+  [_lock unlock];
 }
 
 /**
@@ -1054,6 +1204,12 @@ static NSString	*pathForUser(NSString *user)
 /*************************************************************************
  *** Returning the Search List
  *************************************************************************/
+
+/**
+ * Returns an array listing the domains searched in order to look up
+ * a value in the defaults system.  The order of the names in the
+ * array is the order in which the domains are searched.
+ */
 - (NSArray*) searchList
 {
   NSArray	*copy;
@@ -1064,6 +1220,12 @@ static NSString	*pathForUser(NSString *user)
   return AUTORELEASE(copy);
 }
 
+/**
+ * Sets the list of the domains searched in order to look up
+ * a value in the defaults system.  The order of the names in the
+ * array is the order in which the domains are searched.<br />
+ * On lookup, the first match is used.
+ */
 - (void) setSearchList: (NSArray*)newList
 {
   [_lock lock];
@@ -1115,12 +1277,13 @@ static NSString	*pathForUser(NSString *user)
       [self __changePersistentDomain: domainName];
     }
   [_lock unlock];
-  return;
 }
 
 /**
- * Replaces the persistent-domain specified by domainname with
+ * Replaces the persistent-domain specified by domainName with
  * domain ... a dictionary containing keys and defaults values.
+ * <br />Raises an NSInvalidArgumentException if domainName already
+ * exists as a volatile-domain.
  * <br />Causes a NSUserDefaultsDidChangeNotification to be posted
  * if this is the first change to a persistent-domain since the
  * last -synchronize.
@@ -1128,23 +1291,21 @@ static NSString	*pathForUser(NSString *user)
 - (void) setPersistentDomain: (NSDictionary*)domain
 		     forName: (NSString*)domainName
 {
-  id	dict;
+  NSDictionary	*dict;
 
   [_lock lock];
   dict = [_tempDomains objectForKey: domainName];
-  if (dict)
+  if (dict != nil)
     {
       [_lock unlock];
       [NSException raise: NSInvalidArgumentException
-		   format: @"Persistant domain %@ already exists", domainName];
-      return;
+		  format: @"a volatile domain called %@ exists", domainName];
     }
   domain = [domain mutableCopy];
   [_persDomains setObject: domain forKey: domainName];
   RELEASE(domain);
   [self __changePersistentDomain: domainName];
   [_lock unlock];
-  return;
 }
 
 /**
@@ -1159,27 +1320,23 @@ static NSString	*pathForUser(NSString *user)
   NSFileManager		*mgr = [NSFileManager defaultManager];
   NSMutableDictionary	*newDict;
   NSDictionary		*attr;
+  NSDate		*started = [NSDateClass date];
+  unsigned long		desired;
+  unsigned long		attributes;
+  static BOOL		isLocked = NO;
+  BOOL			wasLocked;
 
   [_lock lock];
 
-  if (_tickingTimer == nil)
-    {
-      _tickingTimer = [NSTimer scheduledTimerWithTimeInterval: 30
-	       target: self
-	       selector: @selector(__timerTicked:)
-	       userInfo: nil
-	       repeats: NO];
-    }
-
   /*
    *	If we haven't changed anything, we only need to synchronise if
    *	the on-disk database has been changed by someone else.
    */
   attr = [mgr fileAttributesAtPath: _defaultsDatabase
 		      traverseLink: YES];
-  if (_changedDomains == NO)
+  if (_changedDomains == nil)
     {
-      BOOL		wantRead = NO;
+      BOOL	wantRead = NO;
 
       if (_lastSync == nil)
 	{
@@ -1195,8 +1352,12 @@ static NSString	*pathForUser(NSString *user)
 	    {
 	      NSDate	*mod;
 
+	      /*
+	       * If the database was modified since the last synchronisation
+	       * we need to read it.
+	       */
 	      mod = [attr objectForKey: NSFileModificationDate];
-	      if ([_lastSync earlierDate: mod] != _lastSync)
+	      if (mod != nil && [_lastSync laterDate: mod] != _lastSync)
 		{
 		  wantRead = YES;
 		}
@@ -1209,105 +1370,153 @@ static NSString	*pathForUser(NSString *user)
 	}
     }
 
+  wasLocked = isLocked;
+  if (isLocked == NO && _fileLock != nil)
+    {
+      while ([_fileLock tryLock] == NO)
+	{
+	  CREATE_AUTORELEASE_POOL(arp);
+	  NSDate	*when;
+	  NSDate	*lockDate;
+
+	  lockDate = [_fileLock lockDate];
+	  when = [NSDateClass dateWithTimeIntervalSinceNow: 0.1];
+
+	  /*
+	   * In case we have tried and failed to break the lock,
+	   * we give up after a while ... 16 seconds should give
+	   * us three lock breaks if we do them at 5 second
+	   * intervals.
+	   */
+	  if ([when timeIntervalSinceDate: started] > 16.0)
+	    {
+	      NSLog(@"Failed to lock user defaults database even after "
+		@"breaking old locks!");
+	      [_lock unlock];
+	      return NO;
+	    }
+
+	  /*
+	   * If lockDate is nil, we should be able to lock again ... but we
+	   * wait a little anyway ... so that in the case of a locking
+	   * problem we do an idle wait rather than a busy one.
+	   */ 
+	  if (lockDate != nil && [when timeIntervalSinceDate: lockDate] > 5.0)
+	    {
+	      [_fileLock breakLock];
+	    }
+	  else
+	    {
+	      [NSThread sleepUntilDate: when];
+	    }
+	  RELEASE(arp);
+	}
+      isLocked = YES;
+    }
+
+  /*
+   * Re-fetch database attributes in cased they changed while obtaining lock.
+   */
+  attr = [mgr fileAttributesAtPath: _defaultsDatabase
+		      traverseLink: YES];
+
   DESTROY(_dictionaryRep);
 
   // Read the persistent data from the stored database
-  if (attr != nil)
+  if (attr == nil)
+    {
+      newDict = [[NSMutableDictionaryClass allocWithZone: [self zone]]
+	initWithCapacity: 1];
+      if (_fileLock != nil)
+	{
+	  NSLog(@"Creating defaults database file %@", _defaultsDatabase);
+	  [newDict writeToFile: _defaultsDatabase atomically: YES];
+	  attr = [mgr fileAttributesAtPath: _defaultsDatabase
+			      traverseLink: YES];
+	}
+    }
+  else
     {
-      unsigned long desired;
-      unsigned long attributes;
-
       newDict = [[NSMutableDictionaryClass allocWithZone: [self zone]]
         initWithContentsOfFile: _defaultsDatabase];
       if (newDict == nil)
 	{
 	  NSLog(@"Unable to load defaults from '%@'", _defaultsDatabase);
+	  if (wasLocked == NO)
+	    {
+	      [_fileLock unlock];
+	      isLocked = NO;
+	    }
 	  [_lock unlock];
 	  return NO;
 	}
-
-      attributes = [attr filePosixPermissions];
-      // We enforce the permission mode 0600 on the defaults database
-#if	!(defined(S_IRUSR) && defined(S_IWUSR))
-      desired = 0600;
-#else
-      desired = (S_IRUSR|S_IWUSR);
-#endif
-      if (attributes != desired)
-	{
-	  NSMutableDictionary	*enforced_attributes;
-	  NSNumber		*permissions;
-
-	  enforced_attributes = [NSMutableDictionary dictionaryWithDictionary:
-	    [mgr fileAttributesAtPath: _defaultsDatabase traverseLink: YES]];
-
-	  permissions = [NSNumber numberWithUnsignedLong: desired];
-	  [enforced_attributes setObject: permissions
-				  forKey: NSFilePosixPermissions];
-
-	  [mgr changeFileAttributes: enforced_attributes
-			     atPath: _defaultsDatabase];
-	}
     }
-  else
+
+  /*
+   * We enforce the permission mode 0600 on the defaults database
+   */
+  attributes = [attr filePosixPermissions];
+#if	!(defined(S_IRUSR) && defined(S_IWUSR))
+  desired = 0600;
+#else
+  desired = (S_IRUSR|S_IWUSR);
+#endif
+  if (attributes != desired)
     {
-      unsigned long	desired;
-      NSNumber		*permissions;
+      NSMutableDictionary	*enforced_attributes;
+      NSNumber			*permissions;
 
-      // We enforce the permission mode 0600 on the defaults database
-#if	!(defined(S_IRUSR) && defined(S_IWUSR))
-      desired = 0600;
-#else
-      desired = (S_IRUSR|S_IWUSR);
-#endif
-      permissions = [NSNumber numberWithUnsignedLong: desired];
-      attr = [NSDictionary dictionaryWithObjectsAndKeys:
-	NSUserName(), NSFileOwnerAccountName,
-	permissions, NSFilePosixPermissions,
-	nil];
-      NSLog(@"Creating defaults database file %@", _defaultsDatabase);
-      [mgr createFileAtPath: _defaultsDatabase
-		   contents: nil
-		 attributes: attr];
-      newDict = [[NSMutableDictionaryClass allocWithZone: [self zone]]
-		  initWithCapacity: 1];
-      [newDict writeToFile: _defaultsDatabase atomically: YES];
+      enforced_attributes = [NSMutableDictionary dictionaryWithDictionary:
+	[mgr fileAttributesAtPath: _defaultsDatabase traverseLink: YES]];
+
+      permissions = [NSNumberClass numberWithUnsignedLong: desired];
+      [enforced_attributes setObject: permissions
+			      forKey: NSFilePosixPermissions];
+
+      [mgr changeFileAttributes: enforced_attributes
+			 atPath: _defaultsDatabase];
     }
 
-  if (_changedDomains)
+  if (_changedDomains != nil)
     {           // Synchronize both dictionaries
       NSEnumerator	*enumerator = [_changedDomains objectEnumerator];
-      IMP		nextImp;
-      IMP		pImp;
-      id		obj, dict;
+      NSString		*domainName;
+      NSDictionary	*domain;
 
-      nextImp = [enumerator methodForSelector: nextObjectSel];
-      pImp = [_persDomains methodForSelector: objectForKeySel];
-      while ((obj = (*nextImp)(enumerator, nextObjectSel)) != nil)
+      DESTROY(_changedDomains);	// Retained by enumerator.
+      while ((domainName = [enumerator nextObject]) != nil)
 	{
-	  dict = (*pImp)(_persDomains, objectForKeySel, obj);
-	  if (dict)       // Domain was added or changed
+	  domain = [_persDomains objectForKey: domainName];
+	  if (domain != nil)	// Domain was added or changed
 	    {
-	      [newDict setObject: dict forKey: obj];
+	      [newDict setObject: domain forKey: domainName];
 	    }
-	  else            // Domain was removed
+	  else			// Domain was removed
 	    {
-	      [newDict removeObjectForKey: obj];
+	      [newDict removeObjectForKey: domainName];
 	    }
 	}
       RELEASE(_persDomains);
       _persDomains = newDict;
-      // Save the changes
-      if (![_persDomains writeToFile: _defaultsDatabase atomically: YES])
+      // Save the changes unless we are in read-only mode.
+      if (_fileLock != nil)
 	{
-	  [_lock unlock];
-	  return NO;
+	  if (![_persDomains writeToFile: _defaultsDatabase atomically: YES])
+	    {
+	      if (wasLocked == NO)
+		{
+		  [_fileLock unlock];
+		  isLocked = NO;
+		}
+	      [_lock unlock];
+	      return NO;
+	    }
 	}
-      ASSIGN(_lastSync, [NSDate date]);
+      ASSIGN(_lastSync, [NSDateClass date]);
     }
   else
     {
-      ASSIGN(_lastSync, [NSDate date]);
+      ASSIGN(_lastSync, [NSDateClass date]);
       if ([_persDomains isEqual: newDict] == NO)
 	{
 	  RELEASE(_persDomains);
@@ -1323,14 +1532,20 @@ static NSString	*pathForUser(NSString *user)
 	}
     }
 
+  if (wasLocked == NO)
+    {
+      [_fileLock unlock];
+      isLocked = NO;
+    }
   [_lock unlock];
   return YES;
 }
 
 
-/*************************************************************************
- *** Maintaining Volatile Domains
- *************************************************************************/
+/**
+ * Removes the volatile domain specified by domainName from the
+ * user defaults.
+ */
 - (void) removeVolatileDomainForName: (NSString*)domainName
 {
   [_lock lock];
@@ -1339,6 +1554,12 @@ static NSString	*pathForUser(NSString *user)
   [_lock unlock];
 }
 
+/**
+ * Sets the volatile-domain specified by domainName to
+ * domain ... a dictionary containing keys and defaults values.<br />
+ * Raises an NSInvalidArgumentException if domainName already
+ * exists as either a volatile-domain or a persistent-domain.
+ */
 - (void) setVolatileDomain: (NSDictionary*)domain
 		   forName: (NSString*)domainName
 {
@@ -1346,21 +1567,30 @@ static NSString	*pathForUser(NSString *user)
 
   [_lock lock];
   dict = [_persDomains objectForKey: domainName];
-  if (dict)
+  if (dict != nil)
     {
       [_lock unlock];
       [NSException raise: NSInvalidArgumentException
-		  format: @"Volatile domain %@ already exists", domainName];
-      return;
+		  format: @"a persistent domain called %@ exists", domainName];
     }
+  dict = [_tempDomains objectForKey: domainName];
+  if (dict != nil)
+    {
+      [_lock unlock];
+      [NSException raise: NSInvalidArgumentException
+		  format: @"the volatile domain %@ already exists", domainName];
+    }
+
   DESTROY(_dictionaryRep);
   domain = [domain mutableCopy];
   [_tempDomains setObject: domain forKey: domainName];
   RELEASE(domain);
   [_lock unlock];
-  return;
 }
 
+/**
+ * Returns the volatile domain specified by domainName.
+ */
 - (NSDictionary*) volatileDomainForName: (NSString*)domainName
 {
   NSDictionary	*copy;
@@ -1486,7 +1716,6 @@ static NSString	*pathForUser(NSString *user)
   [_searchList addObject: NSRegistrationDomain];
 
   [_lock unlock];
-  return;
 }
 
 - (NSDictionary*) __createArgumentDictionary
@@ -1506,7 +1735,7 @@ static NSString	*pathForUser(NSString *user)
 
   while (!done)
     {
-      if ([key hasPrefix: @"-"])
+      if ([key hasPrefix: @"-"] == YES && [key isEqual: @"-"] == NO)
 	{
 	  NSString	*old = nil;
 
@@ -1529,7 +1758,7 @@ static NSString	*pathForUser(NSString *user)
 	      done = YES;
 	      continue;
 	    }
-	  else if ([val hasPrefix: @"-"] == YES)
+	  else if ([val hasPrefix: @"-"] == YES && [val isEqual: @"-"] == NO)
 	    {  // Yet another argument
 	      [argDict setObject: @"" forKey: key];		// arg is empty.
 	      if (old != nil)
@@ -1580,42 +1809,22 @@ static NSString	*pathForUser(NSString *user)
 
 - (void) __changePersistentDomain: (NSString*)domainName
 {
-  NSEnumerator	*enumerator = nil;
-  IMP		nImp;
-  id		obj;
-
   [_lock lock];
   DESTROY(_dictionaryRep);
-  if (!_changedDomains)
+  if (_changedDomains == nil)
     {
-      _changedDomains = [[NSMutableArray alloc] initWithCapacity: 5];
+      _changedDomains = [[NSMutableArray alloc] initWithObjects: &domainName
+							  count: 1];
       updateCache(self);
       [[NSNotificationCenter defaultCenter]
 	postNotificationName: NSUserDefaultsDidChangeNotification
 		      object: self];
     }
-
-  enumerator = [_changedDomains objectEnumerator];
-  nImp = [enumerator methodForSelector: nextObjectSel];
-  while ((obj = (*nImp)(enumerator, nextObjectSel)) != nil)
+  else if ([_changedDomains containsObject: domainName] == NO)
     {
-      if ([obj isEqualToString: domainName])
-	{
-	  [_lock unlock];
-	  return;
-	}
+      [_changedDomains addObject: domainName];
     }
-  [_changedDomains addObject: domainName];
   [_lock unlock];
-  return;
-}
-
-- (void) __timerTicked: (NSTimer*)tim
-{
-  if (tim == _tickingTimer)
-    _tickingTimer = nil;
-
-  [self synchronize];
 }
 @end
 
