@@ -23,20 +23,20 @@
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA.
 
    <title>NSValue class reference</title>
-   $Date: 2002/02/21 13:17:39 $ $Revision: 1.30 $
+   $Date: 2003/07/31 23:49:32 $ $Revision: 1.32 $
 */
 
-#include <config.h>
-#include <base/preface.h>
-#include <Foundation/NSValue.h>
-#include <Foundation/NSCoder.h>
-#include <Foundation/NSDictionary.h>
-#include <Foundation/NSZone.h>
-#include <Foundation/NSException.h>
-#include <Foundation/NSMapTable.h>
-#include <Foundation/NSLock.h>
-#include <Foundation/NSData.h>
-#include <Foundation/NSDebug.h>
+#include "config.h"
+#include "GNUstepBase/preface.h"
+#include "Foundation/NSValue.h"
+#include "Foundation/NSCoder.h"
+#include "Foundation/NSDictionary.h"
+#include "Foundation/NSZone.h"
+#include "Foundation/NSException.h"
+#include "Foundation/NSMapTable.h"
+#include "Foundation/NSLock.h"
+#include "Foundation/NSData.h"
+#include "Foundation/NSDebug.h"
 
 @interface	GSPlaceholderValue : NSValue
 @end
