@@ -22,7 +22,7 @@
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA.
 
    <title>NSZone class reference</title>
-   $Date: 2001/12/18 16:54:15 $ $Revision: 1.46 $
+   $Date: 2003/07/31 23:49:32 $ $Revision: 1.50 $
 */
 
 /*  Design goals:
@@ -85,29 +85,26 @@
 
 #define IN_NSZONE_M 1
 
-#include <config.h>
-#include <base/preface.h>
+#include "config.h"
+#include "GNUstepBase/preface.h"
 #include <stddef.h>
 #include <string.h>
-#include <Foundation/NSObjCRuntime.h>
-#include <Foundation/NSException.h>
-#include <Foundation/NSString.h>
-#include <Foundation/NSZone.h>
-#include <Foundation/NSLock.h>
-
+#include "Foundation/NSObjCRuntime.h"
+#include "Foundation/NSException.h"
+#include "Foundation/NSString.h"
+#include "Foundation/NSZone.h"
+#include "Foundation/NSLock.h"
 
 /*
- *	Try to get more memory - the normal process has failed.
- *	If we can't do anything, bomb out.
+ * Try to get more memory - the normal process has failed.
+ * If we can't do anything, just return a null pointer.
+ * Try to do some logging if possible.
  */
 void *
 GSOutOfMemory(size_t size, BOOL retry)
 {
-  /*
-   *	It would be nice to raise an exception - but how can we if there is
-   *	no memory available?
-   */
-  abort();
+  fprintf(stderr, "GSOutOfMemory ... wanting %u bytes.\n", size);
+  return 0;
 }
 
 #if	GS_WITH_GC == 0
