@@ -22,12 +22,12 @@
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA.
    */ 
 
-#include <config.h>
+#include "config.h"
 #include <stdlib.h>
 #include "callframe.h"
-#include <Foundation/NSException.h>
-#include <Foundation/NSData.h>
-#include <base/GSInvocation.h>
+#include "Foundation/NSException.h"
+#include "Foundation/NSData.h"
+#include "GSInvocation.h"
 
 #if defined(ALPHA) || (defined(MIPS) && (_MIPS_SIM == _ABIN32))
 typedef long long smallret_t;
@@ -35,11 +35,6 @@ typedef long long smallret_t;
 typedef int smallret_t;
 #endif
 
-/* Return the number of arguments that the method MTH expects.  Note
-   that all methods need two implicit arguments `self' and `_cmd'.  
-   From mframe.m */
-extern int method_types_get_number_of_arguments (const char *type);
-
 extern BOOL sel_types_match(const char* t1, const char* t2);
 
 callframe_t *
@@ -47,7 +42,7 @@ callframe_from_info (NSArgumentInfo *info, int numargs, void **retval)
 {
   unsigned      size = sizeof(callframe_t);
   unsigned      align = __alignof(double);
-  unsigned      offset;
+  unsigned      offset = 0;
   void          *buf;
   int           i;
   callframe_t   *cframe;
@@ -488,7 +483,7 @@ callframe_do_call (DOContext *ctxt,
 	     so we can see what it is a pointer to. */
 	  tmptype++;
 	  ctxt->type = tmptype;
-	  ctxt->datum = *(void**)ctxt->datum;
+	  ctxt->datum = *(void**)retval;
 	}
       else
 	{
@@ -565,7 +560,7 @@ callframe_do_call (DOContext *ctxt,
    sets up the invocation with the return value, and updates the
    pass-by-reference arguments.  
 
-   The callback function is finally called with the 'type' set to a nul pointer
+   The callback function is finally called with the 'type' set to a null pointer
    to tell it that the return value and all return parameters have been
    dealt with.  This permits the function to do any tidying up necessary.  */
 
@@ -679,7 +674,7 @@ callframe_build_return (NSInvocation *inv,
 	{
 	  /* Step through all the arguments, finding the ones that were
 	     passed by reference. */
-      for (tmptype = objc_skip_argspec (tmptype), argnum = 0;
+	  for (tmptype = objc_skip_argspec (tmptype), argnum = 0;
 	   *tmptype != '\0';
 	   tmptype = objc_skip_argspec (tmptype), argnum++)
 	    {
@@ -699,9 +694,10 @@ callframe_build_return (NSInvocation *inv,
 	      ctxt->flags = flags;
 
 	      if (*tmptype == _C_PTR
-		  && ((flags & _F_OUT) || !(flags & _F_IN)))
+		&& ((flags & _F_OUT) || !(flags & _F_IN)))
 		{
 		  void *ptr;
+
 		  /* The argument is a pointer (to a non-char), and
 		     the pointer's value is qualified as an OUT
 		     parameter, or it not explicitly qualified as an
@@ -710,11 +706,12 @@ callframe_build_return (NSInvocation *inv,
 		  tmptype++;
 		  ctxt->type = tmptype;
 
-		  (*decoder) (ctxt);
-		  /* Copy the pointed-to data back to the original
-		     pointer */
+                  /* Use the original pointer to find the buffer
+                   * to store the returned data */
 		  [inv getArgument: &ptr atIndex: argnum];
-		  memcpy(ptr, datum, objc_sizeof_type(tmptype));
+		  ctxt->datum = ptr;
+
+		  (*decoder) (ctxt);
 		}
 	      else if (*tmptype == _C_CHARPTR
 		&& ((flags & _F_OUT) || !(flags & _F_IN)))
