@@ -31,21 +31,21 @@
 /* Remove `inline' nested functions if they crash your compiler */
 //#define inline 
 
-#include <config.h>
-#include <base/preface.h>
+#include "config.h"
+#include "GNUstepBase/preface.h"
 #include <mframe.h>
-#include <Foundation/NSObjCRuntime.h>
-#include <Foundation/NSData.h>
-#include <Foundation/NSException.h>
+#include "Foundation/NSObjCRuntime.h"
+#include "Foundation/NSData.h"
+#include "Foundation/NSException.h"
 #include <stdio.h>
 #include <stdlib.h>
 #include <ctype.h>
 
 /* Deal with strrchr: */
-#if STDC_HEADERS || HAVE_STRING_H
+#if STDC_HEADERS || defined(HAVE_STRING_H)
 #include <string.h>
 /* An ANSI string.h and pre-ANSI memory.h might conflict.  */
-#if !STDC_HEADERS && HAVE_MEMORY_H
+#if !STDC_HEADERS && defined(HAVE_MEMORY_H)
 #include <memory.h>
 #endif /* not STDC_HEADERS and HAVE_MEMORY_H */
 #define index strchr
@@ -205,6 +205,7 @@ mframe_next_arg(const char *typePtr, NSArgumentInfo *info)
 {
   NSArgumentInfo	local;
   BOOL			flag;
+  BOOL			negative = NO;
 
   if (info == 0)
     {
@@ -330,9 +331,7 @@ mframe_next_arg(const char *typePtr, NSArgumentInfo *info)
 	  }
 	else
 	  {
-	    typePtr = mframe_next_arg(typePtr, &local);
-	    info->isReg = local.isReg;
-	    info->offset = local.offset;
+	    typePtr = objc_skip_typespec(typePtr);
 	  }
 	break;
 
@@ -359,20 +358,18 @@ mframe_next_arg(const char *typePtr, NSArgumentInfo *info)
 
       case _C_STRUCT_B:
 	{
-	  struct { int x; double y; } fooalign;
-	  int acc_size = 0;
-	  int acc_align = __alignof__(fooalign);
+	  unsigned int acc_size = 0;
+	  unsigned int def_align = objc_alignof_type(typePtr-1);
+	  unsigned int acc_align = def_align;
+	  const char	*ptr = typePtr;
 
 	  /*
 	   *	Skip "<name>=" stuff.
 	   */
-	  while (*typePtr != _C_STRUCT_E)
-	    {
-	      if (*typePtr++ == '=')
-		{
-		  break;
-		}
-	    }
+	  while (*ptr != _C_STRUCT_E && *ptr != '=') ptr++;
+	  if (*ptr == '=') typePtr = ptr;
+	  typePtr++;
+
 	  /*
 	   *	Base structure alignment on first element.
 	   */
@@ -385,7 +382,7 @@ mframe_next_arg(const char *typePtr, NSArgumentInfo *info)
 		}
 	      acc_size = ROUND(acc_size, local.align);
 	      acc_size += local.size;
-	      acc_align = MAX(local.align, __alignof__(fooalign));
+	      acc_align = MAX(local.align, def_align);
 	    }
 	  /*
 	   *	Continue accumulating structure size.
@@ -408,8 +405,8 @@ mframe_next_arg(const char *typePtr, NSArgumentInfo *info)
 
       case _C_UNION_B:
 	{
-	  int	max_size = 0;
-	  int	max_align = 0;
+	  unsigned int	max_size = 0;
+	  unsigned int	max_align = 0;
 
 	  /*
 	   *	Skip "<name>=" stuff.
@@ -452,47 +449,37 @@ mframe_next_arg(const char *typePtr, NSArgumentInfo *info)
     }
 
   /*
-   *	If we had a pointer argument, we will already have gathered
-   *	(and skipped past) the argframe offset information - so we
-   *	don't need to (and can't) do it here.
+   *	May tell the caller if the item is stored in a register.
    */
-  if (info->type[0] != _C_PTR || info->type[1] == '?')
+  if (*typePtr == '+')
     {
-      BOOL	negative = NO;
-
-      /*
-       *	May tell the caller if the item is stored in a register.
-       */
-      if (*typePtr == '+')
-	{
-	  typePtr++;
-	  info->isReg = YES;
-	}
-      else
-	{
-	  info->isReg = NO;
-	}
-      /*
-       * Cope with negative offsets.
-       */
-      if (*typePtr == '-')
-	{
-	  typePtr++;
-	  negative = YES;
-	}
-      /*
-       *	May tell the caller what the stack/register offset is for
-       *	this argument.
-       */
-      info->offset = 0;
-      while (isdigit(*typePtr))
-	{
-	  info->offset = info->offset * 10 + (*typePtr++ - '0');
-	}
-      if (negative == YES)
-	{
-	  info->offset = -info->offset;
-	}
+      typePtr++;
+      info->isReg = YES;
+    }
+  else
+    {
+      info->isReg = NO;
+    }
+  /*
+   * Cope with negative offsets.
+   */
+  if (*typePtr == '-')
+    {
+      typePtr++;
+      negative = YES;
+    }
+  /*
+   *	May tell the caller what the stack/register offset is for
+   *	this argument.
+   */
+  info->offset = 0;
+  while (isdigit(*typePtr))
+    {
+      info->offset = info->offset * 10 + (*typePtr++ - '0');
+    }
+  if (negative == YES)
+    {
+      info->offset = -info->offset;
     }
 
   return typePtr;
@@ -1245,12 +1232,14 @@ mframe_do_call (DOContext *ctxt,
   return;
 }
 
+#if 0
   /* For returning structures etc */
   typedef struct { id many[8];} __big;
 static  __big return_block (void* data)
     {
       return *(__big*)data;
     }
+#endif
   /* For returning a char (or unsigned char) */
 static  char return_char (char data)
     {
@@ -1271,11 +1260,13 @@ static  short return_short (short data)
     {
       return data;
     }
+#if 0
 static  retval_t apply_block(void* data)
     {
       void* args = __builtin_apply_args();
       return __builtin_apply((apply_t)return_block, args, sizeof(void*));
     }
+#endif
 static  retval_t apply_char(char data)
     {
       void* args = __builtin_apply_args();
@@ -1309,7 +1300,7 @@ static  retval_t apply_short(short data)
    alloca()'ed, not malloc()'ed, and therefore is on the stack and can
    be tromped-on by future function calls.
 
-   The callback function is finally called with the 'type' set to a nul pointer
+   The callback function is finally called with the 'type' set to a null pointer
    to tell it that the return value and all return parameters have been
    dealt with.  This permits the function to do any tidying up necessary.
 */
@@ -1324,7 +1315,7 @@ mframe_build_return (arglist_t argframe,
   /* A pointer to the memory that will hold the return value. */
   retval_t retframe = NULL;
   /* The size, in bytes, of memory pointed to by RETFRAME. */
-  int retsize;
+  unsigned int retsize;
   /* Which argument number are we processing now? */
   int argnum;
   /* Type qualifier flags; see <objc/objc-api.h>. */
@@ -1384,7 +1375,7 @@ mframe_build_return (arglist_t argframe,
 		/* We are returning a pointer to something. */
 		/* Increment TYPE so we can see what it is a pointer to. */
 		tmptype++;
-		retLength = objc_sizeof_type(tmptype);
+		retLength = (unsigned int)objc_sizeof_type(tmptype);
 		/* Allocate memory to hold the value we're pointing to. */
 		*(void**)retframe = 
 		  NSZoneMalloc(NSDefaultMallocZone(), retLength);
