@@ -76,11 +76,11 @@ bench_object()
   START_TIMER;
   for (i = 0; i < MAX_COUNT*10; i++)
     {
-      id i = [rootClass class];
+      [rootClass class];
     }
   END_TIMER;
   baseline = [eTime timeIntervalSinceDate: sTime];
-  PRINT_TIMER("Baseline: 10 method calls");
+  PRINT_TIMER("Baseline: 10 method calls\t\t");
 
   START_TIMER;
   for (i = 0; i < MAX_COUNT; i++)
@@ -99,7 +99,7 @@ bench_object()
       i = [NSMutableDictionary class];
     }
   END_TIMER;
-  PRINT_TIMER("Class: 10 class method calls");
+  PRINT_TIMER("Class: 10 class method calls\t\t");
   
   obj = [MyObject new];
   
@@ -110,7 +110,7 @@ bench_object()
       i = [obj self];
     }
   END_TIMER;
-  PRINT_TIMER_NO_BASELINE("Category: 10 super calls");
+  PRINT_TIMER_NO_BASELINE("Category: 10 super calls\t\t");
   
   START_TIMER;
   for (i = 0; i < MAX_COUNT; i++)
@@ -129,7 +129,7 @@ bench_object()
       i = NSClassFromString (@"NSMutableDictionary");
     }
   END_TIMER;
-  PRINT_TIMER("Function: 10 NSClassFromStr");
+  PRINT_TIMER("Function: 10 NSClassFromStr\t\t");
 
   START_TIMER;
   myZone = NSCreateZone(2048, 2048, 1);
@@ -140,7 +140,7 @@ bench_object()
     }
   NSRecycleZone(myZone);
   END_TIMER;
-  PRINT_TIMER("Function: 1 zone alloc/free");
+  PRINT_TIMER("Function: 1 zone alloc/free\t\t");
 
   START_TIMER;
   myZone = NSCreateZone(2048, 2048, 0);
@@ -151,7 +151,7 @@ bench_object()
     }
   NSRecycleZone(myZone);
   END_TIMER;
-  PRINT_TIMER("Function: 1 zone2alloc/free");
+  PRINT_TIMER("Function: 1 zone2alloc/free\t\t");
 
   myZone = NSDefaultMallocZone();
   START_TIMER;
@@ -161,7 +161,7 @@ bench_object()
       NSZoneFree(myZone, mem);
     }
   END_TIMER;
-  PRINT_TIMER("Function: 1 def alloc/free");
+  PRINT_TIMER("Function: 1 def alloc/free\t\t");
 
   START_TIMER;
   myZone = NSCreateZone(2048, 2048, 1);
@@ -172,7 +172,7 @@ bench_object()
     }
   NSRecycleZone(myZone);
   END_TIMER;
-  PRINT_TIMER("NSObject: 1 zone all/init/rel");
+  PRINT_TIMER("NSObject: 1 zone all/init/rel\t\t");
 
   START_TIMER;
   myZone = NSCreateZone(2048, 2048, 0);
@@ -183,7 +183,7 @@ bench_object()
     }
   NSRecycleZone(myZone);
   END_TIMER;
-  PRINT_TIMER("NSObject: 1 zone2all/init/rel");
+  PRINT_TIMER("NSObject: 1 zone2all/init/rel\t\t");
 
   myZone = NSDefaultMallocZone();
   START_TIMER;
@@ -193,7 +193,7 @@ bench_object()
       [obj release];
     }
   END_TIMER;
-  PRINT_TIMER("NSObject: 1 def all/init/rel");
+  PRINT_TIMER("NSObject: 1 def all/init/rel\t\t");
 
   obj = [rootClass new];
   START_TIMER;
@@ -203,7 +203,7 @@ bench_object()
       [obj release];
     }
   END_TIMER;
-  PRINT_TIMER("NSObject: 10 retain/rel");
+  PRINT_TIMER("NSObject: 10 retain/rel\t\t");
   [obj release];
 
   obj = [rootClass new];
@@ -214,16 +214,16 @@ bench_object()
       [obj retain];
     }
   END_TIMER;
-  PRINT_TIMER("NSObject: 10 autorel/ret");
+  PRINT_TIMER("NSObject: 10 autorel/ret\t\t");
   [obj release];
 
   START_TIMER;
   for (i = 0; i < MAX_COUNT*10; i++)
     {
-      BOOL dummy = [rootClass instancesRespondToSelector: @selector(hash)];
+      [rootClass instancesRespondToSelector: @selector(hash)];
     }
   END_TIMER;
-  PRINT_TIMER("ObjC: 10 inst responds to sel");
+  PRINT_TIMER("ObjC: 10 inst responds to sel\t\t");
 
   mutex = objc_mutex_allocate();
   START_TIMER;
@@ -233,7 +233,7 @@ bench_object()
       objc_mutex_unlock(mutex);
     }
   END_TIMER;
-  PRINT_TIMER("ObjC: 10 objc_mutex_lock/unl");
+  PRINT_TIMER("ObjC: 10 objc_mutex_lock/unl\t\t");
   objc_mutex_deallocate(mutex);
 
   obj = [NSLock new];
@@ -244,7 +244,7 @@ bench_object()
       [obj unlock];
     }
   END_TIMER;
-  PRINT_TIMER("NSLock: 10 lock/unlock");
+  PRINT_TIMER("NSLock: 10 lock/unlock\t\t");
   [obj release];
 
 
@@ -273,7 +273,7 @@ bench_array()
       [array addObject: strings[i/10]];
     }
   END_TIMER;
-  PRINT_TIMER("NSArray (10 addObject:)");
+  PRINT_TIMER("NSArray (10 addObject:)\t\t");
 
   START_TIMER;
   for (i = 0; i < MAX_COUNT/100; i++)
@@ -281,7 +281,7 @@ bench_array()
       [array indexOfObject: strings[i]];
     }
   END_TIMER;
-  PRINT_TIMER("NSArray (1/100 indexOfObj)");
+  PRINT_TIMER("NSArray (1/100 indexOfObj)\t\t");
 
   START_TIMER;
   for (i = 0; i < MAX_COUNT/100; i++)
@@ -289,7 +289,7 @@ bench_array()
       [array indexOfObjectIdenticalTo: strings[i]];
     }
   END_TIMER;
-  PRINT_TIMER("NSArray (1/100 indexIdent)");
+  PRINT_TIMER("NSArray (1/100 indexIdent)\t\t");
 
   START_TIMER;
   for (i = 0; i < 1; i++)
@@ -297,7 +297,7 @@ bench_array()
       [array makeObjectsPerformSelector: @selector(hash)];
     }
   END_TIMER;
-  PRINT_TIMER("NSArray (once perform)");
+  PRINT_TIMER("NSArray (once perform)\t\t");
   AUTO_END;
 }
 
@@ -332,7 +332,7 @@ bench_dict()
 	}
     }
   END_TIMER;
-  PRINT_TIMER("NSDict (1 setObject:) ");
+  PRINT_TIMER("NSDict (1 setObject:) \t\t");
 
   START_TIMER;
   for (i = 0; i < MAX_COUNT; i++)
@@ -341,28 +341,28 @@ bench_dict()
 
       for (j = 0; j < 10; j++)
         {
-          id dummy = [dict objectForKey: keys[i/10]];
+          [dict objectForKey: keys[i/10]];
         }
     }
   END_TIMER;
-  PRINT_TIMER("NSDict (10 objectFor:) ");
+  PRINT_TIMER("NSDict (10 objectFor:) \t\t");
 
   START_TIMER;
   for (i = 0; i < MAX_COUNT*10; i++)
     {
-      int dummy = [dict count];
+      [dict count];
     }
   END_TIMER;
-  PRINT_TIMER("NSDictionary (10 count)");
+  PRINT_TIMER("NSDictionary (10 count)\t\t");
 
   obj2 = [dict copy];
   START_TIMER;
   for (i = 0; i < 10; i++)
     {
-      BOOL dummy = [dict isEqual: obj2];
+      [dict isEqual: obj2];
     }
   END_TIMER;
-  PRINT_TIMER("NSDict (ten times isEqual:)");
+  PRINT_TIMER("NSDict (ten times isEqual:)\t\t");
   AUTO_END;
 }
 
@@ -378,9 +378,10 @@ bench_str()
   Class	ser = [NSSerializer class];
   Class	des = [NSDeserializer class];
   Class md = [NSMutableDictionary class];
-
   AUTO_START;
 
+  [[md new] release];
+
   plist = [NSDictionary dictionaryWithObjectsAndKeys:
 	@"Value1", @"Key1",
 	@"", @"Key2",
@@ -413,33 +414,90 @@ bench_str()
       str = [stringClass stringWithCString: "hello world"];
     }
   END_TIMER;
-  PRINT_TIMER("NSString (1 cstring:) ");
+  PRINT_TIMER("NSString (1 cstring:) \t\t");
 
   START_TIMER;
   for (i = 0; i < MAX_COUNT*10; i++)
     {
-      int dummy = [str length];
+      [str length];
     }
   END_TIMER;
-  PRINT_TIMER("NSString (10 length)   ");
+  PRINT_TIMER("NSString (10 length)   \t\t");
+
+  START_TIMER;
+  for (i = 0; i < MAX_COUNT*10; i++)
+    {
+      [str copy];
+    }
+  END_TIMER;
+  PRINT_TIMER("NSString (10 copy) <initWithCString:>   ");
+
+  str = @"ConstantString";
+  START_TIMER;
+  for (i = 0; i < MAX_COUNT*10; i++)
+    {
+      [str copy];
+    }
+  END_TIMER;
+  PRINT_TIMER("NSString (10 copy) <@'ConstantString'>   ");
+
+  str = [[NSString alloc] initWithCStringNoCopy: (char *)[str cString]
+			  length: [str length]
+			  freeWhenDone: NO];
+  START_TIMER;
+  for (i = 0; i < MAX_COUNT*10; i++)
+    {
+      [str copy];
+    }
+  END_TIMER;
+  PRINT_TIMER("NSString (10 copy) <NoCopy:free:NO>   ");
+
+  str = [[NSString alloc] initWithCStringNoCopy: (char *)[str cString]
+			  length: [str length]
+			  freeWhenDone: YES];
+  START_TIMER;
+  for (i = 0; i < MAX_COUNT*10; i++)
+    {
+      [str copy];
+    }
+  END_TIMER;
+  PRINT_TIMER("NSString (10 copy) <NoCopy:free:YES>   ");
+
+  str = [stringClass stringWithCString: "hello world"];
+  START_TIMER;
+  for (i = 0; i < MAX_COUNT*10; i++)
+    {
+      [str hash];
+    }
+  END_TIMER;
+  PRINT_TIMER("NSString (10 hash) <initWithCString:>   ");
+
+  str = @"ConstantString";
+  START_TIMER;
+  for (i = 0; i < MAX_COUNT*10; i++)
+    {
+      [str hash];
+    }
+  END_TIMER;
+  PRINT_TIMER("NSString (10 hash) <@'ConstantString'>   ");
 
   START_TIMER;
   for (i = 0; i < MAX_COUNT/100; i++)
     {
       id arp = [NSAutoreleasePool new];
-      NSString	*s = [plist description];
+      [plist description];
       [arp release];
     }
   END_TIMER;
-  PRINT_TIMER("NSString (1/100 mkplist) ");
+  PRINT_TIMER("NSString (1/100 mkplist) \t\t");
 
   START_TIMER;
   for (i = 0; i < MAX_COUNT/1000; i++)
     {
-      id p = [plstr propertyList];
+      [plstr propertyList];
     }
   END_TIMER;
-  PRINT_TIMER("NSString (1/1000 plparse)");
+  PRINT_TIMER("NSString (1/1000 plparse)\t\t");
 
   START_TIMER;
   for (i = 0; i < MAX_COUNT/1000; i++)
@@ -454,42 +512,63 @@ bench_str()
       [arp release];
     }
   END_TIMER;
-  PRINT_TIMER("NSString (1/1000 plcomp)");
+  PRINT_TIMER("NSString (1/1000 plcomp)\t\t");
 
   START_TIMER;
   for (i = 0; i < MAX_COUNT/100; i++)
     {
       NSData	*d = [ser serializePropertyList: plist];
-      id 	p = [des deserializePropertyListFromData: d
-				       mutableContainers: NO];
+      [des deserializePropertyListFromData: d mutableContainers: NO];
     }
   END_TIMER;
-  PRINT_TIMER("NSString (1/100 ser/des)");
+  PRINT_TIMER("NSString (1/100 ser/des)\t\t");
 
   [NSDeserializer uniquing: YES];
   START_TIMER;
   for (i = 0; i < MAX_COUNT/100; i++)
     {
       NSData	*d = [ser serializePropertyList: plist];
-      id 	p = [des deserializePropertyListFromData: d
-				       mutableContainers: NO];
+      [des deserializePropertyListFromData: d mutableContainers: NO];
     }
   END_TIMER;
-  PRINT_TIMER("NSString (1/100 ser/des - uniquing)");
+  PRINT_TIMER("NSString (1/100 ser/des - uniquing)\t");
   [NSDeserializer uniquing: NO];
 
   START_TIMER;
   for (i = 0; i < MAX_COUNT/100; i++)
     {
       NSData	*d = [arc archivedDataWithRootObject: plist];
-      id 	p = [una unarchiveObjectWithData: d];
+      [una unarchiveObjectWithData: d];
     }
   END_TIMER;
-  PRINT_TIMER("NSString (1/100 arc/una)");
+  PRINT_TIMER("NSString (1/100 arc/una)\t\t");
 
   AUTO_END;
 }
 
+void
+bench_date()
+{
+  int i;
+  id d;
+  AUTO_START;
+  Class	dateClass = [NSCalendarDate class];
+
+  printf("NSCalendarDate\n");
+  START_TIMER;
+  for (i = 0; i < MAX_COUNT/10; i++)
+    { 
+      d = [[dateClass alloc] init];
+      [d description];
+      [d dayOfYear];
+      [d minuteOfHour];
+      [d release];
+    }
+  END_TIMER;
+  PRINT_TIMER("NSCalendarDate (various)\t\t");
+  AUTO_END;
+}
+
 void
 bench_data()
 {
@@ -511,7 +590,7 @@ bench_data()
       [d release];
     }
   END_TIMER;
-  PRINT_TIMER("NSData (various)");
+  PRINT_TIMER("NSData (various)\t\t\t");
   AUTO_END;
 }
 
@@ -531,11 +610,12 @@ int main(int argc, char *argv[], char **env)
   stringClass = [NSString class];
  
   pool = [NSAutoreleasePool new];
-  printf(" Test         	\t time (sec) \t index\n");
+  printf(" Test         	\t\t\t\t time (sec) \t index\n");
   bench_object();
   bench_str();
   bench_array();
   bench_dict();
+  bench_date();
   bench_data();
   AUTO_END;
   return 0;
