@@ -22,6 +22,8 @@
 
 #include	<Foundation/Foundation.h>
 #include        "AGSHtml.h"
+#include "GNUstepBase/GNUstep.h"
+#include "GNUstepBase/GSCategories.h"
 
 static int      XML_ELEMENT_NODE;
 static int      XML_ENTITY_REF_NODE;
@@ -29,15 +31,15 @@ static int      XML_TEXT_NODE;
 
 static GSXMLNode	*firstElement(GSXMLNode *nodes)
 {
-  while (nodes != nil)
+  if (nodes == nil)
     {
-      if ([nodes type] == XML_ELEMENT_NODE)
-	{
-	  return nodes;
-	}
-      nodes = [nodes next];
+      return nil;
     }
-  return nil;
+  if ([nodes type] == XML_ELEMENT_NODE)
+    {
+      return nodes;
+    }
+  return [nodes nextElement];
 }
 
 @implementation	AGSHtml
@@ -183,13 +185,13 @@ static NSMutableSet	*textNodes = nil;
   NSString	*kind = (f == YES) ? @"rel=\"gsdoc\" href" : @"name";
   NSString	*hash = (f == YES) ? @"#" : @"";
 
-  if (f == YES)
+  if (f == NO)
     {
       if (u == nil)
 	{
 	  u = unit;
+	  s = base;
 	}
-      s = base;
     }
   else if (u == nil)
     {
@@ -215,7 +217,7 @@ static NSMutableSet	*textNodes = nil;
 	  s = [globalRefs unitRef: r type: t unit: &u];
 	}
     }
-  else
+  if (s == nil)
     {
       NSString	*tmp = u;
 
@@ -388,11 +390,26 @@ static NSMutableSet	*textNodes = nil;
 			{
 			  NSString	*catName = [catNames objectAtIndex: i];
 			  NSDictionary	*catDict;
+			  NSString	*cName;
+			  NSEnumerator	*enumerator;
+			  NSString	*mname;
 
-			  catName = [classname stringByAppendingFormat: @"(%@)",
+			  cName = [classname stringByAppendingFormat: @"(%@)",
 			    catName];
-			  catDict = [unitDict objectForKey: catName];
-			  [m addEntriesFromDictionary: catDict];
+			  catDict = [unitDict objectForKey: cName];
+			  enumerator = [catDict keyEnumerator];
+			  /*
+			   * Add category references to the dictionary,
+			   * prefixing them with the category they belong to.
+			   */
+			  while ((mname = [enumerator nextObject]) != nil)
+			    {
+			      NSString	*file = [catDict objectForKey: mname];
+			      NSString	*ref = [NSString stringWithFormat:
+				@"(%@)%@", catName, mname];
+
+			      [m setObject: file forKey: ref];
+			    }
 			}
 		      dict = AUTORELEASE(m);
 		    }
@@ -416,6 +433,29 @@ static NSMutableSet	*textNodes = nil;
 	  NSString	*file = [dict objectForKey: ref];
 	  NSString	*text = ref;
 
+	  /*
+	   * If a reference to a method contains a leading category name,
+	   * we don't want it in the visible method name, however if it's
+	   * actually a protocol name, we need to make it look right by
+	   * changing the round brackets to angle brackets.
+	   */
+	  if ([text hasPrefix: @"("] == YES)
+	    {
+	      NSRange	r = [text rangeOfString: @")"];
+
+	      if (NSMaxRange(r) == [text length])	// A protocol
+	        {
+		  text = [text stringByReplacingString: @"("
+					    withString: @"&lt;"];
+		  text = [text stringByReplacingString: @")"
+					    withString: @"&gt;"];
+		}
+	      else	// Category name in brackets followed by class name
+	        {
+		  text = [text substringFromIndex: NSMaxRange(r)];
+		}
+	    }
+
 	  [buf appendString: indent];
 	  [buf appendString: @"<li><a rel=\"gsdoc\" href="];
 	  if (isInUnit == YES)
@@ -438,13 +478,15 @@ static NSMutableSet	*textNodes = nil;
 
 - (void) outputNode: (GSXMLNode*)node to: (NSMutableString*)buf
 {
+#if GS_WITH_GC == 0
   CREATE_AUTORELEASE_POOL(arp);
-  GSXMLNode	*children = [node children];
+#endif
+  GSXMLNode	*children = [node firstChild];
 
   if ([node type] == XML_ELEMENT_NODE)
     {
       NSString		*name = [node name];
-      NSDictionary	*prop = [node propertiesAsDictionary];
+      NSDictionary	*prop = [node attributes];
 
       if ([name isEqual: @"back"] == YES)
 	{
@@ -517,6 +559,7 @@ static NSMutableSet	*textNodes = nil;
 	  ssect = 0;
 	  sssect = 0;
 	  [self outputNodeList: children to: buf];
+	  heading = nil;
 	}
       else if ([name isEqual: @"class"] == YES)
 	{
@@ -528,11 +571,14 @@ static NSMutableSet	*textNodes = nil;
 	  [buf appendString: @"<h2>"];
 	  [buf appendString:
 	    [self makeAnchor: classname ofType: @"class" name: classname]];
-	  sup = [self typeRef: sup];
 	  if (sup != nil)
 	    {
-	      [buf appendString: @" : "];
-	      [buf appendString: sup];
+	      sup = [self typeRef: sup];
+	      if (sup != nil)
+		{
+		  [buf appendString: @" : "];
+		  [buf appendString: sup];
+		}
 	    }
 	  [buf appendString: @"</h2>\n"];
 	  [self outputUnit: node to: buf];
@@ -554,7 +600,7 @@ static NSMutableSet	*textNodes = nil;
 	  nam = [prop objectForKey: @"name"];
 	  str = [prop objectForKey: @"type"];
 	  str = [self typeRef: str];
-	  str = [str stringByAppendingFormat: @" %@;", nam];
+	  str = [str stringByAppendingFormat: @" %@", nam];
 
 	  /*
 	   * Output heading.
@@ -582,10 +628,10 @@ static NSMutableSet	*textNodes = nil;
 	  if ([[node name] isEqual: @"declared"] == YES)
 	    {
 	      [self outputNode: node to: buf];
-	      node = firstElement(node);
+	      node = [node nextElement];
 	    }
 
-	  children = firstElement(node);
+	  children = node;
 	  if ([[children name] isEqual: @"standards"])
 	    {
 	      [self outputNode: children to: buf];
@@ -604,7 +650,7 @@ static NSMutableSet	*textNodes = nil;
 	  NSDictionary	*dict;
 
 	  dict = [[localRefs refs] objectForKey: @"contents"];
-	  if ([dict count] > 0)
+	  if ([dict count] > 1)
 	    {
 	      NSArray	*a;
 	      unsigned	i;
@@ -703,7 +749,7 @@ static NSMutableSet	*textNodes = nil;
 	  [buf appendString: @"<dt><b>Declared in:</b></dt>\n"];
 	  [buf appendString: indent];
 	  [buf appendString: @"<dd>"];
-	  [self outputText: [node children] to: buf];
+	  [self outputText: [node firstChild] to: buf];
 	  [buf appendString: @"</dd>\n"];
 	  [self decIndent];
 	  [buf appendString: indent];
@@ -744,14 +790,21 @@ static NSMutableSet	*textNodes = nil;
 	    {
 	      [buf appendFormat: @"<a href=\"mailto:%@\"><code>", ename];
 	    }
-	  [self outputText: [node children] to: buf];
+	  //
+	  // [node firstChild] doesn't look like it points to
+	  // the mail address.
+	  // not sure _where_ it points to though...
+	  //
+#if 0
+	  [self outputText: [node firstChild] to: buf];
+#endif
 	  if (ename == nil)
 	    {
 	      [buf appendString: @"</code>"];
 	    }
 	  else
 	    {
-	      [buf appendFormat: @"</code></a>", ename];
+	      [buf appendFormat: @"%@</code></a>", ename];
 	    }
 	}
       else if ([name isEqual: @"embed"] == YES)
@@ -763,11 +816,12 @@ static NSMutableSet	*textNodes = nil;
 	  NSString		*text;
 	  NSString		*val;
 
-	  text = [children content];
+	  text = [children escapedContent];
 	  val = [prop objectForKey: @"id"];
 	  if (val == nil)
 	    {
 	      val = text;
+	      if (val == nil) val = @"";
 	    }
 	  [buf appendString: [self makeAnchor: val ofType: @"label" name: @""]];
 	}
@@ -818,7 +872,7 @@ static NSMutableSet	*textNodes = nil;
 		{
 		  if ([[tmp name] isEqual: @"arg"] == YES)
 		    {
-		      GSXMLNode		*t = [tmp children];
+		      GSXMLNode		*t = [tmp firstChild];
 		      NSString		*s;
 
 		      if (hadArg == YES)
@@ -826,7 +880,7 @@ static NSMutableSet	*textNodes = nil;
 			  str = [str stringByAppendingString: @", "];
 			}
 
-		      s = [[tmp propertiesAsDictionary] objectForKey: @"type"];
+		      s = [[tmp attributes] objectForKey: @"type"];
 		      s = [self typeRef: s];
 		      str = [str stringByAppendingString: s];
 
@@ -835,8 +889,9 @@ static NSMutableSet	*textNodes = nil;
 			{
 			  if ([t type] == XML_TEXT_NODE)
 			    {
-			      NSString	*content = [t content];
+			      NSString	*content = [t escapedContent];
 
+			      if (content == nil) content = @"";
 			      str = [str stringByAppendingString: content];
 			    }
 			  t = [t next];
@@ -854,7 +909,7 @@ static NSMutableSet	*textNodes = nil;
 			{
 			  str = [str stringByAppendingString: @"<b>,...</b>"];
 			}
-		      children = [tmp next];
+		      children = [tmp nextElement];
 		      break;
 		    }
 		  else
@@ -863,7 +918,7 @@ static NSMutableSet	*textNodes = nil;
 		      break;
 		    }
 		}
-	      tmp = [tmp next];
+	      tmp = [tmp nextElement];
 	    }
 
 	  /*
@@ -892,10 +947,10 @@ static NSMutableSet	*textNodes = nil;
 	  if ([[node name] isEqual: @"declared"] == YES)
 	    {
 	      [self outputNode: node to: buf];
-	      node = firstElement(node);
+	      node = [node nextElement];
 	    }
 
-	  children = firstElement(node);
+	  children = node;
 	  if ([[children name] isEqual: @"standards"])
 	    {
 	      [self outputNode: children to: buf];
@@ -934,7 +989,7 @@ static NSMutableSet	*textNodes = nil;
 	  [buf appendString: indent];
 	  [buf appendString: @"<title>"];
 	  [self incIndent];
-	  [self outputText: [children children] to: buf];
+	  [self outputText: [children firstChild] to: buf];
 	  [self decIndent];
 	  [buf appendString: @"</title>\n"];
 	  [self decIndent];
@@ -964,103 +1019,111 @@ static NSMutableSet	*textNodes = nil;
 
 	  [buf appendString: indent];
 	  [buf appendFormat: @"<h1><a name=\"title$%@\">", base];
-	  [self outputText: [children children] to: buf];
+	  [self outputText: [children firstChild] to: buf];
 	  [buf appendString: @"</a></h1>\n"];
 
-	  [buf appendString: indent];
-	  [buf appendString: @"<h3>Authors</h3>\n"];
-	  [buf appendString: indent];
-	  [buf appendString: @"<dl>\n"];
-	  [self incIndent];
-	  children = firstElement([children next]);
-	  while ([[children name] isEqual: @"author"] == YES)
+	  children = [children nextElement];
+	  if ([[children name] isEqual: @"author"] == YES)
 	    {
-	      GSXMLNode		*author = children;
-	      GSXMLNode		*tmp;
-	      GSXMLNode		*email = nil;
-	      GSXMLNode		*url = nil;
-	      GSXMLNode		*desc = nil;
-
-	      children = [children next];
-	      children = firstElement(children);
-
-	      tmp = firstElement([author children]);
-	      if ([[tmp name] isEqual: @"email"] == YES)
-		{
-		  email = tmp;
-		  tmp = firstElement([tmp next]);
-		}
-	      if ([[tmp name] isEqual: @"url"] == YES)
-		{
-		  url = tmp;
-		  tmp = firstElement([tmp next]);
-		}
-	      if ([[tmp name] isEqual: @"desc"] == YES)
-		{
-		  desc = tmp;
-		  tmp = firstElement([tmp next]);
-		}
-
 	      [buf appendString: indent];
-	      if (url == nil)
-		{
-		  [buf appendString: @"<dt>"];
-		  [buf appendString: [[author propertiesAsDictionary]
-		    objectForKey: @"name"]];
-		}
-	      else
-		{
-		  [buf appendString: @"<dt><a href=\""];
-		  [buf appendString: [[url propertiesAsDictionary]
-		    objectForKey: @"url"]];
-		  [buf appendString: @"\">"];
-		  [buf appendString: [[author propertiesAsDictionary]
-		    objectForKey: @"name"]];
-		  [buf appendString: @"</a>"];
-		}
-	      if (email != nil)
-		{
-		  [buf appendString: @"("];
-		  [self outputNode: email to: buf];
-		  [buf appendString: @")"];
-		}
-	      [buf appendString: @"</dt>\n"];
+	      [buf appendString: @"<h3>Authors</h3>\n"];
 	      [buf appendString: indent];
-	      [buf appendString: @"<dd>\n"];
-	      if (desc != nil)
+	      [buf appendString: @"<dl>\n"];
+	      [self incIndent];
+	      while ([[children name] isEqual: @"author"] == YES)
 		{
-		  [self incIndent];
-		  while (desc != nil)
+		  GSXMLNode		*author = children;
+		  GSXMLNode		*tmp;
+		  GSXMLNode		*email = nil;
+		  GSXMLNode		*url = nil;
+		  GSXMLNode		*desc = nil;
+
+		  children = [children nextElement];
+
+		  tmp = [author firstChildElement];
+		  if ([[tmp name] isEqual: @"email"] == YES)
 		    {
-		      desc = [self outputBlock: desc to: buf inPara: NO];
+		      email = tmp;
+		      tmp = [tmp nextElement];
 		    }
-		  [self decIndent];
+		  if ([[tmp name] isEqual: @"url"] == YES)
+		    {
+		      url = tmp;
+		      tmp = [tmp nextElement];
+		    }
+		  if ([[tmp name] isEqual: @"desc"] == YES)
+		    {
+		      desc = tmp;
+		      tmp = [tmp nextElement];
+		    }
+
+		  [buf appendString: indent];
+		  if (url == nil)
+		    {
+		      [buf appendString: @"<dt>"];
+		      [buf appendString: [[author attributes]
+			objectForKey: @"name"]];
+		    }
+		  else
+		    {
+		      [buf appendString: @"<dt><a href=\""];
+		      [buf appendString: [[url attributes]
+			objectForKey: @"url"]];
+		      [buf appendString: @"\">"];
+		      [buf appendString: [[author attributes]
+			objectForKey: @"name"]];
+		      [buf appendString: @"</a>"];
+		    }
+		  if (email != nil)
+		    {
+		      //
+		      // Add a beautifier ' ' otherwise we'll get a
+		      //   <dt>John Doe(<a href="mailto:...
+		      // or
+		      //   <dt><a href...>John Doe</a>(<a href...
+		      //
+		      [buf appendString: @" ("];
+		      [self outputNode: email to: buf];
+		      [buf appendString: @")"];
+		    }
+		  [buf appendString: @"</dt>\n"];
+		  [buf appendString: indent];
+		  [buf appendString: @"<dd>\n"];
+		  if (desc != nil)
+		    {
+		      [self incIndent];
+		      while (desc != nil)
+			{
+			  desc = [self outputBlock: desc to: buf inPara: NO];
+			}
+		      [self decIndent];
+		    }
+		  [buf appendString: indent];
+		  [buf appendString: @"</dd>\n"];
 		}
+	      [self decIndent];
 	      [buf appendString: indent];
-	      [buf appendString: @"</dd>\n"];
+	      [buf appendString: @"</dl>\n"];
 	    }
-	  [self decIndent];
-	  [buf appendString: indent];
-	  [buf appendString: @"</dl>\n"];
 	  if ([[children name] isEqual: @"version"] == YES)
 	    {
 	      [buf appendString: indent];
 	      [buf appendString: @"<p><b>Version:</b> "];
-	      [self outputText: [children children] to: buf];
+	      [self outputText: [children firstChild] to: buf];
 	      [buf appendString: @"</p>\n"];
-	      children = firstElement([children next]);
+	      children = [children nextElement];
 	    }
 	  if ([[children name] isEqual: @"date"] == YES)
 	    {
 	      [buf appendString: indent];
 	      [buf appendString: @"<p><b>Date:</b> "];
-	      [self outputText: [children children] to: buf];
+	      [self outputText: [children firstChild] to: buf];
 	      [buf appendString: @"</p>\n"];
-	      children = firstElement([children next]);
+	      children = [children nextElement];
 	    }
 	  if ([[children name] isEqual: @"abstract"] == YES)
 	    {
-	      GSXMLNode	*tmp = [children children];
+	      GSXMLNode	*tmp = [children firstChild];
 
 	      [buf appendString: indent];
 	      [buf appendString: @"<blockquote>\n"];
@@ -1072,29 +1135,36 @@ static NSMutableSet	*textNodes = nil;
 	      [self decIndent];
 	      [buf appendString: indent];
 	      [buf appendString: @"</blockquote>\n"];
-	      children = firstElement([children next]);
+	      children = [children nextElement];
 	    }
 	  if ([[children name] isEqual: @"copy"] == YES)
 	    {
 	      [buf appendString: indent];
 	      [buf appendString: @"<p><b>Copyright:</b> (C) "];
-	      [self outputText: [children children] to: buf];
+	      [self outputText: [children firstChild] to: buf];
 	      [buf appendString: @"</p>\n"];
-	      children = firstElement([children next]);
+	      children = [children nextElement];
 	    }
 	}
       else if ([name isEqual: @"heading"] == YES)
 	{
-	  [buf appendString: indent];
-	  [buf appendString: @"<"];
-	  [buf appendString: heading];
-	  [buf appendString: @">"];
-	  [buf appendFormat: @"<a name=\"%03u%03u%03u%03u\">",
-	    chap, sect, ssect, sssect];
-	  [self outputText: children to: buf];
-	  [buf appendString: @"</a></"];
-	  [buf appendString: heading];
-	  [buf appendString: @">\n"];
+	  if (heading == nil)
+	    {
+	    }
+	  else
+	    {
+	      [buf appendString: indent];
+	      [buf appendString: @"<"];
+	      [buf appendString: heading];
+	      [buf appendString: @">"];
+	      [buf appendFormat: @"<a name=\"%03u%03u%03u%03u\">",
+		chap, sect, ssect, sssect];
+	      [self outputText: children to: buf];
+	      [buf appendString: @"</a></"];
+	      [buf appendString: heading];
+	      [buf appendString: @">\n"];
+	      heading = nil;
+	    }
 	}
       else if ([name isEqual: @"index"] == YES)
         {
@@ -1116,8 +1186,9 @@ static NSMutableSet	*textNodes = nil;
 	  NSString	*t = [prop objectForKey: @"type"];
 	  NSString	*v = [prop objectForKey: @"validity"];
 	  NSString	*s;
-	  GSXMLNode	*tmp = children;
+	  GSXMLNode	*tmp;
 
+	  tmp = children = firstElement(children);
 	  [buf appendString: indent];
 	  [buf appendString: @"<h3>"];
 	  s = [self makeLink: n ofType: @"ivariable" inUnit: nil isRef: NO];
@@ -1140,10 +1211,10 @@ static NSMutableSet	*textNodes = nil;
 
 	  if ([[children name] isEqual: @"desc"] == YES)
 	    {
-	      children = [children next];
+	      children = [children nextElement];
 	    }
 	  /*
-	   * List standards with which method complies
+	   * List standards with which ivar complies
 	   */
 	  if ([[children name] isEqual: @"standards"])
 	    {
@@ -1162,15 +1233,129 @@ static NSMutableSet	*textNodes = nil;
 	  NSString		*text;
 	  NSString		*val;
 
-	  text = [children content];
+	  text = [children escapedContent];
 	  val = [prop objectForKey: @"id"];
 	  if (val == nil)
 	    {
 	      val = text;
+	      if (val == nil) val = @"";
 	    }
 	  [buf appendString:
 	    [self makeAnchor: val ofType: @"label" name: text]];
 	}
+      else if ([name isEqual: @"macro"] == YES)
+	{
+	  NSString	*mac;
+	  NSString	*str;
+	  NSString	*s;
+	  GSXMLNode	*tmp = children;
+	  BOOL		hadArg = NO;
+
+	  mac = [prop objectForKey: @"name"];
+	  str = [NSString stringWithFormat: @" %@", mac];
+	  children = nil;
+	  while (tmp != nil)
+	    {
+	      if ([tmp type] == XML_ELEMENT_NODE)
+		{
+		  if ([[tmp name] isEqual: @"arg"] == YES)
+		    {
+		      GSXMLNode		*t = [tmp firstChild];
+
+		      if (hadArg == YES)
+			{
+			  str = [str stringByAppendingString: @", "];
+			}
+		      else
+		      	{
+			  str = [str stringByAppendingString: @"("];
+			}
+
+		      str = [str stringByAppendingString: @"<b>"];
+		      while (t != nil)
+			{
+			  if ([t type] == XML_TEXT_NODE)
+			    {
+			      NSString	*content = [t escapedContent];
+
+			      if (content == nil) content = @"";
+			      str = [str stringByAppendingString: content];
+			    }
+			  t = [t next];
+			}
+		      str = [str stringByAppendingString: @"</b>"];
+		      hadArg = YES;
+		    }
+		  else if ([[tmp name] isEqual: @"vararg"] == YES)
+		    {
+		      if (hadArg == YES)
+			{
+			  str = [str stringByAppendingString: @"<b>,...</b>"];
+			}
+		      else
+			{
+			  str = [str stringByAppendingString: @"<b>(...</b>"];
+			}
+		      children = [tmp nextElement];
+		      hadArg = YES;
+		      break;
+		    }
+		  else
+		    {
+		      children = tmp;
+		      break;
+		    }
+		}
+	      tmp = [tmp nextElement];
+	    }
+
+	  /*
+	   * Output macro heading.
+	   */
+	  [buf appendString: indent];
+	  [buf appendString: @"<h3>"];
+	  s = [self makeLink: mac ofType: @"macro" isRef: NO];
+	  if (s != nil)
+	    {
+	      [buf appendString: s];
+	      [buf appendString: mac];
+	      [buf appendString: @"</a>"];
+	    }
+	  else
+	    {
+	      [buf appendString: mac];
+	    }
+	  [buf appendString: @"</h3>\n"];
+	  [buf appendString: indent];
+	  [buf appendString: str];
+	  if (hadArg == YES)
+	    {
+	      [buf appendString: @")"];
+	    }
+	  [buf appendString: @"<br />\n"];
+
+	  node = firstElement(children);
+
+	  if ([[node name] isEqual: @"declared"] == YES)
+	    {
+	      [self outputNode: node to: buf];
+	      node = [node nextElement];
+	    }
+
+	  children = node;
+	  if ([[children name] isEqual: @"standards"])
+	    {
+	      [self outputNode: children to: buf];
+	    }
+
+	  if ([[node name] isEqual: @"desc"])
+	    {
+	      [self outputNode: node to: buf];
+	    }
+
+	  [buf appendString: indent];
+	  [buf appendString: @"<hr width=\"25%\" align=\"left\" />\n"];
+	}
       else if ([name isEqual: @"method"] == YES)
 	{
 	  NSString	*sel;
@@ -1199,15 +1384,16 @@ static NSMutableSet	*textNodes = nil;
 		{
 		  if ([[tmp name] isEqual: @"sel"] == YES)
 		    {
-		      GSXMLNode	*t = [tmp children];
+		      GSXMLNode	*t = [tmp firstChild];
 
 		      str = [str stringByAppendingString: @"<b>"];
 		      while (t != nil)
 			{
 			  if ([t type] == XML_TEXT_NODE)
 			    {
-			      NSString	*content = [t content];
+			      NSString	*content = [t escapedContent];
 
+			      if (content == nil) content = @"";
 			      sel = [sel stringByAppendingString: content];
 			      if (hadArg == YES)
 				{
@@ -1221,17 +1407,20 @@ static NSMutableSet	*textNodes = nil;
 		    }
 		  else if ([[tmp name] isEqual: @"arg"] == YES)
 		    {
-		      GSXMLNode	*t = [tmp children];
+		      GSXMLNode	*t = [tmp firstChild];
 		      NSString	*s;
 
-		      s = [[tmp propertiesAsDictionary] objectForKey: @"type"];
+		      s = [[tmp attributes] objectForKey: @"type"];
 		      s = [self typeRef: s];
 		      str = [str stringByAppendingFormat: @" (%@)", s];
 		      while (t != nil)
 			{
 			  if ([t type] == XML_TEXT_NODE)
 			    {
-			      str = [str stringByAppendingString: [t content]];
+			      NSString	*content = [t escapedContent];
+			
+			      if (content == nil) content = @"";
+			      str = [str stringByAppendingString: content];
 			    }
 			  t = [t next];
 			}
@@ -1241,7 +1430,7 @@ static NSMutableSet	*textNodes = nil;
 		    {
 		      sel = [sel stringByAppendingString: @",..."];
 		      str = [str stringByAppendingString: @"<b>,...</b>"];
-		      children = [tmp next];
+		      children = [tmp nextElement];
 		      break;
 		    }
 		  else
@@ -1250,7 +1439,7 @@ static NSMutableSet	*textNodes = nil;
 		      break;
 		    }
 		}
-	      tmp = [tmp next];
+	      tmp = [tmp nextElement];
 	    }
 	  if ([sel length] > 1)
 	    {
@@ -1326,21 +1515,39 @@ static NSMutableSet	*textNodes = nil;
 	{
 	  NSString	*type = [prop objectForKey: @"type"];
 	  NSString	*r = [prop objectForKey: @"id"];
-	  GSXMLNode	*tmp = [node children];
+	  GSXMLNode	*tmp = [node firstChild];
+	  NSString	*c = [prop objectForKey: @"class"];
 	  NSString	*s;
 
 	  if ([type isEqual: @"method"] || [type isEqual: @"ivariable"])
 	    {
-	      s = [prop objectForKey: @"class"];
-	      s = [self makeLink: r ofType: type inUnit: s isRef: YES];
+	      s = [self makeLink: r ofType: type inUnit: c isRef: YES];
 	    }
 	  else
 	    {
 	      s = [self makeLink: r ofType: type isRef: YES];
+	      /**
+	       * As a special case, if we have a reference to a function,
+	       * and we can't find it, we check to see if there is actually
+	       * a macro of that name and refer to that instead.
+	       */
+	      if (s == nil && [type isEqual: @"function"] == YES)
+		{
+		  s = [self makeLink: r ofType: @"macro" isRef: YES];
+		}
 	    }
 	  if (s == nil)
 	    {
-	      NSLog(@"ref '%@' not found for %@", r, type);
+	      if (c == nil)
+		{
+		  NSLog(@"Location of %@ '%@' (referenced from %@) "
+		    @"not found or not unique.", type, r, base);
+		}
+	      else
+		{
+		  NSLog(@"Location of the %@ version of %@ '%@' (referenced "
+		    @"from %@) not found.", c, type, r, base);
+		}
 	      if (tmp == nil)
 		{
 		  [buf appendString: r];
@@ -1367,20 +1574,19 @@ static NSMutableSet	*textNodes = nil;
 	}
       else if ([name isEqual: @"protocol"] == YES)
 	{
-	  NSString	*name = [prop objectForKey: @"name"];
+	  NSString	*value = [prop objectForKey: @"name"];
 
-	  unit = [NSString stringWithFormat: @"(%@)", name];
+	  unit = [NSString stringWithFormat: @"(%@)", value];
 	  [buf appendString: indent];
 	  [buf appendString: @"<h2>"];
 	  [buf appendString:
-	    [self makeAnchor: unit ofType: @"protocol" name: name]];
+	    [self makeAnchor: unit ofType: @"protocol" name: value]];
 	  [buf appendString: @"</h2>\n"];
 	  [self outputUnit: node to: buf];
 	  unit = nil;
 	}
       else if ([name isEqual: @"EOEntity"] == YES
-	|| [name isEqual: @"EOModel"] == YES
-	|| [name isEqual: @"macro"] == YES)
+	|| [name isEqual: @"EOModel"] == YES)
 	{
 	  NSString	*tmp = [prop objectForKey: @"name"];
 
@@ -1393,6 +1599,7 @@ static NSMutableSet	*textNodes = nil;
 	  ssect = 0;
 	  sssect = 0;
 	  [self outputNodeList: children to: buf];
+	  heading = @"h1";
 	}
       else if ([name isEqual: @"site"] == YES)
 	{
@@ -1402,7 +1609,7 @@ static NSMutableSet	*textNodes = nil;
 	}
       else if ([name isEqual: @"standards"])
 	{
-	  GSXMLNode	*tmp = [node children];
+	  GSXMLNode	*tmp = [node firstChild];
 	  BOOL		first = YES;
 
 	  if (tmp != nil)
@@ -1424,7 +1631,7 @@ static NSMutableSet	*textNodes = nil;
 			}
 		      [buf appendString: [tmp name]];
 		    }
-		  tmp = [tmp next];
+		  tmp = [tmp nextElement];
 		}
 	      [buf appendString: @"<br />\n"];
 	    }
@@ -1441,12 +1648,14 @@ static NSMutableSet	*textNodes = nil;
 	  ssect++;
 	  sssect = 0;
 	  [self outputNodeList: children to: buf];
+	  heading = @"h2";
 	}
       else if ([name isEqual: @"subsubsect"] == YES)
 	{
 	  heading = @"h4";
 	  sssect++;
 	  [self outputNodeList: children to: buf];
+	  heading = @"h3";
 	}
       else if ([name isEqual: @"type"] == YES)
 	{
@@ -1485,10 +1694,10 @@ static NSMutableSet	*textNodes = nil;
 	  if (node != nil && [[node name] isEqual: @"declared"] == YES)
 	    {
 	      [self outputNode: node to: buf];
-	      node = firstElement(node);
+	      node = [node nextElement];
 	    }
 
-	  children = firstElement(node);
+	  children = node;
 	  if ([[children name] isEqual: @"standards"])
 	    {
 	      [self outputNode: children to: buf];
@@ -1497,7 +1706,7 @@ static NSMutableSet	*textNodes = nil;
 	  if (node != nil && [[node name] isEqual: @"desc"] == YES)
 	    {
 	      [self outputNode: node to: buf];
-	      node = firstElement(node);
+	      node = [node nextElement];
 	    }
 
 	  [buf appendString: indent];
@@ -1558,10 +1767,10 @@ NSLog(@"Element '%@' not implemented", name); 	    // FIXME
 	  if ([[node name] isEqual: @"declared"] == YES)
 	    {
 	      [self outputNode: node to: buf];
-	      node = firstElement(node);
+	      node = [node nextElement];
 	    }
 
-	  children = firstElement(node);
+	  children = node;
 	  if ([[children name] isEqual: @"standards"])
 	    {
 	      [self outputNode: children to: buf];
@@ -1592,14 +1801,28 @@ NSLog(@"Element '%@' not implemented", name); 	    // FIXME
   RELEASE(arp);
 }
 
+/**
+ * Output all the nodes from this one onwards ... try to output
+ * as text first, if not possible, call the main method to output
+ * each node.
+ */
 - (void) outputNodeList: (GSXMLNode*)node to: (NSMutableString*)buf
 {
   while (node != nil)
     {
-      GSXMLNode	*next = [node next];
+      GSXMLNode	*next = [node nextElement];
+      GSXMLNode	*tmp;
 
-      [self outputNode: node to: buf];
-      node = next;
+      tmp = [self outputText: node to: buf];
+      if (tmp == node)
+        {
+	  [self outputNode: node to: buf];
+	  node = next;
+	}
+      else
+        {
+	  node = tmp;
+        }
     }
 }
 
@@ -1636,7 +1859,7 @@ NSLog(@"Element '%@' not implemented", name); 	    // FIXME
 	  [buf appendString: indent];
 	  [buf appendString: @"<p>\n"];
 	  [self incIndent];
-	  [self outputText: [node children] to: buf];
+	  [self outputText: [node firstChild] to: buf];
 	  [self decIndent];
 	  [buf appendString: indent];
 	  [buf appendString: @"</p>\n"];
@@ -1650,9 +1873,11 @@ NSLog(@"Element '%@' not implemented", name); 	    // FIXME
 	}
       else if ([n isEqual: @"example"] == YES)
 	{
-	  [buf appendString: @"<pre>\n"];
-	  [self outputText: [node children] to: buf];
-	  [buf appendString: @"\n</pre>\n"];
+	  GSXMLNode	*c = [node firstChild];
+
+	  [buf appendString: @"<pre>"];
+	  [self outputText: c to: buf];
+	  [buf appendString: @"</pre>\n"];
 	  return [node next];
 	}
       else if ([n isEqual: @"embed"] == YES)
@@ -1675,7 +1900,7 @@ NSLog(@"Element '%@' not implemented", name); 	    // FIXME
       else
 	{
 	  NSLog(@"Non-block element '%@' in block ...", n);
-	  return node;
+	  return nil;
 	}
     }
 
@@ -1692,7 +1917,7 @@ NSLog(@"Element '%@' not implemented", name); 	    // FIXME
 - (GSXMLNode*) outputList: (GSXMLNode*)node to: (NSMutableString*)buf
 {
   NSString	*name = [node name];
-  GSXMLNode	*children = [node children];
+  GSXMLNode	*children = [node firstChildElement];
 
   if ([name isEqual: @"list"] == YES)
     {
@@ -1701,7 +1926,7 @@ NSLog(@"Element '%@' not implemented", name); 	    // FIXME
       [self incIndent];
       while (children != nil)
 	{
-	  GSXMLNode	*tmp = [children children];
+	  GSXMLNode	*tmp = [children firstChild];
 
 	  [buf appendString: indent];
 	  [buf appendString: @"<li>\n"];
@@ -1713,7 +1938,7 @@ NSLog(@"Element '%@' not implemented", name); 	    // FIXME
 	  [self decIndent];
 	  [buf appendString: indent];
 	  [buf appendString: @"</li>\n"];
-	  children = [children next];
+	  children = [children nextElement];
 	}
       [self decIndent];
       [buf appendString: indent];
@@ -1726,7 +1951,7 @@ NSLog(@"Element '%@' not implemented", name); 	    // FIXME
       [self incIndent];
       while (children != nil)
 	{
-	  GSXMLNode	*tmp = [children children];
+	  GSXMLNode	*tmp = [children firstChild];
 
 	  [buf appendString: indent];
 	  [buf appendString: @"<li>\n"];
@@ -1738,7 +1963,7 @@ NSLog(@"Element '%@' not implemented", name); 	    // FIXME
 	  [self decIndent];
 	  [buf appendString: indent];
 	  [buf appendString: @"</li>\n"];
-	  children = [children next];
+	  children = [children nextElement];
 	}
       [self decIndent];
       [buf appendString: indent];
@@ -1755,13 +1980,13 @@ NSLog(@"Element '%@' not implemented", name); 	    // FIXME
 
 	  [buf appendString: indent];
 	  [buf appendString: @"<dt>"];
-	  [self outputText: [children children] to: buf];
+	  [self outputText: [children firstChild] to: buf];
 	  [buf appendString: @"</dt>\n"];
-	  children = [children next];
+	  children = [children nextElement];
 	  [buf appendString: indent];
 	  [buf appendString: @"<dd>\n"];
 	  [self incIndent];
-	  tmp = [children children];
+	  tmp = [children firstChild];
 	  while (tmp != nil)
 	    {
 	      tmp = [self outputBlock: tmp to: buf inPara: NO];
@@ -1769,7 +1994,7 @@ NSLog(@"Element '%@' not implemented", name); 	    // FIXME
 	  [self decIndent];
 	  [buf appendString: indent];
 	  [buf appendString: @"</dd>\n"];
-	  children = [children next];
+	  children = [children nextElement];
 	}
       [self decIndent];
       [buf appendString: indent];
@@ -1786,13 +2011,13 @@ NSLog(@"Element '%@' not implemented", name); 	    // FIXME
 
 	  [buf appendString: indent];
 	  [buf appendString: @"<dt>"];
-	  [self outputText: [children children] to: buf];
+	  [self outputText: [children firstChild] to: buf];
 	  [buf appendString: @"</dt>\n"];
-	  children = [children next];
+	  children = [children nextElement];
 	  [buf appendString: indent];
 	  [buf appendString: @"<dd>\n"];
 	  [self incIndent];
-	  tmp = [children children];
+	  tmp = [children firstChild];
 	  while (tmp != nil)
 	    {
 	      tmp = [self outputBlock: tmp to: buf inPara: NO];
@@ -1800,7 +2025,7 @@ NSLog(@"Element '%@' not implemented", name); 	    // FIXME
 	  [self decIndent];
 	  [buf appendString: indent];
 	  [buf appendString: @"</dd>\n"];
-	  children = [children next];
+	  children = [children nextElement];
 	}
       [self decIndent];
       [buf appendString: indent];
@@ -1828,7 +2053,10 @@ NSLog(@"Element '%@' not implemented", name); // FIXME
     {
       if ([node type] == XML_TEXT_NODE)
 	{
-	  [buf appendString: [node content]];
+	  NSString	*str = [node escapedContent];
+
+	  if (str == nil) str = @"";
+	  [buf appendString: str];
 	}
       else if ([node type] == XML_ENTITY_REF_NODE)
 	{
@@ -1859,11 +2087,11 @@ NSLog(@"Element '%@' not implemented", name); // FIXME
   GSXMLNode	*t;
   NSArray	*a;
 
-  node = [node children];
+  node = [node firstChildElement];
   if (node != nil && [[node name] isEqual: @"declared"] == YES)
     {
       [self outputNode: node to: buf];
-      node = [node next];
+      node = [node nextElement];
     }
 
   if (node != nil && [[node name] isEqual: @"conform"] == YES)
@@ -1878,13 +2106,14 @@ NSLog(@"Element '%@' not implemented", name); // FIXME
       [buf appendString: @"<dt><b>Conforms to:</b></dt>\n"];
       while (node != nil && [[node name] isEqual: @"conform"] == YES)
 	{
-	  NSString	*text = [[node children] content];
+	  NSString	*text = [[node firstChild] escapedContent];
 
+	  if (text == nil) text = @"";
 	  [buf appendString: indent];
 	  [buf appendString: @"<dd>"];
 	  [buf appendString: [self protocolRef: text]];
 	  [buf appendString: @"</dd>\n"];
-	  node = [node next];
+	  node = [node nextElement];
 	}
       [self decIndent];
       [buf appendString: indent];
@@ -1897,11 +2126,11 @@ NSLog(@"Element '%@' not implemented", name); // FIXME
   t = node;
   while (t != nil && [[t name] isEqual: @"standards"] == NO)
     {
-      t = [t next];
+      t = [t nextElement];
     }
-  if (t != nil && [t children] != nil)
+  if (t != nil && [t firstChild] != nil)
     {
-      t = [t children];
+      t = [t firstChild];
       [buf appendString: indent];
       [buf appendString: @"<blockquote>\n"];
       [self incIndent];
@@ -1919,7 +2148,7 @@ NSLog(@"Element '%@' not implemented", name); // FIXME
 	      [buf appendString: [t name]];
 	      [buf appendString: @"</li>\n"];
 	    }
-	  t = [t next];
+	  t = [t nextElement];
 	}
       [self decIndent];
       [buf appendString: indent];
@@ -1932,7 +2161,7 @@ NSLog(@"Element '%@' not implemented", name); // FIXME
   if (node != nil && [[node name] isEqual: @"desc"] == YES)
     {
       [self outputNode: node to: buf];
-      node = [node next];
+      node = [node nextElement];
     }
 
   if (node != nil && [[node name] isEqual: @"ivariable"] == YES)
@@ -1944,7 +2173,7 @@ NSLog(@"Element '%@' not implemented", name); // FIXME
       while (node != nil && [[node name] isEqual: @"ivariable"] == YES)
 	{
 	  [self outputNode: node to: buf];
-	  node = [node next];
+	  node = [node nextElement];
 	}
     }
 
@@ -1963,7 +2192,7 @@ NSLog(@"Element '%@' not implemented", name); // FIXME
 	    {
 	      [self outputNode: node to: buf];
 	    }
-	  node = [node next];
+	  node = [node nextElement];
 	}
     }
 }
@@ -2012,12 +2241,19 @@ NSLog(@"Element '%@' not implemented", name); // FIXME
  */
 - (NSString*) typeRef: (NSString*)t
 {
-  NSString	*orig = [t stringByTrimmingSpaces];
+  NSString	*str = [t stringByTrimmingSpaces];
   NSString	*s;
-  unsigned	end = [orig length];
+  unsigned	end = [str length];
   unsigned	start;
+  NSMutableString	*ms = nil;
+  NSRange		er;
+  NSRange		sr;
 
-  t = orig;
+  if (end == 0)
+    {
+      return nil;
+    }
+  t = str;
   while (end > 0)
     {
       unichar	c = [t characterAtIndex: end-1];
@@ -2039,7 +2275,7 @@ NSLog(@"Element '%@' not implemented", name); // FIXME
 	}
       start--;
     }
-  t = [orig substringWithRange: NSMakeRange(start, end - start)];
+  t = [str substringWithRange: NSMakeRange(start, end - start)];
 
   s = [self makeLink: t ofType: @"type" isRef: YES];
   if (s == nil)
@@ -2047,16 +2283,88 @@ NSLog(@"Element '%@' not implemented", name); // FIXME
       s = [self makeLink: t ofType: @"class" isRef: YES];
     }
 
+  s = [s stringByAppendingFormat: @"%@</a>", t];
+  if (s != nil && [str length] == [t length])
+    {
+      return s;
+    }
+
+  /*
+   * Look for protocol spec.
+   */
+  sr = [str rangeOfString: @"<"];
+  if (sr.length == 0)
+    {
+      sr = [str rangeOfString: @"&lt;"];
+    }
+  if (sr.length == 0)
+    {
+      sr = [str rangeOfString: @"&#60;"];
+    }
+  er = [str rangeOfString: @">"];
+  if (er.length == 0)
+    {
+      er = [str rangeOfString: @"&gt;"];
+    }
+  if (er.length == 0)
+    {
+      er = [str rangeOfString: @"&#62;"];
+    }
+
+  /*
+   * Substitute in protocol references.
+   */
+  if (sr.length > 0 && er.length > 0 && er.location > sr.location)
+    {
+      NSString	*pString;
+      NSRange	r;
+      NSArray	*protocols;
+      unsigned	i;
+
+      r = NSMakeRange(NSMaxRange(sr), er.location - NSMaxRange(sr));
+      pString = [str substringWithRange: r];
+      protocols = [pString componentsSeparatedByString: @","];
+      ms = [str mutableCopy];
+      pString = @"";
+      for (i = 0; i < [protocols count]; i++)
+	{
+	  NSString	*p = [protocols objectAtIndex: i];
+	  NSString	*l;
+
+	  l = [self makeLink: [NSString stringWithFormat: @"(%@)", p]
+		      ofType: @"protocol"
+		       isRef: YES];
+	  if (l != nil)
+	    {
+	      p = [l stringByAppendingFormat: @"%@</a>", p];
+	    }
+	  if (i > 0)
+	    {
+	      pString = [pString stringByAppendingString: @","];
+	    }
+	  pString = [pString stringByAppendingString: p];
+	}
+      [ms replaceCharactersInRange: r withString: pString];
+    }
+
+
+  /*
+   * Substitute in basic type reference.
+   */
   if (s != nil)
     {
-      s = [s stringByAppendingFormat: @"%@</a>", t];
-      if ([orig length] == [t length])
+      if (ms == nil)
 	{
-	  return s;
+	  ms = [str mutableCopy];
 	}
-      return [orig stringByReplacingString: t withString: s];
+      [ms replaceCharactersInRange: NSMakeRange(start, end - start)
+			withString: s];
     }
-  return orig;
+  if (ms != nil)
+    {
+      str = AUTORELEASE(ms);
+    }
+  return str;
 }
 
 @end
