@@ -22,6 +22,8 @@
 
 #include	<Foundation/Foundation.h>
 #include        "AGSIndex.h"
+#include 	"GNUstepBase/GNUstep.h"
+#include "GNUstepBase/GSCategories.h"
 
 static int      XML_ELEMENT_NODE;
 static int      XML_TEXT_NODE;
@@ -89,6 +91,7 @@ mergeDictionaries(NSMutableDictionary *dst, NSDictionary *src, BOOL override)
 		    {
 		      NSLog(@"String missmatch in merge for %@. S:%@, D:%@",
 			stack, s, d);
+		      [dst setObject: s forKey: k];
 		    }
 		}
 	    }
@@ -216,14 +219,14 @@ setDirectory(NSMutableDictionary *dict, NSString *path)
  */
 - (void) makeRefs: (GSXMLNode*)node
 {
-  GSXMLNode	*children = [node children];
+  GSXMLNode	*children = [node firstChild];
   GSXMLNode	*next = [node next];
   BOOL		newUnit = NO;
 
   if ([node type] == XML_ELEMENT_NODE)
     {
       NSString		*name = [node name];
-      NSDictionary	*prop = [node propertiesAsDictionary];
+      NSDictionary	*prop = [node attributes];
 
       if ([name isEqual: @"category"] == YES)
 	{
@@ -272,6 +275,7 @@ setDirectory(NSMutableDictionary *dict, NSString *path)
 	{
 	  NSMutableDictionary	*d;
 	  NSString		*k;
+	  NSString		*c;
 
 	  d = [refs objectForKey: @"contents"];
 	  if (d == nil)
@@ -283,7 +287,9 @@ setDirectory(NSMutableDictionary *dict, NSString *path)
 
           k = [NSString stringWithFormat: @"%03u%03u%03u%03u",
 	    chap, sect, ssect, sssect];
-	  [d setObject: [[children content] stringByTrimmingSpaces] forKey: k];
+	  c = [[children content] stringByTrimmingSpaces];
+	  if (c == nil) c = @"";
+	  [d setObject: c forKey: k];
 	  children = nil;
 	}
       else if ([name isEqual: @"ivariable"] == YES)
@@ -301,6 +307,7 @@ setDirectory(NSMutableDictionary *dict, NSString *path)
 	  NSString		*val;
 
 	  text = [[children content] stringByTrimmingSpaces];
+	  if (text == nil) text = @"";
 	  children = nil;
 	  all = [refs objectForKey: name];
 	  if (all == nil)
@@ -344,29 +351,32 @@ setDirectory(NSMutableDictionary *dict, NSString *path)
 		{
 		  if ([[tmp name] isEqual: @"sel"] == YES)
 		    {
-		      GSXMLNode	*t = [tmp children];
+		      GSXMLNode	*t = [tmp firstChild];
 
 		      while (t != nil)
 			{
 			  if ([t type] == XML_TEXT_NODE)
 			    {
-			      sel = [sel stringByAppendingString:
-				[[t content] stringByTrimmingSpaces]];
+			      NSString	*s;
+
+			      s = [[t content] stringByTrimmingSpaces];
+			      if (s == nil) s = @"";
+			      sel = [sel stringByAppendingString: s];
 			    }
 			  t = [t next];
 			}
 		    }
-		  else if ([[tmp name] isEqual: @"arg"] == NO)
-		    {
-		      children = tmp;
-		      break;
-		    }
 		  else if ([[tmp name] isEqual: @"vararg"] == YES)
 		    {
 		      sel = [sel stringByAppendingString: @",..."];
 		      children = [tmp next];
 		      break;
 		    }
+		  else if ([[tmp name] isEqual: @"arg"] == NO)
+		    {
+		      children = tmp;
+		      break;
+		    }
 		}
 	      tmp = [tmp next];
 	    }
@@ -414,6 +424,7 @@ setDirectory(NSMutableDictionary *dict, NSString *path)
       else if ([name isEqual: @"title"] == YES)
 	{
 	  NSMutableDictionary	*d;
+	  NSString		*s;
 
 	  d = [refs objectForKey: @"title"];
 	  if (d == nil)
@@ -423,8 +434,9 @@ setDirectory(NSMutableDictionary *dict, NSString *path)
 	      RELEASE(d);
 	    }
 
-	  [d setObject: [[children content] stringByTrimmingSpaces]
-		forKey: base];
+	  s = [[children content] stringByTrimmingSpaces];
+	  if (s == nil) s = @"";
+	  [d setObject: s forKey: base];
 	  children = nil;
 	}
       else
@@ -466,6 +478,21 @@ setDirectory(NSMutableDictionary *dict, NSString *path)
   return [d allKeys];
 }
 
+/**
+ * Return a list of output files for the header
+ */
+- (NSMutableArray*) outputsForHeader: (NSString*)h
+{
+  NSDictionary	*dict = [refs objectForKey: @"output"];
+  NSArray	*array = [dict objectForKey: h];
+
+  if (array == nil)
+    {
+      return [NSMutableArray arrayWithCapacity: 2];
+    }
+  return AUTORELEASE([array mutableCopy]);
+}
+
 - (NSMutableDictionary*) refs
 {
   return refs;
@@ -503,6 +530,23 @@ setDirectory(NSMutableDictionary *dict, NSString *path)
   [t setObject: base forKey: ref];
 }
 
+/**
+ * Set up an array listing the output files for a particular header.
+ */
+- (void) setOutputs: (NSArray*)a forHeader: (NSString*)h
+{
+  NSMutableDictionary	*dict;
+
+  dict = [refs objectForKey: @"output"];
+  if (dict == nil)
+    {
+      dict = [NSMutableDictionary new];
+      [refs setObject: dict forKey: @"output"];
+      RELEASE(dict);
+    }
+  [dict setObject: a forKey: h];
+}
+
 - (void) setRelationship: (NSString*)r from: (NSString*)from to: (NSString*)to
 {
   NSMutableDictionary	*dict;
@@ -585,13 +629,18 @@ setDirectory(NSMutableDictionary *dict, NSString *path)
 }
 
 /**
- * Return a list of source files for the header (or nil)
+ * Return a list of source files for the header.
  */
 - (NSArray*) sourcesForHeader: (NSString*)h
 {
   NSDictionary	*dict = [refs objectForKey: @"source"];
+  NSArray	*array = [dict objectForKey: h];
 
-  return [dict objectForKey: h];
+  if (array == nil)
+    {
+      return [NSMutableArray arrayWithCapacity: 2];
+    }
+  return AUTORELEASE([array mutableCopy]);
 }
 
 /**
@@ -624,7 +673,6 @@ setDirectory(NSMutableDictionary *dict, NSString *path)
       *u = nil;
       return nil;
     }
-
   if (*u == nil)
     {
       /**
@@ -670,6 +718,7 @@ setDirectory(NSMutableDictionary *dict, NSString *path)
       *u = [self globalRef: *u type: @"super"];
       if (*u != nil)
 	{
+	  *u = [*u lastPathComponent];
 	  s = [t objectForKey: *u];
 	  if (s != nil)
 	    {
