@@ -21,6 +21,8 @@
    */
 
 #include "AGSOutput.h"
+#include "GNUstepBase/GNUstep.h"
+#include "GNUstepBase/GSCategories.h"
 
 @interface AGSOutput (Private)
 - (NSString*) mergeMarkup: (NSString*)markup
@@ -69,21 +71,55 @@ static BOOL snuggleStart(NSString *t)
  *  far too special purpose.</p>
  *  <unit />
  *  <p>Here is the afterword for the class.</p>
+ *  <p> Ad here is some automated cross referencing ...
+ *  A method in a protocol: [(NSCopying)-copyWithZone:], a class:
+ *  [NSString], a protocol: [(NSCopying)], and a
+ *  category: [NSRunLoop(GNUstepExtensions)].
+ *  </p>
  * </unit>
  * And finally, here is the actual class description ... outside the chapter.
+ * With a reference to foo() in it.
  */
 @implementation	AGSOutput
 
-+ (void) initialize
+- (NSString*) checkComment: (NSString*)comment
+		      unit: (NSString*)unit
+		      info: (NSDictionary*)d
 {
-  NSUserDefaults	*ud = [NSUserDefaults standardUserDefaults];
+  if ([comment length] == 0)
+    {
+      comment = @"<em>Description forthcoming.</em>";
+      if (warn == YES)
+	{
+	  NSString	*name = [d objectForKey: @"Name"];
+	  NSString	*type = [d objectForKey: @"Type"];
 
-  [ud registerDefaults: [NSDictionary dictionaryWithObjectsAndKeys:
-    @"TypesAndConstants", @"ConstantsTemplate",
-    @"Functions", @"FunctionsTemplate",
-    @"TypesAndConstants", @"TypedefsTemplate",
-    @"TypesAndConstants", @"VariablesTemplate",
-    nil]];
+	  if (unit == nil)
+	    {
+	      if (type == nil)
+	        {
+		  NSLog(@"Warning - No comments for %@", name);
+		}
+	      else
+	        {
+		  NSLog(@"Warning - No comments for %@ %@", type, name);
+		}
+	    }
+	  else
+	    {
+	      if ([d objectForKey: @"ReturnType"] != nil)
+		{
+		  NSLog(@"Warning - No comments for [%@ %@]", unit, name);
+		}
+	      else
+		{
+		  NSLog(@"Warning - No comments for instance variable %@ in %@",
+		    name, unit);
+		}
+	    }
+	}
+    }
+  return comment;
 }
 
 - (void) dealloc
@@ -192,6 +228,8 @@ static BOOL snuggleStart(NSString *t)
     @"_0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"]);
   identStart = RETAIN([NSCharacterSet characterSetWithCharactersInString:
     @"_abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"]);
+  verbose = [[NSUserDefaults standardUserDefaults] boolForKey: @"Verbose"];
+  warn = [[NSUserDefaults standardUserDefaults] boolForKey: @"Warn"];
 
   return self;
 }
@@ -210,6 +248,7 @@ static BOOL snuggleStart(NSString *t)
   NSDictionary		*types;
   NSDictionary		*variables;
   NSDictionary		*constants;
+  NSDictionary		*macros;
   NSMutableArray	*files;
   NSArray		*authors;
   NSString		*base;
@@ -240,11 +279,12 @@ static BOOL snuggleStart(NSString *t)
   types = [info objectForKey: @"Types"];
   variables = [info objectForKey: @"Variables"];
   constants = [info objectForKey: @"Constants"];
+  macros = [info objectForKey: @"Macros"];
 
   [str appendString: @"<?xml version=\"1.0\"?>\n"];
   [str appendString: @"<!DOCTYPE gsdoc PUBLIC "];
-  [str appendString: @"\"-//GNUstep//DTD gsdoc 0.6.7//EN\" "];
-  [str appendString: @"\"http://www.gnustep.org/gsdoc-0_6_7.xml\">\n"];
+  [str appendString: @"\"-//GNUstep//DTD gsdoc 1.0.0//EN\" "];
+  [str appendString: @"\"http://www.gnustep.org/gsdoc-1_0_0.xml\">\n"];
   [str appendFormat: @"<gsdoc"];
 
   if (base != nil)
@@ -497,6 +537,42 @@ static BOOL snuggleStart(NSString *t)
       RELEASE(m);
     }
 
+  if ([macros count] > 0)
+    {
+      NSMutableString	*m = [NSMutableString new];
+      NSArray		*names;
+      unsigned		i;
+      unsigned		c = [macros count];
+
+      [m appendString: @"    <chapter>\n"];
+      [m appendFormat: @"      <heading>%@ macros</heading>\n", base];
+      [m appendString: @"      <p></p>\n"];
+
+      names = [macros allKeys];
+      names = [names sortedArrayUsingSelector: @selector(compare:)];
+      for (i = 0; i < c; i++)
+	{
+	  NSString	*name = [names objectAtIndex: i];
+	  NSDictionary	*d = [macros objectForKey: name];
+
+	  [self outputMacro: d to: m];
+	}
+
+      [m appendString: @"    </chapter>\n"];
+
+      tmp = [self mergeMarkup: m ofKind: @"Macros"];
+      if (tmp == nil)
+	{
+	  [str appendString: m];
+	  chapters++;
+	}
+      else
+	{
+	  [files addObject: tmp];
+	}
+      RELEASE(m);
+    }
+
   if ([variables count] > 0)
     {
       NSMutableString	*m = [NSMutableString new];
@@ -608,7 +684,7 @@ static BOOL snuggleStart(NSString *t)
   NSString	*declared = [d objectForKey: @"Declared"];
   NSString	*standards = nil;
 
-  if ([[d objectForKey: @"Implemented"] isEqual: @"YES"] == NO)
+  if (warn == YES && [[d objectForKey: @"Implemented"] isEqual: @"YES"] == NO)
     {
       NSLog(@"Warning ... %@ %@ is not implemented where expected", kind, name);
     }
@@ -635,10 +711,7 @@ static BOOL snuggleStart(NSString *t)
     }
 
   [str appendString: @"        <desc>\n"];
-  if ([comment length] == 0)
-    {
-      comment = @"<em>Description forthcoming.</em>";
-    }
+  comment = [self checkComment: comment unit: nil info: d];
   [self reformat: comment withIndent: 10 to: str];
   [str appendString: @"        </desc>\n"];
   if (standards != nil)
@@ -662,7 +735,7 @@ static BOOL snuggleStart(NSString *t)
   NSString	*standards = nil;
   unsigned	i = [aa count];
 
-  if ([[d objectForKey: @"Implemented"] isEqual: @"YES"] == NO)
+  if (warn == YES && [[d objectForKey: @"Implemented"] isEqual: @"YES"] == NO)
     {
       NSLog(@"Warning ... function %@ is not implemented where expected", name);
     }
@@ -769,7 +842,7 @@ static BOOL snuggleStart(NSString *t)
     }
   if ([[d objectForKey: @"VarArgs"] boolValue] == YES)
     {
-      [str appendString: @"<vararg />\n"];
+      [str appendString: @"        <vararg />\n"];
     }
 
   if (declared != nil)
@@ -780,10 +853,7 @@ static BOOL snuggleStart(NSString *t)
     }
 
   [str appendString: @"        <desc>\n"];
-  if ([comment length] == 0)
-    {
-      comment = @"<em>Description forthcoming.</em>";
-    }
+  comment = [self checkComment: comment unit: nil info: d];
   [self reformat: comment withIndent: 10 to: str];
   [str appendString: @"        </desc>\n"];
   if (standards != nil)
@@ -797,9 +867,12 @@ static BOOL snuggleStart(NSString *t)
 /**
  * Output the gsdoc code for an instance variable.
  */
-- (void) outputInstanceVariable: (NSDictionary*)d to: (NSMutableString*)str
+- (void) outputInstanceVariable: (NSDictionary*)d
+			     to: (NSMutableString*)str
+			    for: (NSString*)unit
 {
-  NSString	*type = [d objectForKey: @"Type"];
+  NSString	*pref = [d objectForKey: @"Prefix"];
+  NSString	*type = [d objectForKey: @"BaseType"];
   NSString	*validity = [d objectForKey: @"Validity"];
   NSString	*name = [d objectForKey: @"Name"];
   NSString	*comment = [d objectForKey: @"Comment"];
@@ -807,6 +880,10 @@ static BOOL snuggleStart(NSString *t)
 
   [str appendString: @"        <ivariable type=\""];
   [str appendString: escapeType(type)];
+  if ([pref length] > 0)
+    {
+      [str appendString: pref];
+    }
   [str appendString: @"\" name=\""];
   [str appendString: name];
   if (validity != nil)
@@ -817,10 +894,7 @@ static BOOL snuggleStart(NSString *t)
   [str appendString: @"\">\n"];
 
   [str appendString: @"          <desc>\n"];
-  if ([comment length] == 0)
-    {
-      comment = @"<em>Description forthcoming.</em>";
-    }
+  comment = [self checkComment: comment unit: unit info: d];
   [self reformat: comment withIndent: 12 to: str];
   [str appendString: @"          </desc>\n"];
   if (standards != nil)
@@ -830,6 +904,64 @@ static BOOL snuggleStart(NSString *t)
   [str appendString: @"        </ivariable>\n"];
 }
 
+/**
+ * Uses -split: and -reformat:withIndent:to:.
+ */
+- (void) outputMacro: (NSDictionary*)d
+		  to: (NSMutableString*)str
+{
+  NSString	*name = [d objectForKey: @"Name"];
+  NSString	*comment = [d objectForKey: @"Comment"];
+  NSString	*declared = [d objectForKey: @"Declared"];
+  NSString	*standards = nil;
+  unsigned	i;
+
+  if (standards == nil)
+    {
+      standards = [d objectForKey: @"Standards"];
+    }
+
+  [str appendString: @"      <macro name=\""];
+  [str appendString: name];
+  [str appendString: @"\">\n"];
+
+  /*
+   * Storing the argument names array in the 'args' ivar ensures that
+   * when we output comments, the argument names are highlighted.
+   */
+  args = [d objectForKey: @"Args"];
+  for (i = 0; i < [args count]; i++)
+    {
+      NSString		*s = [args objectAtIndex: i];
+
+      [str appendString: @"        <arg>"];
+      [str appendString: s];
+      [str appendString: @"</arg>\n"];
+    }
+  if ([[d objectForKey: @"VarArgs"] boolValue] == YES)
+    {
+      [str appendString: @"        <vararg />\n"];
+    }
+
+  if (declared != nil)
+    {
+      [str appendString: @"        <declared>"];
+      [str appendString: declared];
+      [str appendString: @"</declared>\n"];
+    }
+
+  [str appendString: @"        <desc>\n"];
+  comment = [self checkComment: comment unit: nil info: d];
+  [self reformat: comment withIndent: 10 to: str];
+  [str appendString: @"        </desc>\n"];
+  if (standards != nil)
+    {
+      [self reformat: standards withIndent: 8 to: str];
+    }
+  [str appendString: @"      </macro>\n"];
+  args = nil;
+}
+
 /**
  * Uses -split: and -reformat:withIndent:to:.
  * Also has fun with YES, NO, and nil.
@@ -847,7 +979,8 @@ static BOOL snuggleStart(NSString *t)
   NSString	*override = nil;
   NSString	*standards = nil;
 
-  if (unit != nil && [[d objectForKey: @"Implemented"] isEqual: @"YES"] == NO)
+  if (warn == YES && unit != nil
+    && [[d objectForKey: @"Implemented"] isEqual: @"YES"] == NO)
     {
       NSLog(@"Warning ... method %@ %@ is not implemented where expected",
         unit, name);
@@ -989,14 +1122,11 @@ static BOOL snuggleStart(NSString *t)
     }
   if ([[d objectForKey: @"VarArgs"] boolValue] == YES)
     {
-      [str appendString: @"<vararg />\n"];
+      [str appendString: @"          <vararg />\n"];
     }
 
   [str appendString: @"          <desc>\n"];
-  if ([comment length] == 0)
-    {
-      comment = @"<em>Description forthcoming.</em>";
-    }
+  comment = [self checkComment: comment unit: unit info: d];
   [self reformat: comment withIndent: 12 to: str];
   [str appendString: @"          </desc>\n"];
   if (standards != nil)
@@ -1011,15 +1141,16 @@ static BOOL snuggleStart(NSString *t)
 {
   NSString	*name = [d objectForKey: @"Name"];
   NSString	*type = [d objectForKey: @"Type"];
-  NSDictionary	*methods = [d objectForKey: @"Methods"];
-  NSDictionary	*ivars = [d objectForKey: @"InstanceVariables"];
+  NSString	*kind = type;
+  NSMutableDictionary	*methods = [d objectForKey: @"Methods"];
+  NSMutableDictionary	*ivars = [d objectForKey: @"InstanceVariables"];
   NSString	*comment = [d objectForKey: @"Comment"];
   NSString	*unitName = nil;
   NSArray	*names;
   NSArray	*protocols;
   NSString	*standards = nil;
   NSString	*tmp;
-  NSString	*unit;
+  NSString	*unit = nil;
   NSRange	r;
   unsigned	ind;
   unsigned	i;
@@ -1027,36 +1158,59 @@ static BOOL snuggleStart(NSString *t)
 
   if ([[d objectForKey: @"Implemented"] isEqual: @"YES"] == NO)
     {
-      NSLog(@"Warning ... unit %@ is not implemented where expected", name);
+      if ([name hasPrefix:  @"NSObject("] == YES)
+	{
+	  NSEnumerator		*e = [methods objectEnumerator];
+	  NSMutableDictionary	*m;
+
+	  /*
+	   * Assume an unimplemented category of NSObject is an
+	   * informal protocol, and stop warnings being issued
+	   * about unimplemented methods.
+	   */
+	  unitName = name;
+	  kind = @"informal protocol";
+	  while ((m = [e nextObject]) != nil)
+	    {
+	      [m setObject: @"YES" forKey: @"Implemented"];
+	    }
+	}
+      else if (warn == YES)
+	{
+	  NSLog(@"Warning ... unit %@ is not implemented where expected", name);
+	}
     }
   else
     {
       unitName = name;
     }
 
-  r = [comment rangeOfString: @"<standards>"];
-  if (comment != nil && r.length > 0)
+  if (comment != nil)
     {
-      unsigned  i = r.location;
-
-      r = NSMakeRange(i, [comment length] - i);
-      r = [comment rangeOfString: @"</standards>"
-		     options: NSLiteralSearch
-		       range: r];
+      r = [comment rangeOfString: @"<standards>"];
       if (r.length > 0)
 	{
-	  NSMutableString	*m;
+	  unsigned  i = r.location;
 
-	  r = NSMakeRange(i, NSMaxRange(r) - i);
-	  standards = [comment substringWithRange: r];
-	  m = [comment mutableCopy];
-	  [m deleteCharactersInRange: r];
-	  comment = m;
-	  AUTORELEASE(m);
-	}
-      else
-	{
-	  NSLog(@"unterminated <standards> in comment for %@", name);
+	  r = NSMakeRange(i, [comment length] - i);
+	  r = [comment rangeOfString: @"</standards>"
+			 options: NSLiteralSearch
+			   range: r];
+	  if (r.length > 0)
+	    {
+	      NSMutableString	*m;
+
+	      r = NSMakeRange(i, NSMaxRange(r) - i);
+	      standards = [comment substringWithRange: r];
+	      m = [comment mutableCopy];
+	      [m deleteCharactersInRange: r];
+	      comment = m;
+	      AUTORELEASE(m);
+	    }
+	  else
+	    {
+	      NSLog(@"unterminated <standards> in comment for %@", name);
+	    }
 	}
     }
   if (standards == nil)
@@ -1068,55 +1222,59 @@ static BOOL snuggleStart(NSString *t)
    * Make sure we have a 'unit' part and a class 'desc' part (comment)
    * to be output.
    */
-  r = [comment rangeOfString: @"<unit>"];
-  if (comment != nil && r.length > 0)
+  if (comment != nil)
     {
-      unsigned	pos = r.location;
+      r = [comment rangeOfString: @"<unit>"];
+      if (r.length > 0)
+	{
+	  unsigned	pos = r.location;
 
-      r = [comment rangeOfString: @"</unit>"];
-      if (r.length == 0 || r.location < pos)
-	{
-	  NSLog(@"Unterminated <unit> in comment for %@", name);
-	  return;
-	}
-      
-      if (pos == 0)
-	{
-	  if (NSMaxRange(r) == [comment length])
+	  r = [comment rangeOfString: @"</unit>"];
+	  if (r.length == 0 || r.location < pos)
 	    {
-	      unit = comment;
-	      comment = nil;
+	      NSLog(@"Unterminated <unit> in comment for %@", name);
+	      return;
+	    }
+	  
+	  if (pos == 0)
+	    {
+	      if (NSMaxRange(r) == [comment length])
+		{
+		  unit = comment;
+		  comment = nil;
+		}
+	      else
+		{
+		  unit = [comment substringToIndex: NSMaxRange(r)];
+		  comment = [comment substringFromIndex: NSMaxRange(r)];
+		}
 	    }
 	  else
 	    {
-	      unit = [comment substringToIndex: NSMaxRange(r)];
-	      comment = [comment substringFromIndex: NSMaxRange(r)];
-	    }
-	}
-      else
-	{
-	  if (NSMaxRange(r) == [comment length])
-	    {
-	      unit = [comment substringFromIndex: pos];
-	      comment = [comment substringToIndex: pos];
-	    }
-	  else
-	    {
-	      unsigned	end = NSMaxRange(r);
+	      if (NSMaxRange(r) == [comment length])
+		{
+		  unit = [comment substringFromIndex: pos];
+		  comment = [comment substringToIndex: pos];
+		}
+	      else
+		{
+		  unsigned	end = NSMaxRange(r);
 
-	      r = NSMakeRange(pos, end-pos);
-	      unit = [comment substringWithRange: r];
-	      comment = [[comment substringToIndex: pos]
-		stringByAppendingString: [comment substringFromIndex: end]];
+		  r = NSMakeRange(pos, end-pos);
+		  unit = [comment substringWithRange: r];
+		  comment = [[comment substringToIndex: pos]
+		    stringByAppendingString: [comment substringFromIndex: end]];
+		}
 	    }
+	  unit = [unit stringByReplacingString: @"unit>"
+				    withString: @"chapter>"];
 	}
-      unit = [unit stringByReplacingString: @"unit>" withString: @"chapter>"];
     }
-  else
+  if (unit == nil)
     {
       unit = [NSString stringWithFormat:
         @"    <chapter>\n      <heading>Software documentation "
-	@"for the %@ %@</heading>\n    </chapter>\n", name, type];
+	@"for the %@ %@</heading>\n    </chapter>\n", name, kind];
     }
 
   /*
@@ -1189,10 +1347,7 @@ static BOOL snuggleStart(NSString *t)
 
   for (j = 0; j < ind; j++) [str appendString: @" "];
   [str appendString: @"<desc>\n"];
-  if ([comment length] == 0)
-    {
-      comment = @"<em>Description forthcoming.</em>";
-    }
+  comment = [self checkComment: comment unit: nil info: d];
   [self reformat: comment withIndent: ind + 2 to: str];
   for (j = 0; j < ind; j++) [str appendString: @" "];
   [str appendString: @"</desc>\n"];
@@ -1202,7 +1357,9 @@ static BOOL snuggleStart(NSString *t)
     {
       NSString	*vName = [names objectAtIndex: i];
 
-      [self outputInstanceVariable: [ivars objectForKey: vName] to: str];
+      [self outputInstanceVariable: [ivars objectForKey: vName]
+				to: str
+			       for: unitName];
     }
 
   names = [[methods allKeys] sortedArrayUsingSelector: @selector(compare:)];
@@ -1237,7 +1394,9 @@ static BOOL snuggleStart(NSString *t)
 	   withIndent: (unsigned)ind
 		   to: (NSMutableString*)buf
 {
+#if GS_WITH_GC == 0
   CREATE_AUTORELEASE_POOL(arp);
+#endif
   unsigned	l = [str length];
   NSRange	r = NSMakeRange(0, l);
   unsigned	i = 0;
@@ -1263,7 +1422,7 @@ static BOOL snuggleStart(NSString *t)
 	  i = r.location;
 	}
       /*
-       * Now find the end of the exmple, and output the whole example
+       * Now find the end of the example, and output the whole example
        * literally as it appeared in the comment.
        */
       r = [str rangeOfString: @"</example>"
@@ -1302,7 +1461,7 @@ static BOOL snuggleStart(NSString *t)
   a = [self split: str];
   for (i = 0; i < [a count]; i++)
     {
-      int	j;
+      unsigned int	j;
 
       str = [a objectAtIndex: i];
 
@@ -1360,6 +1519,10 @@ static BOOL snuggleStart(NSString *t)
   return ind;
 }
 
+/**
+ * Split comment text into an array of words (to be reformatted) and
+ * insert markup for cross referencing and highlighting.
+ */
 - (NSArray*) split: (NSString*)str
 {
   NSMutableArray	*a = [NSMutableArray arrayWithCapacity: 128];
@@ -1479,7 +1642,7 @@ static BOOL snuggleStart(NSString *t)
 	       */
 	      if (optr[-2] == '/' && optr[-3] != ' ')
 		{
-		  unsigned	len = ptr - buf;
+		  unsigned	len = optr - buf;
 		  unichar	c[len + 1];
 
 		  memcpy(c, buf, (len+1)*sizeof(unichar));
@@ -1490,7 +1653,7 @@ static BOOL snuggleStart(NSString *t)
 		}
 	      else
 		{
-		  tmp = [NSString stringWithCharacters: buf length: ptr - buf];
+		  tmp = [NSString stringWithCharacters: buf length: optr - buf];
 		}
 	      [a addObject: tmp];
 	    }
@@ -1511,10 +1674,10 @@ static BOOL snuggleStart(NSString *t)
 
   /*
    * Phase 2 ... the array of words is checked to see if a word contains
-   * a well known constant, or a method name specification.
+   * a well known constant, or a method or function name specification.
    * Where these special cases apply, the array of words is modified to
    * insert extra gsdoc markup to highlight the constants and to create
-   * references to where the named methods are documented.
+   * references to where the named methods or functions are documented.
    */
   for (l = 0; l < [a count]; l++)
     {
@@ -1523,7 +1686,6 @@ static BOOL snuggleStart(NSString *t)
       NSString		*tmp = [a objectAtIndex: l];
       unsigned		pos;
       NSRange		r;
-      BOOL		hadMethod = NO;
 
       if (constants == nil)
 	{
@@ -1666,6 +1828,7 @@ static BOOL snuggleStart(NSString *t)
       /*
        * Ensure that methods are rendered as references.
        * First look for format with class name in square brackets.
+       * If that's all there is, we make a class reference.
        */
       r = [tmp rangeOfString: @"["];
       if (r.length > 0)
@@ -1680,7 +1843,8 @@ static BOOL snuggleStart(NSString *t)
 	      unsigned	ePos = r.location;
 	      NSString	*cName = nil;
 	      NSString	*mName = nil;
-	      unichar	c;
+	      unichar	c = 0;
+	      BOOL	isProtocol = NO;
 
 	      if (pos < ePos
 		&& [identStart characterIsMember:
@@ -1737,8 +1901,7 @@ static BOOL snuggleStart(NSString *t)
 		      cName = [tmp substringWithRange: r];
 		    }
 		}
-	      else if (pos < ePos
-		&& (c = [tmp characterAtIndex: pos]) == '(')
+	      else if (pos < ePos && (c = [tmp characterAtIndex: pos]) == '(')
 		{
 		  /*
 		   * Look for protocol name.
@@ -1760,6 +1923,7 @@ static BOOL snuggleStart(NSString *t)
 			}
 		      pos++;
 		    }
+		  isProtocol = YES;
 		}
 
 	      if (pos < ePos && (c == '+' || c == '-'))
@@ -1811,6 +1975,14 @@ static BOOL snuggleStart(NSString *t)
 		  if (sPos > 0)
 		    {
 		      start = [tmp substringToIndex: sPos];
+		      if ([start isEqualToString: @"["] == YES)
+			{
+			  start = nil;
+			}
+		      else if ([start hasSuffix: @"["] == YES)
+			{
+			  start = [start substringToIndex: [start length] - 1];
+			}
 		    }
 		  else
 		    {
@@ -1819,6 +1991,14 @@ static BOOL snuggleStart(NSString *t)
 		  if (ePos < [tmp length])
 		    {
 		      end = [tmp substringFromIndex: ePos];
+		      if ([end isEqualToString: @"]"] == YES)
+			{
+			  end = nil;
+			}
+		      else if ([end hasPrefix: @"]"] == YES)
+			{
+			  end = [end substringFromIndex: 1];
+			}
 		    }
 		  else
 		    {
@@ -1843,11 +2023,30 @@ static BOOL snuggleStart(NSString *t)
 		      ref = [NSString stringWithFormat:
 			@"<ref type=\"method\" id=\"%@\">", mName];
 		    }
+		  else if (isProtocol == YES)
+		    {
+		      ref = [NSString stringWithFormat:
+			@"<ref type=\"method\" id=\"%@\" class=\"%@\">",
+			mName, cName];
+		      if (isProtocol == YES)
+			{
+			  if (sub == nil)
+			    {
+			      sub = tmp;
+			    }
+			  sub = [sub stringByReplacingString: @"("
+						  withString: @"&lt;"];
+			  sub = [sub stringByReplacingString: @")"
+						  withString: @"&gt;"];
+			}
+		    }
 		  else
 		    {
 		      ref = [NSString stringWithFormat:
 			@"<ref type=\"method\" id=\"%@\" class=\"%@\">",
 			mName, cName];
+		      sub = [NSString stringWithFormat: @"[%@&nbsp;%@]",
+			cName, mName];
 		    }
 		  [a insertObject: ref atIndex: l++];
 		  if (sub != nil)
@@ -1861,15 +2060,60 @@ static BOOL snuggleStart(NSString *t)
 		    {
 		      [a insertObject: end atIndex: ++l];
 		    }
-		  hadMethod = YES;
+		}
+	      else if (pos == ePos && cName != nil)
+		{
+		  NSString	*ref;
+
+		  if (isProtocol == YES)
+		    {
+		      NSRange	r;
+
+		      r = NSMakeRange(1, [cName length] - 2);
+		      cName = [cName substringWithRange: r];
+		      ref = [NSString stringWithFormat:
+			@"<ref type=\"protocol\" id=\"(%@)\">&lt;%@&gt;</ref>",
+			cName, cName];
+		    }
+		  else if ([cName hasSuffix: @")"] == YES)
+		    {
+		      ref = [NSString stringWithFormat:
+			@"<ref type=\"category\" id=\"%@\">%@</ref>",
+			cName, cName];
+		    }
+		  else
+		    {
+		      ref = [NSString stringWithFormat:
+			@"<ref type=\"class\" id=\"%@\">%@</ref>",
+			cName, cName];
+		    }
+		  [a replaceObjectAtIndex: l withObject: ref];
+		  if (ePos < [tmp length])
+		    {
+		      NSString	*end = [tmp substringFromIndex: ePos];
+
+		      if ([end isEqualToString: @"]"] == YES)
+			{
+			  end = nil;
+			}
+		      if ([end hasPrefix: @"]"] == YES)
+			{
+			  end = [end substringFromIndex: 1];
+			}
+		      if ([end length] > 0)
+			{
+			  [a insertObject: end atIndex: ++l];
+			}
+		    }
 		}
 	    }
+	  continue;
 	}
       
       /*
        * Now handle bare method names for current class ... outside brackets.
        */
-      if (hadMethod == NO && ([tmp hasPrefix: @"-"] || [tmp hasPrefix: @"+"]))
+      if ([tmp hasPrefix: @"-"] || [tmp hasPrefix: @"+"])
 	{
 	  unsigned	ePos = [tmp length];
 	  NSString	*mName = nil;
@@ -1898,7 +2142,10 @@ static BOOL snuggleStart(NSString *t)
 		  isEqual: @",..."])
 		{
 		  pos += 4;
-		  c = [tmp characterAtIndex: pos];
+		  if (pos < ePos)
+		    {
+		      c = [tmp characterAtIndex: pos];
+		    }
 		}
 	      if (pos > 1 && (pos == ePos || c == ',' || c == '.' || c == ';'))
 		{
@@ -1932,9 +2179,55 @@ static BOOL snuggleStart(NSString *t)
 		    {
 		      [a insertObject: end atIndex: ++l];
 		    }
-		  hadMethod = YES;
 		}
 	    }
+	  continue;
+	}
+      /*
+       * Now handle function names.  Anything ending in '()' is assumed to
+       * be a referencable function name except 'main()' ... which is special.
+       * NB. A comma, fullstop, or semicolon following '()' is counted as if
+       * the text ended in '()'
+       */
+      r = [tmp rangeOfString: @"()"];
+      if (r.length > 0)
+	{
+	  unsigned	c = [tmp characterAtIndex: 0];
+	  unsigned	len = [tmp length];
+	  NSString	*str = [tmp substringToIndex: r.location];
+	  BOOL		ok = NO;
+
+	  if ([identStart characterIsMember: c] == YES
+	    && [str isEqual: @"main"] == NO)
+	    {
+	      ok = YES;
+	      if (len > NSMaxRange(r))
+		{
+		  NSString	*end;
+
+		  end = [tmp substringFromIndex: NSMaxRange(r)];
+		  c = [end characterAtIndex: 0];
+		  if (c == ',' || c == '.' || c == ';')
+		    {
+		      [a insertObject: end atIndex: l + 1];
+		      tmp = [tmp substringToIndex: NSMaxRange(r)];
+		      [a replaceObjectAtIndex: l withObject: tmp];
+		    }
+		  else
+		    {
+		      ok = NO;
+		    }
+		}
+	    }
+	  if (ok == YES)
+	    {
+	      str = [NSString stringWithFormat:
+		@"<ref type=\"function\" id=\"%@\">", str];
+	      [a insertObject: str atIndex: l++];
+	      l++;	// Point past the function name in the array.
+	      [a insertObject: @"</ref>" atIndex: l++];
+	      continue;
+	    }
 	}
     }
 
@@ -1993,6 +2286,8 @@ static BOOL snuggleStart(NSString *t)
 	}
       else
 	{
+	  NSString	*up = [ud stringForKey: @"Up"];
+
 	  /*
 	   * No pre-existing file, and no blank template available ...
 	   * Generate a standard template.
@@ -2001,10 +2296,20 @@ static BOOL snuggleStart(NSString *t)
 
 	  [str appendString: @"<?xml version=\"1.0\"?>\n"];
 	  [str appendString: @"<!DOCTYPE gsdoc PUBLIC "];
-	  [str appendString: @"\"-//GNUstep//DTD gsdoc 0.6.7//EN\" "];
-	  [str appendString: @"\"http://www.gnustep.org/gsdoc-0_6_7.xml\">\n"];
+	  [str appendString: @"\"-//GNUstep//DTD gsdoc 1.0.0//EN\" "];
+	  [str appendString: @"\"http://www.gnustep.org/gsdoc-1_0_0.xml\">\n"];
 	  [str appendString: @"<gsdoc base=\""];
 	  [str appendString: [name lastPathComponent]];
+	  /*
+	   * If a -Up default has been set, create an up link in this
+	   * template file... as long as the specified up link is not
+	   * the template file itsself.
+	   */
+	  if (up != nil && [up isEqual: [name lastPathComponent]] == NO)
+	    {
+	      [str appendString: @"\" up=\""];
+	      [str appendString: up];
+	    }
 	  [str appendString: @"\">\n"];
 	  [str appendString: @"  <head>\n"];
 	  [str appendString: @"    <title>"];
@@ -2016,6 +2321,7 @@ static BOOL snuggleStart(NSString *t)
 	  [str appendString: @"\"></author>\n"];
 	  [str appendString: @"  </head>\n"];
 	  [str appendString: @"  <body>\n"];
+	  [str appendString: @"    <front><contents /></front>\n"];
 	  [str appendString: @"  </body>\n"];
 	  [str appendString: @"</gsdoc>\n"];
 	}
@@ -2028,10 +2334,14 @@ static BOOL snuggleStart(NSString *t)
   start = [str rangeOfString: tmp];
   if (start.length == 0)
     {
-      start = [str rangeOfString: @"</body>"];
+      start = [str rangeOfString: @"<back>"];
       if (start.length == 0)
 	{
-	  NSLog(@"No </body> markup in %@ document", kind);
+	  start = [str rangeOfString: @"</body>"];
+	}
+      if (start.length == 0)
+	{
+	  NSLog(@"No <back> or </body> markup in %@ document", kind);
 	  return NO;
 	}
       [str insertString: tmp atIndex: start.location];
