@@ -54,12 +54,16 @@
   unichar	*buffer;
   unsigned	length;
   unsigned	pos;
+  BOOL		inHeader;
   BOOL		commentsRead;
+  BOOL		haveOutput;
   BOOL		haveSource;
   BOOL		inInstanceVariables;
   BOOL		inArgList;
   BOOL		documentAllInstanceVariables;
   BOOL		verbose;
+  BOOL		warn;
+  BOOL		standards;
   NSDictionary		*wordMap;
   NSString		*declared;	/** Where classes were declared. */
   NSMutableArray	*ifStack;	/** Track preprocessor conditionals. */
@@ -75,36 +79,44 @@
 
 - (NSMutableDictionary*) info;
 - (id) init;	/** <init> Simple initialiser */
+- (NSMutableArray*) outputs;
+- (unsigned) parseComment;
 - (NSMutableDictionary*) parseDeclaration;
 - (NSMutableDictionary*) parseFile: (NSString*)name isSource: (BOOL)isSource;
 - (NSString*) parseIdentifier;
 - (NSMutableDictionary*) parseImplementation;
 - (NSMutableDictionary*) parseInterface;
 - (NSMutableDictionary*) parseInstanceVariables;
+- (NSMutableDictionary*) parseMacro;
 - (NSMutableDictionary*) parseMethodIsDeclaration: (BOOL)flag;
 - (NSMutableDictionary*) parseMethodsAreDeclarations: (BOOL)flag;
 - (NSString*) parseMethodType;
+- (unsigned) parsePreprocessor;
 - (NSMutableDictionary*) parseProtocol;
 - (NSMutableArray*) parseProtocolList;
+- (unsigned) parseSpace: (NSCharacterSet*)spaceSet;
+- (unsigned) parseSpace;
 - (void) reset;
 - (void) setDeclared: (NSString*)name;
 - (void) setDocumentAllInstanceVariables: (BOOL)flag;
 - (void) setGenerateStandards: (BOOL)flag;
 - (void) setStandards: (NSMutableDictionary*)dict;
-- (void) setVerbose: (BOOL)flag;
 - (void) setWordMap: (NSDictionary*)map;
 - (void) setupBuffer;
 - (unsigned) skipArray;
 - (unsigned) skipBlock;
-- (unsigned) skipComment;
 - (unsigned) skipLiteral;
-- (unsigned) skipPreprocessor;
 - (unsigned) skipRemainderOfLine;
 - (unsigned) skipSpaces;
 - (unsigned) skipStatement;
 - (unsigned) skipStatementLine;
+- (unsigned) skipToEndOfLine;
 - (unsigned) skipUnit;
-- (unsigned) skipWhiteSpace;
-- (NSArray*) source;
+- (NSMutableArray*) sources;
 @end
+
+/** Let's document a macro
+ */
+#define	fibble(a,b,c)	feep /** with three arguments: a, b, c */
+
 #endif
