@@ -20,7 +20,7 @@
 
 #include "config.h"
 #include	<string.h>
-#include	<base/preface.h>
+#include	<GNUstepBase/preface.h>
 #include	<Foundation/NSObject.h>
 #include	<Foundation/NSArray.h>
 #include	<Foundation/NSDictionary.h>
@@ -46,7 +46,7 @@ main(int argc, char** argv, char **env)
   NSString		*value;
   NSString		*user = nil;
   BOOL			found = NO;
-  int			i;
+  unsigned int		i;
 
 #ifdef GS_PASS_ARGUMENTS
   [NSProcessInfo initializeWithArguments:argv count:argc environment:env];
@@ -55,9 +55,9 @@ main(int argc, char** argv, char **env)
   proc = [NSProcessInfo processInfo];
   if (proc == nil)
     {
-      NSLog(@"defaults: unable to get process information!\n");
+      GSPrintf(stderr, @"defaults: unable to get process information!\n");
       [pool release];
-      exit(0);
+      exit(EXIT_SUCCESS);
     }
 
   args = [proc arguments];
@@ -116,15 +116,15 @@ main(int argc, char** argv, char **env)
 "defaults help\n"
 "    list options fo the defaults command.\n\n");
 	  [pool release];
-	  exit(0);
+	  exit(EXIT_SUCCESS);
 	}
       else if ([[args objectAtIndex: i] isEqual: @"plist"])
 	{
 	  printf(
 "A property list is a method of providing structured information consisting\n"
 "of strings, arrays, dictionaries, and binary data.\n\n"
-"The defaults system allows you to work with a human-readable form of a
-property list which is set as the value of a default.\n\n");
+"The defaults system allows you to work with a human-readable form of a\n"
+"property list which is set as the value of a default.\n\n");
 	  printf(
 "In a property list, strings appear as plain text (as long as they contain\n"
 "no special characters), and inside quotation marks otherwise.\n"
@@ -171,16 +171,16 @@ property list which is set as the value of a default.\n\n");
 "    Checksum = <01014b5b 123a8b20>\n"
 "}'\n\n");
 	  [pool release];
-	  exit(0);
+	  exit(EXIT_SUCCESS);
 	}
     }
 
   i = 1;
   if ([args count] <= i)
     {
-      NSLog(@"defaults: too few arguments supplied!\n");
+      GSPrintf(stderr, @"defaults: too few arguments supplied!\n");
       [pool release];
-      exit(0);
+      exit(EXIT_SUCCESS);
     }
   if ([[args objectAtIndex: i] isEqual: @"-u"])
     {
@@ -190,9 +190,9 @@ property list which is set as the value of a default.\n\n");
 	}
       else
 	{
-	  NSLog(@"defaults: no name supplied for -u option!\n");
+	  GSPrintf(stderr, @"defaults: no name supplied for -u option!\n");
 	  [pool release];
-	  exit(0);
+	  exit(EXIT_SUCCESS);
 	}
     }
   if (user)
@@ -206,18 +206,18 @@ property list which is set as the value of a default.\n\n");
     }
   if (defs == nil)
     {
-      NSLog(@"defaults: unable to access defaults database!\n");
+      GSPrintf(stderr, @"defaults: unable to access defaults database!\n");
       [pool release];
-      exit(0);
+      exit(EXIT_SUCCESS);
     }
   /* We don't want this tool in the defaults database - so remove it. */
   [defs removePersistentDomainForName: [proc processName]];
 
   if ([args count] <= i)
     {
-      NSLog(@"defaults: too few arguments supplied!\n");
+      GSPrintf(stderr, @"defaults: too few arguments supplied!\n");
       [pool release];
-      exit(0);
+      exit(EXIT_SUCCESS);
     }
 
   if ([[args objectAtIndex: i] isEqual: @"read"] ||
@@ -245,9 +245,9 @@ property list which is set as the value of a default.\n\n");
 	{
 	  if ([args count] == ++i)
 	    {
-	      NSLog(@"defaults: too few arguments supplied!\n");
+	      GSPrintf(stderr, @"defaults: too few arguments supplied!\n");
 	      [pool release];
-	      exit(0);
+	      exit(EXIT_SUCCESS);
 	    }
 	  owner = nil;
 	  name = [args objectAtIndex: i];
@@ -367,7 +367,7 @@ property list which is set as the value of a default.\n\n");
 		{
 		  printf("defaults write: invalid input - nul domain name\n");
 		  [pool release];
-		  exit(0);
+		  exit(EXIT_SUCCESS);
 		}
 	      for (str = start; *str; str++)
 		{
@@ -376,7 +376,7 @@ property list which is set as the value of a default.\n\n");
 		      printf("defaults write: invalid input - "
 				"space in domain name.\n");
 		      [pool release];
-		      exit(0);
+		      exit(EXIT_SUCCESS);
 		    }
 		}
 	      owner = [NSString stringWithCString: start];
@@ -403,7 +403,7 @@ property list which is set as the value of a default.\n\n");
 		  printf("defaults write: invalid input - "
 			"nul default name.\n");
 		  [pool release];
-		  exit(0);
+		  exit(EXIT_SUCCESS);
 		}
 	      for (str = start; *str; str++)
 		{
@@ -412,7 +412,7 @@ property list which is set as the value of a default.\n\n");
 		      printf("defaults write: invalid input - "
 				"space in default name.\n");
 		      [pool release];
-		      exit(0);
+		      exit(EXIT_SUCCESS);
 		    }
 		}
 	      name = [NSString stringWithCString: start];
@@ -448,7 +448,7 @@ property list which is set as the value of a default.\n\n");
 				  printf("defaults write: fatal error - "
 					    "out of memory.\n");
 				  [pool release];
-				  exit(0);
+				  exit(EXIT_SUCCESS);
 				}
 			    }
 			  if (fgets(ptr, BUFSIZ, stdin) == 0)
@@ -456,7 +456,7 @@ property list which is set as the value of a default.\n\n");
 			      printf("defaults write: invalid input - "
 					"no final quote.\n");
 			      [pool release];
-			      exit(0);
+			      exit(EXIT_SUCCESS);
 			    }
 			}
 		      if (*ptr == '\'')
@@ -489,7 +489,7 @@ property list which is set as the value of a default.\n\n");
 		  printf("defaults write: invalid input - "
 			    "empty property list\n");
 		  [pool release];
-		  exit(0);
+		  exit(EXIT_SUCCESS);
 		}
 
 	      /*
@@ -506,7 +506,7 @@ property list which is set as the value of a default.\n\n");
 		      printf("defaults write: invalid input - "
 				    "bad property list\n");
 		      [pool release];
-		      exit(0);
+		      exit(EXIT_SUCCESS);
 		    }
 		  else
 		    {
@@ -528,9 +528,9 @@ property list which is set as the value of a default.\n\n");
 	  owner = [args objectAtIndex: i++];
 	  if ([args count] <= i)
 	    {
-	      NSLog(@"defaults: no dictionary or key for write!\n");
+	      GSPrintf(stderr, @"defaults: no dictionary or key for write!\n");
 	      [pool release];
-	      exit(0);
+	      exit(EXIT_SUCCESS);
 	    }
 	  name = [args objectAtIndex: i++];
 	  if ([args count] > i)
@@ -549,7 +549,7 @@ property list which is set as the value of a default.\n\n");
 		      printf("defaults write: invalid input - "
 				    "bad property list\n");
 		      [pool release];
-		      exit(0);
+		      exit(EXIT_SUCCESS);
 		    }
 		}
 	      else
@@ -571,17 +571,19 @@ property list which is set as the value of a default.\n\n");
 	      if (domain == nil ||
 			[domain isKindOfClass: [NSDictionary class]] == NO)
 		{
-		  NSLog(@"defaults write: domain is not a dictionary!\n");
+		  GSPrintf(stderr,
+		    @"defaults write: domain is not a dictionary!\n");
 		  [pool release];
-		  exit(0);
+		  exit(EXIT_SUCCESS);
 		}
 	    }
 	}
 
       if ([defs synchronize] == NO)
 	{
-	  NSLog(@"defaults: unable to write to defaults database - %s\n",
-		strerror(errno));
+	  GSPrintf(stderr,
+	    @"defaults: unable to write to defaults database - %s\n",
+	    strerror(errno));
 	}
     }
   else if ([[args objectAtIndex: i] isEqual: @"delete"])
@@ -613,7 +615,7 @@ property list which is set as the value of a default.\n\n");
 		{
 		  printf("defaults delete: invalid input\n");
 		  [pool release];
-		  exit(0);
+		  exit(EXIT_SUCCESS);
 		}
 	      owner = [NSString stringWithCString: start];
 	      start = ptr;
@@ -634,7 +636,7 @@ property list which is set as the value of a default.\n\n");
 		{
 		  printf("defaults delete: invalid input\n");
 		  [pool release];
-		  exit(0);
+		  exit(EXIT_SUCCESS);
 		}
 	      name = [NSString stringWithCString: start];
 	      domain = [[defs persistentDomainForName: owner] mutableCopy];
@@ -682,8 +684,9 @@ property list which is set as the value of a default.\n\n");
 	}
       if ([defs synchronize] == NO)
 	{
-	  NSLog(@"defaults: unable to write to defaults database - %s\n",
-		strerror(errno));
+	  GSPrintf(stderr,
+	    @"defaults: unable to write to defaults database - %s\n",
+	    strerror(errno));
 	}
     }
   else if ([[args objectAtIndex: i] isEqual: @"domains"])
@@ -700,9 +703,9 @@ property list which is set as the value of a default.\n\n");
     {
       if ([args count] == ++i)
 	{
-	  NSLog(@"defaults: no arguments for find!\n");
+	  GSPrintf(stderr, @"defaults: no arguments for find!\n");
 	  [pool release];
-	  exit(0);
+	  exit(EXIT_SUCCESS);
 	}
       name = [args objectAtIndex: i];
 
@@ -756,10 +759,10 @@ property list which is set as the value of a default.\n\n");
     }
   else
     {
-      NSLog(@"defaults: unknown option supplied!\n");
+      GSPrintf(stderr, @"defaults: unknown option supplied!\n");
     }
 
   [pool release];
-  exit(0);
+  exit(EXIT_SUCCESS);
 }
 
