@@ -18,14 +18,166 @@
 
    */
 
-#include "config.h"
+#include	"config.h"
 #include	<Foundation/Foundation.h>
 
 #include        <stdio.h>
 #include	<unistd.h>
 
+#ifdef __MINGW__
+#include	"process.h"
+#endif
+
+#include <fcntl.h>
+#ifdef	HAVE_SYSLOG_H
+#include <syslog.h>
+#endif
+
+#include        <signal.h>
+
+#ifndef NSIG
+#define NSIG    32
+#endif
+
+static int	is_daemon = 0;		/* Currently running as daemon.	 */
+static char	ebuf[2048];
+
+#ifdef HAVE_SYSLOG
+
+static int	log_priority;
+
+static void
+gdnc_log (int prio)
+{
+  if (is_daemon)
+    {
+      syslog (log_priority | prio, ebuf);
+    }
+  else if (prio == LOG_INFO)
+    {
+      write (1, ebuf, strlen (ebuf));
+      write (1, "\n", 1);
+    }
+  else
+    {
+      write (2, ebuf, strlen (ebuf));
+      write (2, "\n", 1);
+    }
+
+  if (prio == LOG_CRIT)
+    {
+      if (is_daemon)
+	{
+	  syslog (LOG_CRIT, "exiting.");
+	}
+      else
+     	{
+	  fprintf (stderr, "exiting.\n");
+	  fflush (stderr);
+	}
+      exit(EXIT_FAILURE);
+    }
+}
+#else
+
+#define	LOG_CRIT	2
+#define LOG_DEBUG	0
+#define LOG_ERR		1
+#define LOG_INFO	0
+#define LOG_WARNING	0
+void
+gdnc_log (int prio)
+{
+  write (2, ebuf, strlen (ebuf));
+  write (2, "\n", 1);
+  if (prio == LOG_CRIT)
+    {
+      fprintf (stderr, "exiting.\n");
+      fflush (stderr);
+      exit(EXIT_FAILURE);
+    }
+}
+#endif
+
+static void
+ihandler(int sig)
+{
+  static BOOL	beenHere = NO;
+  BOOL		action;
+  const char	*e;
+
+  /*
+   * Deal with recursive call of handler.
+   */
+  if (beenHere == YES)
+    {
+      abort();
+    }
+  beenHere = YES;
+
+  /*
+   * If asked to terminate, do so cleanly.
+   */
+  if (sig == SIGTERM)
+    {
+      exit(EXIT_FAILURE);
+    }
+
+#ifdef	DEBUG
+  action = YES;		// abort() by default.
+#else
+  action = NO;		// exit() by default.
+#endif
+  e = getenv("CRASH_ON_ABORT");
+  if (e != 0)
+    {
+      if (strcasecmp(e, "yes") == 0 || strcasecmp(e, "true") == 0)
+	action = YES;
+      else if (strcasecmp(e, "no") == 0 || strcasecmp(e, "false") == 0)
+	action = NO;
+      else if (isdigit(*e) && *e != '0')
+	action = YES;
+      else
+	action = NO;
+    }
+
+  if (action == YES)
+    {
+      abort();
+    }
+  else
+    {
+      fprintf(stderr, "gdnc killed by signal %d\n", sig);
+      exit(sig);
+    }
+}
+
+
 #include	"gdnc.h"
 
+/*
+ * The following dummy class is here solely as a workaround for pre 3.3
+ * versions of gcc where protocols didn't work properly unless implemented
+ * in the source where the '@protocol()' directive is used.
+ */
+@interface NSDistributedNotificationCenterDummy : NSObject <GDNCClient>
+- (oneway void) postNotificationName: (NSString*)name
+                              object: (NSString*)object
+                            userInfo: (NSData*)info
+                            selector: (NSString*)aSelector
+                                  to: (unsigned long)observer;
+@end
+@implementation	NSDistributedNotificationCenterDummy
+- (oneway void) postNotificationName: (NSString*)name
+                              object: (NSString*)object
+                            userInfo: (NSData*)info
+                            selector: (NSString*)aSelector
+                                  to: (unsigned long)observer
+{
+}
+@end
+
+
 @interface	GDNCNotification : NSObject
 {
 @public
@@ -212,21 +364,53 @@
 - (id) init
 {
   NSString	*hostname;
+  NSString	*service = GDNC_SERVICE;
+  BOOL		isLocal = NO;
 
   connections = NSCreateMapTable(NSObjectMapKeyCallBacks,
 		NSNonOwnedPointerMapValueCallBacks, 0);
   allObservers = NSCreateHashTable(NSNonOwnedPointerHashCallBacks, 0);
   observersForNames = [NSMutableDictionary new];
   observersForObjects = [NSMutableDictionary new];
-  conn = [NSConnection defaultConnection];
+
+  if ([[NSUserDefaults standardUserDefaults] boolForKey: @"GSNetwork"] == YES)
+    {
+      service = GDNC_NETWORK;
+    }
+  hostname = [[NSUserDefaults standardUserDefaults] stringForKey: @"NSHost"];
+  if ([hostname length] == 0
+    || [hostname isEqualToString: @"localhost"] == YES
+    || [hostname isEqualToString: @"127.0.0.1"] == YES)
+    {
+      isLocal = YES;
+    }
+
+  /*
+   * If this is the local server for the current host,
+   * use the loopback network interface.  Otherwise
+   * create a public connection.
+   */
+  if (0 && isLocal == YES && service != GDNC_NETWORK)
+    {
+      NSPort	*port = [NSSocketPort portWithNumber: 0
+					      onHost: [NSHost localHost]
+					forceAddress: @"127.0.0.1"
+					    listener: YES];
+      conn = [[NSConnection alloc] initWithReceivePort: port sendPort: nil];
+    }
+  else
+    {
+      conn = [NSConnection defaultConnection];
+    }
   [conn setRootObject: self];
 
-  hostname = [[NSUserDefaults standardUserDefaults] stringForKey: @"NSHost"];
-  if ([hostname length] == 0)
+  if (isLocal == YES
+    || [[NSHost hostWithName: hostname] isEqual: [NSHost currentHost]] == YES)
     {
-      if ([conn registerName: GDNC_SERVICE] == NO)
+      if ([conn registerName: service] == NO)
 	{
-	  NSLog(@"gdnc - unable to register with name server - quiting.");
+	  NSLog(@"gdnc - unable to register with name server as %@ - quiting.",
+	    service);
 	  DESTROY(self);
 	  return self;
 	}
@@ -251,7 +435,7 @@
 	{
 	  NSString	*name = [a objectAtIndex: c];
 
-	  name = [GDNC_SERVICE stringByAppendingFormat: @"-%@", name];
+	  name = [service stringByAppendingFormat: @"-%@", name];
 	  if ([ns registerPort: port forName: name] == NO)
 	    {
 	    }
@@ -262,7 +446,7 @@
 	{
 	  NSString	*name = [a objectAtIndex: c];
 
-	  name = [GDNC_SERVICE stringByAppendingFormat: @"-%@", name];
+	  name = [service stringByAppendingFormat: @"-%@", name];
 	  if ([ns registerPort: port forName: name] == NO)
 	    {
 	    }
@@ -408,7 +592,7 @@
   if (connection == conn)
     {
       NSLog(@"argh - gdnc server root connection has been destroyed.");
-      exit(1);
+      exit(EXIT_FAILURE);
     }
   else
     {
@@ -446,6 +630,12 @@
 		  format: @"registration with registered client"];
     }
   info = [GDNCClient new];
+  if ([(id)client isProxy] == YES)
+    {
+      Protocol	*p = @protocol(GDNCClient);
+
+      [(id)client setProtocolForProxy: p];
+    }
   info->client = client;
   NSMapInsert(table, client, info);
   RELEASE(info);
@@ -461,7 +651,7 @@
   NSMutableArray	*byName;
   NSMutableArray	*byObject;
   unsigned		pos;
-  GDNCNotification	*notification;
+  GDNCNotification	*notification = nil;
 
   byName = [observersForNames objectForKey: notificationName];
   byObject = [observersForObjects objectForKey: notificationObject];
@@ -472,8 +662,8 @@
     {
       GDNCObserver	*obs = [byName objectAtIndex: pos - 1];
 
-      if (obs->notificationObject == nil ||
-		[obs->notificationObject isEqual: notificationObject])
+      if (obs->notificationObject == nil
+	|| [obs->notificationObject isEqual: notificationObject])
 	{
 	  [observers addObject: obs];
 	}
@@ -482,8 +672,8 @@
     {
       GDNCObserver	*obs = [byObject objectAtIndex: pos - 1];
 
-      if (obs->notificationName == nil ||
-		[obs->notificationName isEqual: notificationName])
+      if (obs->notificationName == nil
+	|| [obs->notificationName isEqual: notificationName])
 	{
 	  if ([observers indexOfObjectIdenticalTo: obs] == NSNotFound)
 	    {
@@ -550,8 +740,8 @@
 	   *		the queue is empty	([obs->queue count] == 0)
 	   *		the observer is removed	(obs is not in allObservers)
 	   */
-	  while (obs != nil && [obs->queue count] > 0 &&
-		    NSHashGet(allObservers, obs) != 0)
+	  while (obs != nil && [obs->queue count] > 0
+	    && NSHashGet(allObservers, obs) != 0)
 	    {
 	      NS_DURING
 		{
@@ -784,28 +974,52 @@
 int
 main(int argc, char** argv, char** env)
 {
+  int                   c;
   GDNCServer		*server;
   NSString		*str;
   BOOL			shouldFork = YES;
+  BOOL			debugging = NO;
   CREATE_AUTORELEASE_POOL(pool);
 
 #ifdef GS_PASS_ARGUMENTS
   [NSProcessInfo initializeWithArguments:argv count:argc environment:env];
 #endif
+  [NSObject enableDoubleReleaseCheck: YES];
+  if (argc > 1 && strcmp(argv[argc-1], "-f") == 0)
+    {
+      shouldFork = NO;
+    }
   str = [[NSUserDefaults standardUserDefaults] stringForKey: @"debug"];
   if (str != nil && [str caseInsensitiveCompare: @"yes"] == NSOrderedSame)
     {
       shouldFork = NO;
+      debugging = YES;
     }
   RELEASE(pool);
-#ifndef __MINGW__  /* Don't fork on Win32 */
+#ifdef __MINGW__
   if (shouldFork)
     {
+      char	**a = malloc((argc+2) * sizeof(char*));
+
+      memcpy(a, argv, argc*sizeof(char*));
+      a[argc] = "-f";
+      a[argc+1] = 0;
+      if (_spawnv(_P_NOWAIT, argv[0], a) == -1)
+	{
+	  fprintf(stderr, "gdnc - spawn failed - bye.\n");
+	  exit(EXIT_FAILURE);
+	}
+      exit(EXIT_SUCCESS);
+    }
+#else
+  if (shouldFork)
+    {
+      is_daemon = 1;
       switch (fork())
 	{
 	  case -1:
 	    fprintf(stderr, "gdnc - fork failed - bye.\n");
-	    exit(1);
+	    exit(EXIT_FAILURE);
 
 	  case 0:
 	    /*
@@ -819,21 +1033,95 @@ main(int argc, char** argv, char** env)
 	    break;
 
 	  default:
-	    exit(0);
+	    exit(EXIT_SUCCESS);
 	}
     }
+
+  /*
+   *	Ensure we don't have any open file descriptors which may refer
+   *	to sockets bound to ports we may try to use.
+   *
+   *	Use '/dev/null' for stdin and stdout.
+   */
+  for (c = 0; c < FD_SETSIZE; c++)
+    {
+      if (is_daemon || (c != 2))
+	{
+	  (void)close(c);
+	}
+    }
+  if (open("/dev/null", O_RDONLY) != 0)
+    {
+      sprintf(ebuf, "failed to open stdin from /dev/null (%s)\n",
+	strerror(errno));
+      gdnc_log(LOG_CRIT);
+      exit(EXIT_FAILURE);
+    }
+  if (open("/dev/null", O_WRONLY) != 1)
+    {
+      sprintf(ebuf, "failed to open stdout from /dev/null (%s)\n",
+	strerror(errno));
+      gdnc_log(LOG_CRIT);
+      exit(EXIT_FAILURE);
+    }
+  if (is_daemon && open("/dev/null", O_WRONLY) != 2)
+    {
+      sprintf(ebuf, "failed to open stderr from /dev/null (%s)\n",
+	strerror(errno));
+      gdnc_log(LOG_CRIT);
+      exit(EXIT_FAILURE);
+    }
 #endif /* !MINGW */
 
   {
+#if GS_WITH_GC == 0
     CREATE_AUTORELEASE_POOL(pool);
+#endif
+    NSUserDefaults	*defs;
+    int			sym;
+
+    for (sym = 0; sym < NSIG; sym++)
+      {
+	signal(sym, ihandler);
+      }
+#ifndef __MINGW__
+    signal(SIGPIPE, SIG_IGN);
+    signal(SIGTTOU, SIG_IGN);
+    signal(SIGTTIN, SIG_IGN);
+    signal(SIGHUP, SIG_IGN);
+#endif
+    signal(SIGTERM, ihandler);
+
+    /*
+     * Make gdnc logging go to syslog unless overridden by user.
+     */
+    defs = [NSUserDefaults standardUserDefaults];
+    [defs registerDefaults: [NSDictionary dictionaryWithObjectsAndKeys:
+      @"YES", @"GSLogSyslog", nil]];
+
     server = [GDNCServer new];
+
+    /*
+     * Close standard input, output, and error to run as daemon.
+     */
+    [[NSFileHandle fileHandleWithStandardInput] closeFile];
+    [[NSFileHandle fileHandleWithStandardOutput] closeFile];
+#ifndef __MINGW__
+    if (debugging == NO)
+      {
+	[[NSFileHandle fileHandleWithStandardError] closeFile];
+      }
+#endif
+
     RELEASE(pool);
   }
 
   if (server != nil)
     {
+      CREATE_AUTORELEASE_POOL(pool);
       [[NSRunLoop currentRunLoop] run];
+      RELEASE(pool);
     }
-  exit(0);
+  exit(EXIT_SUCCESS);
 }
 
