@@ -27,11 +27,14 @@
 #include	<Foundation/NSUserDefaults.h>
 #include	<Foundation/NSDebug.h>
 #include	<Foundation/NSAutoreleasePool.h>
+#ifdef NeXT_Foundation_LIBRARY
+#include "GNUstepBase/GSCategories.h"
+#endif
 
 int
 convert_unicode(NSArray *args)
 {
-  int i;
+  unsigned int i;
 
   for (i = 2; i < [args count]; i++)
     {
@@ -45,19 +48,23 @@ convert_unicode(NSArray *args)
 
 	  data = [NSData dataWithContentsOfFile: file];
 	  myString = [[NSString alloc] initWithData: data
-				       encoding: NSUTF8StringEncoding];
+					   encoding: NSUTF8StringEncoding];
 	  AUTORELEASE(myString);
 	  if ([myString length] == 0)
-	    myString = [[NSString alloc] initWithData: data
-					 encoding: [NSString defaultCStringEncoding]];
+	    {
+	      myString = [[NSString alloc] initWithData: data
+		encoding: [NSString defaultCStringEncoding]];
+	    }
 	  output = [[file lastPathComponent]
-		     stringByAppendingPathExtension: @"unicode"];
+	    stringByAppendingPathExtension: @"unicode"];
 	  data = [myString dataUsingEncoding: NSUnicodeStringEncoding];
 	  [data writeToFile: output atomically: YES];
 	}
       NS_HANDLER
 	{
-	  NSLog(@"Converting '%@' - %@", file, [localException reason]);
+	  GSPrintf(stderr, @"Converting '%@' - %@\n", file,
+	    [localException reason]);
+	  return 1;
 	}
       NS_ENDHANDLER
     }
@@ -67,7 +74,7 @@ convert_unicode(NSArray *args)
 int
 convert_utf8(NSArray *args)
 {
-  int i;
+  unsigned int i;
 
   for (i = 2; i < [args count]; i++)
     {
@@ -81,13 +88,15 @@ convert_utf8(NSArray *args)
 
 	  myString = [NSString stringWithContentsOfFile: file];
 	  output = [[file lastPathComponent]
-		     stringByAppendingPathExtension: @"utf8"];
+	    stringByAppendingPathExtension: @"utf8"];
 	  data = [myString dataUsingEncoding: NSUTF8StringEncoding];
 	  [data writeToFile: output atomically: YES];
 	}
       NS_HANDLER
 	{
-	  NSLog(@"Converting '%@' - %@", file, [localException reason]);
+	  GSPrintf(stderr, @"Converting '%@' - %@\n", file,
+	    [localException reason]);
+	  return 1;
 	}
       NS_ENDHANDLER
     }
@@ -101,6 +110,7 @@ main(int argc, char** argv, char **env)
   NSProcessInfo		*proc;
   NSArray		*args;
   unsigned		i;
+  int			retval = 0;
 
 #ifdef GS_PASS_ARGUMENTS
   [NSProcessInfo initializeWithArguments:argv count:argc environment:env];
@@ -109,9 +119,9 @@ main(int argc, char** argv, char **env)
   proc = [NSProcessInfo processInfo];
   if (proc == nil)
     {
-      NSLog(@"defaults: unable to get process information!\n");
+      GSPrintf(stderr, @"defaults: unable to get process information!\n");
       [pool release];
-      exit(0);
+      exit(EXIT_FAILURE);
     }
 
   args = [proc arguments];
@@ -125,11 +135,11 @@ main(int argc, char** argv, char **env)
     }
   else if ([[args objectAtIndex: 1] isEqual: @"--unicode"])
     {
-      convert_unicode(args);
+      retval = convert_unicode(args);
     }
   else if ([[args objectAtIndex: 1] isEqual: @"--utf8"])
     {
-      convert_utf8(args);
+      retval = convert_utf8(args);
     }
   else
     {
@@ -145,20 +155,22 @@ main(int argc, char** argv, char **env)
 	      myString = [NSString stringWithContentsOfFile: file];
 	      result = [myString propertyListFromStringsFileFormat];
 	      if (result == nil)
-		NSLog(@"Parsing '%@' - nil property list", file);
+		GSPrintf(stderr, @"Parsing '%@' - nil property list\n", file);
 	      else if ([result isKindOfClass: [NSDictionary class]] == YES)
-		NSLog(@"Parsing '%@' - seems ok", file);
+		GSPrintf(stderr, @"Parsing '%@' - seems ok\n", file);
 	      else
-		NSLog(@"Parsing '%@' - unexpected class - %@",
-			file, [[result class] description]);
+		GSPrintf(stderr, @"Parsing '%@' - unexpected class - %@\n",
+		  file, [[result class] description]);
 	    }
 	  NS_HANDLER
 	    {
-	      NSLog(@"Parsing '%@' - %@", file, [localException reason]);
+	      GSPrintf(stderr, @"Parsing '%@' - %@\n", file,
+		[localException reason]);
+	      retval = 1;
 	    }
 	  NS_ENDHANDLER
 	}
     }
   [pool release];
-  return 0;
+  return retval;
 }
